const { spawn, exec } = require('child_process');

class Spawned {
	/**
	 * Promisified wrapper around child_process.spawn. Inherits the processes
	 * stdout and stderr streams by default.
	 * Rejects the promise if the process throws an error or closes with a
	 * non-zero exit code.
	 * @param {!string} command
	 * @param {!string[]} [args]
	 * @param {!Object} [options]
	 * @param {!string} [stdin]
	 * @param {!stream.Writable} [stdout]
	 * @param {!stream.Writable} [stderr]
	 * @returns {!Promise}
	 */
	static async spawn(command, args = [], options, stdin, stdout = process.stdout, stderr = process.stderr) {
		return new Promise((resolve, reject) => {
			const proc = spawn(command, args, options);
			if (stdin != null) {
				proc.stdin.write(stdin);
				proc.stdin.end();
			}
			if (stdout !== null) {
				proc.stdout.pipe(stdout);
			}
			if (stderr !== null) {
				proc.stderr.pipe(stderr);
			}
			proc.on('error', reject);
			proc.on('close', code => {
				if (code !== 0) {
					reject(new Error(`Process '${command} ${args.join(' ')}' closed with exit code ${code}.`));
				} else {
					resolve();
				}
			});
		});
	}

	/**
	 * Utility wrapper around Spawner.spawn(..) for spawning processes with sudo.
	 * @param {!string} sudoerPassword
	 * @param {!string} command
	 * @param {!string[]} [args]
	 * @param {!Object} [options]
	 * @param {!string} [stdin]
	 * @param {!stream.Writable} [stdout]
	 * @param {!stream.Writable} [stderr]
	 * @returns {!Promise}
	 */
	static async sudo(sudoerPassword, command, args, options = {}, stdin = '', stdout = process.stdout, stderr = process.stderr) {
		const opts = Object.create(options);
		opts.shell = true;

		// TODO: The -k switch forces sudo to prompt for a password, but only if sudo requires a password.
		// Ideally this command should somehow verify sudo is requesting a password before writing it to
		// stdin, as a measure of preventing the password from unexpectedly leaking to the process invoked
		// by sudo.
		// Alternatively it is the callers job to ensure that they are only passing a password for commands
		// that require a password in a given sudo configuration, but this is cumbersome.
		if (sudoerPassword.length > 0) {
			stdin = `${sudoerPassword}\n${stdin}`;
		}
		return Spawned.spawn(`sudo -Sk ${command}`, args, opts, stdin, stdout, stderr);
	}

	/**
	 * Utility function for spawning new detached processes, always sets the
	 * following options:
	 * 	shell: true
	 * 	detached: true
	 * 	stdio: 'ignore'
	 * @param {!string} command
	 * @param {!string[]} [args]
	 * @param {!Object} [options]
	 */
	static detached(command, args, options = {}) {
		const opts = Object.create(options);
		opts.shell = true;
		opts.detached = true;
		opts.stdio = 'ignore';
		const subprocess = spawn(command, args, opts);
		subprocess.unref();
	}

	/**
	 * Utility function for executing a shell command. Output is buffered and
	 * returned as an object containing `stdout` and `stderr` properties.
	 * @param {!string} command
	 * @param {!Object} [options]
	 * @returns {!Promise<!Object>}
	 */
	static async exec(command, options) {
		return new Promise((resolve, reject) => {
			exec(command, options, (error, stdout, stderr) => {
				if (error) {
					reject(error);
				} else {
					resolve({
						stdout: stdout.replace(/\n$/, ''),
						stderr: stderr.replace(/\n$/, '')
					});
				}
			});
		});
	}
}

module.exports = Spawned;
