const pretty = require('prettysize');
const fs = require('fs-extra');
const os = require('os');
const path = require('path');
const Jimp = require('jimp');
const md5File = require('md5-file/promise');
const { http, https } = require('follow-redirects');
const HttpsProxyAgent = require('https-proxy-agent');
const tinycolor = require('tinycolor2');
const Spawned = require('@keywest/spawned');
const net = require('net');
const request = require('request');
const progress = require('request-progress');
const spawn = require('child_process').spawn;
const extract = require('extract-zip');
const EventEmitter = require('events');
const sendkeys = require('sendkeys-js');
const stream = require('stream');

const paths = {};
paths.ASSETS = path.join(process.cwd(), 'src');
paths.DIRECTORY = process.cwd();
paths.CONFIG = path.join(paths.DIRECTORY, 'config');
paths.UPDATES = path.join(paths.CONFIG, 'updates');
paths.CACHE = path.join(paths.DIRECTORY, 'cache');
paths.TMP = path.join(paths.DIRECTORY, 'tmp');
paths.DESKTOP = path.join(os.homedir(), 'Desktop');

let win = null;
let escShortcut = null;
let updateIntervalId = -1;
let registration = null;
let exitHandlers = [];

const http_proxy = process.env['http_proxy'];
let agent = null;
if (http_proxy) {
	try {
		agent = new HttpsProxyAgent(http_proxy);
	} catch (e) {
		console.log(e);
		console.log('Falling back to non-proxied https agent...');
	}
}
if (agent == null) {
	agent = new https.Agent();
}

http.globalAgent.maxSockets = 10;

if (http_proxy) {
	request.defaults({
		proxy: http_proxy
	});
}

/**
 * @returns {!Promise}
 */
async function clearCache() {
	try {
		await fs.remove(paths.CACHE);
	} catch (err) {
		console.log(err.stack || err);
	}
	try {
		await fs.mkdirs(paths.CACHE);
	} catch (err) {
		console.log(err.stack || err);
	}
}

function queryLocalIp() {
	const interfaces = os.networkInterfaces();
	const localIps = [];
	for (const interface in interfaces) {
		for (const address of interfaces[interface]) {
			if (!address.internal && address.family == 'IPv4') {
				localIps.push(address);
			}
		}
	}
	if (localIps.length > 0) {
		return localIps[0].address;
	}
	return undefined;
}

const downloads = new Map();

class ContainerAPI {
	static async QuerySupport(feature) {
		if (feature === 'caching') {
			return true;
		} else if (feature === 'mpv') {
			if (platform === 'Windows') {
				return true;
			} else {
				return false;
			}
		} else if (feature === 'screenshots') {
			return true;
		} else if (feature === 'led') {
			let supported = false;
			try {
				supported = await MeetingMachines.isSupported();
			} catch (err) {
				supported = false;
			}
			return supported;
		} else if (feature === 'videoIn') {
			return false;
		} else if (feature === 'networking') {
			return true;
		} else {
			return null;
		}
	}
	static async RequestScreenshot() {
		try {
			const screenshotPath = path.join(paths.DIRECTORY, 'screenshot.png');
			await Shell.captureScreenshot(screenshotPath);
			return uploadScreenshot(screenshotPath);
		} catch (err) {
			console.log(err);
		}
	}
	static async RequestThumbnailScreenshot(usableWidth, usableHeight, thumbnailWidth, thumbnailHeight) {
		try {
			const thumbnailPath = path.join(paths.DIRECTORY, 'thumbnail.png');
			await Shell.captureScreenshot(thumbnailPath);
			// A screenshot taken at native monitor resolution might not match the rendered scaled resolution. If there is scaling, then scale the screenshot before working with it.
			const screenshot = await Jimp.read(thumbnailPath);
			if (screenshot.bitmap.width != usableWidth || screenshot.bitmap.height  != usableHeight) {
				const scale = usableWidth / screenshot.bitmap.width;
				const newWidth = Math.round(screenshot.bitmap.width * scale);
				const newHeight = Math.round(screenshot.bitmap.height * scale);
				await new Promise(resolve => screenshot.resize(newWidth, newHeight).write(thumbnailPath, resolve));
			}
			return uploadThumbnail(usableWidth, usableHeight, thumbnailWidth, thumbnailHeight, thumbnailPath);
		} catch (err) {
			console.log(err);
		}
	}
	static async RequestHardReboot() {
		try {
			Shell.reboot();
		} catch (err) {
			console.log(err);
		}
	}
	static async RequestSoftReboot() {
		requestSoftReboot();
	}
	static async RequestClearCache() {
		try {
			await clearCache();
			requestSoftReboot();
		} catch (err) {
			console.log(err);
		}
	}
	static async RequestPinchToZoom(pinchToZoom) {
		const packageJsonPath = path.join(paths.DIRECTORY, 'package.json');
		console.log(packageJsonPath);
		fs.readFile(packageJsonPath, 'utf8', (err, data) => {
			if (err) {
				console.log(err);
			} else {
				// If we don't find disable pinch then pinch to zoom is true
				let currentPinchToZoom = data.indexOf('--disable-pinch --overscroll-history-navigation=0') == -1;
				console.log('Current pinch to zoom: ', currentPinchToZoom, ' new value: ', pinchToZoom);
				if (pinchToZoom != currentPinchToZoom) {
					let result;
					if (currentPinchToZoom) {
						console.log('Disable pinch to zoom');
						result = data.replace(/"chromium-args": "/, '"chromium-args": "--disable-pinch --overscroll-history-navigation=0 ');
					} else {
						console.log('Enable pinch to zoom');
						result = data.replace(/--disable-pinch --overscroll-history-navigation=0 /, '');
					}
				
					fs.writeFile(packageJsonPath, result, 'utf8', err => {
						if (err) {
							console.log(err);
						} else {
							closeWindow();
						}
					});
				}
			}
		});
	}
	static async LogVitals() {
		const vitals = process.memoryUsage();
		console.log('totalHeap: ' + pretty(vitals.heapTotal) + ', usedHeap: ' + pretty(vitals.heapUsed));
		if (win !== null) {
			win.focus();
		}
	}
	static async SetFrontLedColor(r, g, b) {
		await Shell.setFrontLedColor(r, g, b);
	}
	static async UnregisteredPlayer() {
		try {
			console.log('Player unregistered by server...');
			registration = null;
			const registrationPath = path.join(paths.CONFIG, 'registration.json');
			await fs.remove(registrationPath);
			clearCache();
			showRegister();
		} catch (err) {
			console.log(err);
		}
	}
	static async UpdateTimezone(timezone) {
		try {
			Shell.updateTimezone(timezone);
		} catch (err) {
			console.log(err);
		}
	}
	static async CheckIn() {
		// Mark the lastCheckIn time stamp
	}
	static async NetworkRequest(options) {
		try {
			NetworkClient.request(options.host, options.port, options.command, options.format);
		} catch (err) {
			console.log(err);
		}
	}
	static async SerialRequest(options) {
		try {
			SerialClient.request(options.port, options.command, options.format, options.baudRate, options.dataBits, options.parity, options.stopBits, options.flowControl);
		} catch (err) {
			console.log(err);
		}
	}
	static async QueryLocalIp() {
		try {
			return queryLocalIp();
		} catch (err) {
			console.log(err);
			return '';
		}
	}
	static async QueryContainerVersion() {
		const version = await Updater.getVersion();
		const containerVersion = `${platform}_${version}`;
		return containerVersion;
	}
	static async QueryContainerType() {
		return 'Node';
	}
	static async QueryRegistration() {
		return registration;
	}
	static async QueryAssetsPath() {
		return `file://${paths.ASSETS}/`;
	}
	static async CreateVideoIn(playlistElementId, videoSource, sourceId, inputWidth, inputHeight) {
		// Not supported
	}
	static async StopVideoIn(playlistElementId, videoSource, sourceId) {
		// Not supported
	}
	static async PlayVideoIn(playlistElementId, videoSource, sourceId, x, y, w, h) {
		// Not supported
	}
	static async PlaylistDestroyed(playlistUniqueId) {
		// Not supported
	}
	static async UpdateSerialNumber(serialNumber, serialNumberModifiedTime) {
		try {
			const serial = await loadSerialNumber();
			if (serialNumberModifiedTime > serial.serialnumbermodifiedtime) {
				await saveSerialNumber({ serialnumber: serialNumber, serialnumbermodifiedtime: serialNumberModifiedTime });
			}
		} catch (exception) {
			console.log('Caught exception while setting serial number.', exception);
		}
	}
	static async QuerySerialNumber() {
		const serial = await loadSerialNumber();
		return serial;
	}
	
	/**
	 * @param {!string} folder
	 * @param {!string} name
	 * @returns {!string}
	 */
	static async getDownloadsKey(folder, name) {
		return `${folder}/${name}`;
	}
	/**
	 * @param {!string} pathsKey
	 * @param {!string} folder
	 * @param {!string} name
	 * @returns {!Promise<!string>}
	 */
	static async resolvePath(pathsKey, folder, name) {
		const folderpath = path.join(paths[pathsKey], folder);
		await fs.mkdirs(folderpath);
		return path.join(folderpath, name);
	}
	static async AbortDownload(name, folder) {
		try {
			const key = ContainerAPI.getDownloadsKey(folder, name);
			const download = downloads.get(key);
			if (download) {
				await download.abort();
			}
		} catch (err) {
			throw err;
		}
	}
	static async DownloadFile(url, name, folder, size) {
		// TODO: Verify the file will fit on disk before downloading.
		try {
			const key = ContainerAPI.getDownloadsKey(folder, name);
			const old = downloads.get(key);
			if (old) {
				console.log(`Aborting download of '${key}' due to new download starting for the same resource!`);
				await old.abort();
			}
			const filepath = await ContainerAPI.resolvePath('CACHE', folder, name);
			const filepathmd5 = await ContainerAPI.resolvePath('CACHE', folder, `${name}.md5`);
			await fs.remove(filepathmd5);
			const download = new FileStreamedRequest(url, filepath);
			await download.promise;
		} catch (err) {
			throw err;
		}
	}
	static async SaveFile(data, name, folder) {
		try {
			await FileUtils.save(name, folder, 'CACHE', data);
		} catch (err) {
			throw err;
		}
	}
	static async DeleteFile(name, folder) {
		try {
			const filepath = path.join(paths.CACHE, folder, name);
			await fs.remove(filepath);
			const filepathmd5 = path.join(paths.CACHE, folder, `${name}.md5`);
			await fs.remove(filepathmd5);
		} catch (err) {
			console.log(err);
			throw err;
		}
	}
	static async DeleteFolder(folder) {
		try {
			const filepath = path.join(paths.CACHE, folder);
			await fs.remove(filepath);
		} catch (err) {
			console.log(err);
			throw err;
		}
	}
	static async CacheLoad(name, folder) {
		try {
			return await FileUtils.load(name, folder, 'CACHE');
		} catch (err) {
			throw err;
		}
	}
	static async CacheStat(name, folder) {
		try {
			const stats = await FileUtils.stat(name, folder, 'CACHE');
			const csr = {
				exists: stats !== null,
				url: null,
				size: null,
				mtime: null
			};
			if (csr.exists) {
				csr.mtime = Math.floor(stats.mtime.getTime() / 1000);
				csr.size = stats.size;
				const filepath = path.normalize(path.join(paths.CACHE, folder, name));
				csr.url = `file://${filepath.replace(/\\/g, '/')}`;
			}
			return csr;
		} catch (err) {
			throw err;
		}
	}
	static async CacheMd5(name, folder) {
		try {
			const filepath = path.normalize(path.join(paths.CACHE, folder, name));
			const filepathmd5 = path.normalize(path.join(paths.CACHE, folder, `${name}.md5`));
			const stats = await fs.stat(filepath);
			try {
				const md5Stats = await fs.stat(filepathmd5);
				if (stats.mtime.getTime() != md5Stats.mtime.getTime()) {
					//throw new Error(`md5 cached value is outdated. Recalculating... media file mtime: ${stats.mtime.getTime()} != md5 file time: ${md5Stats.mtime.getTime()}`);
				}
				// Retrieve the md5 stored in the file
				return await fs.readFile(filepathmd5, 'utf8');
			} catch (exception) {
				const md5 = await md5File(filepath);
				// Save the md5 so we don't have to calculate it more than once
				await fs.writeFile(filepathmd5, md5, 'utf8');
				await fs.utimes(filepathmd5, stats.mtime, stats.mtime); // set the file to the same modified time as the downloaded file so we know they correspond
				return md5;
			}
		} catch (err) {
			throw err;
		}
	}
	static async ListFiles(folder) {
		try {
			const folderpath = path.join(paths.CACHE, folder);
			const files = await fs.readdir(folderpath);
			return files;
		} catch (err) {
			return null;
		}
	}
}

function registerContainerFunctions() {
	/**
	 * Registers a handler that will be called after the Window is fully closed, but prior to exiting the player.
	 * The handlers are required to return a Promise that resolves when the handler has finished executing and is ready for the player to exit.
	 * Each handler will be called in the order they were registered and the next handler will not be called until the current handler has resolved.
	 * @param {function} handler The handler function that will be called prior to exiting once the Window is fully closed. Must return a Promise that resolves when the handler has finished.
	 */
	ContainerMessenger.bus.on('registerExitHandler', handler => {
		if (typeof handler !== 'function') {
			throw new Error('Given handler is not a function: ' + handler);
		}
		exitHandlers.push(handler);
	});
}

async function loadSerialNumber() {
	try {
		// This is the desktop directory on Windows and Linux
		const serialPath = path.join(paths.DESKTOP, 'serial.json');
		const serial = await FileIO.read(serialPath);
		if (serial == null) {
			throw new Error('null');
		}
		return JSON.parse(serial);
	} catch (err) {
		if (err.name == 'SyntaxError') {
			// Might be corrupted, try overwriting with defaults
			saveSerialNumber({ serialnumber: '', serialnumbermodifiedtime: 0 });
			return null;
		} else {
			// Try the config directory for Linux 32-bit
			const serialPath = path.join(paths.CONFIG, 'serial.json');
			let serial = await FileIO.read(serialPath);
			if (serial == null) {
				saveSerialNumber({ serialnumber: '', serialnumbermodifiedtime: 0 });
				serial = await FileIO.read(serialPath);
			}
			try {
				return JSON.parse(serial);
			} catch (err) {
				saveSerialNumber({ serialnumber: '', serialnumbermodifiedtime: 0 });
				return null;
			}
		}
	}
}

async function saveSerialNumber(serial) {
	try {
		// This is the desktop directory on Windows and Linux
		await FileUtils.save('serial.json', '', 'DESKTOP', JSON.stringify(serial));
	} catch (err) {
		if (err.code == 'EACCES') {
			// This is probably a Linux 32-bit, fallback to the config directory
			try {
				await FileUtils.save('serial.json', '', 'CONFIG', JSON.stringify(serial));
			} catch (err) {
				console.log(err);
			}
		} else {
			console.log(err);
		}
	}
}

function closeWindow() {
	if (win !== null) {
		win.close();
	}
}

function uploadScreenshot(filepath) {
	try {
		const formData = {
			playerkey: registration.key,
			file: fs.createReadStream(filepath)
		};
		NetworkIO.post(`https://${registration.subdomain}.breezedigitalsignage.com/rest/uploadscreenshot.php`, formData).catch(e => { throw e; });
	} catch (err) {
		console.log(err);
	}
}

/**
 * @param {!number} thumbnailWidth
 * @param {!number} thumbnailHeight
 * @param {!string} filepath
 * @returns {!Promise}
 */
async function uploadThumbnail(usableWidth, usableHeight, thumbnailWidth, thumbnailHeight, filepath) {
	try {
		const scale = Math.min(1, thumbnailWidth / usableWidth);
		const screenshot = await Jimp.read(filepath);
		const thumbnail = await new Promise(resolve => new Jimp(thumbnailWidth, thumbnailHeight, (err, image) => resolve(image)));
		const newWidth = Math.round(screenshot.bitmap.width * scale);
		const newHeight = Math.round(screenshot.bitmap.height * scale);
		await new Promise(resolve => thumbnail.scan(0, 0, thumbnail.bitmap.width, thumbnail.bitmap.height, function(x, y, idx)
		{
			this.bitmap.data[idx + 0] = 0;
			this.bitmap.data[idx + 1] = 0;
			this.bitmap.data[idx + 2] = 0;
			this.bitmap.data[idx + 3] = 255;
		}).blit(screenshot.resize(newWidth, newHeight), 0, 0, 0, 0, Math.min(thumbnailWidth, newWidth), Math.min(thumbnailHeight, newHeight)).quality(60).write(filepath, resolve));
		const formData = {
			playerkey: registration.key,
			file: fs.createReadStream(filepath)
		};
		return NetworkIO.post(`https://${registration.subdomain}.breezedigitalsignage.com/rest/uploadthumbnailscreenshot.php`, formData);
	} catch (err) {
		console.log(err);
	}
}

function requestSoftReboot() {
	closeWindow();
}

async function checkForUpdates() {
	try {
		await Updater.checkForUpdates(registration.subdomain);
	} catch (err) {
		console.log(err);
	}
}

/**
 * Clean up all video streams and run any exit handlers prior to fully closing.
 */
function registerCloseHandler() {
	win.on('close', async () => {
		try {
			win.hide();
			for (let itr = 0; itr < exitHandlers.length; ++itr) {
				try {
					await exitHandlers[itr]();
				} catch (err) {
					console.log(`An exit handler rejected for reason: ${err.msg}.  Ignoring and continuing.`);
				}
			}
			win.close(true);
		} catch (err) {
			console.log(`Clean up code threw an err: ${err}, exiting anyway..`);
			win.close(true);
		}
	});
}

function registerEscHotkey() {
	escShortcut = new nw.Shortcut({
		key: 'Escape',
		active: function() {
			closeWindow();
		},
		failed: function(msg) {
			console.log('Unable to register Esc hotkey:');
			console.log(msg);
		}
	});
	nw.App.registerGlobalHotKey(escShortcut);
	
	ContainerMessenger.bus.emit('registerExitHandler', () => {
		return async () => {
			if (escShortcut !== null) {
				nw.App.unregisterGlobalHotKey(escShortcut);
				escShortcut = null;
			}
		};
	});
}

async function showRegister() {
	clearInterval(updateIntervalId);
	win.leaveKioskMode();
	const newRegistration = await generateRegistrationPage();
	await FileUtils.save('registration.json', '', 'CONFIG', JSON.stringify(newRegistration));
	const registrationPath = path.join(paths.CONFIG, 'registration.json');
	registration = await FileIO.read(registrationPath);
	if (registration !== null) {
		registration = JSON.parse(registration);
	}
	showPlayer();
}

async function showPlayer() {
	document.body.style.backgroundColor = 'black';
	document.body.innerHTML = '';
	clearInterval(updateIntervalId);
	updateIntervalId = setInterval(checkForUpdates, 1000 * 60 * 60);
	setTimeout(checkForUpdates, 1000 * 5);
	
	chrome.webRequest.onBeforeSendHeaders.addListener(function(details) {
		// On some embeddded videos if there's no referer the API says "Video unavailable".
		for (let index in details.requestHeaders) {
			if (details.requestHeaders[index].name === "Referer") {
				details.requestHeaders[index].value = 'https://example.com';
				return { requestHeaders: details.requestHeaders };
			}
		}
		details.requestHeaders.push({ name: "Referer", value: 'https://example.com' });
		return {
			requestHeaders: details.requestHeaders
		};
	}, { urls: ["<all_urls>"] }, ["blocking", "requestHeaders", "extraHeaders"]);

	chrome.debugger.getTargets(targets => {
		console.warn(targets);
		const found = targets.find(target => target.title == "breeze-2-player");
		const debuggee = {
			targetId: found.id
		};
		chrome.debugger.attach(debuggee, '1.1', () =>{
			chrome.debugger.sendCommand(debuggee, 'Emulation.setTouchEmulationEnabled', {
				enabled: true,
				configuration: 'mobile'
				}, () => {
			});
		});
	});
	
	try {
		Messenger = new MessengerPlayer(window);
		let player = new Player(registration.key);
		player.ready(() => player.start());
	} catch (err) {
		console.log(err.stack || err);
	}
	win.enterKioskMode();
}
