class FileIO {
	/**
	 * @param {!string} filepath
	 * @param {(!string|!Object)} [options]
	 * @returns {!Promise<?string>}
	 */
	static async read(filepath, options = 'utf8') {
		let data = null;
		try {
			data = await fs.readFile(filepath, options);
		} catch (err) {
			if (err.code != 'ENOENT' && err.code != 'EACCES') {
				throw err;
			}
		}
		return data;
	}

	/**
	 * @param {!string} filepath
	 * @returns {!Promise<?fs.Stats>}
	 */
	static async stat(filepath) {
		let stats = null;
		try {
			stats = await fs.stat(filepath);
		} catch (err) {
			if (err.code != 'ENOENT') {
				throw err;
			}
		}
		return stats;
	}

	/**
	 * @param {!string} dir
	 * @returns {!Promise<!string[]>}
	 */
	static async enumerateDirs(dir) {
		const dirs = [];
		const files = await fs.readdir(dir);
		for (const file of files) {
			const filepath = path.join(dir, file);
			const stats = await FileIO.stat(filepath);
			if (stats !== null && stats.isDirectory()) {
				dirs.push(file);
			}
		}
		return dirs;
	}

	/**
	 * Returns a list of the files deleted.
	 * @param {!string} dir
	 * @param {!RegExp} regex
	 * @returns {!Promise<!string[]>}
	 */
	static async deleteFilesMatching(dir, regex) {
		const deletedFiles = [];
		const files = await fs.readdir(dir);
		for (const file of files) {
			if (regex.test(file)) {
				deletedFiles.push(file);
				const filepath = path.join(dir, file);
				await fs.remove(filepath);
			}
		}
		return deletedFiles;
	}

	/**
	 * Returns a list of the files deleted.
	 * @param {!string} dir
	 * @param {!string} ext
	 * @returns {!Promise<!string[]>}
	 */
	static deleteFilesByExt(dir, ext) {
		const pattern = new RegExp(`.${ext}$`);
		return FileIO.deleteFilesMatching(dir, pattern);
	}
}