class AlertFeed {
	/**
	 * @param {!string} feed
	 * @param {!string} type 'BULLETIN' or 'CRAWL'
	 * @param {!number} [headerFontSize]
	 * @param {!number} [titleFontSize]
	 * @param {!number} [descriptionFontSize]
	 * @param {!string} [alertColor]
	 * @param {!string} [alertStripingColor]
	 * @param {!string} [allClearColor]
	 * @param {!string} [allClearStripingColor]
	 * @param {!number} [crawlFontSize]
	 * @param {!string} [crawlForeground]
	 * @param {!string} [crawlBackground]
	 */
	constructor(
		feed,
		type = 'BULLETIN',
		allClearDwell = 15,
		headerFontSize = 32,
		titleFontSize = 64,
		descriptionFontSize = 48,
		alertColor = '#ff0000',
		alertStripingColor = '#ffffff',
		allClearColor = '#008000',
		allClearStripingColor = '#ffffff',
		crawlFontSize = 32,
		crawlForeground = '#ffffff',
		crawlBackground = '#000000') {

		this.feed = feed;
		this.type = type;
		this.allClearDwell = allClearDwell;
		this.headerFontSize = headerFontSize;
		this.titleFontSize = titleFontSize;
		this.descriptionFontSize = descriptionFontSize;
		this.alertColor = alertColor;
		this.alertStripingColor = alertStripingColor;
		this.allClearColor = allClearColor;
		this.allClearStripingColor = allClearStripingColor;
		this.crawlFontSize = crawlFontSize;
		this.crawlForeground = crawlForeground;
		this.crawlBackground = crawlBackground;

		if (this.type !== 'BULLETIN' && this.type !== 'CRAWL') {
			throw new Error(`Invalid type '${this.type}', must be 'BULLETIN' or 'CRAWL'.`);
		}

		/** @type {?string} */
		this.sha1 = null;
	}

	/**
	 * @param {!string} field
	 */
	format(field) {
		if (field === 'type') {
			if (this.type === 'BULLETIN') {
				return 'Full Screen';
			} else if (this.type === 'CRAWL') {
				return 'Crawl';
			}
		}
	}

	/**
	 * Determines if player playback should continue running while Alerts from this feed are active.
	 * @returns {!boolean}
	 */
	shouldPlaybackContinue() {
		return this.type !== 'BULLETIN';
	}

	/**
	 * Retrieves a JSON representation of the AlertFeed as the server would process it.
	 * (Primarily this means the fields are all fully lower cased.)
	 * @returns {!Object}
	 */
	toServerJSON() {
		const json = this.toJSON();
		const serverJson = {};
		Object.keys(json).forEach(key => 
			serverJson[key.toLowerCase()] = json[key]
		);
		return serverJson;
	}

	/**
	 * Checks if the AlertFeed contains only valid values.
	 * @returns {{valid: !boolean, msg: !string, field: !string}}
	 */
	validate() {
		const json = this.toJSON();
		const keys = Object.keys(json);
		for (const key of keys) {
			if (!AlertFeed.validate(json[key], key)) {
				return { valid: false, msg: `'${json[key]}' is an invalid ${AlertFeed.format(key)}.`, field: key };
			}
		}
		return { valid: true, msg: '', field: '' };
	}

	/**
	 * @param {?AlertFeed} alertFeed
	 * @returns {!boolean}
	 */
	equals(alertFeed) {
		if (alertFeed === null || alertFeed === undefined) {
			return false;
		}
		const a = this.toJSON();
		const b = alertFeed.toJSON();
		return Object.keys(a).every(key => a[key] === b[key]);
	}

	/**
	 * Defines the behavior of JSON.stringify requests.
	 * @returns {!Object}
	 */
	toJSON() {
		return {
			feed: this.feed,
			type: this.type,
			allClearDwell: this.allClearDwell,
			headerFontSize: this.headerFontSize,
			titleFontSize: this.titleFontSize,
			descriptionFontSize: this.descriptionFontSize,
			alertColor: this.alertColor,
			alertStripingColor: this.alertStripingColor,
			allClearColor: this.allClearColor,
			allClearStripingColor: this.allClearStripingColor,
			crawlFontSize: this.crawlFontSize,
			crawlForeground: this.crawlForeground,
			crawlBackground: this.crawlBackground
		};
	}

	/**
	 * @returns {!Promise}
	 */
	async hash() {
		this.sha1 = await Utility.hash(JSON.stringify(this), 'SHA-1');
	}

	/**
	 * @param {!string} field
	 * @returns {!string}
	 */
	static format(field) {
		const formatted = {
			feed: 'Feed URL',
			allClearDwell: 'All Clear Dwell',
			headerFontSize: 'Header Font Size',
			titleFontSize: 'Title Font Size',
			descriptionFontSize: 'Description Font Size',
			crawlFontSize: 'Crawl Font Size',
			alertColor: 'Alert Color',
			alertStripingColor: 'Alert Striping Color',
			allClearColor: 'All Clear Color',
			allClearStripingColor: 'All Clear Striping Color',
			crawlForeground: 'Crawl Foreground',
			crawlBackground: 'Crawl Background'
		};
		return formatted[field];
	}

	/**
	 * Checks if the given value is valid for the given field.
	 * @param {!(!string|!number)} value
	 * @param {!string} field
	 * @returns {!boolean}
	 */
	static validate(value, field) {
		const schema = {
			feed: 'url',
			allClearDwell: 'positiveInteger',
			headerFontSize: 'positiveInteger',
			titleFontSize: 'positiveInteger',
			descriptionFontSize: 'positiveInteger',
			crawlFontSize: 'positiveInteger',
			alertColor: 'color',
			alertStripingColor: 'color',
			allClearColor: 'color',
			allClearStripingColor: 'color',
			crawlForeground: 'color',
			crawlBackground: 'color'
		};
		switch (schema[field]) {
		case 'url':
			return AlertFeed.validateUrl(value);
		case 'positiveInteger':
			return AlertFeed.validatePositiveInt(value);
		case 'color':
			return AlertFeed.validateColor(value);
		default:
			return true;
		}
	}

	/**
	 * @param {!string} color
	 * @returns {!boolean}
	 */
	static validateColor(color) {
		const el = document.createElement('div');
		el.style.background = color;
		return el.style.background.length > 0;
	}

	/**
	 * @param {!(!number|!string)} n
	 * @returns {!boolean}
	 */
	static validatePositiveInt(n) {
		const number = Number(n);
		return number > 0 && number == n;
	}

	/**
	 * @param {!string} url
	 * @returns {!boolean}
	 */
	static validateUrl(url) {
		try {
			new URL(url);
			return true;
		} catch (err) {
			return false;
		}
	}

	/**
	 * If an error is thrown while creating the AlertFeed, the error will be logged and
	 * null will be returned.
	 * @param {!string} feed
	 * @param {!string} type 'BULLETIN' or 'CRAWL'
	 * @returns {!Promise<?AlertFeed>}
	 */
	static async create(feed, type) {
		let alertFeed = null;
		try {
			alertFeed = new AlertFeed(feed, type);
			await alertFeed.hash();
		} catch (err) {
			console.log(`Returning null AlertFeed due to error thrown while creating AlertFeed: ${err.stack || err}`);
		}
		return alertFeed;
	}

	/**
	 * @param {?Object[]} alertFeeds
	 * @return {!AlertFeed[]}
	 */
	static parse(alertFeeds = null) {
		if (alertFeeds === null) {
			return [];
		}

		if (!Array.isArray(alertFeeds)) {
			// If alertFeeds is NOT an array, then it is version 1 data that must be converted.
			if (alertFeeds.alertfeeds === null || alertFeeds.alertfeeds === undefined) {
				return [];
			} else {
				return alertFeeds.alertfeeds.map(url => new AlertFeed(
					url,
					'BULLETIN', // Version 1 feeds are always BULLETINs
					Number(alertFeeds.allcleardwell),
					Number(alertFeeds.headerfontsize),
					Number(alertFeeds.titlefontsize),
					Number(alertFeeds.descriptionfontsize),
					alertFeeds.alertcolor,
					alertFeeds.alertstripingcolor,
					alertFeeds.allclearcolor,
					alertFeeds.allclearstripingcolor
				));
			}
		} else {
			return alertFeeds.map(feed => new AlertFeed(
				feed.feed,
				feed.type,
				Number(feed.allcleardwell),
				Number(feed.headerfontsize),
				Number(feed.titlefontsize),
				Number(feed.descriptionfontsize),
				feed.alertcolor,
				feed.alertstripingcolor,
				feed.allclearcolor,
				feed.allclearstripingcolor,
				Number(feed.crawlfontsize),
				feed.crawlforeground,
				feed.crawlbackground
			));
		}
	}
}
