const AlertViewer = (() => {
	/***
	 * Dependencies:
	 * 	BreezeCrawl.js
	 *  DOM.js
	 *  moment.js
	 *  moment-timezone.js
	 */

	/**
	 * @param {!string} rules
	 * @returns {!Element}
	 */
	const insertStyleSheet = rules => {
		return document.head.appendChild($.style({ html: rules }));
	};

	/**
	 * @param {?Element} style
	 */
	const deleteStyleSheet = (style) => {
		if (style !== null) {
			style.disabled = true;
			style.remove();
		}
	};

	class AlertViewer {
		/**
		 * @param {!CAPS.Alert} alert
		 * @param {!CAPS.Filter} filter
		 * @param {!Size} size
		 * @param {!HTMLElement} parent
		 * @param {?string} [timezone]
		 */
		constructor(alert, filter, size, parent, timezone = null) {
			const id = `alert-viewer-${++AlertViewer.uid}`;
			const selector = `#${id}`;
			const rules = AlertViewer.generateStyles(selector, filter);
			this.styles = insertStyleSheet(rules);
			this.displayMode = filter.mode;
			this.tag = $.div({ id, class: 'alert-viewer' },
				$.header({},
					$.span({ class: 'date' }),
					$.span({ class: 'time' })
				),
				$.main({},
					$.h1({ class: 'headline' }),
					$.p({ class: 'description' }),
					$.div({ class: 'crawl-box' },
						$.div({ class: 'crawl' })
					)
				),
				$.footer()
			);

			/** @type {?BreezeCrawl} */
			this.crawl = null;

			this.date = AlertViewer.date(alert, timezone);
			this.time = AlertViewer.time(alert, timezone);
			this.headline = AlertViewer.headline(alert);
			this.description = AlertViewer.description(alert);
			this.tag.querySelector(':scope > header > span.date').innerText = this.date;
			this.tag.querySelector(':scope > header > span.time').innerText = this.time;
			this.tag.querySelector(':scope > main > h1.headline').innerText = this.headline;
			this.tag.querySelector(':scope > main > p.description').innerText = this.description;
			if (this.displayMode === 'Full Screen') {
				this.tag.querySelector(':scope > main > div.crawl-box').style.visibility = 'hidden';
			} else if (this.displayMode === 'Crawl') {
				this.tag.style.background = 'rgba(0,0,0,0)';
				this.tag.querySelector(':scope > header').style.visibility = 'hidden';
				this.tag.querySelector(':scope > main > h1.headline').style.visibility = 'hidden';
				this.tag.querySelector(':scope > main > p.description').style.visibility = 'hidden';
				this.tag.querySelector(':scope > footer').style.visibility = 'hidden';
			}
			// Must be on the DOM to begin crawling
			parent.appendChild(this.tag);
			if (this.displayMode === 'Both' || this.displayMode === 'Crawl') {
				this.startCrawl(this.tag.querySelector(':scope > main > div.crawl-box > div.crawl'), filter, size);
			}
		}

		/**
		 * @returns {!boolean}
		 */
		isFullscreen() {
			return this.displayMode === 'Full Screen' || this.displayMode === 'Both';
		}

		/**
		 * @param {!HTMLElement} tag
		 * @param {!CAPS.Filter} filter
		 * @param {!Size} size
		 */
		startCrawl(tag, filter, size) {
			const lineHeight = filter.crawl.size * 1.5;
			tag.classList.add('widget-crawl');
			this.crawl = new BreezeCrawl(tag);
			this.crawl.foreground = filter.crawl.color;
			this.crawl.startCrawl(
				size.width,
				lineHeight,
				`${this.date} ${this.time} - ${this.headline} - ${this.description}`.replace(/\r?\n?/g, ''),
				filter.crawl.size,
				filter.crawl.family,
				true
			);
		}

		remove() {
			if (this.crawl !== null) {
				this.crawl.destroy();
				this.crawl = null;
			}
			if (this.tag !== null) {
				this.tag.remove();
				this.tag = null;
			}
			deleteStyleSheet(this.styles);
			this.styles = null;
		}

		/**
		 * @param {!string} selector
		 * @param {!CAPS.Filter} filter
		 * @returns {!string}
		 */
		static generateStyles(selector, filter) {
			return DOM.parseHTML`
				${selector} {
					margin: 0;
					padding: 0;
					box-sizing: border-box;
					display: flex;
					flex-flow: column nowrap;
					width: 100%;
					height: 100%;
					background: ${filter.background};
					text-shadow: #000 0 0.1em 0.2em,
								 #000 0 0.1em 0.2em,
								 #000 0 0.1em 0.2em,
								 #000 0 0.1em 0.3em,
								 #000 0 0.1em 0.4em,
								 #000 0 0.1em 0.4em,
								 #000 0 0.1em 0.4em;
					overflow: hidden;
				}
				${selector} > main {
					flex: 1 1 auto;
					display: flex;
					flex-flow: column nowrap;
					justify-content: space-around;
					align-items: center;
				}
				${selector} > main > h1.headline {
					font-family: ${filter.headline.family};
					color: ${filter.headline.color};
					background-color: ${filter.headline.background};
					font-size: ${filter.headline.size}px;
					flex: 0 0 auto;
					padding: 2px 10px;
					box-sizing: border-box;
					max-width: 100%;
					overflow: hidden;
					text-overflow: ellipsis;
					white-space: nowrap;
					margin-bottom: 0;
				}
				${selector} > main > p.description {
					font-family: ${filter.description.family};
					color: ${filter.description.color};
					background-color: ${filter.description.background};
					font-size: ${filter.description.size}px;
					flex: 0 1 auto;
					padding: 2px 10px;
					box-sizing: border-box;
					overflow: hidden;
					margin: 0 2em 1m 2em;
					text-align: center;
				}
				${selector} > main > div.crawl-box {
					position: relative;
					width: 100%;
					box-sizing: border-box;
					margin: 0 0 30px 0;
					height: calc(${filter.crawl.size}px * 1.5);
					line-height: calc(${filter.crawl.size}px * 1.5);
				}
				${selector} > main > div.crawl-box > div.crawl {
					position: absolute;
					width: 100%;
					height: calc(${filter.crawl.size}px * 1.5);
					line-height: calc(${filter.crawl.size}px * 1.5);
					font-size: ${filter.crawl.size}px;
					text-shadow: none;
					font-family: ${filter.crawl.family};
					color: ${filter.crawl.color};
					background-color: ${filter.crawl.background};
					padding: 2px 10px;
					box-sizing: border-box;
					overflow: hidden;
					white-space: nowrap;
				}
				${selector} > header {
					display: flex;
					flex-flow: row nowrap;
					justify-content: space-between;
					padding: 20px;
					box-sizing: border-box;
				}
				${selector} > header > span.date,
				${selector} > header > span.time {
					font-family: ${filter.dateTime.family};
					color: ${filter.dateTime.color};
					font-size: ${filter.dateTime.size}px;
					background-color: ${filter.dateTime.background};
					padding: 2px 10px;
					box-sizing: border-box;
					font-style: italic;
				}
				${selector} > header,
				${selector} > footer {
					background: repeating-linear-gradient(
						-45deg,
						${filter.background},
						${filter.background} 40px,
						${filter.stripes} 40px,
						${filter.stripes} 80px
					);
					width: 100%;
					height: 16%;
					flex: 0 0 auto;
					box-shadow: 0px 0px 18px 3px rgba(0,0,0,0.65);
				}
			`;
		}

		/**
		 * @param {!CAPS.Alert} alert
		 * @param {?string} [timezone]
		 * @returns {!string}
		 */
		static date(alert, timezone = null) {
			const sent = moment.utc(alert.getValue('sent'), 'YYYY-MM-DDTHH:mm:ssZZ').local();
			const localized = timezone === null ? sent.local() : sent.tz(timezone);
			return localized.format('YYYY-MM-DD');
		}

		/**
		 * @param {!CAPS.Alert} alert
		 * @param {?string} [timezone]
		 * @returns {!string}
		 */
		static time(alert, timezone = null) {
			const sent = moment.utc(alert.getValue('sent'), 'YYYY-MM-DDTHH:mm:ssZZ');
			const localized = timezone === null ? sent.local() : sent.tz(timezone);
			return localized.format('hh:mm:ssa');
		}

		/**
		 * @param {!CAPS.Alert} alert
		 * @returns {!string}
		 */
		static headline(alert) {
			const [ headline ] = alert.getValue('headline').filter(headline => headline.length > 0);
			return headline !== undefined ? headline : '';
		}

		/**
		 * @param {!CAPS.Alert} alert
		 * @returns {!string}
		 */
		static description(alert) {
			const [ description ] = alert.getValue('description').filter(description => description.length > 0);
			return description !== undefined ? description : '';
		}
	}
	AlertViewer.uid = 0;

	return AlertViewer;
})();
