class AlertViewerControl {
	/**
	 * @param {!HTMLElement} tag
	 * @param {!(!EASSettings|!EASSettings[]|!AlertFeed[])} [feeds]
	 */
	constructor(tag, feeds = []) {
		this.tag = tag;
		this.feedId = 0;

		/** @type {?HTMLElement[]} */
		this.$entries = null;

		/**
		 * @type {?Minibus}
		 */
		this.bus = Minibus.create();
		this.bus.on('AlertFeedDeleted', this.onAlertFeedDeleted.bind(this));
		this.bus.on('AlertFeedAdded', this.onAlertFeedAdded.bind(this));

		/** @type {?AlertFeedControl[]} */
		this.alertFeedControls = null;

		if (Array.isArray(feeds) && feeds.length > 0) {
			if (feeds[0] instanceof AlertFeed) {
				this.alertFeedControls = feeds.map(af => this.createAlertFeedControl(af));
			}
		}
		if (this.alertFeedControls === null) {
			this.alertFeedControls = AlertFeed.parse(feeds)
				.map(af => this.createAlertFeedControl(af));
		}
	}

	/**
	 * @listens AlertFeedDeleted
	 * @param {!number} feedId
	 */
	onAlertFeedDeleted(feedId) {
		this.alertFeedControls.forEach(afc => afc.update());
		const index = this.alertFeedControls.findIndex(afc => afc.id === feedId);
		if (index !== -1) {
			const [deleted] = this.alertFeedControls.splice(index, 1);
			deleted.remove();
		}
	}

	/**
	 * @listens AddAlertFeed
	 */
	onAlertFeedAdded() {
		const afc = this.createAlertFeedControl();
		this.$entries.querySelector('.formgrid:last-child').scrollIntoView();
		this.alertFeedControls.push(afc);
	}

	/**
	 * @param {!AlertFeed} [alertFeed]
	 * @returns {!AlertFeedControl}
	 */
	createAlertFeedControl(alertFeed = new AlertFeed()) {
		return new AlertFeedControl(this.$entries, this.feedId++, alertFeed, this.bus);
	}

	/**
	 * Validates all of the contained AlertFeed values are OK.
	 * @returns {{valid: !boolean, msg: !string, field: !string}}
	 */
	validate() {
		this.alertFeedControls.forEach(afc => afc.update());
		for (const afc of this.alertFeedControls) {
			const validation = afc.validate();
			if (!validation.valid) {
				afc.setError(validation.field.toLowerCase());
				return validation;
			}
		}
		return { valid: true, msg: '' };
	}

	/**
	 * Returns an array of objects in the form the server expects representing each alert feed.
	 * @returns {!EASSettings[]} feeds
	 */
	toServerJSON() {
		return this.alertFeedControls.map(afc => afc.alertFeed.toServerJSON());
	}

	/**
	 * Renders the control into the tag, overwriting existing markup.
	 */
	render() {
		this.tag.innerHTML = '';
		this.tag.append(
			$.ul({ class: 'iconlist' },
				$.li({ class: 'tooltip-button mdi mdi-plus-circle-outline eas-alertfeedsadd', title: 'Add a feed', click: () => this.bus.emit('AlertFeedAdded') })
			),
			this.$entries = $.div({ class: 'eas-entries' })
		);
		
		this.alertFeedControls.forEach(afc => this.$entries.appendChild(afc.render()));
	}

	/**
	 * Removes the control from the DOM.
	 */
	remove() {
		this.tag.innerHTML = '';
	}
}

/**
 * Provides interactivity for a single AlertFeed control.
 */
class AlertFeedControl {
	/**
	 * @param {!number} id
	 * @param {!AlertFeed} alertFeed
	 * @param {!Minibus} bus
	 */
	constructor(parent, id, alertFeed, bus) {
		this.id = id;
		this.alertFeed = alertFeed;
		this.bus = bus;
		this.controls = new Map();
		this.errors = false;

		const color = {
			config: {
				showPalette: true,
				localStorageKey: 'globalcolor',
				opacity: true
			}
		};
		const text = {
			type: 'text'
		};
		const numerical = {
			type: 'number',
			min: 0,
			step: 1,
			class: 'sampleInput'
		};

		const notify = this.notify.bind(this);
		this.toolbar = $.ul({ class: 'iconlist formgrid-row' },
			$.li({ class: 'mdi mdi-trash-can-outline tooltip-button eas-delete', title: 'Delete' })
		);
		
		const $deleteButton = this.toolbar.querySelector('.eas-delete');
		$deleteButton.addEventListener('click', () => {
			this.bus.emit('AlertFeedDeleted', this.id);
		});
		
		this.controls.set('type', {
			label: this.listboxLabel('Type', 'Full screen alerts take up the entire player screen hiding the scheduled playlist to bring full attention to the emergency alert, whereas crawl alerts take up a smaller portion of the screen and overlay only a part of the scheduled playlist.'),
			input: $.select({ class: 'sampleSelect', change: notify },
				$.option({ value: 'BULLETIN', text: 'Full Screen Alert' }),
				$.option({ value: 'CRAWL', text: 'Crawl Only Alert' })
			)
		});
		
		this.controls.set('url', {
			label: this.listboxLabel('Feed URL', 'An Emergency Alerts feed to monitor.'),
			input: this.listboxTemplate(text, 'eas-url')
		});
		
		this.controls.set('allcleardwell', {
			label: this.listboxLabel('All Clear Dwell', 'The number of minutes to display All clear after an All Clear is published.'),
			input: this.listboxTemplate(numerical, 'eas-allcleardwell')
		});
		
		this.controls.set('headerfontsize', {
			label: this.listboxLabel('Header Font Size', 'The display fontsize of the date and time on the Players during EAS.'),
			input: this.listboxTemplate(numerical, 'eas-headerfontsize')
		});
		
		this.controls.set('titlefontsize', {
			label: this.listboxLabel('Title Font Size', 'The display fontsize of the alert title on the Players during EAS.'),
			input: this.listboxTemplate(numerical, 'eas-titlefontsize')
		});
		
		this.controls.set('descriptionfontsize', {
			label: this.listboxLabel('Description Font Size', 'The display fontsize of the alert description on the Players during EAS.'),
			input: this.listboxTemplate(numerical, 'eas-descriptionfontsize')
		});
		
		this.controls.set('alertcolor', {
			label: this.listboxLabel('Alert Color', 'The background color of the alert.'),
			input: this.listboxTemplate(color, 'eas-alertcolor')
		});
		
		this.controls.set('alertstripingcolor', {
			label: this.listboxLabel('Alert Striping Color', 'The color of the stripes on the alert header and footers (set this to the same color as the Alert Color to hide the stripes.)'),
			input: this.listboxTemplate(color, 'eas-alertstripingcolor')
		});
		
		this.controls.set('allclearcolor', {
			label: this.listboxLabel('All Clear Color', 'The background color of the all clear.'),
			input: this.listboxTemplate(color, 'eas-allclearcolor')
		});
		
		this.controls.set('allclearstripingcolor', {
			label: this.listboxLabel('All Clear Striping Color', 'The color of the stripes on the allclear header and footers (set this to the same color as the AllClear Color to hide the stripes.)'),
			input: this.listboxTemplate(color, 'eas-allclearstripingcolor')
		});
		
		this.controls.set('crawlfontsize', {
			label: this.listboxLabel('Crawl Font Size', 'The display fontsize of the crawl scrolling text on the Players during EAS.'),
			input: this.listboxTemplate(numerical, 'eas-crawlfontsize')
		});
		
		this.controls.set('crawlforeground', {
			label: this.listboxLabel('Text Color', 'The color of the scrolling text showing the alert.'),
			input: this.listboxTemplate(color, 'eas-crawlforeground')
		});
		
		this.controls.set('crawlbackground', {
			label: this.listboxLabel('Background Color', 'The color of the background behind the scrolling text.'),
			input: this.listboxTemplate(color, 'eas-crawlbackground')
		});
		
		this.container = parent.appendChild($.div({ class: 'formgrid', css: {
			backgroundColor: 'var(--subheader-color)',
			border: '1px solid var(--input-border-color)',
			marginBottom: '5px',
			borderRadius: '4px'
		} }, this.toolbar));
		
		for (const control of ['type', 'url', 'allcleardwell', 'headerfontsize', 'titlefontsize', 'descriptionfontsize', 'alertcolor', 'alertstripingcolor', 'allclearcolor', 'allclearstripingcolor', 'crawlfontsize', 'crawlforeground', 'crawlbackground']) {
			this.container.append(
				this.controls.get(control).label,
				this.controls.get(control).input
			);
		}
		
		this.render();
		
		this.views = new Map();
		this.views.set('BasePre', ['url', 'allcleardwell']);
		this.views.set('BasePost', []);
		this.views.set('FullScreen', ['type', 'headerfontsize', 'titlefontsize', 'descriptionfontsize', 'alertcolor', 'alertstripingcolor', 'allclearcolor', 'allclearstripingcolor']);
		this.views.set('Crawl', ['type', 'crawlfontsize', 'crawlforeground', 'crawlbackground']);
		this.selectView();
	}
	
	/**
	 * Returns the primary input field for the control.
	 * @param {!string} control
	 * @returns {!HTMLElement}
	 */
	getInput(control) {
		return this.controls.get(control).input;
	}
	
	/**
	 * @param {!string} control
	 * @returns {?Object}
	 */
	getVal(control) {
		return this.getInput(control).value;
	}
	
	/**
	 * @param {!string} control
	 * @param {?Object} [value]
	 * @param {!string} [type]
	 * @returns {?Object}
	 */
	setVal(control, value = '', type = '') {
		const input = this.getInput(control);
		value = DOM.escape(value);
		input.value = value;
	}

	/**
	 * Fires an event indicating the control was modified.
	 */
	notify() {
		const type = this.alertFeed.type;
		this.update();
		if (type !== this.alertFeed.type) {
			this.selectView();
		}
		this.bus.emit('AlertFeedModified');
	}

	/**
	 * Adds attention grabbing red error styling to the control and scrolls it into view.
	 * @param {!string} control
	 */
	setError(control) {
		this.errors = true;
		if (control === 'feed') {
			control = 'url';
		}
		const el = this.controls.get(control);
		Object.assign(el.querySelector('.eas-label').style, {
			color: 'red'
		});
		Object.assign(el.querySelector('.sampleInput').style, {
			boxShadow: '0 0 0 1px #ff0000',
			zIndex: 1
		});
		el.scrollIntoView();
	}

	/**
	 * Removes all error styling from all controls.
	 */
	clearErrors() {
		if (this.errors) {
			for (const $e of this.container.querySelectorAll('.eas-label')) {
				Object.assign($e.style, {
					color: ''
				});
			}
			for (const $e of this.container.querySelectorAll('.sampleInput')) {
				Object.assign($e.style, {
					boxShadow: '',
					zIndex: 0
				});
			}
			this.errors = false;
		}
	}

	/**
	 * Update the values from the DOM into the object.
	 */
	update() {
		this.clearErrors();
		this.alertFeed.type = this.getVal('type');
		this.alertFeed.feed = this.getVal('url');
		this.alertFeed.allClearDwell = Number(this.getVal('allcleardwell'));
		this.alertFeed.headerFontSize = Number(this.getVal('headerfontsize'));
		this.alertFeed.titleFontSize = Number(this.getVal('titlefontsize'));
		this.alertFeed.descriptionFontSize = Number(this.getVal('descriptionfontsize'));
		this.alertFeed.alertColor = this.getVal('alertcolor');
		this.alertFeed.alertStripingColor = this.getVal('alertstripingcolor');
		this.alertFeed.allClearColor = this.getVal('allclearcolor');
		this.alertFeed.allClearStripingColor = this.getVal('allclearstripingcolor');
		this.alertFeed.crawlFontSize = Number(this.getVal('crawlfontsize'));
		this.alertFeed.crawlForeground = this.getVal('crawlforeground');
		this.alertFeed.crawlBackground = this.getVal('crawlbackground');
	}

	/**
	 * Switches to the correct view for the current alert feed type.
	 */
	selectView() {
		if (this.alertFeed.type === 'BULLETIN') {
			this.showView('FullScreen');
		} else if (this.alertFeed.type === 'CRAWL') {
			this.showView('Crawl');
		}
	}

	/**
	 * Shows all of the controls associated with the given view.
	 * @param {!string} view Views map key e.g. 'FullScreen'
	 */
	showView(view) {
		this.hideAll();
		const mapViewControls = view => this.views.get(view).map(
			id => this.controls.get(id)
		);
		const show = control => {
			if (control.hasOwnProperty('input')) {
				toggleRow(control.input[0], true);
			}
		};
		const foo = mapViewControls('BasePre');
		mapViewControls('BasePre').forEach(show);
		mapViewControls(view).forEach(show);
		mapViewControls('BasePost').forEach(show);
	}

	/**
	 * Hides all of the controls in the view.
	 */
	hideAll() {
		this.controls.forEach(control => {
			if (control.hasOwnProperty('input')) {
				toggleRow(control.input[0], false);
			}
		});
	}

	/**
	 * Validates the AlertFeed values are all OK.
	 * @returns {{valid: !boolean, msg: !string, field: !string}}
	 */
	validate() {
		return this.alertFeed.validate();
	}

	/**
	 * Renders the controls markup into a tag and returns it.
	 * @returns {!HTMLElement}
	 */
	render() {
		this.setVal('type', this.alertFeed.type);
		this.setVal('url', this.alertFeed.feed);
		this.setVal('allcleardwell', this.alertFeed.allClearDwell);
		this.setVal('headerfontsize', this.alertFeed.headerFontSize);
		this.setVal('titlefontsize', this.alertFeed.titleFontSize);
		this.setVal('descriptionfontsize', this.alertFeed.descriptionFontSize);
		this.setVal('alertcolor', this.alertFeed.alertColor, 'color');
		this.setVal('alertstripingcolor', this.alertFeed.alertStripingColor, 'color');
		this.setVal('allclearcolor', this.alertFeed.allClearColor, 'color');
		this.setVal('allclearstripingcolor', this.alertFeed.allClearStripingColor, 'color');
		this.setVal('crawlfontsize', this.alertFeed.crawlFontSize);
		this.setVal('crawlforeground', this.alertFeed.crawlForeground, 'color');
		this.setVal('crawlbackground', this.alertFeed.crawlBackground, 'color');
	}

	/**
	 * Removes the control from the DOM.
	 */
	remove() {
		this.container.remove();
	}

	/**
	 * Wraps the given control inside of a label.
	 * @param {!HTMLElement} control
	 * @param {!string} label
	 * @param {!string} tooltip
	 * @returns {!HTMLElement}
	 */
	listboxLabel(label, tooltip) {
		return $.label({ class: 'tooltip-button eas-label', title: tooltip, text: label });
	}

	/**
	 * @param {!Object} inputAttributes
	 * @returns {!HTMLElement}
	 */
	listboxInput(inputAttributes) {
		const inputChange = this.notify.bind(this);
		if (inputAttributes.hasOwnProperty('config')) {
			return $.uicolor({ ...inputAttributes, input: inputChange, change: inputChange });
		} else {
			return $.input({ ...inputAttributes, input: inputChange, change: inputChange });
		}
	}

	/**
	 * Adds the given classes string to the class property of a newly created shallow clone of the template object and returns the newly modified copy.
	 * @param {!Object} template
	 * @param {!string} classes e.g. 'example-class'
	 * @returns {!Object}
	 */
	fromTemplate(template, classes) {
		const attrs = Object.assign({}, template);
		attrs.class = `${attrs.class !== undefined && attrs.class.length > 0 ? `${attrs.class} ` : ''}${classes}`;
		return attrs;
	}

	/**
	 * Utility function that allows creating listboxInputs from a template easily.
	 * @param {!Object} template
	 * @param {!string} classes e.g. 'example-class'
	 * @returns {!HTMLElement}
	 */
	listboxTemplate(template, classes) {
		return this.listboxInput(this.fromTemplate(template, classes));
	}
}
