class BreezeCrawl {
	/**
	 * @param {!HTMLElement} tag
	 */
	constructor(tag, speed = 180, alternativeScroll = false) {
		/**
		 * Number of pixels the crawl should move per second.
		 * @type {!number}
		 */
		this.crawlSpeed = speed;
		
		this.tag = tag;
		this.front = $.p();
		this.tag.appendChild(this.front);
		this.back = $.p();
		this.tag.appendChild(this.back);
		this.spacer = '                       ';
		this.stock = false;

		this.frontAnimation = null;
		this.backAnimation = null;
		
		this.alternativeScroll = alternativeScroll;
		this.scrollbarAnimation = null;
		this.lastTimestamp = 0;
		this.scrollLeft = 0;
		
		if (this.alternativeScroll) {
			this.tag.style.whiteSpace = 'nowrap';
			this.front.style.position = 'relative';
			this.back.style.position = 'relative';
		}
	}

	/**
	 * @param {!number} width
	 * @param {!number} height
	 * @param {!string} text
	 * @param {!number} fontSize
	 * @param {!string} fontFamily
	 * @param {!boolean} animate
	 * @param {!string} [fallbackFontFamily]
	 */
	async startCrawl(width, height, text, fontSize, fontFamily, animate, fallbackFontFamily) {
		if (this.tag == null) {
			return;
		}
		if (text === null || text === undefined || !text) {
			console.log('[WARNING] Empty text detected in the Crawl - returning early!');
			return;
		}
		fontFamily = fallbackFontFamily != undefined ? fallbackFontFamily : fontFamily;
		let formattedText = text + this.spacer;

		const crawlWidth = await BreezeCrawl.measureCrawlWidth(formattedText, fontFamily, fontSize);
		const crawlText = BreezeCrawl.pad(formattedText, crawlWidth, width);
		const css = BreezeCrawl.generateStyles(fontFamily, fontSize, height, this.foreground);
		this.reset(crawlText, css);
		if (animate) {
			const realWidth = this.front.offsetWidth;
			this.animate(realWidth);
		}
	}

	/**
	 * @param {!string} crawlText
	 * @param {!Object} styles
	 */
	reset(crawlText, styles) {
		if (this.tag == null) {
			return;
		}
		
		if (this.stock) {
			this.front.innerHTML = crawlText;
			this.back.innerHTML = crawlText;
		} else {
			this.front.innerText = crawlText;
			this.back.innerText = crawlText;
		}

		Object.assign(this.front.style, styles);
		Object.assign(this.back.style, styles);
	}

	animateScrollbar(timestamp) {
		if (this.tag == null) {
			return;
		}
		if (this.lastTimestamp == 0) {
			this.lastTimestamp = timestamp;
		}
		let backOffsetWidth = this.back.offsetWidth;
		if (backOffsetWidth != Math.floor(this.tag.offsetWidth)) {
			this.back.style.width = `${Math.floor(this.tag.offsetWidth)}px`;
			backOffsetWidth = this.back.offsetWidth;
		}
		this.scrollLeft += Math.abs(timestamp - this.lastTimestamp) * this.crawlSpeed / 1000;
		const scrollWidth = this.tag.scrollWidth;
		this.tag.scrollLeft = Math.min(scrollWidth - backOffsetWidth, this.scrollLeft);
		if (this.scrollLeft >= scrollWidth - backOffsetWidth) {
			this.scrollLeft = this.scrollLeft % (scrollWidth - backOffsetWidth);
			this.tag.scrollLeft = this.scrollLeft;
		}
		this.lastTimestamp = timestamp;
		this.scrollbarAnimation = window.requestAnimationFrame(this.animateScrollbar.bind(this));
	}

	/**
	 * @param {!number} crawlWidth
	 */
	animate(crawlWidth) {
		if (this.tag == null) {
			return;
		}
		if (this.alternativeScroll) {
			this.scrollbarAnimation = window.requestAnimationFrame(this.animateScrollbar.bind(this));
		} else {
			const duration = BreezeCrawl.calculateDuration(this.crawlSpeed, crawlWidth);
			this.frontAnimation = this.front.animate({
				transform: ['translate(0,0)', `translate(${-crawlWidth}px,0)`]
			}, {
				duration: duration,
				iterations: Infinity
			});
			this.backAnimation = this.back.animate({
				transform: [`translate(${crawlWidth}px,0)`, 'translate(0,0)']
			}, {
				duration: duration,
				iterations: Infinity
			});
		}
	}

	play() {
		if (this.tag == null) {
			return;
		}
		if (this.alternativeScroll) {
			this.scrollbarAnimation = window.requestAnimationFrame(this.animateScrollbar.bind(this));
		} else {
			if (this.frontAnimation) {
				this.frontAnimation.play();
				this.backAnimation.play();
			}
		}
	}

	pause() {
		if (this.tag == null) {
			return;
		}
		if (this.scrollbarAnimation) {
			window.cancelAnimationFrame(this.scrollbarAnimation);
			this.scrollbarAnimation = null;
			this.lastTimestamp = 0;
		}
		if (this.frontAnimation) {
			this.frontAnimation.pause();
			this.backAnimation.pause();
		}
	}

	destroy() {
		this.tag = null;
		if (this.scrollbarAnimation) {
			window.cancelAnimationFrame(this.scrollbarAnimation);
			this.scrollbarAnimation = null;
			this.lastTimestamp = 0;
		}
		if (this.frontAnimation) {
			this.frontAnimation.cancel();
			this.frontAnimation = null;
			this.backAnimation.cancel();
			this.backAnimation = null;
		}
	}

	/**
	 * @param {!string} crawlText
	 * @param {!string} fontFamily
	 * @param {!number} fontSize
	 * @returns {!number}
	 */
	static async measureCrawlWidth(crawlText, fontFamily, fontSize) {
		const $p = $.p({ css: {
			position: 'absolute',
			font: fontFamily,
			fontSize: `${fontSize}px`,
			whiteSpace: 'pre'
		} });
		if (this.stock) {
			$p.innerHTML = crawlText;
		} else {
			$p.innerText = crawlText;
		}
		return (await Measure($.div({ class: 'widget-crawl' }, $p), $p)).width;
	}

	/**
	 * @param {!string} crawlText
	 * @param {!number} crawlWidth
	 * @param {!number} desiredWidth
	 * @returns {!string}
	 */
	static pad(crawlText, crawlWidth, desiredWidth) {
		let text = crawlText;
		if (crawlWidth < desiredWidth) {
			const widthDifference = desiredWidth - crawlWidth;
			const numberOfRepeats = Math.round(widthDifference / crawlWidth) + 1;
			for (let i = 0; i < numberOfRepeats; ++i) {
				text += crawlText;
			}
		}
		return text;
	}

	/**
	 * @param {!string} fontFamily
	 * @param {!number} fontSize
	 * @param {!number} height
	 * @param {?string} [foreground]
	 * @returns {!Object}
	 */
	static generateStyles(fontFamily, fontSize, height, foreground = '#000000') {
		return {
			fontFamily: fontFamily,
			fontSize: `${fontSize}px`,
			color: foreground,
			lineHeight: `${height}px`
		};
	}

	/**
	 * @param {!number} crawlSpeed
	 * @param {!number} frontWidth
	 * @returns {!number}
	 */
	static calculateDuration(crawlSpeed, frontWidth) {
		return frontWidth / crawlSpeed * 1000;
	}
}
