class EAS {
	/**
	 * @typedef {Object} EASSettings
	 * @property {!string[]} alertFeeds Enable or Disable EAS by setting a alert feed URL.
	 * @property {!number} allClearDwell The number of minutes to display All Clear after an All Clear is published.
	 * @property {!number} headerFontSize The display fontsize of the date and time on the Players during EAS.
	 * @property {!number} titleFontSize The display fontsize of the alert title on the Players during EAS.
	 * @property {!number} descriptionFontSize The display fontsize of the alert description on the Players during EAS.
	 * @property {!string} alertColor The background color of the alert.
	 * @property {!string} alertStripingColor The color of the stripes on the alert header and footer (set to the same color as the Alert Color to hide the stripes).
	 * @property {!string} allClearColor The background color of the allclear.
	 * @property {!string} allClearStripingColor The color of the stripes on the allclear header and footer (set to the same color as the AllClear Color to hide the stripes).
	 */

	/**
	 * Returns with the default values or the parsed values given from a server response.
	 * @param {!Object[]} [data]
	 * @returns {!EASSettings}
	 */
	static getDefaults(data = []) {
		const defaults = {
			type: 'BULLETIN',
			url: '',
			allcleardwell: 15,
			headerfontsize: 32,
			titlefontsize: 64,
			descriptionfontsize: 48,
			alertcolor: '#ff0000',
			alertstripingcolor: '#ffffff',
			allclearcolor: '#008000',
			allclearstripingcolor: '#ffffff',
			crawlfontsize: 32,
			crawlforeground: '#ffffff',
			crawlbackground: '#000000'
		};

		if (!Array.isArray(data)) {
			// If data is NOT an array, then it's the old version 1 data which was an object with an array of AlertFeeds,
			// and it must be converted into an array of Feed objects (where each entry includes the full settings.)
			if (data.alertfeeds === null || data.alertfeeds === undefined) {
				return [];
			} else {
				return data.alertfeeds.map(url => ({
					type: 'BULLETIN', // Version 1 feeds are always BULLETINs.
					url: url,
					allClearDwell: data.allcleardwell != null ? Number(data.allcleardwell) : defaults.allcleardwell,
					headerFontSize: data.headerfontsize != null ? Number(data.headerfontsize) : defaults.headerfontsize,
					titleFontSize: data.titlefontsize != null ? Number(data.titlefontsize) : defaults.titlefontsize,
					descriptionFontSize: data.descriptionfontsize != null ? Number(data.descriptionfontsize) : defaults.descriptionfontsize,
					alertColor: data.alertcolor != null ? data.alertcolor : defaults.alertcolor,
					alertStripingColor: data.alertstripingcolor != null ? data.alertstripingcolor : defaults.alertstripingcolor,
					allClearColor: data.allclearcolor != null ? data.allclearcolor : defaults.allclearcolor,
					allClearStripingColor: data.allclearstripingcolor != null ? data.allclearstripingcolor : defaults.allclearstripingcolor,
					// The following crawl options didn't exist in version 1 feeds
					crawlFontSize: defaults.crawlfontsize,
					crawlForeground: defaults.crawlforeground,
					crawlBackground: defaults.crawlbackground
				}));
			}
		} else {
			return data.map(feed => ({
				type: feed.type != null ? feed.type : defaults.type,
				url: feed.url != null ? feed.url : defaults.url,
				allClearDwell: feed.allcleardwell != null ? Number(feed.allcleardwell) : defaults.allcleardwell,
				headerFontSize: feed.headerfontsize != null ? Number(feed.headerfontsize) : defaults.headerfontsize,
				titleFontSize: feed.titlefontsize != null ? Number(feed.titlefontsize) : defaults.titlefontsize,
				descriptionFontSize: feed.descriptionfontsize != null ? Number(feed.descriptionfontsize) : defaults.descriptionfontsize,
				alertColor: feed.alertcolor != null ? feed.alertcolor : defaults.alertcolor,
				alertStripingColor: feed.alertstripingcolor != null ? feed.alertstripingcolor : defaults.alertstripingcolor,
				allClearColor: feed.allclearcolor != null ? feed.allclearcolor : defaults.allclearcolor,
				allClearStripingColor: feed.allclearstripingcolor != null ? feed.allclearstripingcolor : defaults.allclearstripingcolor,
				crawlFontSize: feed.crawlfontsize != null ? Number(feed.crawlfontsize) : defaults.crawlfontsize,
				crawlForeground: feed.crawlforeground != null ? feed.crawlforeground : defaults.crawlforeground,
				crawlBackground: feed.crawlbackground != null ? feed.crawlbackground : defaults.crawlbackground
			}));
		}
	}
}
