/* exported FilterControl */
/* global CAPS, AlertViewer, post, get */
const FilterControl = (() => {
	const enumTypes = 
		Object.keys(CAPS.Enums)
			.map(key => CAPS.Enums[key].Name);
	const datumTypes =
		Object.keys(CAPS.Datums)
			.map(key => CAPS.Datums[key].Name);

	class AlertViewerPreview {
		/**
		 * @param {!HTMLElement} preview
		 * @param {!Object} filter
		 * @param {!Size} resolution
		 * @param {!Alert} alert
		 */
		constructor(preview, filter, resolution, alert) {
			this.preview = preview;
			this.resolution = resolution;
			const oldViewer = this.preview.alertviewerpreview;
			if (oldViewer !== undefined) {
				oldViewer.remove();
			}
			this.layout();
			this.alertViewer = new AlertViewer(alert, filter, this.resolution, this.preview);
			this.preview.alertviewerpreview = this;
		}

		remove() {
			this.alertViewer.remove();
		}

		layout() {
			Object.assign(this.preview.parentNode.style, {
				width: 0,
				height: 0
			});
			Object.assign(this.preview.style, {
				width: `${this.resolution.width}px`,
				height: `${this.resolution.height}px`,
				transform: `scale(${this.preview.closest('.display-options').offsetHeight / this.resolution.height})`
			});
			const dimensions = this.preview.getBoundingClientRect();
			Object.assign(this.preview.parentNode.style, {
				width: `${dimensions.width}px`,
				height: `${dimensions.height}px`
			});
		}

		/**
		 * @param {!HTMLElement} preview
		 * @param {!Object} filter
		 * @param {!Size} resolution
		 * @returns {!Promise<!AlertViewerPreview>}
		 */
		static async enable(preview, filter, resolution) {
			const alert = await CAPS.CAPSLoader.fromXml(AlertViewerPreview.xml(resolution));
			return new AlertViewerPreview(preview, filter, resolution, alert);
		}

		/**
		 * @param {!Size} resolution
		 * @returns {!string}
		 */
		static xml(resolution) {
			const now = moment();
			const expires = now.add(4, 'hours');
			return `
				<alert xmlns="urn:oasis:names:tc:emergency:cap:1.2">
					<identifier>TEST-ALERT</identifier>
					<sender>BREEZE</sender>
					<sent>${now.format('YYYY-MM-DDTHH:mm:ssZZ')}</sent>
					<status>Actual</status>
					<msgType>Alert</msgType>
					<scope>Public</scope>
					<source>null</source>
					<info>
						<category>Other</category>
						<event>Display Options Preview</event>
						<responseType>Assess</responseType>
						<urgency>Unknown</urgency>
						<severity>Unknown</severity>
						<certainty>Observed</certainty>
						<expires>${expires.format('YYYY-MM-DDTHH:mm:ssZZ')}</expires>
						<headline>Alert Preview - ${resolution.width}x${resolution.height}</headline>
						<description>Alert Preview generated for ${resolution.width}x${resolution.height} resolution, real alerts may appear different depending on the players display resolution.</description>
					</info>
				</alert>
			`;
		}
	}
	
	function FilterItemConditionList($tag, conditions) {
		const render = () => {
			$tag.innerHTML = '';
			$tag.append(
				$.header({},
					$.label({ text: 'Field' }),
					$.label({ text: 'Operator' }),
					$.label({ text: 'Value' }),
					$.i({ class: 'add-item mdi mdi-plus-circle-outline tooltip-button', title: 'Add a filter condition', click: () => {
						conditions.push({
							field: 'sender',
							operation: 'EXACTLY',
							value: ''
						});
						render();
					} })
				),
				...conditions.map((condition, index) => [
					$.select({ class: 'field', value: condition.field, change: function() {
						condition.field = this.value;
						const schema = CAPS.FieldsUtils.getFieldSchema(condition.field);
						if (datumTypes.includes(schema.type)) {
							condition.value = { valueName: '', value: '' };
						}
						render();
					} },
						...CAPS.FieldsUtils.getAllFieldNames().map(name => $.option({ value: name, text: name }))
					),
					$.select({ class: 'operation', value: condition.operation, change: function() { condition.operation = this.value; } },
						$.option({ text: 'INCLUDES' }),
						$.option({ text: 'EXCLUDES' }),
						$.option({ text: 'EXACTLY' }),
						$.option({ text: 'NOT' })
					),
					(() => {
						const schema = CAPS.FieldsUtils.getFieldSchema(condition.field);
						if (schema.type == 'string') {
							const inputChange = function() { condition.value = this.value; };
							return $.input({ type: 'text', class: 'value', value: condition.value, input: inputChange, change: inputChange });
						} else if (enumTypes.includes(schema.type)) {
							return $.select({ class: 'value', value: condition.value, change: function() { condition.value = this.value; } },
								...(new CAPS.Enums[schema.type](null)).allowed.map(value => $.option({ value: value, text: value }))
							);
						} else if (datumTypes.includes(schema.type)) {
							const nameInputChange = function() { condition.value.valueName = this.value; };
							const valueInputChange = function() { condition.value.value = this.value; };
							return $.span({ class: 'datum-container' },
								$.input({ type: 'text', class: 'datum-value-name value', placeholder: 'valueName', value: condition.value.valueName, input: nameInputChange, change: nameInputChange }),
								$.input({ type: 'text', class: 'datum-value value', placeholder: 'value', value: condition.value.value, input: valueInputChange, change: valueInputChange })
							);
						}
						throw new Error(`Unable to find valid control for schema: ${schema.field}/${schema.type}/${schema.isArray}`);
					})(),
					$.i({ class: 'mdi mdi-trash-can-outline tooltip-button delete', title: 'Delete', click: () => {
						conditions.splice(index, 1);
						render();
					} })
				]).flat()
			)
		};
		render();
	}
	
	function FilterItems($tag, filters) {
		const render = () => {
			$tag.innerHTML = '';
			$tag.append(
				$.header({ text: 'CAPS Feeds' }),
				$.ul({ class: 'iconlist' },
					$.li({ class: 'filter-add-item tooltip-button mdi mdi-plus-circle-outline', title: 'Add Filter', click: () => {
						filters.push({
							displayMode: 'Full Screen',
							dwell: null,
							forceDwell: false,
							displayOptions: {
								background: 'rgba(255, 0, 0, 1)',
								stripes: 'rgba(255, 255, 255, 1)',
								dateTime: { family: null, size: 32, color: 'rgba(255, 255, 255, 1)', background: 'rgba(0, 0, 0, 0)' },
								headline: { family: null, size: 64, color: 'rgba(255, 255, 255, 1)', background: 'rgba(0, 0, 0, 0)' },
								description: { family: null, size: 48, color: 'rgba(255, 255, 255, 1)', background: 'rgba(0, 0, 0, 0)' },
								crawl: { family: null, size: 32, color: 'rgba(255, 255, 255, 1)', background: 'rgba(0, 0, 0, 1)' }
							},
							conditions: []
						});
						render();
					} })
				),
				$.div({ class: 'filter-list list-formgrid' }),
				$.div({ class: 'panel-buttons' },
					$.button({ class: 'primary', text: 'Save', click: async () => {
						LockBody();
						try {
							const response = await post('/post/updatecapsfilters.php', {
								capsfilters: JSON.stringify(filters)
							});
							CheckSessionExpired(response);
							if (response.success) {
								ConfirmationBox('CAPS Filters Updated', 'CAPS Filters Updated.');
							} else {
								ConfirmationBox('Error', `CAPS Filters Error: ${response.error}.`);
							}
						} catch (err) {
							NetworkingError();
						} finally {
							UnlockBody();
						}
					} })
				)
			);
			
			for (const filter of filters) {
				const preview = async () => {
					await AlertViewerPreview.enable($filter.querySelector('.preview-box > .alert-preview'), CAPS.Filter.create(filter), { width: 1280, height: 720 });
				};
				const filterDisplayOptions = () => {
					const displayMode = $filter.querySelector('.display-mode').value;
					for (const $e of [$filter.querySelector('.background-color'), $filter.querySelector('.background-color').previousElementSibling]) $e.hidden = displayMode == 'Crawl';
					for (const $e of [$filter.querySelector('.stripes-color'), $filter.querySelector('.stripes-color').previousElementSibling]) $e.hidden = displayMode == 'Crawl';
					for (const $e of [...document.querySelectorAll('.dateTime, .dateTime ~ *')].slice(0, 5)) $e.hidden = displayMode == 'Crawl';
					for (const $e of [...document.querySelectorAll('.headline, .headline ~ *')].slice(0, 5)) $e.hidden = displayMode == 'Crawl';
					for (const $e of [...document.querySelectorAll('.description, .description ~ *')].slice(0, 5)) $e.hidden = displayMode == 'Crawl';
					for (const $e of [...document.querySelectorAll('.crawl, .crawl ~ *')].slice(0, 5)) $e.hidden = !(displayMode == 'Crawl' || displayMode == 'Both');
				};
				const index = filters.indexOf(filter);
				const fontFamilyOptions = {
					options: fonts,
					history: 10,
					processOption: function() {
						this.style.fontFamily = `'${this.value}'`;
					}
				};
				const fontFamilyTextChanged = function() {
					this.$input.style.fontFamily = `'${this.value}'`;
				};
				const colorConfig = {
					showPalette: true,
					localStorageKey: 'globalcolor',
					opacity: true,
					previewChange: function() {
						this.dispatchEvent(new Event('change'));
					}
				};
				const $filter = $tag.querySelector('.filter-list').appendChild($.div({ class: 'item' },
					$.div({ class: 'formgrid' },
						$.label({ text: 'Display Mode' }),
						$.select({ class: 'display-mode sampleSelect', value: filter.displayMode, change: function()
						{
							filter.displayMode = this.value;
							filterDisplayOptions();
							preview();
						} },
							$.option({ text: 'Full Screen' }),
							$.option({ text: 'Crawl' }),
							$.option({ text: 'Both' })
						),
						$.label({ text: 'Dwell' }),
						$.input({ type: 'number', class: 'dwell sampleInput', min: 0, value: filter.dwell, input: function() { filter.dwell = this.value; }, change: function() { filter.dwell = this.value; } }),
						$.label({ text: 'Force Dwell' }),
						$.uicheckbox({ id: `force-dwell-${index}`, class: 'force-dwell', checked: filter.forceDwell, change: function() { filter.forceDwell = this.checked; } }),
						$.label({ text: 'Filter Conditions' }),
						$.div({ class: 'filter-condition-list formgrid compact' }),
						$.label({ text: 'Display Options' }),
						$.div({ class: 'display-options' },
							$.div({ class: 'formgrid' },
								$.label({ text: 'Background' }),
								$.uicolor({ type: 'text', class: 'background-color color-picker', css: { gridColumn: 'content / end' }, config: colorConfig, value: 'rgba(255, 0, 0, 1)' }),
								
								$.label({ text: 'Stripes' }),
								$.uicolor({ type: 'text', class: 'stripes-color color-picker', css: { gridColumn: 'content / end' }, config: colorConfig, value: 'rgba(255, 255, 255, 1)' }),
								
								$.label({ class: 'dateTime', text: 'DateTime' }),
								$.uiinputselect({ class: 'font-family', config: fontFamilyOptions, textchanged: fontFamilyTextChanged, value: defaultFont }),
								$.input({ type: 'number', class: 'font-size sampleInput', min: '0', value: '32', title: 'Font Size' }),
								$.uicolor({ type: 'text', class: 'font-color color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(255, 255, 255, 1)', title: 'Font Color' }),
								$.uicolor({ type: 'text', class: 'background color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(0, 0, 0, 0)', title: 'Background Color' }),
								
								$.label({ class: 'headline', text: 'Headline' }),
								$.uiinputselect({ class: 'font-family', config: fontFamilyOptions, textchanged: fontFamilyTextChanged, value: defaultFont }),
								$.input({ type: 'number', class: 'font-size sampleInput', min: '0', value: '64', title: 'Font Size' }),
								$.uicolor({ type: 'text', class: 'font-color color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(255, 255, 255, 1)', title: 'Font Color' }),
								$.uicolor({ type: 'text', class: 'background color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(0, 0, 0, 0)', title: 'Background Color' }),
								
								$.label({ class: 'description', text: 'Description' }),
								$.uiinputselect({ class: 'font-family', config: fontFamilyOptions, textchanged: fontFamilyTextChanged, value: defaultFont }),
								$.input({ type: 'number', class: 'font-size sampleInput', min: '0', value: '48', title: 'Font Size' }),
								$.uicolor({ type: 'text', class: 'font-color color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(255, 255, 255, 1)', title: 'Font Color' }),
								$.uicolor({ type: 'text', class: 'background color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(0, 0, 0, 0)', title: 'Background Color' }),
								
								$.label({ class: 'crawl', text: 'Crawl' }),
								$.uiinputselect({ class: 'font-family', config: fontFamilyOptions, textchanged: fontFamilyTextChanged, value: defaultFont }),
								$.input({ type: 'number', class: 'font-size sampleInput', min: '0', value: '32', title: 'Font Size' }),
								$.uicolor({ type: 'text', class: 'font-color color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(255, 255, 255, 1)', title: 'Font Color' }),
								$.uicolor({ type: 'text', class: 'background color-picker', config: Object.assign({}, colorConfig, { compact: true }), value: 'rgba(0, 0, 0, 1)', title: 'Background Color' })
							),
							$.div({ class: 'preview-box' },
								$.div({ class: 'alert-preview transparency-checkerboard' })
							)
						)
					),
					$.div({ class: 'controls' },
						$.i({ class: 'mdi mdi-trash-can-outline tooltip-button delete', title: 'Delete', click: () => {
							filters.splice(index, 1);
							render();
						} })
					)
				));
				FilterItemConditionList($filter.querySelector('.filter-condition-list'), filter.conditions);
				
				const $displayOptions = $filter.querySelector('.display-options > .formgrid');
				$displayOptions.querySelector('.background-color').value = filter.displayOptions.background;
				$displayOptions.querySelector('.background-color').addEventListener('change', function() {
					filter.displayOptions.background = this.value;
					preview();
				});
				$displayOptions.querySelector('.stripes-color').value = filter.displayOptions.stripes;
				$displayOptions.querySelector('.stripes-color').addEventListener('change', function() {
					filter.displayOptions.stripes = this.value;
					preview();
				});
				const setFont = ($font, data) => {
					// Leave the font to the default if it isn't set
					$.next($font, '.font-family').value = data.family !== null ? data.family : defaultFont;
					$.next($font, '.font-family').addEventListener('change', function() { data.family = this.value; preview(); });
					$.next($font, '.font-size').value = data.size;
					$.next($font, '.font-size').addEventListener('input', function() { data.size = this.value; preview(); });
					$.next($font, '.font-size').addEventListener('change', function() { data.size = this.value; preview(); });
					$.next($font, '.font-color').value = data.color;
					$.next($font, '.font-color').addEventListener('change', function() { data.color = this.value; preview(); });
					$.next($font, '.background').value = data.background;
					$.next($font, '.background').addEventListener('change', function() { data.background = this.value; preview(); });
				};
				setFont($displayOptions.querySelector('.dateTime'), filter.displayOptions.dateTime);
				setFont($displayOptions.querySelector('.headline'), filter.displayOptions.headline);
				setFont($displayOptions.querySelector('.description'), filter.displayOptions.description);
				setFont($displayOptions.querySelector('.crawl'), filter.displayOptions.crawl);
				filterDisplayOptions();
				preview();
			}
		};
		render();
	}
	
	async function FilterControl($tag) {
		try {
			LockBody();
			const response = await get('/get/capsfilters.php');
			CheckSessionExpired(response);
			if (response.success) {
				const data = response.data.length > 0 ? JSON.parse(response.data[0].capfilters) : [];
				FilterItems($tag, data);
			}
		} catch (exception) {
			NetworkingError();
		} finally {
			UnlockBody();
		}
	}
	
	return FilterControl;
})();