customElements.define('ui-color', class UIColor extends HTMLElement {
	constructor() {
		super();
		this.optionsSet = false;
		this.lastHidePickerHandler = null;
		this.attachShadow({mode: 'open'});
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/materialdesignicons.min.css');

:host {
	display: flex;
	flex-flow: row nowrap;
	align-items: stretch;
	color: var(--input-color);
	background: url("ui/images/transparent.png") 0% 0% / 10px 10px white;
	border: 1px solid var(--input-border-color);
	border-radius: 4px;
	box-sizing: border-box;
	height: 32px;
	line-height: 32px;
	overflow: hidden;
}
:host(:focus) {
	outline: none;
}
#input {
	flex: 1 1 0;
	min-width: 0;
	width: 0;
	overflow: hidden;
	padding: 0 5px;
	font-size: 16px;
	white-space: nowrap;
}
#input:focus {
	outline: none;
}
#original-color {
	flex: 0 0 30px;
	cursor: pointer;
	background-color: transparent;
	border: none;
	border-right: 1px solid black;
}
#original-color:focus {
	border: none;
	outline: none;
}
#show-picker {
	flex: 0 0 auto;
	font-size: 14px;
	align-self: stretch;
	display: flex;
	align-items: center;
	cursor: pointer;
	padding: 0 2px;
	user-select: none;
	background-color: var(--input-background-color);
	border: none;
}
#show-picker:focus {
	border: none;
	outline: none;
}
#show-picker:hover:before {
	opacity: 0.8;
}
</style>`;
		this.shadowRoot.append(
			this.$input = $.span({ id: 'input', contenteditable: 'true', spellcheck: false, keydown: onlyPoundHex, paste: e => {
				e.preventDefault();
				const text = e.clipboardData.getData('text/plain').toLowerCase().replace(/[^#0-9a-f]/g, '');
				const selection = this.shadowRoot.getSelection();
				if (selection.rangeCount) {
					const range = selection.getRangeAt(0);
					range.deleteContents();
					const textNode = document.createTextNode(text);
					range.insertNode(textNode);
					const newRange = document.createRange();
					newRange.setStart(textNode, textNode.length);
					newRange.collapse(true);
					selection.removeAllRanges();
					selection.addRange(newRange);
					this.hexInputChange();
				}
			}, focus: () => {
				const { r, g, b } = rgbaToColor(this.value);
				this.$input.innerText = `#${rgbToHex(r, g, b)}`;
				this.showPicker();
			}, blur: () => {
				this.setValueText(this.value);
				document.getSelection().removeAllRanges();
			}, input: e => {
				this.hexInputChange();
			} }),
			this.$originalColor = $.button({ id: 'original-color', title: 'Reset Color', hidden: true, click: () => {
				this.hidePicker(true);
			} }),
			$.button({ id: 'show-picker', class: 'mdi mdi-play mdi-rotate-90', pointerdown: () => {
				if (this.$colorPicker) {
					this.hidePicker();
				} else {
					this.showPicker();
				}
			} })
		);
		this.$colorPicker = null;
		this.lastPreviewChange = 0;
		this.bouncePreviewChange = null;
	}
	get value() {
		// If the color picker is open then return the current temporary new value
		if (this.$colorPicker) {
			return this.internalNewValue;
		} else {
			return this.internalValue;
		}
	}
	set value(value) {
		if (/^rgb\(.*$/.test(value)) {
			const { r, g, b } = rgbToColor(value);
			value = `rgba(${r}, ${g}, ${b}, 1)`;
		}
		if (value != this.internalValue) {
			this.internalValue = value;
			this.internalNewValue = value;
			this.setValueText(value);
		
			this.$colorPicker?.update('hex');
		}
	}
	set newValue(value) {
		this.internalNewValue = value;
		this.setValueText(value);
		this.callBouncePreviewChange();
	}
	callBouncePreviewChange() {
		// Bounce update this.previewChange() every 100 ms
		if (this.bouncePreviewChange == null) {
			const currentTime = Date.now();
			const bounceDelay = 100; // 100 ms
			this.bouncePreviewChange = setTimeout(() => {
				this.bouncePreviewChange = null;
				this.previewChange();
				this.lastPreviewChange = currentTime;
			}, Math.max(0, this.lastPreviewChange + bounceDelay - currentTime));
		}
	}
	setValueText(value) {
		if (!this.compact) {
			const { r, g, b, a } = rgbaToColor(value);
			this.$input.innerText = a == 0 ? 'Transparent' : `#${rgbToHex(r, g, b)}${a == 100 ? '' : ` ${a}%`}`;
		}
		this.setBackgroundColor(value);
	}
	setBackgroundColor(value) {
		this.$input.style.background = value;
		if (!this.compact) {
			const { r, g, b, a } = rgbaToColor(value);
			// Choose either a black or white font color that contrasts with the background.
			const { l } = RGBtoHSL(r, g, b);
			// The background is white so as the color becomes transparent it becomes lighter.
			Object.assign(this.$input.style, l * a / 100 + (100 - a) < 50 ? {
				color: 'white',
				textShadow: '-1px 0 1px black,1px 0 black,0 -1px black,0 1px 1px black,0 0px 10px black,0 0px 10px black,0 0px 10px black'
			} : {
				color: 'black',
				textShadow: '-1px 0 1px white,1px 0 white,0 -1px white,0 1px 1px white,0 0px 10px white,0 0px 10px white,0 0px 10px white'
			});
		}
	}
	connectedCallback() {
		this.tabIndex = -1;
	}
	options({
		previewChange = () => {},
		opacity = true,
		palette = null,
		compact = false,
		showPalette = false,
		showPaletteOnly = false,
		localStorageKey = false,
		transparentButton = false } = {}) {
		if (this.optionsSet) {
			throw new Error('options already called');
		}
		this.optionsSet = true;
		
		this.previewChange = previewChange;
		this.opacity = opacity;
		this.palette = palette;
		this.compact = compact;
		this.showPalette = showPalette;
		this.showPaletteOnly = showPaletteOnly;
		this.localStorageKey = localStorageKey;
		this.transparentButton = transparentButton;

		if (this.compact) {
			this.$input.style.pointerEvents = 'none';
			this.style.width = `${18 + 60}px`;
			this.style.minWidth = `${18 + 60}px`;
		}
		
		this.movedLeft = null;
		this.movedTop = null;
	}
	showPicker() {
		if (this.$colorPicker != null) {
			return;
		}
		this.$colorPicker = $.uicolorpicker();
		this.$colorPicker.$color = this;
		this.$colorPicker?.update('hex');
		document.body.appendChild(this.$colorPicker);
		const thisRectangle = this.getBoundingClientRect();
		
		if (this.movedLeft != null) {
			// Use the moved position when reopening the picker
			Object.assign(this.$colorPicker.style, {
				left: `${this.movedLeft}px`,
				top: `${this.movedTop}px`
			});
		} else {
			const left = thisRectangle.right + this.$colorPicker.offsetWidth > document.documentElement.clientWidth ? thisRectangle.left - this.$colorPicker.offsetWidth : thisRectangle.right;
			const top = thisRectangle.top + this.$colorPicker.offsetHeight < document.documentElement.clientHeight ? thisRectangle.top : thisRectangle.bottom - this.$colorPicker.offsetHeight;
			Object.assign(this.$colorPicker.style, {
				left: `${left}px`,
				top: `${top}px`
			});
		}
		this.$originalColor.style.background = this.internalValue;
		this.$originalColor.hidden = false;
		// If the user clicks on anything outside of the picker then close it
		document.addEventListener('pointerdown', this.lastHidePickerHandler = e => {
			if (e.target !== this && e.target !== this.$colorPicker) {
				e.stopPropagation();
				this.hidePicker();
			}
		}, true);
		this.dispatchEvent(new CustomEvent('showpicker'));
	}
	hidePicker(reset = false) {
		document.removeEventListener('pointerdown', this.lastHidePickerHandler, true);
		this.$colorPicker?.remove();
		this.$colorPicker = null;
		if (reset) {
			if (this.internalNewValue != this.internalValue) {
				this.internalNewValue = this.internalValue;
				this.setValueText(this.internalValue);
				this.dispatchEvent(new Event('change'));
			}
		} else if (this.internalNewValue != this.internalValue) {
			this.internalValue = this.internalNewValue;
			this.setValueText(this.internalValue);
			this.saveColorToPalette();
			this.dispatchEvent(new Event('change'));
		}
		this.$originalColor.hidden = true;
		this.dispatchEvent(new CustomEvent('hidepicker'));
	}
	hexInputChange() {
		const hex = this.$input.innerText;
		// rrggbb, #rrggbb, rrggbbaa, or #rrggbbaa
		const matches = hex.match(/^#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})?$/i);
		if (matches) {
			let [r, g, b, a] = matches.slice(1).map(v => v == null ? null : parseInt(v, 16));
			if (!isNaN(r) && !isNaN(g) && !isNaN(b) && (a == null || !isNaN(a))) {
				if (a == null) {
					this.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(this.$colorPicker.$opacity.value / 100).toFixed(3)})`;
				} else {
					a = Math.max(0, Math.min(Math.round(a / 255 * 100), 100));
					this.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(a / 100).toFixed(3)})`;
				}
				this.$colorPicker.update('hex');
			}
		}
	}
	saveColorToPalette() {
		if (this.localStorageKey) {
			let palette = [];
			try {
				palette = JSON.parse(localStorage[this.localStorageKey]);
			} catch (exception) {
				localStorage[this.localStorageKey] = JSON.stringify([]);
			}
			const index = palette.indexOf(this.value);
			if (index == -1) {
				palette.unshift(this.value);
				// Truncate to 10 colors
				if (palette.length > 10) {
					palette.length = 10;
				}
			} else {
				// Move the color to the front
				palette.splice(index, 1);
				palette.unshift(this.value);
			}
			localStorage[this.localStorageKey] = JSON.stringify(palette);
		}
	}
});

customElements.define('ui-colorpicker', class UIColorPicker extends HTMLElement {
	constructor() {
		super();
		
		this.attachShadow({mode: 'open'});
	}
	async options({ value }) {
		this.internalValue = value;
		
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/materialdesignicons.min.css');

[hidden] {
	display: none !important;
}

:host {
	position: absolute;
	background-color: var(--background-contrast-color);
	padding: 5px;
	border: 1px solid var(--input-border-color);
	box-sizing: border-box;
	font-family: var(--font);
	font-size: 16px;
	color: var(--font-color);
	user-select: none;
}
#options {
	display: flex;
	flex-flow: row nowrap;
}
#show-hsv,
#show-hsl {
	cursor: pointer;
	flex: 1 0 100px;
	font-size: 12px;
	padding: 7px 5px;
	background-color: var(--tab-color);
	color: var(--font-color);
	border: none;
}
#show-hsv:focus,
#show-hsl:focus {
	outline: none;
}
#show-hsv:hover,
#show-hsl:hover {
	opacity: 0.8;
	background-color: var(--subheader-color);
}
#show-hsv.selected,
#show-hsl.selected {
	background-color: var(--tab-selected-color);
}
#body {
	display: inline-grid;
	grid-template-columns: max-content max-content 50px max-content 50px;
	row-gap: 2px;
	column-gap: 5px;
	padding: 10px;
}
#hsv-hue-picker,
#hsl-hue-picker {
	grid-column: 1 / span 5;
	cursor: pointer;
	margin-bottom: 5px;
	position: relative;
}
#hsv-saturationvalue-picker,
#hsl-saturationlightness-picker {
	grid-column: 1;
	grid-row: 2 / span 3;
	cursor: pointer;
	position: relative;
}
#opacity-picker {
	grid-column: 1;
	grid-row: 5;
	cursor: pointer;
	position: relative;
	width: 100px;
	height: 20px;
	background: linear-gradient(to right, rgba(255,255,255,0), #fff), url("ui/images/transparent.png") 0% 0% / 10px 10px, white;
}
label {
	margin-right: 4px;
	display: flex;
	flex-flow: row nowrap;
	align-items: center;
	justify-content: flex-end;
}
input {
	padding: 4px;
	border: 1px solid var(--input-border-color);
	background-color: var(--input-background-color);
	border-radius: 4px;
	box-sizing: border-box;
	height: 32px;
	font-family: var(--font);
	font-size: 16px;
}
input:focus {
	outline: none;
}
#hsv-hue-picker .handle,
#hsl-hue-picker .handle {
	position: absolute;
	margin: -2px 0 0 -4px;
	border: 1px solid white;
	width: 8px;
	height: 24px;
	left: 0;
	top: 0;
	pointer-events: none;
	box-sizing: border-box;
	box-shadow: 0px 0px 5px #000;
}
#hsv-saturationvalue-picker .handle,
#hsl-saturationlightness-picker .handle {
	position: absolute;
	margin: -4px 0 0 -4px;
	border: 1px solid white;
	width: 8px;
	height: 8px;
	left: 0;
	top: 0;
	pointer-events: none;
	box-sizing: border-box;
	box-shadow: 0px 0px 5px #000;
}
#opacity-picker .handle {
	position: absolute;
	margin: -2px 0 0 -4px;
	border: 1px solid white;
	width: 8px;
	height: 24px;
	left: 0;
	top: 0;
	pointer-events: none;
	box-sizing: border-box;
	box-shadow: 0px 0px 5px #000;
}
label[for="hsv-hue"],
label[for="hsl-hue"] {
	grid-row: 2;
	grid-column: 4;
}
#hsv-hue,
#hsl-hue {
	grid-row: 2;
	grid-column: 5;
}
label[for="hsv-saturation"],
label[for="hsl-saturation"] {
	grid-row: 3;
	grid-column: 4;
}
#hsv-saturation,
#hsl-saturation {
	grid-row: 3;
	grid-column: 5;
}
label[for="hsv-value"],
label[for="hsl-lightness"] {
	grid-row: 4;
	grid-column: 4;
}
#hsv-value,
#hsl-lightness {
	grid-row: 4;
	grid-column: 5;
}
#opacity {
	width: calc(100% - 1ch - 4px);
	grid-column: 3;
	grid-row: 5;
}
#percent {
	display: flex;
	flex-flow: row nowrap;
	align-items: center;
	justify-content: flex-end;
	grid-row: 5;
	grid-column: 3;
	pointer-events: none;
}
#pound {
	margin: 0 5px;
}
#hex {
	width: 9ch;
}
#transparent-button {
	display: block;
	border: 1px solid black;
	background: url(ui/images/transparent.png) 0% 0% / 10px 10px, white;
	padding: 2px;
	margin: 0 12px 5px 12px;
	box-sizing: border-box;
	width: 24px;
	height: 24px;
	cursor: pointer;
}
#color-palette {
	display: flex;
	flex-flow: row wrap;
	padding: 0 10px 10px 10px;
}
#color-palette > button {
	width: 24px;
	height: 24px;
	cursor: pointer;
	border: 1px solid black;
	margin: 2px;
	box-sizing: border-box;
}
#transparent-button:hover,
#color-palette > button:hover {
	border: 1px solid white;
}
#transparent-button:focus,
#color-palette > button:focus {
	outline: none;
}
#move {
	cursor: move;
	font-size: 22px;
	margin: 0 2px;
}
</style>`;
		const clamp = (min, max) => {
			return function() {
				const value = parseInt(this.value, 10);
				if (isNaN(value)) {
					this.value = 0;
				} else {
					this.value = Math.max(min, Math.min(value, max));
				}
			};
		};
		let lastColor;
		this.shadowRoot.append(
			this.$options = $.div({ id: 'options' },
				this.$showHSV = $.button({ id: 'show-hsv', text: 'HSV', click: this.showHSV.bind(this) }),
				this.$showHSL = $.button({ id: 'show-hsl', text: 'HSL', click: this.showHSL.bind(this) }),
				this.$move = $.i({ id: 'move', class: 'mdi mdi-cursor-move' })
			),
			this.$body = $.div({ id: 'body' },
				this.$hsvHuePicker = $.div({ id: 'hsv-hue-picker', pointerdown: e => {
					const documentPointerMove = e => {
						const canvasRect = this.$hsvHueCanvas.getBoundingClientRect();
						const x = Math.max(0, Math.min(Math.round(e.clientX - canvasRect.left), 359));
						this.$hsvHue.value = x;
						this.$hsvHueHandle.style.left = `${x}px`;
						this.hsvInputChange(null, true);
					};
					const documentPointerUp = e => {
						document.removeEventListener('pointermove', documentPointerMove);
						document.removeEventListener('pointerup', documentPointerUp);
					};
					document.addEventListener('pointermove', documentPointerMove);
					document.addEventListener('pointerup', documentPointerUp);
					documentPointerMove(e);
				} },
					this.$hsvHueCanvas = $.canvas({ width: 360, height: 20 }),
					this.$hsvHueHandle = $.span({ class: 'handle' })
				),
				this.$hsvSaturationValuePicker = $.div({ id: 'hsv-saturationvalue-picker', pointerdown: e => {
					const documentPointerMove = e => {
						const canvasRect = this.$hsvSaturationValueCanvas.getBoundingClientRect();
						const x = Math.max(0, Math.min(Math.round(e.clientX - canvasRect.left), 100));
						const y = Math.max(0, Math.min(Math.round(e.clientY - canvasRect.top), 100));
						this.$hsvSaturation.value = x;
						this.$hsvValue.value = 100 - y;
						this.hsvInputChange(null, true);
					};
					const documentPointerUp = e => {
						document.removeEventListener('pointermove', documentPointerMove);
						document.removeEventListener('pointerup', documentPointerUp);
					};
					document.addEventListener('pointermove', documentPointerMove);
					document.addEventListener('pointerup', documentPointerUp);
					documentPointerMove(e);
				} },
					this.$hsvSaturationValueCanvas = $.canvas({ width: 100, height: 100 }),
					this.$hsvSaturationValueHandle = $.span({ class: 'handle' })
				),
				this.$hslHuePicker = $.div({ id: 'hsl-hue-picker', pointerdown: e => {
					const documentPointerMove = e => {
						const canvasRect = this.$hslHueCanvas.getBoundingClientRect();
						const x = Math.max(0, Math.min(Math.round(e.clientX - canvasRect.left), 359));
						this.$hslHue.value = x;
						this.$hslHueHandle.style.left = `${x}px`;
						this.hslInputChange(null, true);
					};
					const documentPointerUp = e => {
						document.removeEventListener('pointermove', documentPointerMove);
						document.removeEventListener('pointerup', documentPointerUp);
					};
					document.addEventListener('pointermove', documentPointerMove);
					document.addEventListener('pointerup', documentPointerUp);
					documentPointerMove(e);
				} },
					this.$hslHueCanvas = $.canvas({ width: 360, height: 20 }),
					this.$hslHueHandle = $.span({ class: 'handle' })
				),
				this.$hslSaturationLightnessPicker = $.div({ id: 'hsl-saturationlightness-picker', pointerdown: e => {
					const documentPointerMove = e => {
						const canvasRect = this.$hslSaturationLightnessCanvas.getBoundingClientRect();
						const x = Math.max(0, Math.min(Math.round(e.clientX - canvasRect.left), 100));
						const y = Math.max(0, Math.min(Math.round(e.clientY - canvasRect.top), 100));
						this.$hslSaturation.value = x;
						this.$hslLightness.value = 100 - y;
						this.hslInputChange(null, true);
					};
					const documentPointerUp = e => {
						document.removeEventListener('pointermove', documentPointerMove);
						document.removeEventListener('pointerup', documentPointerUp);
					};
					document.addEventListener('pointermove', documentPointerMove);
					document.addEventListener('pointerup', documentPointerUp);
					documentPointerMove(e);
				} },
					this.$hslSaturationLightnessCanvas = $.canvas({ width: 100, height: 100 }),
					this.$hslSaturationLightnessHandle = $.span({ class: 'handle' })
				),
				this.$opacityPicker = $.div({ id: 'opacity-picker', pointerdown: e => {
					const documentPointerMove = e => {
						const canvasRect = this.$opacityPicker.getBoundingClientRect();
						const x = Math.max(0, Math.min(Math.round(e.clientX - canvasRect.left), 100));
						this.$opacity.value = x;
						this.opacityInputChange();
					};
					const documentPointerUp = e => {
						document.removeEventListener('pointermove', documentPointerMove);
						document.removeEventListener('pointerup', documentPointerUp);
					};
					document.addEventListener('pointermove', documentPointerMove);
					document.addEventListener('pointerup', documentPointerUp);
					documentPointerMove(e);
				} },
					this.$opacityHandle = $.span({ class: 'handle' })
				),
				// RGB
				$.label({ for: 'red', text: 'Red' }),
				this.$red = $.input({ id: 'red', type: 'text', value: '0', title: '0 to 255', keydown: onlyPositiveIntegers, input: this.rgbInputChange.bind(this), change: this.rgbInputChange.bind(this), blur: clamp(0, 255) }),
				$.label({ for: 'green', text: 'Green' }),
				this.$green = $.input({ id: 'green', type: 'text', value: '0', title: '0 to 255', keydown: onlyPositiveIntegers, input: this.rgbInputChange.bind(this), change: this.rgbInputChange.bind(this), blur: clamp(0, 255) }),
				$.label({ for: 'blue', text: 'Blue' }),
				this.$blue = $.input({ id: 'blue', type: 'text', value: '0', title: '0 to 255', keydown: onlyPositiveIntegers, input: this.rgbInputChange.bind(this), change: this.rgbInputChange.bind(this), blur: clamp(0, 255) }),
				// HSV
				$.label({ for: 'hsv-hue', text: 'Hue' }),
				this.$hsvHue = $.input({ id: 'hsv-hue', type: 'text', value: '0', title: '0 to 359', keydown: onlyPositiveIntegers, input: this.hsvInputChange.bind(this), change: this.hsvInputChange.bind(this), blur: clamp(0, 359) }),
				$.label({ for: 'hsv-saturation', text: 'Saturation' }),
				this.$hsvSaturation = $.input({ id: 'hsv-saturation', type: 'text', value: '0', title: '0 to 100', keydown: onlyPositiveIntegers, input: this.hsvInputChange.bind(this), change: this.hsvInputChange.bind(this), blur: clamp(0, 100) }),
				$.label({ for: 'hsv-value', text: 'Value' }),
				this.$hsvValue = $.input({ id: 'hsv-value', type: 'text', value: '0', title: '0 to 100', keydown: onlyPositiveIntegers, input: this.hsvInputChange.bind(this), change: this.hsvInputChange.bind(this), blur: clamp(0, 100) }),
				// HSL
				$.label({ for: 'hsl-hue', text: 'Hue' }),
				this.$hslHue = $.input({ id: 'hsl-hue', type: 'text', value: '0', title: '0 to 359', keydown: onlyPositiveIntegers, input: this.hslInputChange.bind(this), change: this.hslInputChange.bind(this), blur: clamp(0, 359) }),
				$.label({ for: 'hsl-saturation', text: 'Saturation' }),
				this.$hslSaturation = $.input({ id: 'hsl-saturation', type: 'text', value: '0', title: '0 to 100', keydown: onlyPositiveIntegers, input: this.hslInputChange.bind(this), change: this.hslInputChange.bind(this), blur: clamp(0, 100) }),
				$.label({ for: 'hsl-lightness', text: 'Lightness' }),
				this.$hslLightness = $.input({ id: 'hsl-lightness', type: 'text', value: '0', title: '0 to 100', keydown: onlyPositiveIntegers, input: this.hslInputChange.bind(this), change: this.hslInputChange.bind(this), blur: clamp(0, 100) }),
				// Opacity
				$.label({ for: 'opacity', text: 'Alpha' }),
				this.$opacity = $.input({ id: 'opacity', type: 'text', value: '0', title: '0 to 100', keydown: onlyPositiveIntegers, input: this.opacityInputChange.bind(this), change: this.opacityInputChange.bind(this), blur: clamp(0, 100) }),
				$.span({ id: 'percent', text: '%' }),
				// Hex
				$.div({ css: {
					gridColumn: 'span 2',
					display: 'flex',
					flexFlow: 'row nowrap',
					alignItems: 'center',
					justifyContent: 'flex-end'
				} },
					$.span({ id: 'pound', text: '#' }),
					this.$hex = $.input({ id: 'hex', type: 'text', value: '', title: 'Hexadecimal format, example: ffffff', keydown: onlyHex, paste: onlyHex, input: this.hexInputChange.bind(this), change: this.hexInputChange.bind(this), blur: () => {
						if (this.$hex.value.length > 6) {
							this.$hex.value = this.$hex.value.slice(0, 6);
						} else if (this.$hex.value.length != 6) {
							const { r, g, b } = rgbaToColor(this.$color.internalNewValue);
							this.$hex.value = rgbToHex(r, g, b);
						}
					} })
				)
			),
			this.$transparentButton = $.button({ id: 'transparent-button', title: 'Transparent', click: () => {
				this.$color.internalNewValue = 'rgba(255, 255, 255, 0)';
				this.update();
				this.$color.hidePicker();
			}, pointerenter: () => {
				lastColor = this.$color.internalNewValue;
				this.$color.internalNewValue = 'rgba(255, 255, 255, 0)';
				this.update();
			}, pointerleave: () => {
				this.$color.internalNewValue = lastColor;
				this.update();
			} }),
			this.$palette = $.div({ id: 'color-palette' })
		);
		const hsvHueContext = this.$hsvHueCanvas.getContext('2d');
		const hslHueContext = this.$hslHueCanvas.getContext('2d');
		for (var x = 0; x < 360; ++x) {
			const { r, g, b } = HSVtoRGB(x / 360, 1, 1);
			hsvHueContext.fillStyle = `rgb(${r},${g},${b})`;
			hsvHueContext.fillRect(x, 0, 1, 20);
			hslHueContext.fillStyle = `rgb(${r},${g},${b})`;
			hslHueContext.fillRect(x, 0, 1, 20);
		}
		movable(this, { handle: this.$move, move: (left, top) => {
			this.$color.movedLeft = left;
			this.$color.movedTop = top;
		} });
	}
	connectedCallback() {
		if (!localStorage.hasOwnProperty('colordefaultview')) {
			localStorage['colordefaultview'] = 'hsv';
		}
		this.$options.hidden = this.$color.showPaletteOnly;
		this.$body.hidden = this.$color.showPaletteOnly;
		this.$transparentButton.hidden = !this.$color.transparentButton;
		this.$palette.hidden = this.$color.palette == null && !this.$color.showPalette;
		this.loadPalette();
		if (this.$color.showPaletteOnly) {
			// The palette is already visible
		} else if (localStorage['colordefaultview'] == 'hsv') {
			this.showHSV();
		} else {
			this.showHSL();
		}
	}
	renderSaturationValue() {
		const context = this.$hsvSaturationValueCanvas.getContext('2d');
		const imageData = context.getImageData(0, 0, 100, 100);
		const pixels = imageData.data;
		const hue = parseInt(this.$hsvHue.value, 10) / 360;
		for (var x = 0; x < 100; ++x) {
			for (var y = 0; y < 100; ++y) {
				const offset = (y * 100 + x) * 4;
				const { r, g, b } =  HSVtoRGB(hue, x / 100, 1 - y / 100);
				pixels[offset] = r;
				pixels[offset + 1] = g;
				pixels[offset + 2] = b;
				pixels[offset + 3] = 255;
			}
		}
		context.putImageData(imageData, 0, 0);
	}
	renderSaturationLightness() {
		const context = this.$hslSaturationLightnessCanvas.getContext('2d');
		const imageData = context.getImageData(0, 0, 100, 100);
		const pixels = imageData.data;
		const hue = parseInt(this.$hslHue.value, 10) / 360;
		for (var x = 0; x < 100; ++x) {
			for (var y = 0; y < 100; ++y) {
				const offset = (y * 100 + x) * 4;
				const { r, g, b } =  HSLtoRGB(hue, x / 100, 1 - y / 100);
				pixels[offset] = r;
				pixels[offset + 1] = g;
				pixels[offset + 2] = b;
				pixels[offset + 3] = 255;
			}
		}
		context.putImageData(imageData, 0, 0);
	}
	rgbInputChange() {
		const components = [this.$red, this.$green, this.$blue].map($input => parseInt($input.value, 10));
		if (isNaN(parseFloat(this.$opacity.value))) {
			this.$opacity.value = 100;
		}
		if (components.every(component => !isNaN(component))) {
			let [r, g, b] = components;
			r = Math.max(0, Math.min(r, 255));
			g = Math.max(0, Math.min(g, 255));
			b = Math.max(0, Math.min(b, 255));
			this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(parseFloat(this.$opacity.value) / 100).toFixed(3)})`;
			this.update('rgb');
		}
	}
	hsvInputChange(e, lockHue = false) {
		const components = [this.$hsvHue, this.$hsvSaturation, this.$hsvValue].map($input => parseInt($input.value, 10));
		if (isNaN(parseFloat(this.$opacity.value))) {
			this.$opacity.value = 100;
		}
		if (components.every(component => !isNaN(component))) {
			let [h, s, v] = components;
			h = Math.max(0, Math.min(h, 359));
			s = Math.max(0, Math.min(s, 100));
			v = Math.max(0, Math.min(v, 100));
			const { r, g, b } = HSVtoRGB(h / 360, s / 100, v / 100);
			this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(parseFloat(this.$opacity.value) / 100).toFixed(3)})`;
			this.update('hsv', { lockHue });
		}
	}
	hslInputChange(e, lockHue = false) {
		const components = [this.$hslHue, this.$hslSaturation, this.$hslLightness].map($input => parseInt($input.value, 10));
		if (isNaN(parseFloat(this.$opacity.value))) {
			this.$opacity.value = 100;
		}
		if (components.every(component => !isNaN(component))) {
			let [h, s, l] = components;
			h = Math.max(0, Math.min(h, 359));
			s = Math.max(0, Math.min(s, 100));
			l = Math.max(0, Math.min(l, 100));
			const { r, g, b } = HSLtoRGB(h / 360, s / 100, l / 100);
			this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(parseFloat(this.$opacity.value) / 100).toFixed(3)})`;
			this.update('hsl', { lockHue });
		}
	}
	opacityInputChange() {
		const components = [this.$red, this.$green, this.$blue].map($input => parseInt($input.value, 10));
		let opacity = parseFloat(this.$opacity.value);
		if (components.every(component => !isNaN(component)) && !isNaN(opacity)) {
			const [r, g, b] = components;
			opacity = Math.max(0, Math.min(opacity, 100));
			this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(opacity / 100).toFixed(3)})`;
			this.update('opacity');
		}
	}
	hexInputChange() {
		const hex = this.$hex.value;
		// rrggbb, #rrggbb, rrggbbaa, or #rrggbbaa
		const matches = hex.match(/^#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})?$/i);
		if (matches) {
			let [r, g, b, a] = matches.slice(1).map(v => v == null ? null : parseInt(v, 16));
			if (!isNaN(r) && !isNaN(g) && !isNaN(b) && (a == null || !isNaN(a))) {
				if (a == null) {
					this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(this.$opacity.value / 100).toFixed(3)})`;
				} else {
					a = Math.max(0, Math.min(Math.round(a / 255 * 100), 100));
					this.$color.internalNewValue = `rgba(${r}, ${g}, ${b}, ${+(a / 100).toFixed(3)})`;
				}
				this.update('pickerhex');
			}
		}
	}
	update(from = null, { lockHue = false } = {}) {
		const { r, g, b, a } = rgbaToColor(this.$color.internalNewValue);
		// Opacity
		this.$opacityHandle.style.left = `${a}px`;
		if (from == 'opacity') {
			this.$color.setValueText(this.$color.internalNewValue);
			return;
		} else {
			this.$opacity.value = a;
		}
		// RGB
		if (from != 'rgb') {
			this.$red.value = r;
			this.$green.value = g;
			this.$blue.value = b;
		}
		// HSV
		const { h:hsvHue, s:hsvSaturation, v:hsvValue } = RGBtoHSV(r, g, b);
		if (from != 'hsv') {
			if (!lockHue) {
				this.$hsvHue.value = hsvHue;
				this.$hsvHueHandle.style.left = `${hsvHue}px`;
			}
			this.$hsvSaturation.value = hsvSaturation;
			this.$hsvValue.value = hsvValue;
			Object.assign(this.$hsvSaturationValueHandle.style, {
				left: `${hsvSaturation}px`,
				top: `${100 - hsvValue}px`
			});
		} else {
			this.$hsvHueHandle.style.left = `${Math.max(0, Math.min(parseInt(this.$hsvHue.value, 10), 359))}px`;
			Object.assign(this.$hsvSaturationValueHandle.style, {
				left: `${Math.max(0, Math.min(parseInt(this.$hsvSaturation.value, 10), 100))}px`,
				top: `${Math.max(0, Math.min(100 - parseInt(this.$hsvValue.value, 10), 100))}px`
			});
		}
		if (from == 'hsv' && lockHue) {
			this.$hslHue.value = this.$hsvHue.value;
			this.$hslHueHandle.style.left = this.$hsvHueHandle.style.left;
		}
		if (this.$showHSV.matches('.selected')) {
			this.renderSaturationValue();
		}
		// HSL
		const { h:hslHue, s:hslSaturation, l:hslLightness } = RGBtoHSL(r, g, b);
		if (from != 'hsl') {
			if (!lockHue) {
				this.$hslHue.value = hslHue;
				this.$hslHueHandle.style.left = `${hslHue}px`;
			}
			this.$hslSaturation.value = hslSaturation;
			this.$hslLightness.value = hslLightness;
			Object.assign(this.$hslSaturationLightnessHandle.style, {
				left: `${hslSaturation}px`,
				top: `${100 - hslLightness}px`
			});
		} else {
			this.$hslHueHandle.style.left = `${Math.max(0, Math.min(parseInt(this.$hslHue.value, 10), 359))}px`;
			Object.assign(this.$hslSaturationLightnessHandle.style, {
				left: `${Math.max(0, Math.min(parseInt(this.$hslSaturation.value, 10), 100))}px`,
				top: `${Math.max(0, Math.min(100 - parseInt(this.$hslLightness.value, 10), 100))}px`
			});
		}
		if (from == 'hsl' && lockHue) {
			this.$hsvHue.value = this.$hslHue.value;
			this.$hsvHueHandle.style.left = this.$hslHueHandle.style.left;
		}
		if (this.$showHSL.matches('.selected')) {
			this.renderSaturationLightness();
		}
		if (from != 'hex') {
			this.$color.setValueText(this.$color.internalNewValue);
		}
		if (from != 'pickerhex') {
			const { r, g, b } = rgbaToColor(this.$color.internalNewValue);
			this.$hex.value = rgbToHex(r, g, b);
		}
		this.$color.setBackgroundColor(this.$color.internalNewValue);
		this.$color.callBouncePreviewChange();
	}
	showHSVHSL(showHSV) {
		for (const $e of [
			this.$hsvHuePicker,
			this.$hsvSaturationValuePicker,
			this.$hsvHue.previousElementSibling, this.$hsvHue,
			this.$hsvSaturation.previousElementSibling, this.$hsvSaturation,
			this.$hsvValue.previousElementSibling, this.$hsvValue
		]) {
			$e.hidden = !showHSV;
		}
		for (const $e of [
			this.$hslHuePicker,
			this.$hslSaturationLightnessPicker,
			this.$hslHue.previousElementSibling, this.$hslHue,
			this.$hslSaturation.previousElementSibling, this.$hslSaturation,
			this.$hslLightness.previousElementSibling, this.$hslLightness
		]) {
			$e.hidden = showHSV;
		}
		this.$showHSV.classList.toggle('selected', showHSV);
		this.$showHSL.classList.toggle('selected', !showHSV);
	}
	showHSV() {
		localStorage['colordefaultview'] = 'hsv';
		this.showHSVHSL(true);
		this.renderSaturationValue();
	}
	showHSL() {
		localStorage['colordefaultview'] = 'hsl';
		this.showHSVHSL(false);
		this.renderSaturationLightness();
	}
	loadPalette() {
		this.$palette.innerHTML = '';
		let lastColor;
		const generateButton = color => {
			const { r, g, b, a } = rgbaToColor(color);
			return $.button({ title: `#${rgbToHex(r, g, b)}${a == 100 ? '' : ` ${a}%`} ${color}`, css: { backgroundColor: color }, click: () => {
				this.$color.internalNewValue = color;
				this.update();
				this.$color.hidePicker();
			}, pointerenter: () => {
				lastColor = this.$color.internalNewValue;
				this.$color.internalNewValue = color;
				this.update();
			}, pointerleave: () => {
				this.$color.internalNewValue = lastColor;
				this.update();
			} })
		};
		if (this.$color.palette) {
			this.$palette.append(
				...this.$color.palette.map(color => generateButton(color))
			);
		}
		if (!this.$color.showPaletteOnly && this.$color.localStorageKey) {
			if (!localStorage.hasOwnProperty(this.$color.localStorageKey)) {
				localStorage[this.$color.localStorageKey] = JSON.stringify([]);
			}
			let palette = [];
			try {
				palette = JSON.parse(localStorage[this.$color.localStorageKey]);
			} catch (exception) {
				localStorage[this.$color.localStorageKey] = JSON.stringify([]);
			}
			this.$palette.append(
				...palette.map(color => generateButton(color))
			);
		}
	}
});

function rgbToHex(r, g, b) {
	return `${('0' + r.toString(16)).slice(-2)}${('0' + g.toString(16)).slice(-2)}${('0' + b.toString(16)).slice(-2)}`;
}

function RGBtoHSV(r, g, b) {
	let rabs, gabs, babs, rr, gg, bb, h, s, v, diff, diffc;
	rabs = r / 255;
	gabs = g / 255;
	babs = b / 255;
	v = Math.max(rabs, gabs, babs),
	diff = v - Math.min(rabs, gabs, babs);
	diffc = c => (v - c) / 6 / diff + 1 / 2;
	if (diff == 0) {
		h = s = 0;
	} else {
		s = diff / v;
		rr = diffc(rabs);
		gg = diffc(gabs);
		bb = diffc(babs);

		if (rabs === v) {
			h = bb - gg;
		} else if (gabs === v) {
			h = (1 / 3) + rr - bb;
		} else if (babs === v) {
			h = (2 / 3) + gg - rr;
		}
		if (h < 0) {
			h += 1;
		} else if (h >= 1) {
			h -= 1;
		}
	}
	return {
		h: Math.floor(h * 360),
		s: Math.round(s * 100),
		v: Math.round(v * 100)
	};
}

function HSVtoRGB(h, s, v) {
	var r, g, b, i, f, p, q, t;
	if (h && s === undefined && v === undefined) {
		s = h.s, v = h.v, h = h.h;
	}
	i = Math.floor(h * 6);
	f = h * 6 - i;
	p = v * (1 - s);
	q = v * (1 - f * s);
	t = v * (1 - (1 - f) * s);
	switch (i % 6) {
		case 0: r = v, g = t, b = p; break;
		case 1: r = q, g = v, b = p; break;
		case 2: r = p, g = v, b = t; break;
		case 3: r = p, g = q, b = v; break;
		case 4: r = t, g = p, b = v; break;
		case 5: r = v, g = p, b = q; break;
	}
	return {
		r: Math.min(Math.max(0, Math.round(r * 255)), 255),
		g: Math.min(Math.max(0, Math.round(g * 255)), 255),
		b: Math.min(Math.max(0, Math.round(b * 255)), 255)
	};
}

/**
 * Converts an HSL color value to RGB. Conversion formula
 * adapted from http://en.wikipedia.org/wiki/HSL_color_space.
 * Assumes h, s, and l are contained in the set [0, 1] and
 * returns r, g, and b in the set [0, 255].
 *
 * @param   {number}  h       The hue
 * @param   {number}  s       The saturation
 * @param   {number}  l       The lightness
 * @return  {object}           The RGB representation
 */
function hue2rgb(p, q, t) {
	if (t < 0) t += 1;
	if (t > 1) t -= 1;
	if (t < 1 / 6) return p + (q - p) * 6 * t;
	if (t < 1 / 2) return q;
	if (t < 2 / 3) return p + (q - p) * (2 / 3 - t) * 6;
	return p;
}
function HSLtoRGB(h, s, l) {
	var r, g, b;

	if (s == 0) {
		r = g = b = l; // achromatic
	} else {
		var q = l < 0.5 ? l * (1 + s) : l + s - l * s;
		var p = 2 * l - q;
		r = hue2rgb(p, q, h + 1 / 3);
		g = hue2rgb(p, q, h);
		b = hue2rgb(p, q, h - 1 / 3);
	}
	
	return {
		r: Math.round(r * 255),
		g: Math.round(g * 255),
		b: Math.round(b * 255)
	};
}

/**
 * Converts an RGB color value to HSL. Conversion formula
 * adapted from http://en.wikipedia.org/wiki/HSL_color_space.
 * Assumes r, g, and b are contained in the set [0, 255] and
 * returns h, s, and l in the set [0, 1].
 *
 * @param   {number}  r       The red color value
 * @param   {number}  g       The green color value
 * @param   {number}  b       The blue color value
 * @return  {Array}           The HSL representation
 */
function RGBtoHSL(r, g, b) {
	r /= 255, g /= 255, b /= 255;
	var max = Math.max(r, g, b), min = Math.min(r, g, b);
	var h, s, l = (max + min) / 2;
	
	if(max == min){
		h = s = 0; // achromatic
	}else{
		var d = max - min;
		s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
		switch(max){
			case r:
				h = (g - b) / d + (g < b ? 6 : 0);
				break;
			case g:
				h = (b - r) / d + 2;
				break;
			case b:
				h = (r - g) / d + 4;
				break;
		}
		h /= 6;
	}

	return {
		h: Math.floor(h * 360),
		s: Math.round(s * 100),
		l: Math.round(l * 100)
	};
}

function hexToRGBA(hex, alpha) {
	return `rgba(${[parseInt(hex.slice(1, 3), 16), parseInt(hex.slice(3, 5), 16), parseInt(hex.slice(5, 7), 16), alpha == undefined ? 1 : alpha].join(',')})`;
}

function rgbToColor(rgb) {
	const matches = rgb.match(/^rgb\(([0-9]+) *, *([0-9]+) *, *([0-9]+)\)$/);
	if (matches && matches.length == 4) {
		return {
			r: Math.min(Math.max(0, parseInt(matches[1], 10)), 255),
			g: Math.min(Math.max(0, parseInt(matches[2], 10)), 255),
			b: Math.min(Math.max(0, parseInt(matches[3], 10)), 255)
		};
	} else {
		return {
			r: 0,
			g: 0,
			b: 0
		};
	}
}