customElements.define('ui-inputselect', class UIInputSelect extends HTMLElement {
	constructor() {
		super();
		this.optionsSet = false;
	}
	options({ options, ...config } = {}) {
		if (options == undefined) {
			throw new Error('options must be set for ui-inputselect');
		}
		this.options = options;
		config = Object.assign({
			history: 0,
			processOption: () => {},
			allowInput: false
		}, config);
		for (const key in config) {
			this[key] = config[key];
		}
		// TODO: Restructure to allow options to be called again. For now this is easier as no input select needs this
		if (this.optionsSet) {
			throw new Error('options already called');
		}
		this.optionsSet = true;
		this.promise = null;
		
		this.tabIndex = -1;
		
		let uniqueCharacters = '';
		for (const option of options) {
			for (const character of [...option.name.toString()]) {
				if (uniqueCharacters.indexOf(character) == -1) {
					uniqueCharacters += character;
				}
			}
		}
		const hash = s => {
			var hash = 0, i, chr;
			if (s.length === 0) {
				return hash;
			}
			for (i = 0; i < s.length; i++) {
				chr = s.charCodeAt(i);
				hash = ((hash << 5) - hash) + chr;
				hash |= 0; // Convert to 32bit integer
			}
			return hash;
		};
		// Stringify the options and use it as a kind of hash to deduplicate pickers. We don't want to create 100 font size pickers as that would create thousands of DOM elements
		this.optionsHash = hash(JSON.stringify(options));
		this.localStorageHistoryKey = `${this.optionsHash}-history`;
		
		// Get the unique picker
		this.picker = null;
		for (const $inputSelectPicker of document.querySelectorAll('ui-inputselectpicker')) {
			if ($inputSelectPicker.optionsHash == this.optionsHash) {
				this.picker = $inputSelectPicker;
				break;
			}
		}
		
		// If there isn't an existing picker then create a new one
		if (this.picker == null) {
			this.picker = new UIInputSelectPicker(this);
			document.body.appendChild(this.picker);
		}
		this.picker.register(this);
		
		let lastKey = null;
		this.attachShadow({mode: 'open', delegatesFocus: true });
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/materialdesignicons.min.css');

:host {
	display: flex;
	flex-flow: row nowrap;
	align-items: center;
	color: var(--input-color);
	background-color: var(--input-background-color);
	border: 1px solid var(--input-border-color);
	height: 32px;
	border-radius: 4px;
	box-sizing: border-box;
}
:host([disabled]) {

}
:host(:focus) {
	outline: none;
}
:host > span {
	flex: 1 0 0;
	min-width: 0;
	padding: 0 5px;
	font-size: 16px;
	white-space: nowrap;
}

:host > span:empty::before {
	content: attr(placeholder);
	opacity: 0.5;
	pointer-events: none;
	width: 0;
}
:host > span:focus {
	outline: none;
}
:host > button {
	flex: 0 0 auto;
	font-size: 14px;
	align-self: stretch;
	display: flex;
	align-items: center;
	cursor: pointer;
	padding: 0 2px;
	user-select: none;
	background-color: transparent;
	border: none;
}
:host > button:focus {
	border: none;
	outline: none;
}
:host > button:hover {
	opacity: 0.8;
}
</style>`;
		this.shadowRoot.append(
			this.$input = $.span({ contenteditable: 'true', spellcheck: false, tabindex: -1, placeholder: this.getAttribute('placeholder') ? this.getAttribute('placeholder') : '', css: {
				flex: `1 1 ${this.picker.clientWidth}px`
			}, keydown: e => {
				lastKey = e.key;
				if (e.key == 'Enter') {
					e.preventDefault();
					this.$input.blur();
					return;
				}
				let name = this.$input.innerText;
				if (e.key == 'ArrowUp') {
					let optionIndex = options.findIndex(option => option.name == name);
					if (optionIndex != -1 && optionIndex != 0) {
						this.value = options[optionIndex - 1].value;
						this.showPicker();
					}
				}
				if (e.key == 'ArrowDown') {
					let optionIndex = options.findIndex(option => option.name == name);
					if (optionIndex != -1 && optionIndex != options.length - 1) {
						this.value = options[optionIndex + 1].value;
						this.showPicker();
					}
				}
				if (e.key != 'Backspace' && e.key != 'Delete' && e.key != 'ArrowLeft' && e.key != 'ArrowRight' && !this.allowInput && uniqueCharacters.indexOf(e.key) == -1) {
					e.preventDefault();
				}
			}, input: () => {
				if (lastKey == 'Backspace' || lastKey == 'Delete') {
					return;
				}
				const name = this.$input.innerText;
				for (const option of options) {
					const optionName = option.name.toString();
					if (optionName.toLowerCase().indexOf(name.toLowerCase()) == 0 && name.length <= optionName.length) {
						if (optionName.toLowerCase() != name.toLowerCase()) {
							// Replace the text with the full option name and then select the last part
							this.value = option.value;
							var selection = window.getSelection();
							selection.removeAllRanges();
							var range = document.createRange();
							range.setStart(this.$input.firstChild, name.length);
							range.setEnd(this.$input.firstChild, optionName.length);
							selection.addRange(range);
						}
						this.picker.setSelected(optionName);
						break;
					}
				}
			}, paste: function(e) {
				e.preventDefault();
				let pastedText = e.clipboardData.getData('text/plain');
				if (!this.allowInput) {
					pastedText = pastedText.replace(new RegExp(`[^${uniqueCharacters.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&')}]`, 'g'),'');
				}
				const oldText = this.innerText;
				const selection = window.getSelection();
				const range = selection.getRangeAt(0);
				const beforeSelection = oldText.substring(0, range.startOffset);
				const afterSelection = oldText.substring(range.endOffset, oldText.length);
				this.innerText = `${beforeSelection}${pastedText}${afterSelection}`;
				selection.removeAllRanges();
				const newRange = document.createRange();
				newRange.setStart(this.firstChild, (`${beforeSelection}${pastedText}`).length);
				newRange.collapse();
				selection.addRange(newRange);
			} }),
			this.$arrow = $.button({ class: 'mdi mdi-play mdi-rotate-90', pointerdown: () => {
				if (this.promise) {
					this.picker.hide();
				} else {
					this.showPicker();
				}
			} })
		);
		this.addEventListener('blur', () => {
			const name = this.$input.innerText;
			// Convert the string name to a value. If allowInput is set then use the name if it can't be converted to a value
			if (this.picker.nameToValue.hasOwnProperty(name.toLowerCase())) {
				const value = this.picker.nameToValue[name.toLowerCase()];
				this.picker.addValueToHistory(value);
				this.value = value;
			} else if (this.allowInput) {
				this.value = name;
			}
			this.dispatchEvent(new CustomEvent('change'));
			this.picker.hide();
		});
		this.picker.hide();
	}
	connectedCallback() {
	}
	disconnectedCallback() {
		this.picker.unregister(this);
	}
	attributeChangedCallback(name, oldValue, newValue) {
		if (name == 'placeholder') {
			this.$input.setAttribute('placeholder', newValue);
		}
	}
	get value() {
		return this.innerValue;
	}
	set value(value) {
		this.innerValue = value;
		this.$input.innerText = value == '' && !this.picker.valueToName.hasOwnProperty(value) ? '' : (this.picker.valueToName.hasOwnProperty(value) ? this.picker.valueToName[value] : value);
		this.dispatchEvent(new CustomEvent('textchanged'));
	}
	async showPicker() {
		this.picker.setSelected(this.$input.innerText);
		if (this.promise) {
			return this.promise;
		}
		const thisOffset = $.offset(this);
		this.promise = this.picker.show(thisOffset.left, thisOffset.top, this.offsetWidth, this.offsetHeight);
		const selection = window.getSelection();
		selection.removeAllRanges();
		if (this.$input.firstChild) {
			var range = document.createRange();
			range.setStart(this.$input.firstChild, 0);
			range.setEnd(this.$input.firstChild, this.$input.innerText.length);
			selection.addRange(range);
		} else {
			selection.selectAllChildren(this.$input);
		}
		const result = await this.promise;
		this.promise = null;
		if (result != null) {
			this.$input.innerText = result.name;
			this.value = result.value;
			this.picker.addValueToHistory(result.value);
		} else {
			// User closed the dialog without selecting anything
		}
	}
});

class UIInputSelectPicker extends HTMLElement {
	constructor($inputSelect) {
		super();
		this.inputSelects = new Set();
		this.optionsHash = $inputSelect.optionsHash;
		this.processOption = $inputSelect.processOption;
		this.localStorageHistoryKey = $inputSelect.localStorageHistoryKey;
		this.history = $inputSelect.history;
		this.attachShadow({ mode: 'open' });
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/style.css');
:host {
	position: absolute;
	max-width: 300px;
	min-width: 50px;
	max-height: 50%;
	display: flex;
	flex-flow: column nowrap;
	overflow-y: auto;
	border: 1px solid var(--input-border-color);
	background-color: var(--input-background-color);
	color: var(--input-color);
	z-index: 1;
}

:host(:not(:hover)) ::slotted(ui-inputselect-option.selected) {
	background-color: var(--select-selection-color);
	color: var(--font-color);
}
</style>
<slot></slot>`;
		this.nameToValue = {};
		this.valueToName = {};
		this.nameToTag = {};
		this.valueToTag = {};
		for (const option of $inputSelect.options) {
			this.nameToValue[option.name.toString().toLowerCase()] = option.value;
			this.valueToName[option.value] = option.name;
			const $option = new UIInputSelectOption(option.name, option.value);
			this.processOption.call($option);
			this.appendChild($option);
			this.nameToTag[option.name.toString().toLowerCase()] = $option;
			this.valueToTag[option.value] = $option;
		}
		this.addEventListener('click', e => e.stopPropagation());
		this.rebuildHistory();
		
		this.addEventListener('pointerdown', e => {
			if (e.target instanceof UIInputSelectOption) {
				this.resolve({ name: e.target.name, value: e.target.value });
				return true;
			}
			return false;
		});
	}
	// Reference count the input select instances that use this picker
	register($inputSelect) {
		this.inputSelects.add($inputSelect);
	}
	unregister($inputSelect) {
		this.inputSelects.delete($inputSelect);
		if (this.inputSelects.size == 0) {
			this.remove();
		}
	}
	async show(x, y, width, height) {
		this.style.display = '';
		this.style.left = `${x}px`,
		this.style.top = `${y + (y + height + this.offsetHeight < document.documentElement.clientHeight ? height : -this.offsetHeight)}px`;
		return new Promise(resolve => this.resolve = resolve);
	}
	hide() {
		this.style.display = 'none';
		if (this.resolve) {
			this.resolve(null);
		}
	}
	getLocalStorageHistory() {
		if (!localStorage.hasOwnProperty(this.localStorageHistoryKey)) {
			localStorage[this.localStorageHistoryKey] = JSON.stringify([]);
		}
		return JSON.parse(localStorage[this.localStorageHistoryKey]);
	}
	setLocalStorageHistory(value) {
		localStorage[this.localStorageHistoryKey] = JSON.stringify(value);
	}
	rebuildHistory() {
		this.historyNameToTag = {};
		if (this.history != 0) {
			for (const $option of this.querySelectorAll('.history, .history-spacer')) {
				$option.remove();
			}
			const history = this.getLocalStorageHistory();
			if (history.length != 0) {
				const $option = new UIInputSelectOption('-------------', '');
				$option.classList.add('history-spacer');
				this.prepend($option);
			}
			history.reverse();
			
			for (const value of history) {
				const $option = new UIInputSelectOption(this.valueToName[value], value);
				$option.classList.add('history');
				this.processOption.call($option);
				this.prepend($option);
				this.historyNameToTag[this.valueToName[value].toString().toLowerCase()] = $option;
			}
		}
	}
	addValueToHistory(value) {
		if (this.history != 0) {
			let history = this.getLocalStorageHistory();
			var index = history.indexOf(value);
			if (index != -1) {
				history.splice(index, 1);
			}
			history.unshift(value);
			history = history.slice(0, this.history);
			this.setLocalStorageHistory(history);
			this.rebuildHistory();
		}
	}
	setSelected(name) {
		// Set the selected value in the picker and scroll to it
		this.querySelector(':scope > .selected')?.classList.remove('selected');
		const nameLowerCase = name.toString().toLowerCase();
		if (this.nameToTag.hasOwnProperty(nameLowerCase)) {
			// History has a higher precedence than other options
			const $option = this.historyNameToTag.hasOwnProperty(nameLowerCase) ? this.historyNameToTag[nameLowerCase] : this.nameToTag[nameLowerCase];
			$option.classList.add('selected');
			const optionTop = $option.offsetTop;
			if (optionTop < 0) {
				this.scrollTop = optionTop + this.scrollTop;
			} else if (optionTop + $option.offsetHeight > this.offsetHeight) {
				this.scrollTop = optionTop + $option.offsetHeight + this.scrollTop - this.offsetHeight;
			}
		}
	}
}
customElements.define('ui-inputselectpicker', UIInputSelectPicker);

class UIInputSelectOption extends HTMLElement {
	constructor(name, value) {
		super();
		this.innerText = name;
		this.value = value;
		this.attachShadow({ mode: 'open' });
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/style.css');
:host {
	flex: 0 0 auto;
	padding: 0px 4px !important;
	font-size: 16px;
	min-width: 50px;
	line-height: 24px;
	cursor: pointer;
	white-space: nowrap;
	overflow: hidden;
	text-overflow: ellipsis;
}
:host(:hover) {
	background-color: var(--select-selection-color);
	color: var(--font-color);
}
:host(.history-spacer) {
	pointer-events: none;
}
</style>
<slot></slot>`;
	}
}
customElements.define('ui-inputselect-option', UIInputSelectOption);