// Options: fullSelect, multiselect, checkbox, propagateCheckbox, groupIcon, leafIcon, tooltip, filter, disable, dragAppendTo, deleteConfirmation, click, doubleClick, saveState, thumbnail, editable, showRoot, defaultOptions, largeImages, filterDraggable, dragHelper, nameTransform, dispatchChangeOnLoad
// Events:
// load: Fired after load and render, can be fired multiple times
// change: Fired when the selection or check state changes
// nochange: Fired when the user selects an already selected item

const uitreeStyle = document.createElement('style');
uitreeStyle.type = 'text/css';
uitreeStyle.append(`
:host {
	list-style: none;
	background-color: var(--background-contrast-color);
	overflow: auto;
	position: relative;
	display: flex;
	flex-flow: column nowrap;
	color: var(--tree-font-color);
	z-index: 0;
	user-select: none;
}
`);

customElements.define('ui-tree', class UITree extends HTMLElement {
	constructor() {
		super();
		this.optionsSet = false;
		this.attachShadow({mode: 'open'}).append(
			uitreeStyle.cloneNode(true),
			this.$slot = $.slot()
		);
		// TODO: Rename this to like $treeselect or something?
		// The ui-treeselect if this tree is in a picker
		this.input = null;
	}
	async options({
		tree:treeName,
		src = null,
		fullSelect = false,
		multiselect = false,
		checkbox = false,
		// 'up' unchecks children if the parent is checked and unchecks parents if the child is checked
		// 'down' checks children if the parent is checked and requires all children to be checked if the parent is checked
		// propagateCheckbox = 'down',
		groupTypes = [1],
		groupIcon = () => '',
		leafIcon = () => '',
		tooltip = null,
		disable = () => false,
		filter = [() => true],
		dragAppendTo = this,
		deleteConfirmation = () => {},
		click = null,
		doubleClick = null,
		saveState = true,
		thumbnail = false,
		overrideThumbnailDimensions = null,
		editable = () => false,
		showRoot = true,
		defaultOptions = [],
		largeImages = false,
		filterDraggable = () => true,
		filterDroppable = () => true,
		dragHelper = (e, dropClass) => {}, // returning undefined will use the default helper which is the cloned selected nodes
		showMinimizeMaximizeAll = true,
		showSmallLargeThumbnail = false,
		nameTransform = node => node.name,
		dispatchChangeOnLoad = false }) {
		// TODO: Restructure to allow options to be called again. For now this is easier as no tree needs this
		if (this.optionsSet) {
			throw new Error('options already called');
		}
		this.optionsSet = true;
		
		const filters = Array.isArray(filter) ? filter : [filter];
		Object.assign(this, {
			treeName,
			src,
			fullSelect,
			multiselect,
			checkbox,
			// propagateCheckbox,
			groupTypes,
			groupIcon,
			leafIcon,
			tooltip,
			disable,
			filters,
			dragAppendTo,
			deleteConfirmation,
			click,
			doubleClick,
			saveState,
			thumbnail,
			overrideThumbnailDimensions,
			editable,
			showRoot,
			defaultOptions,
			largeImages,
			filterDraggable,
			filterDroppable,
			dragHelper,
			showMinimizeMaximizeAll,
			showSmallLargeThumbnail,
			nameTransform,
			dispatchChangeOnLoad
		});
		
		if (!localStorage.hasOwnProperty(`${this.treeName}-thumbnail`)) {
			localStorage[`${this.treeName}-thumbnail`] = true;
		}
		this.thumbnailEnabled = localStorage[`${this.treeName}-thumbnail`] == 'true';
		
		if (this.showSmallLargeThumbnail) {
			if (!localStorage.hasOwnProperty(`${this.treeName}-imagesize`)) {
				localStorage[`${this.treeName}-imagesize`] = false;
			}
			this.largeImages = localStorage[`${this.treeName}-imagesize`] == 'true';
		}
		
		var oldLeafIcon = this.leafIcon;
		this.leafIcon = node => {
			// defaultOptions do not use the leafIcon function
			if (this.defaultOptions.map(node => node.name).indexOf(node.name) == -1) {
				return oldLeafIcon(node);
			} else {
				return node.hasOwnProperty('icon') ? node.icon : '';
			}
		};
		this.classList.add(`tree-${this.treeName}`);
		
		this.tree = [];
		this.selection = [];
		this.focusedNode = null;
		this.lastFocusedNode = null;
		this.shiftNode = null;
		this.currentSyncId;
		this.temporaryOpenStates = null;
		this.lockCount = 0;
		this.lockScrollLeft = 0;
		this.lockScrollTop = 0;
		this.activeRequests = [];
		this.activeXHRs = [];
		this.processingAdds = false;
		this.treeFilterView = null;
		this.searchString = '';
		this.searchStringTimer = null;
		
		this.editTimer = null;
		
		this.loading = null;
		await this.load();
	}
	async connectedCallback() {
		this.addEventListener('keydown', async e => {
			if (e.keyCode == 38) { // up arrow
				this.querySelector('[contenteditable]')?.blur();
				await this.setSelected('previous', e.ctrlKey, e.shiftKey, this.input == null);
				e.preventDefault();
			} else if (e.keyCode == 40) { // down arrow
				this.querySelector('[contenteditable]')?.blur();
				await this.setSelected('next', e.ctrlKey, e.shiftKey, this.input == null);
				e.preventDefault();
			}
		}, false);
		
		this.addEventListener('keyup', e => {
			if (this.querySelector('[contenteditable]') == null) {
				if (e.keyCode == 46) {
					this.deleteSelected();
				}
			}
		}, false);
		
		this.addEventListener('keydown', async e => {
			if (e.key.length == 1 && this.querySelector('[contenteditable]') == null && this.getNonFilteredRoots().length > 0) {
				this.searchString += e.key;
				const selectedNodes = await this.getSelected();
				if (selectedNodes.length > 0) {
					var startNode = selectedNodes[selectedNodes.length - 1];
					// Start searching at the next node if the user is typing one character
					let node = startNode;
					let skipFirstCheck = false;
					if (this.searchString.length == 1) {
						node = node.next ? node.next : this.getNonFilteredRoots()[0];
					} else {
						skipFirstCheck = true;
					}
					for (; skipFirstCheck || node !== startNode; node = node.next ? node.next : this.getNonFilteredRoots()[0]) {
						skipFirstCheck = false;
						if (node.name.toLowerCase().indexOf(this.searchString.toLowerCase()) == 0) {
							await this.setSelected([node.id]);
							break;
						}
					}
				}
				this.bounceSearchString();
			}
		}, false);
		
		// If the user clicks away from the tree then unfocus the focusedNode
		this.addEventListener('focusout', e => {
			if (this.focusedNode && this.focusedNode.$handle === e.target) {
				this.lastFocusedNode = this.focusedNode;
				this.focusedNode = null;
			}
		});
	}
	disconnectedCallback() {
		if (this.hasOwnProperty('activeRequests')) {
			for (const controller of this.activeRequests) {
				controller.abort();
			}
		}
		if (this.hasOwnProperty('activeXHRs')) {
			for (const xhr of this.activeXHRs) {
				xhr.abort();
			}
		}
	}
	attributeChangedCallback() {
		
	}
	get syncId() {
		return this.currentSyncId;
	}
	set syncId(value) {
		this.currentSyncId = value;
	}
	TreeSort(a, b) {
		if (!a.group && this.defaultOptions.map(node => node.name).indexOf(a.name) != -1) {
			return -1;
		} else if (!b.group && this.defaultOptions.map(node => node.name).indexOf(b.name) != -1) {
			return 1;
		} else {
			if (!a.group && !b.group) {
				return TreeSort(a.name, b.name);
			} else if (a.group && !b.group) {
				return -1;
			} else if (!a.group && b.group) {
				return 1;
			} else if (a.group && b.group && a.type != b.type) {
				return a.type - b.type;
			} else {
				return TreeSort(a.name, b.name);
			}
		}
	}
	// This includes roots that are disabled. Essentially this is an array of all visible roots
	getNonFilteredRoots() {
		return this.tree.filter(node => (this.showRoot || !this.showRoot && node.id != 1) && !node.filtered && (this.fullSelect || node.permitted) && node.parentNode == null).sort(this.TreeSort.bind(this));
	}
	// This is all visible roots excluding the ones not selectable
	getSelectableRoots() {
		return this.getNonFilteredRoots().filter(node => !node.disabled).sort(this.TreeSort.bind(this));
	}
	ConstructTreeStructure() {
		for (const node of this.tree) {
			if (!node.hasOwnProperty('filtered')) {
				node.filtered = false;
			}
			if (!node.hasOwnProperty('group')) {
				Object.defineProperty(node, 'group',
				{
					get: () => {
						return this.groupTypes.includes(node.type);
					}
				});
			}
			if (!node.hasOwnProperty('done')) {
				Object.defineProperty(node, 'done',
				{
					get: function() {
						if (!this.hasOwnProperty('data')) {
							return false;
						}
						return this.data.status.done;
					}
				});
			}
			if (!node.hasOwnProperty('error')) {
				Object.defineProperty(node, 'error',
				{
					get: function() {
						if (!this.hasOwnProperty('data')) {
							return false;
						}
						return this.data.status.upload.error || ['original', ...this.data.status.output].some(output => this.data.status[output].convert.error || this.data.status[output].cdn.error);
					}
				});
			}
			if (!node.hasOwnProperty('errorMessage')) {
				Object.defineProperty(node, 'errorMessage',
				{
					get: function() {
						if (this.data.status.upload.error) {
							return this.data.status.upload.error;
						}
						const output = ['original', ...this.data.status.output].find(output => this.data.status[output].convert.error || this.data.status[output].cdn.error);
						if (output == null) {
							return '';
						}
						const convertError = this.data.status[output].convert.error;
						if (convertError) {
							return convertError;
						}
						const cdnError = this.data.status[output].cdn.error;
						if (cdnError) {
							return cdnError;
						}
						return '';
					}
				});
			}
			if (!node.hasOwnProperty('busy')) {
				Object.defineProperty(node, 'busy',
				{
					get: function() {
						return this.treeName == 'content' && !(this.group || this.done || this.error);
					}
				});
			}
		}
		
		for (const node of this.tree) {
			// All subtree nodes
			node.allChildren = [];
			// Direct filtered children with permissions
			node.children = [];
			// The children with no filter or permissions
			node.realChildren = [];
			if (node.group) {
				for (const child of this.tree) {
					if (child.parent == node.id) {
						// node is the parent of child
						if (this.fullSelect || node.permitted && child.permitted) {
							child.parentNode = node;
							if (!child.filtered) {
								node.children.push(child);
							}
						}
						child.realParentNode = node;
						node.realChildren.push(child);
					}
				}
				node.children.sort(this.TreeSort.bind(this));
			}
		}
		
		// If for whatever reason a node doesn't end up with a parent set then set the parent to null
		for (const node of this.tree) {
			if (!node.hasOwnProperty('parentNode')) {
				node.parentNode = null;
			}
			if (!node.hasOwnProperty('realParentNode')) {
				node.realParentNode = null;
			}
		}
		
		if (!this.showRoot) {
			// Hide the real root by setting all the children's parents to null
			const realRoot = this.tree.find(node => node.id == 1);
			if (realRoot) {
				for (const child of realRoot.children) {
					child.parentNode = null;
				}
			}
		}
		
		for (const node of this.tree) {
			if (node.group) {
				for (const child of node.children) {
					child.previousSibling = null;
					child.nextSibling = null;
				}
				for (var i = 0; i < node.children.length - 1; ++i) {
					node.children[i + 1].previousSibling = node.children[i];
					node.children[i].nextSibling = node.children[i + 1];
				}
			}
			let parent = node.parentNode;
			while (parent != null) {
				parent.allChildren.push(node);
				parent = parent.parentNode;
			}
		}
		
		var previousNode = null;
		var setPreviousNext = node => {
			if (previousNode != null) {
				previousNode.next = node;
			}
			node.previous = previousNode;
			previousNode = node;
			if (node.group) {
				for (const child of node.children) {
					setPreviousNext(child);
				}
			}
		};
		// Iterate the root nodes
		const rootNodes = this.tree.filter(node => (this.fullSelect || node.permitted) && node.parentNode == null);
		for (const node of rootNodes.sort(this.TreeSort.bind(this))) {
			setPreviousNext(node);
		}
		if (previousNode != null) {
			previousNode.next = null;
		}
		// Set the root nodes previousSibling and nextSibling
		for (const node of rootNodes) {
			node.previousSibling = null;
			node.nextSibling = null;
		}
		for (var i = 0; i < rootNodes.length - 1; ++i) {
			rootNodes[i + 1].previousSibling = rootNodes[i];
			rootNodes[i].nextSibling = rootNodes[i + 1];
		}
	}
	saveTemporaryOpenStates() {
		if (this.temporaryOpenStates == null) {
			this.temporaryOpenStates = [...this.querySelectorAll('.nodeContainer.collapsed')].map(element => element.data.id);
		}
	}
	loadTemporaryOpenStates() {
		if (this.temporaryOpenStates) {
			// The map here is just to decrease lookup times
			var temporaryOpenStatesMap = {};
			for (const id of this.temporaryOpenStates) {
				temporaryOpenStatesMap[id] = true;
			}
			for (const node of this.tree.filter(node => node.group)) {
				const collapsed = temporaryOpenStatesMap.hasOwnProperty(node.id);
				if (node.hasOwnProperty('tag')) {
					node.tag.classList.toggle('collapsed', collapsed);
					node.tag.querySelector(':scope > .group').classList.toggle('collapsed', collapsed);
				}
			}
			this.temporaryOpenStates = null;
		}
	}
	expandAll() {
		for (const node of this.tree.filter(node => node.group)) {
			if (node.hasOwnProperty('tag')) {
				node.tag.classList.remove('collapsed');
				node.tag.querySelector(':scope > .group').classList.remove('collapsed');
			}
		}
	}
	collapseAll() {
		for (const node of this.tree.filter(node => node.group)) {
			if (node.hasOwnProperty('tag')) {
				node.tag.classList.add('collapsed');
				node.tag.querySelector(':scope > .group').classList.add('collapsed');
			}
		}
	}
	enableThumbnail() {
		this.thumbnailEnabled = true;
		localStorage[`${this.treeName}-thumbnail`] = this.thumbnailEnabled;
		this.render();
	}
	disableThumbnail() {
		this.thumbnailEnabled = false;
		localStorage[`${this.treeName}-thumbnail`] = this.thumbnailEnabled;
		this.render();
	}
	toggleThumbnail() {
		if (this.thumbnailEnabled) {
			this.disableThumbnail();
		} else {
			this.enableThumbnail();
		}
	}
	saveOpenStates() {
		if (this.saveState && this.treeFilterView == null) {
			localStorage[`${this.treeName}-collapsed`] = JSON.stringify([...this.querySelectorAll('.nodeContainer.collapsed')].map($node => $node.data.id));
		}
	}
	loadOpenStates() {
		if (this.saveState) {
			if (localStorage.hasOwnProperty(`${this.treeName}-collapsed`)) {
				var states = JSON.parse(localStorage[`${this.treeName}-collapsed`]);
				var statesMap = {};
				for (const id of states) {
					statesMap[id] = true;
				}
				for (const node of this.tree.filter(node => node.group)) {
					var collapsed = statesMap.hasOwnProperty(node.id);
					if (node.hasOwnProperty('tag')) {
						node.tag.classList.toggle('collapsed', collapsed);
						node.tag.querySelector(':scope > .group').classList.toggle('collapsed', collapsed);
					}
				}
			} else {
				this.querySelector('.nodeContainer.collapsed')?.classList.remove('collapsed');
				this.saveOpenStates();
			}
		}
	}
	lock() {
		if (this.lockCount == 0) {
			this.lockScrollLeft = this.scrollLeft;
			this.lockScrollTop = this.scrollTop;
			this.scrollLeft = 0;
			this.scrollTop = 0;
			this.classList.add('busy');
		}
		this.lockCount++;
	}
	unlock() {
		this.lockCount--;
		if (this.lockCount == 0) {
			this.classList.remove('busy');
			this.scrollLeft = this.lockScrollLeft;
			this.scrollTop = this.lockScrollTop;
		}
	}
	isBusyForUploads() {
		return this.classList.contains('busy');
	}
	isBusy() {
		return this.classList.contains('busy') || this.tree.some(node => node.busy);
	}
	bounceSearchString() {
		clearTimeout(this.searchStringTimer);
		this.searchStringTimer = setTimeout(() => this.searchString = '', 1000);
	}
	mergeModifications(data) {
		// Added
		if (data.hasOwnProperty('added')) {
			for (const addedNode of data.added) {
				addedNode.permitted = true;
				this.tree.push(addedNode);
			}
		}
		// Updated
		if (data.hasOwnProperty('updated')) {
			for (const node of data.updated) {
				for (var treeItr = 0; treeItr < this.tree.length; ++treeItr) {
					if (this.tree[treeItr].id == node.id) {
						for (var property in node) {
							this.tree[treeItr][property] = node[property];
						}
						break;
					}
				}
			}
		}
		// Deleted
		if (data.hasOwnProperty('deleted')) {
			for (const id of data.deleted) {
				var index = this.tree.findIndex(node => node.id == id);
				if (index != -1) {
					this.tree.splice(index, 1);
				}
				this.selection = this.selection.filter(selectedNode => this.tree.some(node => node === selectedNode));
			}
		}
		// Error
		if (data.hasOwnProperty('errors')) {
			// Do nothing
		}
		this.render();
	}
	editName($handle, selectAll = false) {
		const node = $handle.closest('.nodeContainer').data;
		if (!this.editable(node)) {
			return;
		}
		if (this.isBusy()) {
			ConfirmationBox('Tree busy', 'Editing is disabled until pending operations complete.');
			return;
		}
		const inputTag = $handle.querySelector('.name');
		inputTag.innerText = node.name;
		if (inputTag.contentEditable  !== 'true') {
			inputTag.contentEditable = 'true';
			var saveGroupName = async e => {
				if (e.type == 'blur' || e.type == 'keydown' && e.key == 'Enter') {
					e.preventDefault();
					inputTag.removeEventListener('blur', saveGroupName);
					inputTag.removeEventListener('keydown', saveGroupName);
					inputTag.removeAttribute('contenteditable');
					if (e.type == 'keydown') {
						inputTag.blur();
					}
					if (node.name != inputTag.innerText) {
						this.lock();
						const controller = new AbortController();
						this.activeRequests.push(controller);
						try {
							const data = await post('/post/treeupdatename.php', {
								tree: this.treeName,
								syncid: this.currentSyncId,
								id: node.id,
								name: inputTag.innerText
							}, controller);
							this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
							CheckSessionExpired(data);
							if (data.success) {
								this.currentSyncId = data.syncid;
								this.mergeModifications(data);
								const updatedNode = this.tree.find(node => node.id == data.updated[0].id);
								this.dispatchEvent(new CustomEvent('change'));
								if (this.tooltip) {
									this.tooltip.call(this, node);
								} else {
									inputTag.title = node.name;
								}
								inputTag.innerText = this.nameTransform(node);
								this.unlock();
								updatedNode.$handle.focus();
							} else {
								if (data.hasOwnProperty('treeoutdated')) {
									TreeOutOfDate(this.treeName);
									await this.load();
									this.unlock();
								} else {
									ConfirmationBox('Error editing name', data.error);
									this.unlock();
									inputTag.innerText = this.nameTransform(node);
								}
							}
						} catch (exception) {
							this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
							NetworkingError();
							this.unlock();
							inputTag.innerText = this.nameTransform(node);
						}
					} else {
						inputTag.innerText = this.nameTransform(node);
					}
				}
			};
			inputTag.addEventListener('blur', saveGroupName);
			inputTag.addEventListener('keydown', saveGroupName);
			inputTag.focus();
			if (selectAll) {
				document.execCommand('selectAll', false, null);
			}
		}
	}
	async getAllNodes() {
		await this.loading;
		return this.tree.map(node => node);
	}
	applySelection(preserveScroll, dispatchChanged = false) {
		// Select the corresponding nodes in the selection
		for (const node of this.tree) {
			if (node.hasOwnProperty('tag')) {
				const selected = this.selection.includes(node);
				node.tag.classList.toggle('selected', selected);
				if (selected && node === this.focusedNode) {
					const scrollLeft = this.scrollLeft;
					const scrollTop = this.scrollTop;
					if (dispatchChanged || !IsFocusedElementAnInput()){
						node.$handle.focus();
					}
					if (preserveScroll) {
						this.scrollLeft = scrollLeft;
						this.scrollTop = scrollTop;
					}
				}
			}
		}
	}
	async setSelected(ids, ctrl = false, shift = false, dispatchChanged = false) {
		await this.loading;
		let ctrlArrowKeys = false;
		let nodes = [];
		if (Array.isArray(ids)) {
			nodes = this.tree.filter(node => ids.some(id => node.id == id));
		} else if (ids === 'previous' || ids === 'next') {
			if (this.selection.length == 0) {
				return;
			}
			var node = this.selection[this.selection.length - 1];
			if (shift) {
				if (this.shiftNode == null) {
					this.shiftNode = node;
				}
				if (this.focusedNode && this.focusedNode[`${ids}Sibling`]) {
					nodes = [this.focusedNode[`${ids}Sibling`]];
				}
			} else {
				if (ctrl) {
					ctrlArrowKeys = true;
				}
				if (this.focusedNode && this.focusedNode[`${ids}Sibling`]) {
					nodes = [this.focusedNode[`${ids}Sibling`]];
				} else {
					return;
				}
			}
		}
		else if (typeof ids == 'function') {
			nodes = this.tree.filter(ids);
		} else {
			throw new Error(`Expected arrays, 'previous', 'next', or filter function`);
		}
		
		const oldSelection = this.selection.map(node => node.id).sort();
		
		if (this.multiselect && ctrl) {
			var node = nodes[0];
			var index = this.selection.indexOf(node);
			if (index == -1) {
				this.focusedNode = node;
				this.selection.push(node);
			} else {
				this.selection.splice(index, 1);
				this.focusedNode = null;
				if (ctrlArrowKeys) {
					this.selection.push(node);
				}
				if (this.selection.length != 0) {
					this.focusedNode = this.selection[this.selection.length - 1];
				}
			}
			this.shiftNode = null;
		} else if (this.multiselect && shift && this.selection.length != 0) {
			if (this.shiftNode == null) {
				this.shiftNode = this.selection[this.selection.length - 1]
			}
			const start = this.shiftNode;
			const end = nodes[0];
			const getDirection = node => {
				let direction = '';
				for (var searchNode = start; searchNode != null; searchNode = searchNode.previousSibling) {
					if (searchNode === node) {
						direction = 'up';
						break;
					}
				}
				for (var searchNode = start; searchNode != null; searchNode = searchNode.nextSibling) {
					if (searchNode === node) {
						direction = 'down';
						break;
					}
				}
				return direction;
			};
			var direction = getDirection(end);
			
			if (direction != '') {
				if (this.focusedNode == null) {
					this.focusedNode = this.lastFocusedNode;
				}
				if (this.focusedNode != null) {
					var oldDirection = getDirection(this.focusedNode);
					for (var searchNode = this.focusedNode; searchNode !== start; searchNode = oldDirection == 'up' ? searchNode.nextSibling : searchNode.previousSibling) {
						this.selection.splice(this.selection.indexOf(searchNode), 1);
					}
					for (var searchNode = end; searchNode !== start; searchNode = direction == 'up' ? searchNode.nextSibling : searchNode.previousSibling) {
						// Nodes are removed from the selection to preserve last selection order
						var index = this.selection.indexOf(searchNode);
						if (index != -1) {
							this.selection.splice(index, 1);
						}
						this.selection.push(searchNode);
					}
					this.focusedNode = end;
				}
			}
		} else {
			this.shiftNode = null;
			if (this.selection.length == 1 && this.selection.includes(nodes[0]) && dispatchChanged) {
				var node = this.selection[0];
				if (this.editable(node)) {
					const pointerMove = () => clearInterval(this.editTimer);
					this.editTimer = setTimeout(() => {
						document.removeEventListener('pointermove', pointerMove, { once: true });
						this.editName(node.$handle);
					}, 500);
					document.addEventListener('pointermove', pointerMove, { once: true });
				}
			} else {
				this.selection.length = 0;
				for (const node of nodes) {
					this.focusedNode = node;
					this.selection.push(node);
				}
				if (this.click && this.selection.length != 0) {
					const clickedNode = this.selection[this.selection.length - 1];
					this.click.call(clickedNode.tag, e, clickedNode);
				}
			}
		}
		this.applySelection(false, dispatchChanged);
		
		if (dispatchChanged) {
			if (oldSelection.join(' ') != this.selection.map(node => node.id).sort().join(' ')) {
				// The selection has changed
				if (this.checkbox) {
					this.dispatchEvent(new CustomEvent('selectionchange'));
				} else {
					this.dispatchEvent(new CustomEvent('change'));
				}
			} else {
				// The selection hasn't changed
				if (this.checkbox) {
					if (this.dispatchEvent(new CustomEvent('selectionnochange'))) {
						clearInterval(this.editTimer);
					}
				} else {
					this.closest('.treeselect-modal')?.dispatchEvent(new Event('click'));
					// If the last nochange handler returns true then cancel the edit. Gets rid of a visual bug where the node is edited even though the user just meant to click on it again
					if (this.dispatchEvent(new CustomEvent('nochange'))) {
						clearInterval(this.editTimer);
					}
				}
			}
		}
	}
	async getSelected() {
		await this.loading;
		return this.selection.map(node => node);
	}
	getAllChildrenFromNodes(nodes) {
		const allChildren = [];
		for (const node of nodes) {
			if (allChildren.indexOf(node) == -1) {
				allChildren.push(node);
			}
			if (node.group) {
				for (const child of node.allChildren) {
					if (allChildren.indexOf(child) == -1) {
						allChildren.push(child);
					}
				}
			}
		}
		allChildren.sort(this.TreeSort.bind(this));
		return allChildren;
	}
	async getSelectedAllChildren() {
		return this.getAllChildrenFromNodes(await this.getSelected());
	}
	async getSelectedFilteredChildren() {
		return (await this.getSelectedAllChildren()).filter(this.filterDraggable);
	}
	UpdateCheckState() {
		if (this.checkbox) {
			// Reset all indeterminate states
			for (const node of this.tree) {
				if (node.$checkbox.indeterminate) {
					node.$checkbox.checked = false;
					node.$checkbox.indeterminate = false;
				}
			}
			// Set the folder indeterminate state if all the children are checked
			// This must be done as a post-order traversal of nodes as children folders must be updated first
			const postOrderTraverseNode = node => {
				if (node.group) {
					for (const child of node.children) {
						postOrderTraverseNode(child);
					}
					if (node.children.length != 0 && node.children.every(child => child.$checkbox.checked)) {
						if (!node.$checkbox.checked) {
							node.$checkbox.checked = true;
							node.$checkbox.indeterminate = true;
						}
					}
				}
			};
			for (const node of this.tree) {
				postOrderTraverseNode(node);
			}
			// Bit naive, but set indeterminate state of children
			for (const node of this.tree) {
				if (node.$checkbox.checked && !node.$checkbox.indeterminate) {
					for (const child of node.allChildren) {
						if (!child.$checkbox.checked) {
							child.$checkbox.checked = true;
							child.$checkbox.indeterminate = true;
						}
					}
				}
			}
			for (const node of this.tree) {
				node.previousChecked = node.$checkbox.checked;
				node.previousIndeterminate = node.$checkbox.indeterminate;
			}
		}
	}
	async setChecked(filter, indeterminateFilter = () => false, dispatchChanged = false) {
		await this.loading;
		const oldChecked = this.tree.map(node => node.$checkbox.checked && !node.$checkbox.indeterminate);
		const oldIndeterminate = this.tree.map(node => node.$checkbox.indeterminate);
		
		for (const node of this.tree) {
			node.$checkbox.checked = filter(node);
			node.$checkbox.indeterminate = false;
			if (indeterminateFilter(node)) {
				node.$checkbox.checked = true;
				node.$checkbox.indeterminate = true;
			}
		}
		if (dispatchChanged) {
			if (oldChecked.concat(oldIndeterminate).join(' ') != this.tree.map(node => node.$checkbox.checked && !node.$checkbox.indeterminate).concat(this.tree.map(node => node.$checkbox.indeterminate)).join(' ')) {
				// The selection has changed
				this.dispatchEvent(new CustomEvent('change'));
			} else {
				// The selection hasn't changed
				this.closest('.treeselect-modal')?.dispatchEvent(new Event('click'));
				// If the last nochange handler returns true then cancel the edit. Gets rid of a visual bug where the node is edited even though the user just meant to click on it again
				if (this.dispatchEvent(new CustomEvent('nochange'))) {
					clearInterval(this.editTimer);
				}
			}
		}
		this.UpdateCheckState();
	}
	async getChecked() {
		await this.loading;
		return !this.checkbox ? [] : this.tree.filter(node => !node.filtered && node.$checkbox.checked && !node.$checkbox.indeterminate);
	}
	async getCheckedAllChildren() {
		if (!this.checkbox) {
			return [];
		}
		var allChildren = [];
		for (const node of await this.getChecked()) {
			if (allChildren.indexOf(node) == -1) {
				allChildren.push(node);
			}
			if (node.group) {
				for (const child of node.allChildren) {
					if (allChildren.indexOf(child) == -1) {
						allChildren.push(child);
					}
				}
			}
		}
		allChildren.sort(this.TreeSort.bind(this));
		return allChildren;
	}
	async getCheckedFilteredChildren() {
		return (await this.getCheckedAllChildren()).filter(this.filterDraggable);
	}
	async load() {
		if (this.loading) {
			// Don't allow overlapping loads
			await this.loading;
		}
		this.loading = new Promise(async (resolve, reject) => {
			const oldChecked = this.tree.filter(node => node.$checkbox.checked && !node.$checkbox.indeterminate).map(node => node.id);
			const oldIndeterminate = this.tree.filter(node => node.$checkbox.indeterminate).map(node => node.id);
			const oldFileUploads = this.tree.filter(node => node.hasOwnProperty('file'));
			this.tree.length = 0; // Truncate the tree length so that any setSelected calls are queued up and executed after load. This is important because if the code calls reload() and attempts to select a recently created node, not rendered, then it would fail. Forcing selections to be queued waits until the tree structure is up to date.
			this.lock();
			var controller = new AbortController();
			this.activeRequests.push(controller);
			try {
				let data;
				if (this.src) {
					data = {
						success: true,
						syncid: 0,
						tree: await this.src()
					};
					for (const node of data.tree) {
						node.permitted = true;
					}
				} else {
					for (let tryCount = 0; tryCount < 3; ++tryCount) {
						try {
							data = await get('/get/tree.php',
							{
								tree: this.treeName
							}, controller);
							CheckSessionExpired(data);
							break;
						} catch (exception) {
							// Give up trying to get data
							if (tryCount == 2) {
								throw exception;
							}
						} finally {
							this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
						}
					}
				}
			
				if (data.success) {
					this.unlock();
					this.currentSyncId = data.syncid;
					this.tree.length = 0;
					for (const node of this.defaultOptions) {
						this.tree.push({ id: node.id, type: 0, name: node.name, parent: null, children: [], networkcommandid: 0, scheduleid: 0, value: node.value, defaultOption: true, permitted: true });
					}
					for (const node of data.tree) {
						this.tree.push(node);
					}
					for (let itr = 0; itr < this.selection.length; ++itr) {
						var replacementNode = this.tree.find(searchNode => searchNode.id == this.selection[itr].id);
						if (replacementNode) {
							this.selection[itr] = replacementNode;
						}
					}
					this.render();
					for (const id of oldChecked) {
						const node = this.tree.find(searchNode => searchNode.id == id);
						if (node) {
							node.$checkbox.checked = true;
						}
					}
					for (const id of oldIndeterminate) {
						const node = this.tree.find(searchNode => searchNode.id == id);
						if (node) {
							node.$checkbox.checked = true;
							node.$checkbox.indeterminate = true;
						}
					}
					for (const oldNode of oldFileUploads) {
						const node = this.tree.find(searchNode => searchNode.id == oldNode.id);
						if (node) {
							node.file = oldNode.file;
							if (oldNode.hasOwnProperty('overwrite') && oldNode.overwrite) {
								node.overwrite = oldNode.overwrite;
							}
						}
					}
					this.loadOpenStates();
					this.dispatchEvent(new CustomEvent('load'));
					resolve();
				} else {
					reject(new Error('load failed'));
				}
			} catch (exception) {
				NetworkingError();
				reject(new Error('load failed'));
			}
		});
		await this.loading;
		if ((await this.getSelected()).length == 0) {
			const roots = this.getSelectableRoots();
			if (roots.length != 0 && this.input == null) {
				await this.setSelected([roots[0].id], false, false, this.dispatchChangeOnLoad);
			}
		}
	}
	async reload() {
		// Ensure any previous load has finished before starting a new one
		await this.loading;
		await this.load();
	}
	async processFileuploads() {
		// Only allow one processing adds loop
		if (this.processingAdds) {
			return;
		}
		this.processingAdds = true;
		while (true) {
			// Get all the nodes that are waiting to be uploaded
			const nodes = this.tree.filter(node => node.hasOwnProperty('file'));
			if (nodes.length == 0) {
				// No nodes to upload
				break;
			}
			// Try to upload the first file found
			const node = nodes[0];
			const request = new XMLHttpRequest();
			node.xhr = request;
			const isBusy = await new Promise(resolve => {
				const data = new FormData();
				data.append('tree', this.treeName);
				data.append('syncid', this.currentSyncId);
				data.append('id', node.id);
				data.append('size', node.file.size);
				if (node.hasOwnProperty('uploaddata')) {
					data.append('data', JSON.stringify(node.uploaddata));
				}
				data.append('file0', node.file);
				request.onload = async () => {
					console.log('Here on upload onload');
					delete node.xhr;
					try {
						const data = JSON.parse(request.responseText);
						CheckSessionExpired(data);
						this.mergeModifications(data);
						if (data.success) {
							this.currentSyncId = data.syncid;
							if (this.treeName != 'content' && this.treeName != 'csv' && this.treeName != 'customwebpage') {
								if (data.updated.length != 0) {
									await this.setSelected([data.updated[0].id]);
								}
							}
							delete node.file;
							delete node.uploaddata;
							resolve();
						} else {
							if (data.hasOwnProperty('treeoutdated')) {
								TreeOutOfDate(this.treeName);
								await this.load();
							} else if (data.hasOwnProperty('busy')) {
								resolve('busy');
							} else if (data.hasOwnProperty('error')) {
								toastError(data.error);
								delete node.file;
								delete node.uploaddata;
							}
							resolve();
						}
					} catch (exception) {
						// Error parsing the json response probably
						console.log(`${node.name} exception: ${exception.toString()}`);
						delete node.xhr;
						delete node.file;
						delete node.uploaddata;
						resolve();
					}
				};
				// These two events fire seemingly randomly. I think there's a race condition on the server responding with a busy json and the upload being cancelled.
				request.onerror = error => {
					console.log('request.onerror, treating as busy', error);
					resolve('busy');
				};
				request.upload.onerror = error => {
					console.log('request.upload.onerror, treating as busy', error);
					resolve('busy');
				};
				const abort = () => {
					console.log(`file upload aborted for ${node.name}`);
					delete node.xhr;
					delete node.file;
					delete node.uploaddata;
					this.render();
					resolve();
				};
				request.onabort = abort;
				request.upload.onabort = abort;
				request.upload.onprogress = e => {
					if (e.lengthComputable && node.hasOwnProperty('progress')) {
						node.progress(node, e.loaded, e.total);
					}
				};
				request.open('post', this.treeName == 'content' ? '/uploader' : '/post/treefileupload.php');
				this.activeXHRs.push(request);
				request.send(data);
			});
			this.activeXHRs.splice(this.activeXHRs.indexOf(request), 1);
			if (isBusy === 'busy') {
				// The server is busy, so wait 5 seconds and try again. Only the content tree has this behavior as files are processed sequentially from all sessions.
				await new Promise(resolve => setTimeout(resolve, 5000));
			}
		}
		this.processingAdds = false;
	}
	async addItems(nodes) {
		const nodesToOverwrite = nodes.filter(node => node.hasOwnProperty('overwrite') && node.overwrite);
		for (const node of nodesToOverwrite) {
			const existingNode = this.tree.find(treeNode => treeNode.id == node.id);
			if (existingNode) {
				// TODO: This should be simplified. Transferring data from created nodes or overwritten nodes below is similar
				existingNode.file = node.file;
				existingNode.overwrite = true; // This is used to track on the client-side so that contentfileviewupdate doesn't overwrite the status
				if (existingNode?.data?.status && existingNode.data.status.hasOwnProperty('done')) {
					existingNode.data.status.done = false;
					if (existingNode.data.status?.output) {
						for (const output of ['original', ...existingNode.data.status.output]) {
							existingNode.data.status[output].convert.error = null;
							existingNode.data.status[output].cdn.error = null;
						}
					}
				}
				if (node.hasOwnProperty('progress')) {
					existingNode.progress = node.progress;
				}
			}
		}
		if (nodesToOverwrite.length != 0) {
			this.processFileuploads();
		}
		for (const node of nodes) {
			// TODO: This is naive and could fail randomly.
			node.addid = Math.floor(Math.random() * 1000000);
		}
		nodes = nodes.filter(node => !node.hasOwnProperty('overwrite') || !node.overwrite);
		if (nodes.length == 0) {
			return { success: true };
		}
		const controller = new AbortController();
		this.activeRequests.push(controller);
		try {
			const data = await post('/post/treeadd.php',
			{
				tree: this.treeName,
				syncid: this.currentSyncId,
				nodes: JSON.stringify(nodes)
			}, controller);
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			CheckSessionExpired(data);
			if (data.success) {
				this.currentSyncId = data.syncid;
				// Transfer the file and uploaddata to the newly created node
				for (const node of nodes) {
					const addedNode = data.added.find(addedNode => addedNode.addid == node.addid);
					if (addedNode) {
						if (node.hasOwnProperty('file')) {
							addedNode.file = node.file;
						}
						if (node.hasOwnProperty('uploaddata')) {
							addedNode.uploaddata = node.uploaddata;
						}
						if (node.hasOwnProperty('progress')) {
							addedNode.progress = node.progress;
						}
					}
				}
				this.mergeModifications(data);
				
				this.processFileuploads();
			} else {
				if (data.hasOwnProperty('treeoutdated')) {
					TreeOutOfDate(this.treeName);
					await this.load();
				} else {
					if (data.hasOwnProperty('errors')) {
						const nonOverwriteErrors = data.errors.filter(error => !error.hasOwnProperty('overwrite'));
						if (nonOverwriteErrors.length != 0) {
							ConfirmationBox('Error updating', `There was a problem adding items: ${nonOverwriteErrors.map(error => {
								const node = nodes.find(node => node.addid == error.addid);
								return `${node.name}: ${error.error}`;
							}).join(', ')}`);
						}
					} else {
						ConfirmationBox('Error updating', `There was a problem adding items: ${data.error}`);
					}
				}
			}
			return data;
		} catch (exception) {
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			NetworkingError();
		}
		return { success: false };
	}
	async updateItems(nodes) {
		if (this.isBusy()) {
			ConfirmationBox('Tree busy', 'Updating is disabled until pending operations complete.');
			return false;
		}
		this.lock();
		const controller = new AbortController();
		this.activeRequests.push(controller);
		try {
			const data = await post('/post/treeupdate.php',
			{
				tree: this.treeName,
				syncid: this.currentSyncId,
				nodes: JSON.stringify(nodes)
			}, controller);
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			CheckSessionExpired(data);
			if (data.success) {
				this.currentSyncId = data.syncid;
				this.mergeModifications(data);
				this.unlock();
				return true;
			} else {
				if (data.hasOwnProperty('treeoutdated')) {
					TreeOutOfDate(this.treeName);
					await this.load();
					this.unlock();
				} else {
					ConfirmationBox('Error updating', `There was a problem updating items: ${data.error}`);
					this.unlock();
				}
			}
		} catch (exception) {
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			NetworkingError();
			this.unlock();
			throw exception;
		}
		return false;
	}
	async addGroup() {
		let selectedNode = (await this.getSelected())[0];
		if (selectedNode == undefined) {
			ConfirmationBox('Select a group', 'Select a group.');
			return;
		}
		if (!selectedNode.group && selectedNode.parentNode) {
			selectedNode = selectedNode.parentNode;
		}
		if (!selectedNode.group) {
			ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
			return;
		}
		if (this.isBusy()) {
			ConfirmationBox('Cannot add group', 'The tree is currently busy and groups cannot be added.');
			return;
		}
		
		if (selectedNode.tag.classList.contains('collapsed')) {
			selectedNode.tag.querySelector(':scope > li:first-child > i:first-child').click();
		}
		this.lock();
		var controller = new AbortController();
		this.activeRequests.push(controller);
		try {
			var data = await post('/post/treeaddgroup.php', {
				tree: this.treeName,
				syncid: this.currentSyncId,
				parent: selectedNode.id
			}, controller);
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			CheckSessionExpired(data);
			if (data.success) {
				this.currentSyncId = data.syncid;
				this.mergeModifications(data);
				await this.setSelected([data.added[0].id], false, false, true);
				this.unlock();
				this.editName(this.querySelector('.selected > .node > .handle'), true);
			} else {
				if (data.hasOwnProperty('treeoutdated')) {
					TreeOutOfDate(this.treeName);
					await this.load();
					this.unlock();
				} else {
					ConfirmationBox('Error adding group', `There was a problem adding a group: ${data.error}`);
					this.unlock();
				}
			}
		} catch (exception) {
			this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
			NetworkingError();
			this.unlock();
		}
	}
	async duplicateSelected() {
		let selectedNode = (await this.getSelected())[0];
		if (selectedNode == undefined || selectedNode.group) {
			ConfirmationBox('Select an item', 'Select an item to duplicate.');
			return;
		}
		if (this.isBusy()) {
			ConfirmationBox('Tree busy', 'Duplicating is not allowed until pending operations complete.');
			return false;
		}
		await InputBox('Duplicate', 'Enter a name for the duplicated item.', 'Ok', 'Cancel', '', async name => {
			this.lock();
			var controller = new AbortController();
			this.activeRequests.push(controller);
			try {
				var data = await post('/post/treeduplicate.php',
				{
					tree: this.treeName,
					syncid: this.currentSyncId,
					name: name,
					id: selectedNode.id
				}, controller);
				this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
				CheckSessionExpired(data);
				if (data.success) {
					this.currentSyncId = data.syncid;
					this.mergeModifications(data);
					await this.setSelected([data.added[0].id], false, false, true);
					this.unlock();
					return true;
				} else {
					if (data.hasOwnProperty('treeoutdated')) {
						TreeOutOfDate(this.treeName);
						await this.load();
					} else {
						ConfirmationBox('Error duplicating', `There was a problem duplicating: ${data.error}`);
					}
					this.unlock();
				}
			} catch (exception) {
				this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
				NetworkingError();
				this.unlock();
			}
		}, () => true, null);
	}
	async deleteSelected() {
		const selectedNodes = await this.getSelected();
		const selectedAllChildren = await this.getSelectedAllChildren();
		try{
			// Throwing or not calling the callback in the deleteConfirmation function is how one cancels the delete operation
			// confirm is an optional field used by some trees for further questions. Note this is not the result of the initial confirmation
			await this.deleteConfirmation.call(this, selectedNodes, selectedAllChildren, async (nodes, confirm = false) => {
				this.lock();
				for (var itr = 0; itr < nodes.length; ++itr) {
					const node = nodes[itr];
					if (node.hasOwnProperty('xhr')) {
						node.xhr.abort();
					}
				}
				// If none of the nodes were on the server then stop here
				if (nodes.length == 0) {
					this.unlock();
					return;
				}
				var controller = new AbortController();
				this.activeRequests.push(controller);
				try {
					var data = await post('/post/treedelete.php', {
						tree: this.treeName,
						syncid: this.currentSyncId,
						ids: JSON.stringify(nodes.map(node => node.id)),
						confirm
					}, controller);
					this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
					CheckSessionExpired(data);
					if (data.success) {
						this.currentSyncId = data.syncid;
						this.mergeModifications(data);
						await this.setSelected([1], false, false, true);
						this.unlock();
					} else {
						if (data.hasOwnProperty('treeoutdated')) {
							TreeOutOfDate(this.treeName);
							await this.load();
							this.unlock();
						} else if (data.hasOwnProperty('error')) {
							ConfirmationBox('Error deleting', `There was a problem deleting: ${data.error}`);
							this.unlock();
						} else {
							this.unlock();
							// Pass the data up the to the catch below this so it can throw it and the outer code can do its own thing (like display a confirmation box for additional questions)
							throw data;
						}
					}
				} catch (exception) {
					this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
					if (exception.hasOwnProperty('success')) {
						// Allow the deleteConfirmation handler to get the data
						throw exception;
					} else {
						NetworkingError();
					}
					this.unlock();
				}
			});
		} catch (exception) {
			// User cancelled delete
		}
	}
	filterView(filter) {
		this.treeFilterView = filter;
		this.expandAll();
		this.render();
	}
	removeFilterView() {
		this.treeFilterView = null;
		this.loadOpenStates();
		this.render();
	}
	// Select dropdowns don't have multiselect drag and drop. They could, but right now nothing requires it
	async multiSelectDragStart(e) {
		const node = e.currentTarget.closest('.nodeContainer').data;
		const inputTag = e.currentTarget.querySelector('.name');
		if (!inputTag.isContentEditable) {
			// Only run the setSelected with dispatchChanged as true if the user clicked on the name field
			const index = this.selection.indexOf(node);
			if (index == -1 || e.ctrlKey || this.input) {
				await this.setSelected([node.id], e.ctrlKey, e.shiftKey, e.target.closest('.handle') != null);
			}
		}
		if (this.input) {
			return false;
		}
		const selection = await this.getSelected();
		// Root elements aren't draggable just because it results in annoying user errors
		if (selection.some(node => node.$name.isContentEditable || node.parentNode == null)) {
			return false;
		}
		return selection;
	}
	async multiSelectDragStop(e) {
		const node = e.target.closest('.nodeContainer')?.data;
		const index = this.selection.indexOf(node);
		if (index != -1 && !e.ctrlKey) {
			await this.setSelected([node.id], e.ctrlKey, e.shiftKey, e.target.closest('.handle') != null);
		}
	}
	removeRedundantSelections(selectedNodes) {
		// Remove any node if the parent of the node already exists in the selection
		// This is kind of expensive for selections in deep trees
		for (var itr = 0; itr < selectedNodes.length; ++itr) {
			var node = selectedNodes[itr];
			var parentInSelection = false;
			while (!parentInSelection && node.parentNode != null) {
				node = node.parentNode;
				for (var itr2 = 0; itr2 < selectedNodes.length; ++itr2) {
					if (itr != itr2 && node === selectedNodes[itr2]) {
						selectedNodes.splice(itr--, 1);
						parentInSelection = true;
						break;
					}
				}
			}
		}
	}
	async multiSelectDraggingHelper($node, e, dropRegion) {
		let selectedNodes = await this.getSelected();
		if (!selectedNodes.some(node => !node.$name.isContentEditable && !node.busy)) {
			return false;
		}
		
		const dropTag = dropRegion == '' ? undefined : await this.dragHelper.call(this, e, dropRegion, this.thumbnail);
		if (dropTag == undefined) {
			selectedNodes = selectedNodes.filter((node, index, nodes) => !node.busy && this.filterDraggable(node, index, nodes)).sort(this.TreeSort.bind(this));
			this.removeRedundantSelections(selectedNodes);
			if (selectedNodes.length == 0) {
				// All the nodes were filtered
				return false;
			}
			const $drop = $.ul({ class: ['tree-dragging', this.thumbnail !== false, 'leafs'] });
			for (const node of selectedNodes) {
				const $clonedNode = node.tag.querySelector(':scope > .node').cloneNode(true);
				$clonedNode.classList.add('default');
				$drop.appendChild($clonedNode);
			}
			return $drop;
		} else {
			return dropTag;
		}
	}
	async multiSelectFilterDroppable($droppable) {
		// This is rather expensive and might cause issues on massive content trees with a lot selected
		let selected = false;
		const $nodeContainer = $droppable.closest('.nodeContainer');
		if ($nodeContainer) {
			const droppableNode = $nodeContainer.data;
			const selectedAllChildren = await this.getSelectedAllChildren();
			for (const node of selectedAllChildren) {
				if (droppableNode === node) {
					selected = true;
				}
			}
			return !selected && this.filterDroppable(selectedAllChildren, droppableNode);
		} else {
			return true;
		}
	}
	updateIcon(node) {
		if (node.group) {
			const groupClass = this.groupIcon(node);
			if (groupClass != '') {
				node.tag.querySelector(':scope > .group').classList.add(groupClass);
			}
		} else {
			const leafClass = this.leafIcon(node);
			if (leafClass != '') {
				node.tag.querySelector(':scope > .leaf').classList.add(leafClass);
			}
		}
	}
	updateIcons() {
		for (const node of this.tree) {
			this.updateIcon(node);
		}
	}
	updateName(node) {
		if (node.hasOwnProperty('tag')) {
			node.$name.innerText = this.nameTransform(node);
		}
	}
	updateNames() {
		for (const node of this.tree) {
			this.updateName(node);
		}
	}
	buildTree(node) {
		const isRoot = node.parentNode == null;
		
		const hasChildren = node.children.length != 0;
		const hasLeaves = node.children.filter(node => !node.group).length != 0;
		
		const drop = async (e, source) => {
			if (this.isBusy()) {
				ConfirmationBox('Moving files not allowed.', 'Moving files is disabled until pending operations complete.');
				return false;
			}
			const target = node;
			if (target.tag.classList.contains('collapsed')) {
				target.tag.querySelector(':scope > li:first-child > i:first-child').click();
			}
			this.removeRedundantSelections(source);
			let parent = null;
			if (!source.every(node => {
				if (parent == null) {
					parent = node.parentNode;
					return true;
				} else {
					return node.parentNode === parent;
				}
			})) {
				ConfirmationBox('Invalid drag selection', 'To move multiple files between folders all selected files must share the same parent folder.');
				return false;
			}
			if (source.some(node => node === target)) {
				// Cannot drop onto itself
				return false;
			}
			// Check if the parent doesn't change and if so do nothing
			if (source.some(node => node.parentNode != null && node.parentNode === target)) {
				return false;
			}
			this.lock();
			var controller = new AbortController();
			this.activeRequests.push(controller);
			try {
				const data = await post('/post/treeupdateparent.php',
				{
					tree: this.treeName,
					syncid: this.currentSyncId,
					source: JSON.stringify(source.map(node => node.id)),
					target: target.id
				}, controller);
				this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
				CheckSessionExpired(data);
				if (data.success) {
					this.currentSyncId = data.syncid;
					this.mergeModifications(data);
					this.unlock();
				} else {
					if (data.hasOwnProperty('treeoutdated')) {
						TreeOutOfDate(this.treeName);
						await this.load();
						this.unlock();
					} else {
						ConfirmationBox('Error adding group', `There was a problem moving the selected items: ${data.error}`);
						this.unlock();
					}
				}
			} catch (exception) {
				this.activeRequests.splice(this.activeRequests.indexOf(controller), 1);
				NetworkingError();
				this.unlock();
			}
			return true;
		};
		
		const nodeDoubleClick = async e => {
			const $nodeContainer = e.currentTarget.closest('.nodeContainer');
			var node = $nodeContainer.data;
			// The order of these statements with editable matters. Do not change them or try to simplify them.
			// Basically editName should only be called if it's editable and the user is clicking on the name field and if not the doubleClick event should be called, but editName should only be called then if it's editable
			if (this.editable(node) && e.target.classList.contains('name') || this.doubleClick && !(await this.doubleClick.call($nodeContainer, e, node)) && this.editable(node)) {
				this.editName(e.currentTarget);
			}
		};
		
		const BuildHandle = ($handle, node) => {
			if (this.checkbox) {
				$handle.prepend(
					node.$checkbox = $.uicheckbox({ click: e => e.stopPropagation(), change: () => {
						if (node.group) {
							for (const child of node.allChildren) {
								child.$checkbox.checked = false;
								child.$checkbox.indeterminate = false;
							}
						}
						if (node.parentNode?.$checkbox?.checked) {
							if (node.previousChecked && node.previousIndeterminate) {
								for (const child of node.parentNode.children) {
									child.$checkbox.checked = true;
									child.$checkbox.indeterminate = false;
								}
								node.$checkbox.checked = false;
							}
						}
						for (let parentNode = node.parentNode; parentNode != null; parentNode = parentNode.parentNode) {
							parentNode.$checkbox.checked = false;
							parentNode.$checkbox.indeterminate = false;
						}
						this.UpdateCheckState();
					} })
				);
			}
			if (this.tooltip) {
				this.tooltip.call($handle, node);
			} else {
				$handle.title = node.name;
			}
			
			if (node.hasOwnProperty('data') && node.data.hasOwnProperty('status')) {
				if (node.done) {
					$handle.style.removeProperty('background');
				} else {
					// TODO: Should use --var colors for this instead.
					let progressBarColor = 'rgba(255,255,255,0.4)';
					if (node.error) {
						progressBarColor = 'rgba(255,0,0,0.4)';
					}
					let percent = node.data.status.upload.progress ? node.data.status.upload.progress : 0;
					for (const output of ['original', ...node.data.status.output]) {
						percent += node.data.status[output].convert.progress ? node.data.status[output].convert.progress : 0;
						percent += node.data.status[output].cdn.progress ? node.data.status[output].cdn.progress : 0;
					}
					percent = Math.round(percent / ((['original', ...node.data.status.output].length * 2 + 1) * 100) * 100);
					$handle.style.background = `linear-gradient(to right, rgba(1,1,1,0.2), rgba(1,1,1,0.2)), linear-gradient(to right, ${progressBarColor}, ${progressBarColor} ${percent}%, transparent ${percent}%)`;
				}
			}
		};
		
		if (node.group) {
			return $.li({},
				node.tag = $.ul({
					class: [
						'groupContainer nodeContainer',
						isRoot, 'root',
						node.previousSibling != null && !node.previousSibling.filtered, 'hasprevioussibling',
						node.nextSibling != null && !node.nextSibling.filtered, 'hasnextsibling',
						hasChildren, 'haschildren',
						hasLeaves, 'hasleaves'
					],
					process: $element => {
						// TODO: Can I rewrite this?
						// Set this here so it's available to updateIcon
						node.tag = $element;
						$element.data = node;
						this.updateIcon(node);
					}
				},
					$.li({},
						$.i({ click: e => {
							const $nodeContainer = e.currentTarget.closest('.nodeContainer');
							$nodeContainer.classList.toggle('collapsed');
							$nodeContainer.querySelector(':scope > .group').classList.toggle('collapsed');
							this.saveOpenStates();
						} }),
						$.i()
					),
					$.li({ class: `group node ${this.treeName}-group`, process: $node => {
						draggable($node, {
							handle: ':scope > .handle',
							helper: this.multiSelectDraggingHelper.bind(this),
							appendTo: this.dragAppendTo,
							containment: this.dragAppendTo,
							thresholdX: 3,
							thresholdY: 3,
							start: this.multiSelectDragStart.bind(this),
							filterDroppable: this.multiSelectFilterDroppable.bind(this),
							stop: this.multiSelectDragStop.bind(this)
						});
					} },
						node.$handle = $.div({ class: ['handle', node.disabled, 'disabled'], tabindex: -1, dblclick: nodeDoubleClick, droppable: {
							accept: [`${this.treeName}-group`, `${this.treeName}-leaf`],
							drop: drop
						}, process: $handle => BuildHandle($handle, node) },
							$.i(),
							node.$name = $.span({ text: this.nameTransform(node), class: 'name', spellcheck: false })
						),
						$.ul({ process: $node => {
							if (this.thumbnail && this.thumbnailEnabled) {
								$node.appendChild($.li({ class: 'leafs', hidden: !hasLeaves, droppable: {
									accept: [`${this.treeName}-group`, `${this.treeName}-leaf`],
									drop: drop
								} },
									$.ul({},
										...node.children.filter(node => !node.group).map(node => {
											let overrideThumbnailWidth = null;
											let overrideThumbnailHeight = null;
											if (this.overrideThumbnailDimensions) {
												var overrideDimension = this.overrideThumbnailDimensions(this.tree.filter(node => !node.group), this.largeImages);
												overrideThumbnailWidth = overrideDimension.width;
												overrideThumbnailHeight = overrideDimension.height;
											}
											node.$handle = this.thumbnail(node, this.largeImages, overrideThumbnailWidth, overrideThumbnailHeight);
											node.$handle.addEventListener('dblclick', nodeDoubleClick);
											BuildHandle(node.$handle, node);
											return $.li({ title: this.nameTransform(node), process: $element => $element.data = node },
												node.tag = $.ul({ class: 'leafContainer nodeContainer', process: $element => {
													// TODO: Can I rewrite this?
													// Set this here so it's available to updateIcon
													node.tag = $element;
													$element.data = node;
													this.updateIcon(node);
												} },
													$.li({ class: `leaf node ${this.treeName}-leaf`, draggable: {
														handle: ':scope > .handle',
														helper: this.multiSelectDraggingHelper.bind(this),
														appendTo: this.dragAppendTo,
														containment: this.dragAppendTo,
														thresholdX: 3,
														thresholdY: 3,
														start: this.multiSelectDragStart.bind(this),
														filterDroppable: this.multiSelectFilterDroppable.bind(this),
														stop: this.multiSelectDragStop.bind(this)
													} },
														node.$handle
													)
												)
											)
										})
									)
								));
							} else {
								$node.append(...node.children.map(node => this.buildTree(node)));
							}
						} }, ...node.children.filter(node => this.thumbnail && this.thumbnailEnabled && node.group).map(node => this.buildTree(node)))
					)
				)
			);
		} else {
			return $.li({},
				node.tag = $.ul({ class: [
					'leafContainer',
					'nodeContainer',
					isRoot, 'root',
					node.previousSibling != null && !node.previousSibling.filtered, 'hasprevioussibling',
					node.nextSibling != null && !node.nextSibling.filtered, 'hasnextsibling'
				], process: $element => {
					// TODO: Can I rewrite this?
					// Set this here so it's available to updateIcon
					node.tag = $element;
					$element.data = node;
					this.updateIcon(node);
				} },
					$.li({},
						$.i()
					),
					$.li({ class: `leaf node ${this.treeName}-leaf`, draggable: {
						handle: ':scope > .handle',
						helper: this.multiSelectDraggingHelper.bind(this),
						appendTo: this.dragAppendTo,
						containment: this.dragAppendTo,
						thresholdX: 3,
						thresholdY: 3,
						start: this.multiSelectDragStart.bind(this),
						filterDroppable: this.multiSelectFilterDroppable.bind(this),
						stop: this.multiSelectDragStop.bind(this)
					} },
						node.$handle = $.div({ class: ['handle', node.disabled, 'disabled'], tabindex: -1, dblclick: nodeDoubleClick, process: $handle => BuildHandle($handle, node) },
							$.i(),
							node.$name = $.span({ text: this.nameTransform(node), class: 'name', spellcheck: false })
						)
					)
				)
			);
		}
	}
	async render(options = {}) {
		// Allow the user to overwrite options
		Object.assign(this, options);
		
		this.classList.toggle('thumbnails', this.thumbnail && this.thumbnailEnabled);
		
		// Store check state into the nodes
		if (this.checkbox) {
			for (const node of this.tree) {
				if (node.hasOwnProperty('$checkbox')) {
					node.checked = node.$checkbox.checked;
				}
				if (node.hasOwnProperty('$checkbox')) {
					node.indeterminate = node.$checkbox.indeterminate;
				}
				if (!node.hasOwnProperty('checked')) {
					node.checked = false;
				}
				if (!node.hasOwnProperty('indeterminate')) {
					node.indeterminate = false;
				}
			}
		}
		
		// If the selected node no longer exists then remove it from the selection
		this.selection = this.selection.filter(node => this.tree.includes(node));
		
		this.ConstructTreeStructure();
		
		for (const node of this.tree) {
			node.disabled = !node.hasOwnProperty('defaultOption') && this.disable(node);
			node.filtered = !node.hasOwnProperty('defaultOption') && (!this.filters.every(filter => filter(node)) || this.treeFilterView && !this.treeFilterView(node));
		}
		
		// All children of a parent must be filtered in order to filter the parent
		for (const node of this.tree) {
			for (let parent = node.parentNode; parent != null; parent = parent.parentNode) {
				if (!node.filtered) {
					if (parent.filtered) {
						parent.disabled = true;
					}
					parent.filtered = false;
				}
			}
		}
		
		this.ConstructTreeStructure();
		
		this.saveTemporaryOpenStates();
		this.innerHTML = '';
		if (this.showMinimizeMaximizeAll || this.showSmallLargeThumbnail) {
			const treeOperations = $.li({ class: 'tree-operations' });
			if (this.showSmallLargeThumbnail) {
				treeOperations.append(
					$.i({ title: 'Small Images', class: 'tooltip-button mdi mdi-image-size-select-small thumbnailshrink', click:() => {
						localStorage[`${this.treeName}-imagesize`] = false;
						this.largeImages = false;
						this.render();
					} }),
					$.i({ title: 'Large Images', class: 'tooltip-button mdi mdi-image-size-select-large thumbnailexpand', click: () => {
						localStorage[`${this.treeName}-imagesize`] = true;
						this.largeImages = true;
						this.render();
					} })
				);
			}
			if (this.thumbnail) {
				treeOperations.appendChild(
					$.i({ title: 'Toggle Thumbnails', class: 'tooltip-button mdi mdi-image-outline collapseall', click: this.toggleThumbnail.bind(this) }),
				);
			}
			if (this.showMinimizeMaximizeAll) {
				treeOperations.append(
					$.i({ title: 'Collapse All', class: 'tooltip-button mdi mdi-collapse-all-outline collapseall', click: this.collapseAll.bind(this) }),
					$.i({ title: 'Expand All', class: 'tooltip-button mdi mdi-expand-all-outline expandall', click: this.expandAll.bind(this) })
				);
			}
			this.appendChild(treeOperations);
		}
		this.append(...this.getNonFilteredRoots().map(node => this.buildTree(node)));
		
		for (const node of this.tree) {
			if (!node.hasOwnProperty('$checkbox')) {
				node.$checkbox = $.uicheckbox();
			}
		}
		this.loadTemporaryOpenStates();
		this.applySelection(true);
		// Reapply the check states
		if (this.checkbox) {
			for (const node of this.tree) {
				if (node.hasOwnProperty('$checkbox')) {
					node.$checkbox.checked = node.checked;
					node.$checkbox.indeterminate = node.indeterminate;
				}
			}
		}
		
		this.UpdateCheckState();
	}
});