/**
 * Manages audio ducking for the background music from some background music provider.
 */
class BackgroundMusicService {
	constructor() {
		/**
		 * Represents the element ID of the currently active playlist, or null if there is no playlist currently scheduled to play (e.g. a blank schedule.)
		 * Used to filter out stray events that may be emitted by playlists that are no longer the active playlist.
		 * @type {?string}
		 */
		this.playlistElementId = null;

		/**
		 * @type {!boolean}
		 */
		this.ducked = false;

		/**
		 * @type {!BackgroundMusicProviders[]}
		 */
		this.providers = [];
	}

	/**
	 * @param {?string} playlistElementId
	 */
	ActivePlaylistChanged(playlistElementId) {
		this.playlistElementId = playlistElementId;
	}

	/**
	 * @param {!string} playlistElementId
	 */
	DuckAudioDown(playlistElementId) {
		if (this.playlistElementId !== playlistElementId) {
			return;
		}
		this.duckDown();
	}

	/**
	 * @param {!string} playlistElementId
	 */
	DuckAudioUp(playlistElementId) {
		if (this.playlistElementId !== playlistElementId) {
			return;
		}
		this.duckUp();
	}

	duckDown() {
		this.ducked = true;
		this.providers
			.filter(provider => provider.playing)
			.forEach(provider => provider.duckDown());
	}

	duckUp() {
		this.ducked = false;
		this.providers
			.filter(provider => provider.playing)
			.forEach(provider => provider.duckUp());
	}

	/**
	 * Attempts to play the first provider, if it throws an error then it tries to play the next provider, repeating until there are no providers left to try or a provider plays successfully.
	 */
	play() {
		let tryToPlay = true;
		for (const provider of this.providers) {
			if (tryToPlay) {
				try {
					provider.play();
					tryToPlay = false;
					if (this.ducked) {
						provider.mute();
					}
				} catch (err) {
					console.log(`Warning, BackgroundMusicService provider failed to play.  Skipping it.  ${err.stack || err}`);
					provider.stop();
				}
			} else {
				// Only one provider may play at a time
				provider.stop();
			}
		}
	}

	stop() {
		this.providers.forEach(provider => provider.stop());
	}

	/**
	 * Adds a BackgroundMusicProvider to the list.
	 * Returns true if the provider was added or false if it was already added.
	 * @param {!Object} provider
	 * @returns {!boolean}
	 */
	add(provider) {
		if (this.providers.find(p => p.url === provider) !== undefined) {
			return false;
		}
		this.providers.push(new CCMProvider(provider));
		return true;
	}

	/**
	 * Removes all providers.
	 */
	clear() {
		this.stop();
		this.providers = [];
	}
}
