/* exported BreezeFonts */
/* global FileInfo */

class BreezeFonts {
	/**
	 * @param {!Player} player
	 */
	constructor(player) {
		this.player = player;

		/**
		 * User uploaded custom fonts.
		 * @type {!FontFaceData[]}
		 */
		this.customFonts = [];

		/**
		 * Builtin/google/etc fonts.
		 * @type {!FontFaceData[]}
		 */
		this.fonts = [];

		this.file = new FileInfo('fonts/custom.fonts', false);

		/** @type {!string} */
		this.styleSheetId = `breezeFonts-${Utility.getUID()}`;
	}

	/**
	 * @returns {!Promise}
	 */
	async initialize() {
		try {
			await this.initializeCustomFonts();

			const builtinStylesheet = BreezeFonts.getBuiltinFontsStylesheet();
			const builtIn = await BreezeFonts.getFontFaceData(builtinStylesheet);
			builtIn.forEach(ffd => this.fonts.push(ffd));
		} catch (err) {
			console.log(err.stack || err);
		}
	}

	/**
	 * @param {!boolean} [ignoreCache]
	 * @returns {!Promise}
	 */
	async initializeCustomFonts(ignoreCache = false) {
		try {
			const customFontsMetadata = await this.getCustomFontsMetadata(ignoreCache);
			await this.downloadCustomFonts(customFontsMetadata);
			const customFontsCSS = await this.generateStyles(customFontsMetadata);

			const el = document.querySelector(`#${this.styleSheetId}`);
			Utility.deleteStyleSheet(el);
			Utility.insertStyleSheet(customFontsCSS, this.styleSheetId);

			const customFontsStyle = document.querySelector(`#${this.styleSheetId}`);
			const customStylesheet = BreezeFonts.getCustomFontsStylesheet(customFontsStyle);
			this.customFonts = [];
			const custom = await BreezeFonts.getFontFaceData(customStylesheet);
			custom.forEach(ffd => this.customFonts.push(ffd));
		} catch (err) {
			console.log(err.stack || err);
		}
	}

	/**
	 * Attempts to load the custom fonts metadata from the cache if available or
	 * from the server.
	 * @param {!boolean} [ignoreCache]
	 * @returns {!Promise<?FontDetails>}
	 */
	async getCustomFontsMetadata(ignoreCache = false) {
		try {
			let metadata = null;
			if (this.player.container.features.caching && !ignoreCache) {
				try {
					const buffer = await this.file.load();
					console.log('font buffer: ', buffer);
					if (buffer !== null) {
						metadata = JSON.parse(buffer);
					}
				} catch (err) {
					console.log(`Ignoring error while loading local custom fonts metadata: ${err.stack || err}`);
				}
			}
			if (metadata === null) {
				const url = `${this.player.container.hostUrl}get/customfonts.php?account=${this.player.settings.data.account}`;
				const response = await Network.download(url, 'json');
				if (!response.success) {
					throw new Error('Downloading custom fonts metadata failed.');
				}
				metadata = response.customfonts;
				if (this.player.container.features.caching) {
					try {
						await this.file.save(JSON.stringify(metadata));
					} catch (err) {
						console.log(`Skipping caching custom fonts metadata due to error: ${err.stack || err}`);
					}
				}
			}
			if (metadata === null) {
				throw new Error('Custom fonts metadata failed to load from cache or server.');
			}
			return metadata;
		} catch (err) {
			console.log(`Unable to retrieve custom fonts metadata: ${err.stack || err}`);
		}
	}

	/**
	 * @param {!FontDetails[]} customFontData
	 * @returns {!Promise}
	 */
	async downloadCustomFonts(customFontData) {
		if (this.player.container.features.caching) {
			await Promise.all(
				customFontData.map(async font => {
					const basename = `${font.filename}.${font.extension}`;
					const url = `${this.player.container.hostUrl}customfonts/${basename}`;
					try {
						const file = new FileInfo(
							`${this.file.getParentName()}/${basename}`,
							false,
							null,
							font.mtime
						);
						await this.player.downloader.download(url, file);
					} catch (err) {
						console.log(`Skipping caching font '${basename}' from '${url}' due to error thrown while downloading: ${err.stack || err}`);
					}
				})
			);
		}
	}

	/**
	 * @param {!FontDetails[]} customFontsMetadata
	 * @returns {!Promise<!string>}
	 */
	async generateStyles(customFontsMetadata) {
		const fonts = customFontsMetadata.map(
			font => new FontDetails(font.filename, font.extension,  'fonts', font.mtime)
		);
		const cssArr = await Promise.all(fonts.map(async font => {
			let format = 'woff';
			if (font.extension === 'otf') {
				format = 'opentype';
			} else if (font.extension === 'ttf') {
				format = 'truetype';
			}
			const basename = `${font.filename}.${font.extension}`;
			let url = '';
			if (this.player.container.features.caching) {
				try {
					const file = new FileInfo(
						`${this.file.getParentName()}/${basename}`,
						false
					);
					const fileUrl = await file.getLocalUrl();
					url = `${fileUrl}?mtime=${font.mtime}`;
				} catch (err) {
					console.log(`Error loading font '${basename}' from cache, falling back to downloading from server: ${err.stack || err}`);
				}
			}
			if (url.length === 0) {
				url = `${this.player.container.hostUrl}customfonts/${basename}?mtime=${font.mtime}`;
			}
			return await font.generateFontFace(
				this.player.container.hostUrl,
				this.file.getParentName(),
				this.player.container.features.caching
			);
		}));
		return cssArr.join('');
	}

	/**
	 * Ensures all the custom fonts have a @font-face definition and that the custom
	 * fonts are downloaded into the cache if supported.
	 * @param {!string[]} usedFonts List of font names excluding extension in use by
	 * a Playlist.
	 * @returns {!Promise} Returns a promise that fulfills after verifying 
	 * all of the usedFonts are in {@link BreezeFonts#customFonts} or 
	 * {@link BreezeFonts#fonts}, or until after pulling custom fonts down from the
	 * server and the new fonts have finished caching.
	 */
	async verifyCustomFonts(usedFonts) {
		const preloaded = usedFonts.every(usedFont => {
			const isCustomFont = this.customFonts.some(font => font.family === usedFont);
			const isFont = this.fonts.some(font => font.family === usedFont);
			return isCustomFont || isFont;
		});
		if (!preloaded) {
			await this.initializeCustomFonts(true);
		}
	}

	/**
	 * Ensures all used fonts are layout-ready.
	 * @param {!string[]} usedFonts List of font family names.
	 * @returns {!Promise}
	 */
	async using(usedFonts) {
		await this.verifyCustomFonts(usedFonts);
		const loaders = [];
		this.fonts
			.filter(ffd => usedFonts.includes(ffd.family))
			.map(ffd => new FontFaceObserver(ffd.family, {
				style: ffd.style,
				weight: ffd.weight
			}))
			.forEach(ffo => loaders.push(ffo.load()));
		this.customFonts
			.filter(ffd => usedFonts.includes(ffd.family))
			.map(ffd => new FontFaceObserver(ffd.family))
			.forEach(ffo => loaders.push(ffo.load()));

		await Promise.all(loaders);
	}

	/**
	 * @param {!Element} customFontsStyle
	 * @returns {?CSSStyleSheet}
	 */
	static getCustomFontsStylesheet(customFontsStyle) {
		for (let itr = 0; itr < document.styleSheets.length; ++itr) {
			const stylesheet = document.styleSheets[itr];
			if (stylesheet.ownerNode === customFontsStyle) {
				return stylesheet;
			}
		}
		return null;
	}

	/**
	 * @returns {!CSSStyleSheet}
	 */
	static getBuiltinFontsStylesheet() {
		for (let itr = 0; itr < document.styleSheets.length; ++itr) {
			const stylesheet = document.styleSheets[itr];
			if (stylesheet.href !== null && stylesheet.href.includes('ui/css/googlefonts/googlefonts.css')) {
				return stylesheet;
			}
		}
		throw new Error('Builtin fonts stylesheet is missing!');
	}

	/**
	 * Removes the outer quotes from font family names that contain space(s).
	 * @param {!string} family
	 * @returns {!string}
	 */
	static stripQuotes(family) {
		if (family.includes(' ')) {
			const quoted = family.startsWith('"') && family.endsWith('"');
			if (quoted) {
				return family.substring(1, family.length - 1);
			}
		}
		return family;
	}

	/**
	 * @param {!CSSStyleSheet} stylesheet
	 * @returns {!FontFaceData[]}
	 */
	static getFontFaceData(stylesheet) {
		const ffd = [];
		for (let itr = 0; itr < stylesheet.cssRules.length; ++itr) {
			const rule = stylesheet.cssRules[itr];
			ffd.push({
				family: BreezeFonts.stripQuotes(rule.style.fontFamily),
				style: rule.style.fontStyle,
				weight: rule.style.fontWeight,
				src: rule.style.src
			});
		}
		return ffd;
	}
}

class FontDetails {
	/**
	 * @param {!string} filename The filename excluding the extension.
	 * @param {!string} extension The extension excluding the leading dot. Set to
	 * null if the font file is not a custom font (the built-in fonts only list 
	 * font family names.)
	 * @param {!string} directory
	 * @param {?number} mtime
	 */
	constructor(filename, extension, directory, mtime) {
		this.filename = filename;
		this.extension = extension;
		this.mtime = mtime;

		this.file = new FileInfo(
			`${directory}/${this.filename}.${this.extension}`,
			false,
			null,
			mtime
		);
	}

	/**
	 * The font in 'filename.extension' format.
	 * @returns {!string}
	 */
	getBasename() {
		return `${this.filename}.${this.extension}`;
	}

	/**
	 * @returns {!Promise<?string>}
	 */
	async getCachedResourceUrl() {
		try {
			const fileUrl = await this.file.getLocalUrl();
			return `${fileUrl}?mtime=${this.mtime}`;
		} catch (err) {
			console.log(`Error finding font '${this.file}' in cache, falling back to downloading from server: ${err}`);
		}
		return null;
	}

	/**
	 * @param {!string} hostUrl
	 * @returns {!string}
	 */
	getResourceUrl(hostUrl) {
		return `${hostUrl}customfonts/${this.getBasename()}?mtime=${this.mtime}`;
	}

	/**
	 * @returns {!string}
	 */
	getFormat() {
		let format = 'woff';
		if (this.extension === 'otf') {
			format = 'opentype';
		} else if (this.extension === 'ttf') {
			format = 'truetype';
		}
		return format;
	}

	/**
	 * @param {!string} hostUrl
	 * @param {!boolean} caching
	 * @returns {!string}
	 */
	async generateFontFace(hostUrl, cacheFolder, caching) {
		let url = this.getResourceUrl(hostUrl);
		if (caching) {
			try {
				url = await this.getCachedResourceUrl();
			} catch (err) {
				console.log(`Failed to find '${this.getBasename()}' in cache, falling back to downloading from server: ${err.stack || err}`);
			}
		}
		return `@font-face { font-family: '${this.filename}'; src: url('${url}') format('${this.getFormat()}'); }\n`;
	}
}

/**
 * @typedef {Object} FontFaceData
 * @property {!string} family
 * @property {?string} style
 * @property {?number} weight
 * @property {!string} src
 */

/**
 * @typedef {Object} CustomFontsDetailsResponse
 * @property {!boolean} success Whether the API call was succesfully or not.
 * @property {!FontDetails[]} customfonts An array containing one entry for each
 * custom font on the server, will be an empty array if there are no custom 
 * fonts on the server.
 */

/**
 * @typedef {Object} CustomFontData
 * @property {!string} css The stylesheet text for the cached fonts.
 * @property {!FontDetails[]} fonts The list of cached fonts.
 */
