class BreezeMPV {
	/**
	 * @param {!Node} tag The parent DOM Node that MPV should append itself to.
	 * @param {!number} width The width that MPV should render at.
	 * @param {!number} height The height that MPV should render at.
	 * @param {!Object} options The attribute options
	 */
	constructor(tag, width, height, { url, fillstyle, subtitles, inputtype, inputwidth, inputheight, inputpixelformat, inputframerate }) {
		/*
		 * @param {!boolean} subtitles
		 * @param {!number} inputwidth BlackMagic: The width of the incoming video signal.
		 * @param {!number} inputheight BlackMagic: The height of the incoming video signal.
		 * @param {!string} inputpixelformat BlackMagic: The pixel format of the incoming video signal.
		 * e.g. 'uyvy422'.
		 * @param {!number|!string} inputframerate BlackMagic: The exact frame rate of the incoming video signal.
		*/
		
		/**
		 * The display resolution MPV should render at.
		 * @type {{w: !number, h: !number}}
		 */
		this.displayRes = {
			w: width,
			h: height
		};

		/**
		 * The input signals metadata.
		 * @type {{w: ?number, h: ?number, pixelFormat: ?string, frameRate: ?string}}
		 */
		this.inputFormat = {
			w: null,
			h: null
		};

		this.subtitles = subtitles;

		this.tag = tag;

		/**
		 * Set to false until MPV is initialized and ready for playback.
		 * @type {!boolean}
		 */
		this.ready = false;

		this.mpvToDelete = null;

		/**
		 * Enumeration values copied from editor/database.
		 */
		this.FILL_STYLES = { NONE: 1, COVER: 2, ASPECT_RATIO: 3 };
		
		/**
		 * The currently set fillstyle, may not reflect the latest fillstyle set by
		 * the user.
		 */
		this.fillStyle = this.FILL_STYLES.ASPECT_RATIO;

		/**
		 * The last set fillstyle by the user.
		 */
		this.requestedFillStyle = fillstyle;

		this.volume = 100;
		
		console.log('Breeze MPV Create');
		/**
		 * The DOM node that the MPV plugin will attach to. The postMessage API can
		  * be used on this Node to send and receive messages from the MPV plugin.
		 * @type {?Node}
		*/
		this.mpv = this.tag.appendChild($.object({
			type: 'application/x-mpvjs',
			width: this.displayRes.w,
			height: this.displayRes.h
		}));
		this.mpv.addEventListener('message', msg => {
			if (msg.data.type === 'ready') {
				this.ready = true;
				
				// Configure an IPC socket that we can communicate with
				const socketId = Math.floor(Math.random() * 99999);
				this.setProperty('input-ipc-server', `/mpvsocket${socketId}`);
				
				// These help with performance on the 352s
				this.setProperty('opengl-swapinterval', 0);
				this.setProperty('opengl-glfinish', true);
				this.setProperty('display-fps', 30);
				this.setProperty('ao', 'wasapi');

				// Stylize subtitles
				this.setProperty('sub-color', '#FFFFFF');
				this.setProperty('sub-back-color', '#000000');
				this.setProperty('sub-border-size', 0);
				this.setProperty('sub-font-size', 38);

				// Performance fixes for subtitles
				this.setProperty('sub-fix-timing', true);
				this.setProperty('sub-fps', 1);
				this.setProperty('sub-speed', 10.0);
				this.setProperty('force-rgba-osd-rendering', true);
				this.setProperty('vf', 'format=420p,scale,sub');
				
				if (this.subtitles) {
					this.setProperty('sid', 'auto');
					this.setProperty('sub-create-cc-track', true);
					this.setProperty('sub-visibility', true);
				} else {
					this.setProperty('sub-visibility', false);
				}
				
				this.setProperty('volume', this.volume);
				
				// MPV starts out paused, so unpause it
				this.setProperty('pause', false);
				
				/**
				 * Refer to the MPV documentation (mpv.io/manual/) for supported protocols.
				 */
				let mrl = '';
				
				if (inputtype === 'blackmagicinput') {
					// DO NOT USE UNTIMED WITH RTP OR RTSP. The audio definitely desyncs instantly
					// untimed might mess up audio syncing, but it stops stuttering when the user unplugs hdmi or switches between playlists with hdmi
					this.setProperty('untimed', true);

					this.inputFormat.w = inputwidth;
					this.inputFormat.h = inputheight;
					
					this.setPropertyString('demuxer', 'lavf');
					// fflags=+genpts:fflags=+igndts:use_wallclock_as_timestamps=1 probably aren't valid.
					const avOptions = `video_size=${inputwidth}x${inputheight}:pixel_format=${inputpixelformat}:framerate=${inputframerate}:rtbufsize=702000k`;
					this.setPropertyString('demuxer-lavf-o', avOptions);
					mrl = 'av://dshow:video=Decklink Video Capture:audio=Decklink Audio Capture';
				} else {
					mrl = url;
				}
				
				this.runCommand('loadfile', mrl);
				
				['width', 'height', 'track-list', 'track-list/count', 'track-list/17/id'].forEach(this.observeProperty, this);
				const xpipe = require('xpipe');
				const net = require('net');
				
				this.client = null;
				
				// Wait 1 second just to be sure MPV is ready to listen
				setTimeout(() => {
					if (this.mpv) {
						// Using an IPC call to communicate with mpv
						this.client = net.connect(xpipe.eq(`/mpvsocket${socketId}`), () => {
							console.log('connected to IPC server');
							let command = JSON.stringify({ 'command': ['request_log_messages', 'info'], 'request_id': 100 });
							this.client.write(Buffer.from(command + '\n'));
							command = JSON.stringify({ 'command': ['enable_event', 'end-file'], 'request_id': 101 });
							this.client.write(Buffer.from(command + '\n'));
						});
						
						this.client.on('data', response => {
							response = response.toString('utf8');
							response = response.trim();
							response = `[${response}]`;
							response = response.replace(/(\r\n|\n|\r)/g, ',');
							response = JSON.parse(response);
							for (const event of response) {
								if (event.event === 'log-message') {
									if (event.hasOwnProperty('level') && event.hasOwnProperty('text')) {
										console.log(`log-message: ${event.level} - ${event.text}`);
									}
								} else if (event.event === 'end-file') {
									if (this.mpvToDelete) {
										this.mpvToDelete.remove();
										this.mpvToDelete = null;
										this.client.end();
									}
								}
							}
						});
					}
				}, 1000);
			} else if (msg.data.type === 'property_change') {
				console.log(`${msg.data.data.name}: ${msg.data.data.value}`);
				if (msg.data.data.name === 'track-list/count') {
					const trackCount = msg.data.data.value;
					for (let i = 0; i <= trackCount; ++i) {
						this.observeProperty(`track-list/${i}/type`);
					}
				}
				if (/^track-list\/\d*\/type$/.test(msg.data.data.name)) {
					if (msg.data.data.value === 'sub') {
						const matches = /\/(\d*)\//.exec(msg.data.data.name);
						if (matches !== null) {
							if (matches.length == 2) {
								const id = Number(matches[1]);
								//console.log(`Observing property track-list/${id}/id`);
								this.observeProperty(`track-list/${id}/id`);
							}
						}
					}
				}
				const checkFillStyle = () => {
					if (this.requestedFillStyle != this.fillStyle) {
						this.fillStyle = this.requestedFillStyle;
						switch (this.fillStyle) {
							case 1: // this.FILL_STYLES.NONE
								this.setVideoZoom(0);
								this.setAspectRatio(this.displayRes.w / this.displayRes.h);
								break;
							case 2: // this.FILL_STYLES.COVER
								if (this.inputFormat.w !== null && this.inputFormat.h !== null) {
									const srcW = this.inputFormat.w;
									const srcH = this.inputFormat.h;
									const dstW = this.displayRes.w;
									const dstH = this.displayRes.h;
									const srcRatio = srcW / srcH;
									const dstRatio = dstW / dstH;
									if (srcRatio > dstRatio) {
										this.setAspectRatio((dstH * srcRatio) / dstH);
									} else if (dstRatio > srcRatio) {
										this.setAspectRatio(dstW / (dstW * srcRatio));
									}
									const deltaW = srcW >= dstW ? srcW - dstW : dstW - srcW;
									const deltaH = srcH >= dstH ? srcH - dstH : dstH - srcH;
									if (deltaW >= deltaH) {
										this.setVideoZoom(Math.log2(1 + (deltaW / srcW)));
									} else {
										this.setVideoZoom(Math.log2(1 + (deltaH / srcH)));
									}
								}
								break;
							case 3: // this.FILL_STYLES.ASPECT_RATIO
								if (this.inputFormat.w !== null && this.inputFormat.h !== null) {
									this.setVideoZoom(0);
									this.setAspectRatio(this.inputFormat.w / this.inputFormat.h);
								}
								break;
							default:
								throw new Error(`Invalid fillstyle requested: ${this.requestedFillStyle}`);
						}
					}
				};
				if (msg.data.data.name == 'width') {
					this.inputFormat.w = msg.data.data.value;
					checkFillStyle();
				} else if (msg.data.data.name == 'height') {
					this.inputFormat.h = msg.data.data.value;
					checkFillStyle();
				}
			}
		});
	}
	
	/**
	 * @param {!string} type
	 * @param {?Object} data
	 */
	post(type, data) {
		if (this.mpv) {
			this.mpv.postMessage({
				type: type,
				data: data
			});
		}
	}
	
	/**
	 * @param {!string} name
	 */
	observeProperty(name) {
		this.post('observe_property', name);
	}
	
	/**
	 * @param {!string} name
	 * @param {?Object} value
	 */
	setProperty(name, value) {
		this.post('set_property', {
			name: name,
			value: value
		});
	}
	
	/**
	 * @param {!string} name
	 * @param {!string} value
	 */
	setPropertyString(name, value) {
		this.post('set_property_string', [name, value]);
	}
	
	/**
	 * @param {!string} command
	 * @param {!string} [parameter]
	 */
	runCommand(command, parameter) {
		this.post('command', [command, parameter]);
	}
	
	/**
	 * Stops any and all playback from MPV, and clears the internal MPV playlist.
	 */
	destroy() {
		console.log('Breeze MPV Destroy');
		if (this.ready) {
			// These must be called or Chrome locks up when the HDMI is playing when the object tag is removed
			this.setProperty('pause', true);
			this.runCommand('stop');
		}
		if (this.mpv) {
			this.mpvToDelete = this.tag.removeChild(this.mpv);
			this.mpv = null;
		}
	}

	/**
	 * Sets the MPV master volume to the given value.
	 * @param {!number} value Set to 0 to mute, or 100 for normal volume. May exceed
	 * 100 at cost of potential artifacting.
	 */
	setVolume(value) {
		this.volume = Math.round(value * 100);
		if (this.ready) {
			this.setProperty('volume', this.volume);
		}
	}

	/**
	 * Sets the aspect ratio that MPV should display the video at. MPV defaults to 
	 * source materials aspect ratio.
	 * @param {!number} value
	 */
	setAspectRatio(value) {
		this.setPropertyString('video-aspect', value.toString());
	}

	/**
	 * Sets the display scale factor. MPV will calculate the zoom scale by 2^value.
	 * E.g., passing the value 0 results in a 1x zoom scale as 2^0 = 1. A value of 
	 * 1 results in a 2x zoom scale as 2^1 = 2. A value of -1 results in a 0.5x 
	 * scale as 2^-1 = 0.5, etc.
	 * @param {!number} value
	 */
	setVideoZoom(value) {
		this.setPropertyString('video-zoom', value.toString());
	}
}
