/**
 * CloudCoverMusic streaming and playback/volume controls.
 */
class CCMProvider {
	/**
	 * @param {!string} url CloudCoverMusic auto-login streaming URL that points to the background audio
	 * location stream.
	 */
	constructor(url) {
		this.url = url;

		/**
		 * An element that holds a hidden CloudCoverMusic player.
		 * @type {?HTMLIFrameElement}
		 */
		this.el = null;

		/**
		 * The volume of the player, 0 is muted and 1 is max volume.
		 * @type {!number}
		 */
		this.volume = 0;

		/**
		 * @type {!boolean}
		 */
		this.playing = false;

		this.duckAmount = 0.05;
	}

	/**
	 * Attaches the CCMProvider to the DOM. This will cause the CCM player to begin
	 * playing music.
	 */
	play() {
		if (this.playing) {
			this.stop();
		}
		this.el = CCMProvider.createHiddenPlayer(this.url);
		document.body.appendChild(this.el);
		this.playing = true;
		setTimeout(() => { this.setVolume(1); }, 1000);
	}

	/**
	 * Detaches (removes) the CCMProvider from the DOM. This will cause the CCM 
	 * player to stop playing music.
	 */
	stop() {
		if (this.playing) {
			if (this.el !== null) {
				this.el.remove();
				this.el = null;
			}
			this.playing = false;
		}
		// Stopping should reset the ducked state
		this.volume = 1;
	}

	/**
	 * Sets the volume to a value on the interval [0, 1].
	 * @param {!number} value
	 */
	setVolume(value) {
		const update = this.el !== null && this.volume !== value;
		this.volume = value;
		if (update) {
			this.el.contentWindow.postMessage(`set volume=${value}`, '*');
		}
	}

	mute() {
		this.setVolume(0);
	}

	duckDown() {
		this.setVolume(0); // Math.max(0, this.volume - this.duckAmount)
	}

	duckUp() {
		this.setVolume(1); // Math.min(1, this.volume + this.duckAmount)
	}

	/**
	 * Creates an iframe that loads a CloudCoverMusic background streaming
	 * auto-login URL.
	 * @returns {!Element} Returns the new hidden iframe element.
	 */
	static createHiddenPlayer(url) {
		return $.iframe({ src: url, allow: 'autoplay', hidden: true });
	}
}
