/* exported Container */
/* global WidgetsSchema, FileInfo, Utility */

async function addStyle(src) {
	if (document.querySelector(`link[href="${src}"]`) == null) {
		await new Promise(resolve => {
			const link = document.createElement('link');
			link.onload = resolve;
			link.rel = 'stylesheet';
			link.href = src;
			document.head.appendChild(link);
		});
	}
}

async function addScript(src) {
	if (document.querySelector(`script[src="${src}"]`) == null) {
		await new Promise(resolve => {
			const script = document.createElement('script');
			script.setAttribute('src', src);
			script.onload = resolve;
			document.head.appendChild(script);
		});
	}
}

async function LoadScripts() {
	await Promise.all(
	[
		'ui/css/style.css',
		'ui/css/player.css',
		'ui/css/googlefonts/googlefonts.css',
		'javascript/fullcalendar/main.css',
		'ui/css/materialdesignicons.min.css',
		'ui/css/ol.css'
	].map(path => addStyle(path)));
	
	// shared libraries
	await addScript('javascript/moment.min.js'); // These 3 libraries have very specific load orders because of fullcalendar-moment.js
	await addScript('javascript/fullcalendar/main.js');
	await addScript('javascript/fullcalendar-moment.js');
	await Promise.all(
	[
		'javascript/moment-timezone-with-data.min.js',
		'javascript/minibus.js',
		'javascript/csv.min.js',
		'javascript/pdf.js',
		'javascript/hls.min.js',
		'javascript/fontfaceobserver.standalone.js',
		'javascript/ol.js',
		'javascript/caps/caps.capsloader.js',
		'javascript/caps/caps.filter.js'
	].map(path => addScript(path)));
	
	// shared modules
	await Promise.all(
	[
		'javascript/common.js',
		'javascript/web-animations.min.js',
		'javascript/dom.js',
		'javascript/widget.js',
		'javascript/alertviewer.js'
	].map(path => addScript(path)));
	
	// third party libraries
	await Promise.all(
	[
		'player/thirdparty/base64.js',
		'player/thirdparty/dexie.js',
		'player/thirdparty/es6-object.assign.js',
		'player/thirdparty/fetch-polyfill.js',
		'player/thirdparty/web-animations.min.js',
		'player/thirdparty/breezelib-schedule.js'
	].map(path => addScript(path)));
	
	// player modules
	await Promise.all(
	[
		'player/routemap.js',
		'player/timermap.js',
		'player/timesync.js',
		'player/timer.js',
		'player/network.js',
		'player/utility.js',
		'player/type.js',
		'player/loadingscreen.js',
		'player/messenger.js',
		'player/files.js',
		'player/downloader.js',
		'player/schedule/playerschedule.js',
		'player/schedule/breezeschedule.js',
		'player/schedule/scheduleloader.js',
		'player/schedule/schedulepreloader.js',
		'player/playlist/loadingplaylist.js',
		'player/breezempv.js',
		'player/breezefonts.js',
		'player/transitions.js',
		'player/widgets.js',
		//'player/container.js',
		'player/watermark.js',
		'player/backgroundmusicservice.js',
		'player/ccmprovider.js',
		'player/media.js',
		'player/zone.js',
		'player/playlist.js',
		'player/playersettings.js',
		'player/checkin.js',
		'player/player.js'
	].map(path => addScript(path)));
	
	// widgets
	await Promise.all(
	[
		'javascript/breezecrawl.js',
		'javascript/powerbi.js'
	].map(path => addScript(path)));
	
	const baseURL = await ContainerAPI.QueryAssetsPath();
	const widgetsSchema = await WidgetsSchema.load(baseURL);
	const imports = [];
	for (const widget of widgetsSchema) {
		const path = `${baseURL}widgets/${widget.name}`;
		imports.push(addScript(`${path}.js`));
		imports.push(addStyle(`${path}.css`));
	}
	await Promise.all(imports);
}

// The Container shall return a null value for any unsupported
// operation that expects a response.
class Container {
	/**
	 * @param {?Player} player
	 */
	constructor(player) {
		this.player = player;

		this.baseUrl = '';
		this.hostUrl = '';
		this.type = '';
		this.widgetsSchema = [];
		this.fonts = null;
		this.metadata = null;
		this.features = {
			caching: null,
			mpv: null,
			screenshots: null,
			led:  null,
			videoIn: null,
			networking: null
		};
	}

	/** @returns {!Promise} */
	async initialize() {
		const registration = await ContainerAPI.QueryRegistration();
		this.hostUrl = `https://${registration.subdomain}.breezedigitalsignage.com/`;
		this.baseUrl = await ContainerAPI.QueryAssetsPath();
		this.type = await ContainerAPI.QueryContainerType();
		this.widgetsSchema = await WidgetsSchema.load(this.baseUrl);
		this.disableFrontLed();
		await this.queryFeatures();
	}

	/** @returns {!Promise} */
	async queryFeatures() {
		await Promise.all(
			Object.keys(this.features)
				.map(async feature => {
					this.features[feature] = await ContainerAPI.QuerySupport(feature);
				}));
	}

	/* -----------------------------------
	 *  Legacy
	 * ----------------------------------- */

	/**
	 * @param {!number} widgetId
	 * @param {?Object} data
	 */
	async writeWidget(widgetId, data) {
		if (this.features.caching) {
			try {
				const json = JSON.stringify(data);
				const file = new FileInfo(
					`${Media.WIDGET_DATA_DIRECTORY}/${widgetId}.json`,
					false,
					json.length
				);
				await file.save(json);
				return;
			} catch (err) {
				console.log(`Skipping saving data into cache for widget (id => ${widgetId}) due to error: ${err}`);
			}
		}
		return;
	}

	/**
	 * @param {!number} widgetId
	 * @returns {!Promise<?Object>}
	 */
	async readWidget(widgetId) {
		if (this.features.caching) {
			try {
				const file = new FileInfo(
					`${Media.WIDGET_DATA_DIRECTORY}/${widgetId}.json`,
					false
				);
				const buffer = await file.load();
				if (buffer !== null) {
					return JSON.parse(buffer);
				}
			} catch (err) {
				console.log(`Skipping reading data from cache for widget (id => ${widgetId}) due to error: ${err}`);
			}
		}
		return null;
	}

	/**
	 * @param {!Media} media
	 * @returns {!Promise<?string>}
	 */
	async cacheMediaObject(media) {
		if (media === null || media === undefined) {
			return null;
		}
		if (this.features.caching) {
			try {
				const type = Media.parseType(media);
				const dir = Media.getLocalDirectory(type);
				const name = Media.getLocalName(media.cdnurl, type, media.id);
				const file = new FileInfo(
					`${dir}/${name}`,
					false,
					media.filesize,
					Utility.parseMtime(media.uploadedtimestamp),
					null,
					media.md5
				);
				await this.player.downloader.download(media.cdnurl, file);
				const stats = await file.stat();
				return stats.getLocalUrl();
			} catch (err) {
				console.log(`Skipping caching of media object due to error: ${err.stack || err}`);
			}
		}
		return null;
	}

	/** @param {!string} rgb */
	async setFrontLedColor(rgb) {
		if (this.features.led) {
			try {
				const {r, g, b} = rgbaToColor(rgb);
				await ContainerAPI.SetFrontLedColor(r, g, b);
			} catch (err) {
				console.log(`Ignoring error thrown while attempting to set front LED color: ${err.stack || err}`);
			}
		}
	}

	disableFrontLed() {
		this.setFrontLedColor(Utility.ledOffColor);
	}

	/**
	 * Represents either an entry or exit time for a given Media object.
	 * @typedef {Object} MediaEvent
	 * @property {Media} media The media the MediaEvent refers to.
	 * @property {number} defaultTime The untransformed playback time of the 
	 * Event, refer to {@link Media#updateMediaEvents} to see how they are 
	 * calculated.
	 * @property {number} nthTime The transformed playback time of the Event, 
	 * refer to {@link Zone#generateMediaEvents} to see how they are 
	 * calculated.
	 */

	/**
	 * Contains network request details.
	 * @typedef {Object} NetworkCommand
	 * @property {string} command The data to send once a network
	 * connection has successfully been established.
	 * @property {string} commandformat Either "hex" or "ascii",
	 * specifies the format that command should be interpreted in.
	 * @property {string} protocol "rs232", "tcp" or "udp"
	 * @property {string} port Represents the serial path (e.g. 'COM1' or 
	 * '/dev/ttyS0') when protocol is "rs232", alternatively represents
	 * the port that will be used when establishing a TCP/UDP connection
	 * to ip.
	 * @property {string} ip The IP to try to connect when protocol is
	 * "tcp" or "udp", requires a port to be specified.
	 * @property {string} baudrate The serial baudrate
	 * @property {string} databits The serial databits
	 * @property {string} parity The serial parity
	 * @property {string} stopbits The serial stopbits
	 * @property {string} flowcontrol The serial flowcontrol
	 */

	/** @param {NetworkCommand} command */
	static async networkCommand(command) {
		const port = command.protocol === 'rs232' ?
			command.port :
			parseInt(command.port, 10);
		const baudrate = command.baudrate !== undefined ?
			parseInt(command.baudrate, 10) :
			undefined;
		const databits = command.databits !== undefined ?
			parseInt(command.databits, 10) :
			undefined;
		const stopbits = command.stopbits !== undefined ?
			parseFloat(command.stopbits) :
			undefined;
		let flowcontrol = undefined;
		if (command.flowcontrol === 'true') {
			flowcontrol = true;
		} else if (command.flowcontrol === 'false') {
			flowcontrol = false;
		}
		if (command.protocol === 'tcp' || command.protocol === 'udp') {
			await ContainerAPI.NetworkRequest({
				host: command.ip,
				port: port,
				command: command.command,
				format: command.commandformat
			});
		} else if (command.protocol === 'rs232') {
			await ContainerAPI.SerialRequest({
				port: port,
				command: command.command,
				format: command.commandformat,
				baudRate: baudrate,
				dataBits: databits,
				parity: command.parity,
				stopBits: stopbits,
				flowControl: flowcontrol
			});
		}
	}
}
