/* exported Downloader */
/* global FileIntegrity */

const Downloader = (() => {
	return class Downloader {
		/**
		 */
		constructor() {
			this.downloads = new Map();
		}

		/**
		 * @param {!FileInfo} file
		 */
		abort(file) {
			const key = file.computeKey();
			console.log(`this.downloads.get(${key})`);
			const download = this.downloads.get(key);
			if (download) {
				console.log(`aborting download: ${key}`);
				download.abort(new Error('download aborted'));
				console.log(`this.downloads.delete(${key})`);
				this.downloads.delete(key);
			}
		}

		/**
		 * This can be called as many times as needed. Will only download a single file and resolves when a valid file is local.
		 * Note: The cache system deletes the least recently used file. It's possible for this to run then another set of downloads to fill the cache and this local file is deleted.
		 * This is why codepaths that get the local file path check for null.
		 * @param {!string} url
		 * @param {!FileInfo} file
		 * @returns {!Promise}
		 */
		async download(url, file) {
			const key = file.computeKey();
			console.log(`this.downloads.get(${key})`);
			let download = this.downloads.get(key);
			if (download == undefined) {
				let abort;
				download = { promise: null, abortPromise: new Promise((resolve, reject) => abort = reject) };
				download.abort = abort;
				console.log(`this.downloads.set(${key}, ...)`);
				this.downloads.set(key, download);
			} else {
				console.log(`Download from '${url}' to '${file}' requested, but integrity check or download already in progress.`);
			}
			
			if (download.promise == null) {
				download.promise = (async () => {
					// Limit the amount of download and hash checks before giving up
					// Expected sequence when the file is in the cache: integrity -> resolve
					// Expected sequence when file is not in the cache: integrity, download, integrity -> resolve
					// Expected sequence if there's a hash or download problem: integrity, download, integrity, download, integrity -> reject
					for (let failureCount = 0; failureCount < 3; ++failureCount) {
						const integrityStart = Date.now();
						// Note the abort can throw here as expected
						const integrity = await Promise.race([await FileIntegrity.verify(file), download.abortPromise]);
						console.log(`Integrity for '${key}' took ${Date.now() - integrityStart}ms`);
						if (integrity) {
							// The file exists and the hash matches
							console.log(`this.downloads.delete(${key})`);
							this.downloads.delete(key);
							return;
						} else {
							if (failureCount == 2) {
								break;
							}
							// The download does not exist or the hash does not match, download the file
							console.log(`Download from '${url}' to '${file}' requested, starting download!`);
							// Note the abort can throw here as expected
							try {
								await Promise.race([await ContainerAPI.DownloadFile(url, file.getName(), file.getParentName(), file.size), download.abortPromise]);
							} catch (exception) {
								console.log(`Failed to download file '${file}' from '${url}', container responded with: ${exception}`);
							}
						}
					}
					console.log(`this.downloads.delete(${key})`);
					this.downloads.delete(key);
					throw new Error('Failed twice to download and/or verify the hash.');
				})();
			}
			await Promise.race([download.promise, download.abortPromise]);
		}
	};
})();
