/* exported FileInfo, LocalFile, FilePath, FileIntegrity */
/* global Utility */

class FilePath {
	/**
	 * @param {?(string|FilePath)} [path]
	 */
	constructor(path = '') {
		if (path instanceof FilePath) {
			this.path = path.path;
		} else if (path === null) {
			this.path = '';
		} else {
			this.path = path.trim();
		}
		// Remove trailing slashes
		while (this.path.endsWith('/')) {
			this.path = this.path.substring(0, this.path.length - 1);
		}
	}

	/**
	 * @returns {!string}
	 */
	getExt() {
		const index = this.path.lastIndexOf('.');
		return this.path.substring(index + 1);
	}

	/**
	 * @returns {!string}
	 */
	getName() {
		const index = this.path.lastIndexOf('/');
		return this.path.substring(index + 1);
	}

	/**
	 * @returns {!string}
	 */
	getParentName() {
		const parent = this.getParent();
		return parent.getName();
	}

	/**
	 * @returns {!FilePath}
	 */
	getParent() {
		const name = this.getName();
		const path = this.path.substring(0, this.path.length - name.length - 1);
		return new FilePath(path);
	}

	/**
	 * Converts 'dir/name' to ['', 'dir', 'name']
	 * @returns {!string[]}
	 */
	toArray() {
		const arr = [];
		let path = this;
		let name;
		do {
			name = path.getName();
			arr.unshift(name);
			path = path.getParent();
		} while (name !== '');
		return arr;
	}

	/**
	 * @param {!FilePath} path
	 * @returns {!boolean}
	 */
	includes(path) {
		const thisArr = this.toArray();
		const pathArr = path.getParent().toArray();
		if (thisArr.length > pathArr.length) {
			return false;
		}
		return thisArr.every((name, index) => {
			return name === pathArr[index];
		});
	}

	/**
	 * @param {!FilePath} parent
	 * @returns {?FilePath}
	 */
	relativeFrom(parent) {
		if (!parent.includes(this)) {
			return null;
		}
		return this.toArray().slice(parent.toArray().length).join('/');
	}

	/**
	 * @param {?(string|FilePath)} [path]
	 * @returns {!FilePath}
	 */
	append(path = '') {
		if (path === null) {
			path = '';
		}
		if (path instanceof FilePath) {
			path = path.path;
		}
		return new FilePath(`${this.path}/${path}`);
	}

	/**
	 * @param {!FilePath} file
	 * @returns {!boolean}
	 */
	equals(file) {
		return this.path === file.path;
	}

	/**
	 * @returns {!string}
	 * @override
	 */
	toString() {
		return this.path;
	}
}

class FileInfo extends FilePath {
	/**
	 * @param {!string} path
	 * @param {!boolean} isDirectory
	 * @param {?number} [size]
	 * @param {?number} [mtime]
	 * @param {?number} [atime]
	 * @param {?string} [md5Hash]
	 * @param {?string} [localUrl]
	 * @param {?boolean} [exists]
	 */
	constructor(path, isDirectory, size = null, mtime = null, atime = null, md5Hash = null, localUrl = null, exists = null) {
		super(path);
		this.isDirectory = isDirectory;
		this.size = size;
		this.mtime = mtime;
		this.atime = atime;
		this.md5Hash = md5Hash;
		this.url = localUrl;
		this._exists = exists;
		this.saving = false;
		this.promise = null;
	}

	/**
	 * @returns {!string}
	 */
	computeKey() {
		return `${this.path}${this.mtime}`;
	}

	/**
	 * @returns {!Promise<?string>}
	 */
	async md5() {
		try {
			return await ContainerAPI.CacheMd5(this.getName(), this.getParentName());
		} catch (exception) {
			console.log(exception);
			return null;
		}
	}

	/**
	 * @returns {!Promise<!boolean>}
	 */
	async exists() {
		const stats = await this.stat();
		return stats !== null;
	}

	/**
	 * @returns {!Promise<?string>}
	 */
	async getLocalUrl() {
		const stats = await this.stat();
		if (stats === null) {
			return null;
		}
		return stats.url;
	}

	/**
	 * Returns null if this does not represent a directory.
	 * @returns {!Promise<?string[]>}
	 */
	async listFiles() {
		if (!this.isDirectory) {
			return null;
		}
		return await ContainerAPI.ListFiles(this.getName());
	}

	/**
	 * Returns a new FileInfo object.
	 * @returns {!Promise<?FileInfo>}
	 */
	async stat() {
		const name = this.getName();
		const dir = this.getParentName();
		try {
			const csr = await ContainerAPI.CacheStat(name, dir);
			return new FileInfo(
				this,
				csr.isDirectory,
				csr.size,
				csr.mtime,
				csr.atime,
				csr.md5,
				csr.url,
				csr.exists
			);
		} catch (exception) {
			return null;
		}
	}

	/**
	 * @returns {!Promise<?string>}
	 */
	async load() {
		const name = this.getName();
		const dir = this.getParentName();
		try {
			const data = await ContainerAPI.CacheLoad(name, dir);
			if (data == null) {
				throw new Error('Invalid response from Container for CacheLoad request.');
			}
			return data;
		} catch (exception) {
			throw new Error(`Failed to load file '${dir}/${name}': ${exception}`);
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async delete() {
		if (this.isDirectory) {
			await ContainerAPI.DeleteFolder(this.getName());
		} else {
			await ContainerAPI.DeleteFile(this.getName(), this.getParentName());
		}
	}

	/**
	 * @param {!string} data
	 */
	async save(data) {
		if (this.saving) {
			await this.promise;
			return;
		}
		this.saving = true;
		const dir = this.getParentName();
		const name = this.getName();
		this.promise = ContainerAPI.SaveFile(data, name, dir);
		try {
			await this.promise;
		} catch (exception) {
			throw new Error(`SaveFile failed for '${dir}/${name}': ${exception}`);
		} finally {
			this.saving = false;
		}
		return;
	}
}

class FileIntegrity {
	/**
	 * @param {!FileInfo} file
	 * @returns {!Promise<!boolean>}
	 */
	static async verify(file) {
		if (file.md5Hash) {
			const md5 = await file.md5();
			return file.md5Hash === md5;
		} else {
			const stats = await file.stat();
			if (stats) {
				if (file.mtime !== null) {
					return stats.mtime >= file.mtime;
				}
				return true;
			}
			return false;
		}
	}
}
