/* exported Media, Widget, Composite */
/* global WidgetsSchema, Point, Size, TimerMap, FileInfo, Type, Utility, BreezeSchedule, CreateWidget, Minibus, Transition, BreezeMPV, moment, CreateWidget */

class Composite {
	/**
	 * @param {!Type.Transition} type
	 * @param {!string} id
	 * @param {!Size} size
	 * @param {!Point} position
	 * @returns {?HTMLElement}
	 */
	static generate(type, id, size, position) {
		switch (type) {
		case Type.Transition.FADE_TO_BLACK:
			return Composite.generateSolid(id, '#000000', size, position);
		case Type.Transition.FADE_TO_WHITE:
			return Composite.generateSolid(id, '#ffffff', size, position);
		}
		return null;
	}

	/**
	 * Generates a solid composite DOM element.
	 * @param {!string} mediaId 
	 * @param {!string} color A valid HTML/CSS color string, e.g. '#FF0000'
	 * @param {!Size} size
	 * @param {!Point} position
	 * @returns {!HTMLElement}
	 */
	static generateSolid(mediaId, color, size, position) {
		return $.div({ id: `${mediaId}-composite-${Utility.getUID()}`, css: {
			position: 'absolute',
			pointerEvents: 'none',
			width: `${size.width}px`,
			height: `${size.height}px`,
			left: `${position.x}px`,
			top: `${position.y}px`,
			backgroundColor: color,
			visibility: 'hidden',
			transform: 'translate(-10000px,-10000px)',
			transformOrigin: 'top left'
		} });
	}
}

class MiscMedia {
	/**
	 * @param {!string} url
	 * @param {!number} width
	 * @param {!number} height
	 * @returns {!HTMLElement}
	 */
	static generateElement(url, width, height) {
		const $object = $.object({
			data: url,
			type: 'application/x-shockwave-flash',
			width,
			height
		});
		const createParam = (name, value) => {
			$object.appendChild($.param({ name, value }));
		};
		createParam('quality', 'high');
		createParam('bgcolor', '#000000');
		createParam('play', 'true');
		createParam('loop', 'true');
		createParam('wmode', 'opaque');
		createParam('scale', 'showall');
		createParam('menu', 'true');
		createParam('devicefont', 'false');
		createParam('salign', '');
		createParam('allowScriptAccess', 'sameDomain');
		return $object;
	}
}

class Media {
	/**
	 * @param {!Zone} zone
	 * @param {!Object} data
	 */
	constructor(
		zone,
		data
	) {
		this.player = zone.player;
		this.zone = zone;
		WidgetsSchema.regenerateWidgetAttributes(this.player.container.widgetsSchema, data.widgettype, data.attributes);

		/**
		 * A unique value across Media items generated when a playlist is saved in the Breeze editor.
		 * Underlying resources (e.g. cache data) that can be safely shared between Media instances (especially
		 * when they refer to the same Media object) should be based on this value instead of the 
		 * Media#id which refers to resources that cannot be safely shared (e.g. a DOM elements ID which must
		 * be unique throughout the entire DOM.)
		 * @type {!number}
		 */
		this.id = data.id;

		this.name = data.name;

		this.maxVolume = this.zone.maxVolume * Media.parseVolume(data.volume);
		this.duckedVolume = this.maxVolume * Media.parseVolume(data.ducking);

		this.duration = Media.parseTimeUnit(data.runlength);
		this.startTime = Media.parseTimeUnit(data.starttime);

		this.widget = null;
		this.widgettype = data.widgettype;
		this.attributes = data.attributes;
		this.type = Media.parseType(data);
		this.animationType = Media.parseAnimation(data.zoompan);

		this.url = Media.parseUrl(data.cdnurl, this.type, this.player.isPreview);
		this.uploadedTimestamp = Utility.parseMtime(data.uploadedtimestamp);
		this.fileSize = data.filesize;

		this.position = new Point(data.x, data.y);
		this.size = new Size(data.width, data.height);

		this.entryTransitionType = Media.parseTransitionType(data.entrytransitiontype);
		this.entryTransitionDuration = Media.parseTransitionDuration(data.entrytransitionduration, this.duration, this.entryTransitionType);
		this.exitTransitionType = Media.parseTransitionType(data.exittransitiontype);
		this.exitTransitionDuration = Media.parseTransitionDuration(data.exittransitionduration, this.duration, this.exitTransitionType);

		this.schedule = new BreezeSchedule(this.player, data);

		this.videoSource = 'HDMI';
		this.videoSourceId = 0;

		const dir = Media.getLocalDirectory(this.type);
		const name = Media.getLocalName(this.url, this.type, this.id);
		this.file = new FileInfo(`${dir}/${name}`, false, this.fileSize, this.uploadedTimestamp, null, data.hash);
		
		// Assign the player metadata to this media by making a copy. Each media has its own metadata table that is overwritten by buttons and other actions
		this.metadata = JSON.parse(JSON.stringify(this.player.container.hasOwnProperty('metadata') && this.player.container.metadata != undefined ? this.player.container.metadata : []));
		
		/**
		 * Based on the underlying Media#id, acts as a unique ID across all Media instances.
		 * @type {!string}
		 */
		this.elId = `${this.id}-${Utility.getUID()}`;

		/**
		 * Indicates the play-state of video media.
		 * @type {!Type.PlayState}
		 */
		this.playing = Type.PlayState.STOPPED;

		/** @type {?moment} */
		this.startedAt = null;

		// This is either the server mediaurl.php url, the CDN url, or the local file url
		this.mediaURL = this.getMediaUrl();

		this.lastMediaInPlaylist = false;
		this.currentVolume = 1;
		this.visibility = Type.Visibility.HIDDEN;

		this.timers = new TimerMap();

		/** @type {!number} */
		this.startTime = this.startTime;

		/** @type {?number} */
		this.seekTo = null;

		/** @type {?moment} */
		this.seekToTime = null;

		/** @type {?HTMLElement} */
		this.el = Media.createMediaElement(this.type);

		this.breezeMpv = null;

		this.destroyPromise = new Promise(resolve => this.destroyResolve = resolve);
		this.cacheRequestedPromise = new Promise(resolve => this.cacheRequestedResolve = resolve);
		this.cachePromise = null;
		this.loadedPromise = null;

		this.entry = {
			transition: new Transition(
				this.el,
				this.entryTransitionDuration,
				this.size,
				this.position,
				zone.size,
				zone.position,
				this.entryTransitionType,
				Type.PlaybackEvent.ENTRY
			),
			composite: Composite.generate(
				this.entryTransitionType,
				this.elId,
				this.size,
				this.position
			),
			event: {
				type: Type.PlaybackEvent.ENTRY,
				media: this,
				defaultTime: null,
				nthTime: null
			},
			visibility: Type.Visibility.HIDDEN,
			start: 0
		};
		this.exit = {
			transition: new Transition(
				this.el,
				this.exitTransitionDuration,
				this.size,
				this.position,
				zone.size,
				zone.position,
				this.exitTransitionType,
				Type.PlaybackEvent.EXIT
			),
			composite: Composite.generate(
				this.exitTransitionType,
				this.elId,
				this.size,
				this.position
			),
			event: {
				type: Type.PlaybackEvent.EXIT,
				media: this,
				defaultTime: null,
				nthTime: null
			},
			visibility: Type.Visibility.HIDDEN,
			start: 0
		};
		this.exit.transition.onFinish(this.onExitFinish.bind(this));
		this.animation = {
			effect: new Animation(this.el, this.duration, this.animationType),
			visibility: Type.Visibility.HIDDEN,
			start: 0
		};

		const willChange = Transition.mergeWillChange(
			this.entry.transition.willChange,
			this.exit.transition.willChange,
			this.animation.effect.willChange
		);
		if (willChange.length > 0 && this.el) {
			this.el.style.willChange = willChange;
		}

		this.enabled = this.schedule.isScheduledNow();
		this.scheduleMedia();
		this.updateMediaEvents();

		// Container-level objects are not constrained by the Zone view, so it is up to Media to apply the Zone transformation manually:
		if (this.type === Type.Resource.VIDEO_INPUT) {
			this.position.x += this.zone.position.x;
			this.position.y += this.zone.position.y;
		}

		if (this.el) {
			for (const kv of Media.getMediaElementAttributes(this.type, this.elId)) {
				this.el[kv.key] = kv.value;
			}
			Object.assign(this.el.style, Media.getMediaElementStyles(
				this.type,
				this.size,
				this.position,
				this.widgettype,
				this.attributes
			));
		}
	}

	promiseState(p) {
		const t = {};
		return Promise.race([p, t]).then(v => (v === t)? "pending" : "fulfilled", () => "rejected");
	}

	async promiseSettled(p) {
		return await this.promiseState(p) != 'pending';
	}

	/**
	 * @returns {!boolean}
	 */
	shouldCache() {
		return this.type === Type.Resource.AUDIO ||
			this.type === Type.Resource.VIDEO ||
			this.type === Type.Resource.IMAGE ||
			this.type === Type.Resource.MISC;
	}

	/**
	 * @param {!Map} fontSchemas
	 * @returns {!string[]}
	 */
	getFonts(fontSchemas) {
		const fonts = [];
		if (this.widgettype === 'buttongroup' && this.attributes.buttons !== undefined) {
			for (const button of this.attributes.buttons) {
				if (button.font !== undefined) {
					fonts.push(button.font);
				}
			}
		}
		const schema = fontSchemas.get(this.widgettype);
		if (schema !== undefined) {
			for (const schemaAttribute of schema) {
				for (const key of Object.keys(this.attributes)) {
					if (key === schemaAttribute.name) {
						fonts.push(this.attributes[key]);
					}
				}
			}
		}
		return fonts;
	}

	/**
	 * @returns {!string}
	 */
	getMediaUrl() {
		switch (this.type) {
		case Type.Resource.AUDIO:
		case Type.Resource.VIDEO:
		case Type.Resource.IMAGE:
		case Type.Resource.MISC:
			return this.getServerCachedMediaUrl();
		default:
			return '';
		}
	}

	/**
	 * @returns {!string}
	 */
	getServerCachedMediaUrl() {
		return `${this.player.container.hostUrl}get/mediaurl.php?account=${this.player.settings.data.account}&cdnurl=${encodeURIComponent(this.url)}&t=${this.uploadedTimestamp}&size=${this.fileSize}`;
	}

	/**
	 * @returns {!Promise}
	 */
	static async cleanCache() {
		const widgetDataDirectory = new FileInfo(
			Media.WIDGET_DATA_DIRECTORY,
			true
		);
		await widgetDataDirectory.delete();
	}

	/**
	 * @param {!Type.Resource}
	 * @returns {!string}
	 */
	static getLocalDirectory(type) {
		return 'media';
	}

	/**
	 * @param {!string} url
	 * @param {!Type.Resource} type
	 * @param {!number} id
	 * @returns {!string}
	 */
	static getLocalName(url, type, id) {
		return Media.generateLocalName(url);
	}

	/**
	 * @param {!string} cdnUrl
	 * @returns {!string}
	 */
	static generateLocalName(cdnUrl) {
		let name = cdnUrl;
		let index = name.lastIndexOf('/');
		if (index > 0) {
			name = name.substring(index + 1);
		}
		name = name.replace(/\W+/g, '_');
		index = name.lastIndexOf('_');
		if (index > 0) {
			name = `${name.substring(0, index)}.${name.substring(index + 1)}`;
		}
		return name;
	}

	/**
	 * @param {!number} tu
	 * @returns {!number}
	 */
	static parseTimeUnit(tu) {
		return tu * 1000;
	}

	/**
	 * @param {!number} animation
	 * @returns {!Type.Animation}
	 */
	static parseAnimation(animation) {
		if (animation === 1) {
			return Type.Animation.ZOOM_PAN_STANDARD;
		}
		return Type.Animation.NONE;
	}

	/**
	 * @param {!object}
	 * @return {!Type.Resource}
	 */
	static parseType({ type, name, attributes }) {
		switch (type) {
			case Type.Media.AUDIO:
				return Type.Resource.AUDIO;
			case Type.Media.IMAGE:
				return Type.Resource.IMAGE;
			case Type.Media.VIDEO:
				return Type.Resource.VIDEO;
			case Type.Media.MISC:
				return Type.Resource.MISC;
			case Type.Media.WIDGET:
				if (name === 'Video Input') {
					if (attributes.inputtype === 'hdmi') {
						return Type.Resource.VIDEO_INPUT;
					} else if (attributes.inputtype === 'blackmagicinput') {
						return Type.Resource.VIDEO_STREAMING;
					} else {
						return Type.Resource.UNSUPPORTED;
					}
				} else if (name === 'Video Streaming' && attributes.format != 'hls') {
					return Type.Resource.VIDEO_STREAMING;
				} else if (name === 'Button Group') {
					return Type.Resource.BUTTON_GROUP;
				}
				return Type.Resource.WIDGET;
			default:
				return Type.Resource.UNSUPPORTED;
		}
	}

	/**
	 * @param {!string} type
	 * @return {!Type.Transition}
	 */
	static parseTransitionType(type) {
		switch (type) {
		case 'None':
			return Type.Transition.NONE;
		case 'Fade':
			return Type.Transition.FADE;
		case 'Fade To Black':
			return Type.Transition.FADE_TO_BLACK;
		case 'Fade To White':
			return Type.Transition.FADE_TO_WHITE;
		case 'Slide Left':
			return Type.Transition.SLIDE_L;
		case 'Slide Right':
			return Type.Transition.SLIDE_R;
		case 'Slide Up':
			return Type.Transition.SLIDE_T;
		case 'Slide Down':
			return Type.Transition.SLIDE_B;
		case 'Expand Right':
			return Type.Transition.EXPAND_R;
		case 'Expand Left':
			return Type.Transition.EXPAND_L;
		case 'Expand Up':
			return Type.Transition.EXPAND_T;
		case 'Expand Down':
			return Type.Transition.EXPAND_B;
		default:
			return Type.Transition.NONE;
		}
	}

	/**
	 * @param {!number} duration
	 * @param {!number} limit
	 * @param {!Type.Transition} type
	 */
	static parseTransitionDuration(duration, limit, type) {
		if (type === Type.Transition.NONE) {
			return 0;
		}
		return duration > limit ? limit : duration;
	}

	/**
	 * @param {!string} url
	 * @param {!Type.Resource} type
	 * @param {!boolean} preview
	 * @returns {!string}
	 */
	static parseUrl(url, type, preview) {
		// Can't use default parameter as url is explicitly passed as undefined
		url = url === undefined ? '' : url;
		if (type === Type.Resource.VIDEO && preview) {
			let index = url.lastIndexOf('.');
			index = url.lastIndexOf('.', index - 1);
			url = `${url.substring(0, index)}.preview${url.substring(index)}`;
		}
		return url;
	}

	/**
	 * @param {!number} volume
	 * @returns {!number}
	 */
	static parseVolume(volume) {
		return volume === undefined ? 1 : volume / 100;
	}

	/**
	 * @returns {!string}
	 */
	static get WIDGET_DATA_DIRECTORY() {
		return 'widget_data';
	}

	/**
	 * Updates the Medias schedule status and emits 'UpdateView' if this Medias enabled status changes.
	 * @param {!boolean} [emitUpdateView=true] When set to true this function will emit 'UpdateView' if the Medias enabled status changes. Set to false if you are doing a bulk update and wish to trigger the 'UpdateView' event afterwards.
	 */
	scheduleMedia(emitUpdateView = true) {
		this.timers.off('scheduleMedia');
		const delay = this.schedule.delayFromNow();
		if (delay !== Infinity) {
			this.timers.on('scheduleMedia', delay, this.scheduleMedia.bind(this));
		}
		const enabled = this.schedule.isScheduledNow();
		if (this.enabled !== enabled) {
			this.enabled = enabled;
			if (emitUpdateView) {
				this.zone.playlist.UpdateView();
			}
		}
	}

	/**
	 * Attempts to enable the Media if it is disabled and scheduled to play.
	 * @param {!boolean} [emitUpdateView=true] When set to true this function will emit 'UpdateView' if the Medias enabled status changes. Set to false if you are doing a bulk update and wish to trigger the 'UpdateView' event afterwards.
	 */
	enable(emitUpdateView = true) {
		this.scheduleMedia(emitUpdateView);
	}

	/**
	 * Disables the Media.
	 * @param {boolean} [emitUpdateView=true] When set to true this function will 
	 * emit 'UpdateView' if the Medias enabled status changes. Set to false if you 
	 * are doing a bulk update and wish to trigger the 'UpdateView' event 
	 * afterwards.
	 */
	disable(emitUpdateView = true) {
		this.timers.off('scheduleMedia');
		if (this.enabled) {
			this.enabled = false;
			if (emitUpdateView) {
				this.zone.playlist.UpdateView();
			}
		}
	}

	sync() {
		try {
			if (this.startedAt) {
				if (this.el && this.type === Type.Resource.VIDEO && this.isPlaying()) {
					const now = this.player.currentTime;
					const currentTime = now.diff(this.startedAt, 'ms');
					const realCurrentTime = this.el.currentTime * 1000;
					const diff = Math.abs(realCurrentTime - currentTime);
					if (diff > 1000) {
						this.seek(currentTime / 1000);
						console.log(`Synced video drift of ${diff}ms.`);
					}
				}
			}
		} catch (err) {
			console.log(err.stack);
		}
	}

	/**
	 * Calculates the {@link Media#entry} and {@link Media#exit} event default timestamps. These are the timestamps on the first iteration without considering other Zones and may start before 0. The default timestamps are used to calculate the nthTimestamps and to determine which transition state the Media is in.
	 */
	updateMediaEvents() {
		this.entry.event.defaultTime = this.startTime - this.entry.transition.duration / 2;
		this.exit.event.defaultTime = this.startTime + this.duration - this.exit.transition.duration / 2;
	}

	/**
	 * @param {!HTMLElement} $parent
	 */
	attach($parent) {
		if (this.el) {
			$parent.appendChild(this.el);
		}
	}

	/**
	 * @param {!HTMLElement} $parent
	 */
	attachTransitionalMedia($parent) {
		if (this.entry.composite) {
			$parent.appendChild(this.entry.composite);
		}
		if (this.exit.composite) {
			$parent.appendChild(this.exit.composite);
		}
	}

	/**
	 * Performs all the necessary tasks for the Media object to be ready to show.
	 * @returns {!Promise}
	 */
	async initialize() {
		console.log(`start media cache: ${this.id} ${this.name}`);
		let failedToCache = false;
		try {
			await this.cache();
			
			console.log(`end media cache: ${this.id} ${this.name}`);
			try {
				if (!await this.promiseSettled(this.destroyPromise)) {
					await this.load();
				}
			} catch (err) {
				console.log(`Media ${this.id} ${this.name} was unable to load, hiding the media temporarily, reason: ${err.stack || err}`);
			}
		} catch (exception) {
			failedToCache = true;
		}
		
		if (!await this.promiseSettled(this.destroyPromise)) {
			if (failedToCache || await this.promiseState(this.loadedPromise) == 'rejected') {
				this.timers.on('reinitialize', 30 * 1000, this.initialize.bind(this));
			}
		}
	}

	/**
	 * Returns a value between 0 and 100, where 100 is fully loaded.
	 * TODO: Make the progress reflect download/load progress.
	 * @returns {!number}
	 */
	async getLoadingProgress() {
		if (this.cachePromise == null || await this.promiseState(this.cachePromise) == 'pending') {
			return 0; // Cache is still processing
		} else if (this.loadedPromise != null && await this.promiseState(this.loadedPromise) == 'pending') {
			return 50; // Cache is done processing and failed or succeeded to download the file. The media has not loaded yet though from any source.
		} else {
			return 100; // Either the media failed to load or loaded using the server, CDN, or local file.
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async cache() {
		this.cachePromise = (async () => {
			try {
				if (this.shouldCache() && this.player.container.features.caching) {
					await this.player.downloader.download(this.getMediaUrl(), this.file);
					this.cacheRequestedResolve();
				}
				const localUrl = await this.file.getLocalUrl();
				if (localUrl != null) {
					this.mediaURL = localUrl;
				}
			} catch (err) {
				console.log(`Media ${this.id} ${this.name} was unable to cache, hiding the media temporarily, reason: ${err.stack || err}`);
				throw err;
			}
		})();
		return this.cachePromise;
	}

	/**
	 * @returns {!Promise}
	 */
	async load() {
		let callback = null;
		try {
			this.loadedPromise = Promise.race([
				new Promise(async (resolve, reject) => {
					await this.destroyPromise;
					reject(new Error('media destroyed'));
				}),
				new Promise(resolve => {
					callback = () => {
						console.log(`canplaythrough or load called for ${this.id} ${this.name}`);
						resolve();
					};
					switch (this.type) {
						case Type.Resource.AUDIO:
						case Type.Resource.VIDEO:
							this.el.addEventListener('canplaythrough', callback);
							this.el.src = this.mediaURL;
							this.el.load();
							break;
						case Type.Resource.IMAGE:
							this.el.addEventListener('load', callback);
							this.el.src = this.mediaURL;
							break;
						case Type.Resource.WIDGET:
							this.widget = CreateWidget(this.el, this, this.size.width, this.size.height, false, this.player.settings.data.widgetData, this.player.container, this.player.settings.data.account, this.player.settings.data.key, this.player, resolve);
							break;
						case Type.Resource.BUTTON_GROUP:
							this.widget = CreateWidget(this.el, this, this.size.width, this.size.height, false, this.player.settings.data.widgetData, this.player.container, this.player.settings.data.account, this.player.settings.data.key, this.player, () => {
								for (let index = 0; index < this.el.childNodes.length; ++index) {
									const $button = this.el.childNodes[index];
									$button.style.pointerEvents = 'all';
									// Assign the buttongroup to the buttons
									this.attributes.buttons[index].$button = $button;
									this.attributes.buttons[index].buttongroup = this;
									this.zone.playlist.AddButton(this.attributes.buttons[index], $button);
								}
								resolve();
							});
							break;
						case Type.Resource.VIDEO_INPUT:
							console.log('CreateVideoIn');
							ContainerAPI.CreateVideoIn(this.zone.playlist.uniqueId, this.videoSource, this.videoSourceId, this.attributes.inputwidth, this.attributes.inputheight);
							resolve();
							break;
						case Type.Resource.VIDEO_STREAMING:
							resolve();
							break;
						case Type.Resource.MISC:
							// TODO: Figure out what DOM load event is fired for 'MISC' file types and use it here.
							this.el.appendChild(MiscMedia.generateElement(this.mediaURL, this.size.width, this.size.height));
							resolve();
							break;
					}
				}),
				new Promise((resolve, reject) => {
					// We don't cancel this timer since it can reject without hurting anything
					setTimeout(() => {
						reject(new Error('Media took too long to load and timed out!'));
					}, 30 * 1000);
				})
			]);
			await this.loadedPromise;
		} catch (exception) {
			throw exception;
		} finally {
			// If destroy is called then this.el will be null
			if (this.el) {
				// destroy or the above timeout was called, so clean up any event handlers that might have been attached
				this.el.removeEventListener('canplaythrough', callback);
				this.el.removeEventListener('load', callback);
			}
		}
		this.resetVolume();
	}

	TimeSync() {
		if (this.widget) {
			this.widget.TimeSync();
		}
	}

	destroy() {
		this.hide();
		this.render();
		this.timers.clear();
		if (this.el) {
			if (this.type === Type.Resource.AUDIO || this.type === Type.Resource.VIDEO) {
				this.stop();
				this.el.src = '';
				if (this.el.load !== undefined) {
					this.el.load();
				}
			}
			this.el.remove();
		}
		this.el = null;
		if (this.entry.composite) {
			this.entry.composite.remove();
		}
		if (this.exit.composite) {
			this.exit.composite.remove();
		}
		this.entry.transition.destroy();
		this.exit.transition.destroy();
		this.animation.effect.destroy();
		if (this.widget) {
			this.widget.Destroy();
		}
		this.destroyResolve();
	}

	/**
	 * Shows the Media representation, has no effect on the DOM until next 
	 * {@link Media#render} call.
	 */
	show() {
		this.visibility = Type.Visibility.WILL_SHOW;
	}

	/**
	 * Hides the Media representation, has no effect on the DOM until next 
	 * {@link Media#render} call.
	 */
	hide() {
		this.visibility = Type.Visibility.WILL_HIDE;
	}

	/**
	 * @param {!number} stamp
	 */
	beginAnimation(stamp) {
		this.animation.start = stamp;
		this.animation.visibility = Type.Visibility.WILL_SHOW;
	}

	endAnimation() {
		this.animation.visibility = Type.Visibility.WILL_HIDE;
	}

	/**
	 * @param {!Type.PlaybackEvent} type
	 * @param {!number} stamp
	 */
	beginTransition(type, stamp) {
		if (type === Type.PlaybackEvent.ENTRY) {
			this.entry.start = stamp;
			this.entry.visibility = Type.Visibility.WILL_SHOW;
		} else if (type === Type.PlaybackEvent.EXIT) {
			this.exit.start = stamp;
			this.exit.visibility = Type.Visibility.WILL_SHOW;
		}
	}

	/**
	 * @param {!Type.PlaybackEvent} type
	 */
	endTransition(type) {
		if (type === Type.PlaybackEvent.ENTRY) {
			this.entry.visibility = Type.Visibility.WILL_HIDE;
		} else if (type === Type.PlaybackEvent.EXIT) {
			this.exit.visibility = Type.Visibility.WILL_HIDE;
		}
	}

	/*
	 * Returns true if there is any frequency in the audio channel producing sound
	 */
	/*
	// Brandon: Created this as a proof of concept. It works alright, but requires a few seconds of audio to figure out if there is no audio. Not an ideal method. We moved to doing the audio check during upload
	hasVolume() {
		if (this.type === Type.Resource.AUDIO || this.type === Type.Resource.VIDEO) {
			const maxVolumeSamples = 20;
			if (!this.hasOwnProperty('volumeSamples')) {
				this.volumeSamples = 0;
				this.hasAudio = false;
			}
			if (this.volumeSamples < maxVolumeSamples) {
				const element = this.el;
				// Reuse all the nodes and data structures for every call
				if (!this.hasOwnProperty('audioContext')) {
					this.audioContext = new AudioContext();
					this.source = this.audioContext.createMediaElementSource(element);
					this.analyser = this.audioContext.createAnalyser();
					this.analyser.fftSize = 16384;
					this.source.connect(this.analyser);
					this.analyser.connect(this.audioContext.destination);
					this.audioVolumeArray = new Uint8Array(this.analyser.fftSize);
				}
				this.analyser.getByteTimeDomainData(this.audioVolumeArray);
				let max = 0;
				for (let a of this.audioVolumeArray) {
					a = Math.abs(a - 128);
					max = Math.max(max, a);
				}
				this.volumeSamples++;
				if (max > 1) {
					this.volumeSamples = maxVolumeSamples;
					this.hasAudio = true;
				}
				// Assume audio until we have 20 samples without audio
				return true;
			} else {
				console.log(this.hasAudio);
				return this.hasAudio;
			}
		} else {
			return true;
		}
	}
	*/

	seek(time) {
		if (this.el) {
			this.seekTo = null;
			this.el.currentTime = time;
		} else {
			this.seekTo = time;
			this.seekToTime = this.player.currentTime;
		}
	}

	/**
	 * @param {!moment} startTime
	 */
	play() {
		this.playing = Type.PlayState.WILL_PLAY;
		this.startedAt = this.player.currentTime;
	}

	stop() {
		this.playing = Type.PlayState.WILL_STOP;
	}

	resetVolume() {
		if (this.currentVolume !== this.maxVolume) {
			this.currentVolume = this.maxVolume;
			if (this.type === Type.Resource.VIDEO_STREAMING) {
				if (this.breezeMpv) {
					this.breezeMpv.setVolume(this.currentVolume);
				}
			} else if (this.el) {
				this.el.volume = this.currentVolume;
			}
		}
	}

	increaseVolume(delta) {
		if (this.currentVolume < this.maxVolume) {
			this.currentVolume = Math.min(this.currentVolume + delta, this.maxVolume);
			if (this.type === Type.Resource.VIDEO_STREAMING) {
				if (this.breezeMpv) {
					this.breezeMpv.setVolume(this.currentVolume);
				}
			} else if (this.type === Type.Resource.WIDGET) {
				this.widget.SetVolume(this.currentVolume);
			} else if (this.el) {
				this.el.volume = this.currentVolume;
			}
		}
	}

	decreaseVolume(delta) {
		if (this.currentVolume > 0) {
			this.currentVolume = Math.max(this.currentVolume - delta, 0);
			if (this.type === Type.Resource.VIDEO_STREAMING) {
				if (this.breezeMpv) {
					this.breezeMpv.setVolume(this.currentVolume);
				}
			} else if (this.type === Type.Resource.WIDGET) {
				this.widget.SetVolume(this.currentVolume);
			} else if (this.el) {
				this.el.volume = this.currentVolume;
			}
		}
	}

	duckVolume() {
		if (this.currentVolume !== this.duckedVolume) {
			this.currentVolume = this.duckedVolume;
			if (this.type === Type.Resource.VIDEO_STREAMING) {
				if (this.breezeMpv) {
					this.breezeMpv.setVolume(this.currentVolume);
				}
			} else if (this.type === Type.Resource.WIDGET) {
				this.widget.SetVolume(this.currentVolume);
			} else if (this.el) {
				this.el.volume = this.currentVolume;
			}
		}
	}

	duckDown() {
		if (this.currentVolume > this.duckedVolume) {
			this.decreaseVolume(0.01);
		}
	}

	duckUp() {
		this.increaseVolume(0.01);
	}

	reset() {
		this.hide();
		this.endTransition(Type.PlaybackEvent.ENTRY);
		this.endTransition(Type.PlaybackEvent.EXIT);
		this.endAnimation();
		if (this.type === Type.Resource.AUDIO ||
			this.type === Type.Resource.VIDEO) {
			this.stop();
			this.seek(0);
		}
		this.render();
	}

	/**
	 * Modern browsers will return a promise when you play() that rejects if
	 * the playback could not start for whatever reason.
	 */
	async _play() {
		if (typeof this.el.play != 'function') {
			return;
		}
		for (let tryPlay = 0; tryPlay < 3; ++tryPlay) {
			if (this.playing !== Type.PlayState.PLAYING) {
				break;
			}
			try {
				const p = this.el.play();
				if (p != undefined) {
					await p;
					break;
				}
			} catch (err) {
				if (this.playing === Type.PlayState.PLAYING) {
					// Android very randomly throws when calling play() on a cached media object. This might have just been a Chrome 53 bug as I (Brandon) have never seen it on Windows or Linux
					// Try to play three more times
					console.log('play try:', tryPlay + 1, err.stack || err);
					this.el.src = '';
					if (this.el.load !== undefined) {
						this.el.load();
					}
					// Delay 1 ms so there's time to unload
					await new Promise(resolve => setTimeout(resolve, 1));
					if (this.playing === Type.PlayState.PLAYING) {
						await this.load();
					}
				}
			}
		}
	}

	/**
	 * A hack used to force hide media after finishing an exit transition. Should always
	 * hide the Media unless it has already entered an ENTRY transition.
	 */
	onExitFinish() {
		if (this.el && this.entry.visibility === Type.Visibility.HIDDEN) {
			this.el.style.visibility = 'hidden';
			this.el.style.transform = 'translate(-10000px,-10000px)';
		}
	}

	render() {
		if (this.el) {
			if (this.seekTo) {
				const offsetMS = this.player.currentTime.diff(this.seekToTime, 'ms');
				this.el.currentTime = this.seekTo + offsetMS / 1000;
				this.seekTo = null;
			}
			if (this.playing === Type.PlayState.WILL_PLAY) {
				this.playing = Type.PlayState.PLAYING;
				this._play();
			} else if (this.playing === Type.PlayState.WILL_STOP) {
				this.playing = Type.PlayState.STOPPED;
				this.el.pause();
			}
		}

		// Show/hide animation effects
		if (this.animation.visibility === Type.Visibility.WILL_SHOW) {
			this.animation.effect.start(this.animation.start);
			this.animation.visibility = Type.Visibility.SHOWN;
		} else if (this.animation.visibility === Type.Visibility.WILL_HIDE) {
			this.animation.effect.finish();
			this.animation.visibility = Type.Visibility.HIDDEN;
		}

		// Show/hide transitions and composites
		if (this.entry.visibility === Type.Visibility.WILL_SHOW) {
			if (this.entry.composite) {
				this.entry.composite.style.visibility = 'visible';
				this.entry.composite.style.removeProperty('transform');
			}
			this.entry.transition.start(this.entry.start);
			this.entry.visibility = Type.Visibility.SHOWN;
		} else if (this.entry.visibility === Type.Visibility.WILL_HIDE) {
			if (this.entry.composite) {
				this.entry.composite.style.visibility = 'hidden';
				this.entry.composite.style.transform = 'translate(-10000px,-10000px)';
			}
			this.entry.transition.finish();
			this.entry.visibility = Type.Visibility.HIDDEN;
		}
		if (this.exit.visibility === Type.Visibility.WILL_SHOW) {
			if (this.exit.composite) {
				this.exit.composite.style.visibility = 'visible';
				this.exit.composite.style.removeProperty('transform');
				
			}
			this.exit.transition.start(this.exit.start);
			this.exit.visibility = Type.Visibility.SHOWN;
		} else if (this.exit.visibility === Type.Visibility.WILL_HIDE) {
			if (this.exit.composite) {
				this.exit.composite.style.visibility = 'hidden';
				this.exit.composite.style.transform = 'translate(-10000px,-10000px)';
			}
			this.exit.transition.finish();
			this.exit.visibility = Type.Visibility.HIDDEN;
		}

		// Show/hide media
		if (this.visibility === Type.Visibility.WILL_SHOW) {
			this.visibility = Type.Visibility.SHOWN;
			if (this.widget) {
				this.widget.Show();
			}
			if (this.type === Type.Resource.VIDEO_INPUT) {
				ContainerAPI.PlayVideoIn(this.zone.playlist.uniqueId, this.videoSource, this.videoSourceId, Math.round(this.position.x * this.zone.scale.width), Math.round(this.position.y * this.zone.scale.height), this.size.width * this.zone.scale.width, this.size.height * this.zone.scale.height);
			} else if (this.type === Type.Resource.VIDEO_STREAMING && this.el) {
				if (this.type === Type.Resource.VIDEO_STREAMING) {
					// If for whatever reason this is called twice in a row destroy the previous mpv instance
					if (this.breezeMpv) {
						this.breezeMpv.destroy();
						this.breezeMpv = null;
					}
					this.breezeMpv = new BreezeMPV(
						this.el,
						this.size.width,
						this.size.height,
						this.attributes
					);
				}
				this.el.style.visibility = 'visible';
				this.el.style.removeProperty('transform');
			} else if (this.el && this.type !== Type.Resource.AUDIO) {
				this.el.style.visibility = 'visible';
				this.el.style.removeProperty('transform');
			}
		} else if (this.visibility === Type.Visibility.WILL_HIDE) {
			this.visibility = Type.Visibility.HIDDEN;
			if (this.widget) {
				this.widget.Hide();
			}
			if (this.type === Type.Resource.VIDEO_INPUT) {
				ContainerAPI.StopVideoIn(this.zone.playlist.uniqueId, this.videoSource, this.videoSourceId);
			} else if (this.type === Type.Resource.VIDEO_STREAMING && this.el) {
				if (this.breezeMpv) {
					this.breezeMpv.destroy();
					this.breezeMpv = null;
				}
				this.el.style.visibility = 'hidden';
				this.el.style.transform = 'translate(-10000px,-10000px)';
			} else if (this.el && this.type !== Type.Resource.AUDIO) {
				this.el.style.visibility = 'hidden';
				this.el.style.transform = 'translate(-10000px,-10000px)';
			}
		}
	}

	/**
	 * @returns {!boolean}
	 */
	isPlaying() {
		return this.playing === Type.PlayState.PLAYING || this.playing === Type.PlayState.WILL_PLAY;
	}

	/**
	 * @returns {!boolean}
	 */
	isShowing() {
		return this.visibility === Type.Visibility.SHOWN || this.visibility === Type.Visibility.WILL_SHOW;
	}

	/**
	 * @returns {!boolean}
	 */
	isWidget() {
		return this.widgettype !== undefined;
	}

	/**
	 * @returns {!boolean}
	 */
	isSignwave() {
		return this.widgettype === 'signwave';
	}

	/**
	 * @param {!Type.Resource} type
	 * @returns {?Element}
	 */
	static createMediaElement(type) {
		switch(type) {
		case Type.Resource.AUDIO:
			return $.audio();
		case Type.Resource.IMAGE:
			return $.img();
		case Type.Resource.VIDEO:
			return $.video();
		case Type.Resource.VIDEO_STREAMING:
		case Type.Resource.WIDGET:
		case Type.Resource.MISC:
		case Type.Resource.BUTTON_GROUP:
		case Type.Resource.UNSUPPORTED:
			return $.div();
		default:
			return null;
		}
	}

	/**
	 * @param {!Type.Resource} type
	 * @param {!string} id
	 * @returns {!{key: !string, value: !string}[]}
	 */
	static getMediaElementAttributes(type, id) {
		let attribs = [];
		switch(type) {
		case Type.Resource.AUDIO:
		case Type.Resource.IMAGE:
		case Type.Resource.VIDEO:
		case Type.Resource.VIDEO_STREAMING:
		case Type.Resource.WIDGET:
		case Type.Resource.MISC:
		case Type.Resource.BUTTON_GROUP:
		case Type.Resource.UNSUPPORTED:
			attribs.push({
				key: 'id',
				value: id
			});
			break;
		}
		if (type === Type.Resource.AUDIO ||
			type === Type.Resource.IMAGE ||
			type === Type.Resource.VIDEO) {
			attribs.push({
				key: 'crossOrigin',
				value: 'anonymous'
			});
		}
		if (type === Type.Resource.VIDEO) {
			attribs.push({
				key: 'type',
				value: 'video/mp4'
			});
		}
		return attribs;
	}

	/**
	 * @param {!Type.Resource} type
	 * @param {!Size} size
	 * @param {!Point} position
	 * @param {!string} widgetType
	 * @param {?Object} attributes
	 * @returns {?Object}
	 */
	static getMediaElementStyles(type, size, position, widgetType, attributes) {
		let css = null;
		switch(type) {
		case Type.Resource.AUDIO:
		case Type.Resource.IMAGE:
		case Type.Resource.VIDEO:
		case Type.Resource.VIDEO_STREAMING:
		case Type.Resource.WIDGET:
		case Type.Resource.MISC:
		case Type.Resource.BUTTON_GROUP:
		case Type.Resource.UNSUPPORTED:
			css = {
				position: 'absolute',
				width: `${size.width}px`,
				height: `${size.height}px`,
				left: `${position.x}px`,
				top: `${position.y}px`,
				pointerEvents: 'none',
				visibility: 'hidden',
				transform: 'translate(-10000px,-10000px)',
				transformOrigin: 'top left'
			};
			break;
		}
		if (type === Type.Resource.VIDEO) {
			css.objectFit = 'fill';
		} else if (type === Type.Resource.WIDGET) {
			if (widgetType === 'signwave' ||
				widgetType === 'readerboard' ||
				(attributes != null && attributes.live)) {
				css.pointerEvents = 'all';
			}
		}
		return css;
	}
}
