class MessengerPlayer {
	constructor(frame = window.parent) {
		this.frame = frame;
		this.postbacks = {};
		this.bus = Minibus.create();
		this.token = 0;

		window.addEventListener('message', this.listener.bind(this));
	}

	/**
	 * @param {!string} type
	 * @param {!number} token
	 * @param {...[]} args
	 */
	_post(type, token, ...args) {
		this.frame.postMessage({
			type: 'bus',
			event: type,
			args: args,
			token: token
		}, '*');
	}

	/**
	 * @param {!string} type
	 * @param {...[]} args
	 */
	post(type, ...args) {
		this._post(type, undefined, ...args);
	}

	/**
	 * Waits for a response and resolves with the response value.
	 * @param {!string} type
	 * @param {...[]} args
	 * @returns {!Promise<?Object>}
	 */
	postBack(type, ...args) {
		return new Promise(resolve => {
			const token = this.token++;
			if (!this.postbacks[type]) {
				this.postbacks[type] = [];
			}
			this.postbacks[type].push({
				token: token,
				callback: response => {
					resolve(response);
				}
			});
			this._post(type, token, ...args);
		});
	}

	/**
	 * @param {!MessageEvent} e
	 */
	listener(e) {
		if (e.data.type === 'bus') {
			this.bus.emit(e.data.event, e.data.args);
		} else if (e.data.type === 'bus_response') {
			const arr = this.postbacks[e.data.event];
			if (!arr) {
				console.log(`Received unrequested 'bus_response' of type '${e.data.event}' (token => ${e.data.token}).  Could not find postback array entry for given type.`);
				return;
			}
			const index = arr.findIndex(x => e.data.token === x.token);
			if (index === -1) {
				console.log(`Received unrequested 'bus_response' of type '${e.data.event}' (token => ${e.data.token}).  Could not find the token in the postback array.`);
				return;
			}
			arr[index].callback(e.data.response);
			arr.splice(index, 1);
		}
	}
}

class MessengerContainer {
	constructor(win, frame) {
		this.bus = Minibus.create();
		this.frame = frame.contentWindow !== undefined ? frame.contentWindow : frame;
		win.addEventListener(
			'message',
			this.listener.bind(this)
		);
	}

	/**
	 * @param {!string} event
	 * @param {!string} token
	 * @param {?Object} response
	 */
	postBack(event, token, response) {
		this.frame.postMessage({
			type: 'bus_response',
			event: event,
			response: response,
			token: token
		}, '*');
	}

	/**
	 * @param {!MessageEvent} e
	 */
	listener(e) {
		if (e.data.type === 'bus') {
			let callback = null;
			if (e.data.token !== undefined) {
				callback = response => {
					this.postBack(
						e.data.event,
						e.data.token,
						response
					);
				};
			}
			this.bus.emit(
				e.data.event,
				e.data.args,
				callback
			);
		}
	}
}
