/* exported Network */
/* global Timer */

class Network {
	/**
	 * @param {!string} url
	 * @returns {!boolean}
	 */
	static canUseFetch(url) {
		return !url.startsWith('file://');
	}

	/**
	 * @param {!string} key
	 * @param {!string} value
	 * @returns {!string}
	 */
	static getParam(key, value) {
		return `${encodeURIComponent(key)}=${encodeURIComponent(value)}`;
	}

	/**
	 * Annoyingly 404s will be logged to the console, hopefully this will
	 * be changed in the future: 
	 * https://bugs.chromium.org/p/chromium/issues/detail?id=124534
	 * @param {!string} url
	 * @returns {!Promise<!boolean>}
	 */
	static async exists(url) {
		let status = 0;
		if (Network.canUseFetch(url)) {
			const response = await fetch(url, { method: 'HEAD' });
			status = response.status;
		} else {
			const p = new Promise((resolve, reject) => {
				$.ajax({
					url: url,
					method: 'HEAD',
				}).done((data, textStatus, jqXHR) => {
					resolve(jqXHR.status);
				}).fail((jqXHR, textStatus, errorThrown) => {
					reject(new Error(`XHR request failed: ${textStatus} - ${errorThrown}`));
				});
			});
			status = await p;
		}
		return status !== 404;
	}

	/**
	 * @param {(!string|!Request)} request
	 * @param {!string} [dataType]
	 * @returns {!Promise<!Response>}
	 */
	static async fetch(request, dataType) {
		const response = await fetch(request);
		if (!response.ok) {
			throw new Error(`Response returned failure code ${response.status}`);
		}
		if (dataType === 'json') {
			return response.json();
		} else {
			return response;
		}
	}

	/**
	 * As fetch is restricted from local file requests (at least in Chrome currently), an xhr-based method
	 * is provided as an alternative for loading resources over the file:// protocol.
	 * An explanation can be found in the comments at https://github.com/github/fetch/pull/92
	 * @param {!string} url
	 * @param {!string} [dataType]
	 * @returns {!Promise<!Object>}
	 */
	static xhr(url, dataType = 'json') {
		return new Promise((resolve, reject) => {
			$.ajax({
				url: url,
				dataType: dataType
			}).done(data =>
				resolve(data)
			).fail((jqXHR, textStatus, errorThrown) => {
				reject(new Error(`XHR request failed: ${textStatus} - ${errorThrown}`));
			});
		});
	}

	/**
	 * Downloads a file (using fetch where possible, or xhr if the protocol is file://).
	 * Retries the download until attempts is exceeded, or infinitely if attempts is 0.
	 * Waits the given delay between retries.
	 * @param {!string} url
	 * @param {!string} [dataType] different options depending on if the download is made
	 * over xhr or fetch
	 * @param {!number} [attempts]
	 * @param {!number} [delay]
	 * @returns {!Promise<!Object>}
	 */
	static async download(url, dataType = 'json', attempts = 10, delay = 5000) {
		const request = Network.canUseFetch(url) ? Network.fetch : Network.xhr;
		let tries = 1;
		while (true) {
			try {
				return await request(url, dataType);
			} catch (err) {
				if (++tries > attempts) {
					throw err;
				}
			}
			if (delay > 0) {
				await Timer.sleep(delay);
			}
		}
	}
}
