/* exported Player */
/* global ContainerAPI, Messenger, MessengerPlayer, BackgroundMusicService, TimeSync, Watermark, AlertViewer, CAPS, TimerMap, Downloader, Container, Checkin, PlayerSettings, BreezeFonts, Utility, moment, BreezeSchedule, PlayerSchedule */
class Player {
	/**
	 * @param {!string} key
	 * @param {!Object} [defaults]
	 */
	constructor(key, defaults = {}) {
		/** @type {!boolean} */
		this._ready = false;
		
		/** @type {?Function[]} */
		this._readyFns = [];
		
		// If the Container didn't create the global Messenger assume it
		// is in a parent context.
		if (window.Messenger === undefined) {
			window.Messenger = new MessengerPlayer(window.parent);
		}
		
		const $playlists = Player.createPlaceholder();
		document.body.appendChild($playlists);
		
		/** @type {!HTMLElement} */
		this.alertsDiv = Player.createPlaceholder();
		document.body.appendChild(this.alertsDiv);
		
		/** @type {!BackgroundMusicService} */
		this.backgroundMusicService = new BackgroundMusicService();
		
		this.serverOffset = 0;
		this._checkin = this.checkin.bind(this);
		this.checkinDelay = 15000;
		this.networkCommands = null;
		this.networkCommandSchedules = null;
		this.playlistUniqueIds = 0;
		this.downloader = new Downloader();
		this.container = new Container(this);
		const startTime = defaults.startTime != undefined ? defaults.startTime : 0;
		this.settings = new PlayerSettings(key, defaults.account);
		this.schedule = new PlayerSchedule(this, startTime, $playlists);
		this.timers = new TimerMap();
		
		this.timeSync = new TimeSync(diff => {
			this.serverOffset = diff;
			this.TimeSync();
		});
		
		/** @type {?Watermark} */
		this.watermark = null;
		
		this.alerts = new Map();
		this.filters = [];
		
		this.audioContext = null;
		this.gainNode = null;
		
		this.initialize();
	}

	get isPreview() {
		return this.container.type == 'Preview';
	}

	get currentTime() {
		// Windows timezones are unparsable and local time should be forced instead
		return (this.timezone ? moment().tz(this.timezone) : moment()).add(this.serverOffset, 'ms');
	}

	get timezone() {
		return this.container.type === 'Node' ? null : this.settings.data.timezone;
	}
	
	applyTimezone(input) {
		return this.timezone ? input.clone().tz(this.timezone) : input.clone();
	}

	/**
	 * @returns {!Promise}
	 */
	async initialize() {
		try {
			await this.container.initialize();
			if (!this.isPreview) {
				try {
					await this.settings.initialize(this.container.hostUrl);
				} catch (err) {
					if (err.message.includes('Invalid key')) {
						await this.unregister();
						return;
					} else {
						throw err;
					}
				}
			}
			this.setDailyUpdate();
			this.setDailyReboot();
			// This should revert the zoom, but it's simply not working...
			/*
			this.timers.off('ResetPinchToZoom');
			const resetPinchToZoom = () => {
				const viewport = document.querySelector('meta[name="viewport"]');
				const original = document.querySelector("meta[name='viewport']").getAttribute('content');
				viewport.content = '';
				if (viewport) {
					setTimeout(() => {
						viewport.content = "initial-scale=1";
						viewport.content = `width=device-width`;
					}, 500);
				}
				setTimeout(() => {
					document.querySelector("meta[name='viewport']").setAttribute('content', original);
				}, 1000);
				this.timers.on(
					'ResetPinchToZoom',
					1 * 60 * 1000,
					resetPinchToZoom
				);
			};
			resetPinchToZoom();
			*/
			this.schedule.initialize(
				this.settings.data.widgetData
			);
			if (Watermark.shouldWatermark(this.container.type)) {
				this.watermark = new Watermark(this.container.hostUrl);
				this.watermark.attach(document.body);
			}
			this.container.fonts = new BreezeFonts(this);
			await this.container.fonts.initialize();
			await this.timeSync.sync(this.container.hostUrl);

			const ccmProvider = this.settings.data.widgetData !== null ? this.settings.data.widgetData.ccmProvider : '';
			this.setBackgroundMusicProviders(ccmProvider);

			if (this.container && this.container.type != 'Preview' && this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('audioin') && this.settings.data.widgetData.audioin) {
				navigator.mediaDevices.ondevicechange = event => this.updateLineInAudio();
				this.updateLineInAudio();
			}

			if (this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('keyboardoverlay') ? this.settings.data.widgetData.keyboardoverlay : false) {
				ShowKeyboard();
			}

			this._ready = true;
			this._readyFns.forEach(fn => fn());
			this._readyFns = null;
		} catch (err) {
			if (err != null && err.inner != null) {
				console.log(`${err}: ${err.inner}`);
			} else {
				console.log(err.stack || err);
			}
		}
	}

	/**
	 * @param {!Function} callback
	 */
	ready(callback) {
		if (this._ready) {
			callback();
		} else {
			this._readyFns.push(callback);
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async checkin() {
		try {
			await ContainerAPI.LogVitals();
			await ContainerAPI.CheckIn(this.settings.data.widgetData);
			const localIp = await ContainerAPI.QueryLocalIp();
			const containerVersion = await ContainerAPI.QueryContainerVersion();
			const serial = await ContainerAPI.QuerySerialNumber();
			await this.settings.updateLocals(localIp, containerVersion, serial);
		} catch (err) {
			console.warn(err.stack || err);
		}

		let registered = true;
		try {
			const response = await Checkin.request(
				this.container.hostUrl,
				this,
				this.settings.data.account,
				this.settings.data.key,
				this.settings.data.localIp,
				this.settings.data.containerVersion,
				this.schedule.getActiveScheduleName(),
				this.settings.data.serialnumber,
				this.settings.data.serialnumbermodifiedtime
			);
			if (!response.success) {
				throw new Error(response.error);
			}

			/*
			 * Update/save the settings.
			 */
			let reloadSchedule = await this.settings.updateDisplay(response.screenwidth, response.screenheight);
			if (await this.settings.modifyTimezone(response.timezone)) {
				reloadSchedule = true;
			}
			if (await this.settings.updateWidgetData(response.widgetdata)) {
				reloadSchedule = true;
			}
			this.setDailyUpdate();
			this.setDailyReboot();

			this.filters = Utility.flat(response.capsfilters
				.map(row => {
					const filters = JSON.parse(row.capfilters);
					return filters.map(filter => {
						const f = CAPS.Filter.create(filter);
						f.id = row.capfilterid;
						return f;
					});
				}));
			await this.updateCapsAlerts(response.capsalerts);

			/*
			 * Execute any queued requests or commands.
			 */
			const timestamp = moment();
			response.testnetworkcommands.forEach(command => this.networkCommand(command, timestamp));

			if (response.clearcache) {
				this.destroy();
				await ContainerAPI.RequestClearCache();
			}
			if (response.update || this.schedule.shouldUpdate) {
				reloadSchedule = false;
				await this.updateSchedule(false, false);
			}
			if (reloadSchedule) {
				await this.updateSchedule(false, true);
			}
			if (response.screenshot) {
				await ContainerAPI.RequestScreenshot();
			}
			if (response.thumbnailscreenshot) {
				await ContainerAPI.RequestThumbnailScreenshot(
					this.settings.data.usableSize.width,
					this.settings.data.usableSize.height,
					this.settings.data.thumbnailSize.width,
					this.settings.data.thumbnailSize.height
				);
			}
			if (response.reboot) {
				await ContainerAPI.RequestHardReboot();
			}
			if (response.restartapp) {
				await ContainerAPI.RequestSoftReboot();
			}
			// Only updates if it changes, but this requires a file access to see if it changes
			const pinchToZoom = this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('pinchtozoom') ? this.settings.data.widgetData.pinchtozoom : false;
			await ContainerAPI.RequestPinchToZoom(pinchToZoom);
		} catch (err) {
			if (err.message.includes('Invalid key')) {
				registered = false;
				await this.unregister();
			} else {
				console.warn(`Checkin unsuccessful: ${err.stack || err}`);
			}
		} finally {
			if (registered) {
				this.timers.on(
					'Checkin',
					this.checkinDelay,
					this._checkin
				);
			}
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async unregister() {
		try {
			await this.settings.delete();
		} catch (err) {
			console.log(err.stack || err);
		}
		try {
			await ContainerAPI.UnregisteredPlayer();
		} catch (err) {
			console.log(err.stack || err);
		}
	}

	TimeSync() {
		if (this.schedule) {
			this.schedule.TimeSync();
			for (const playlist of this.schedule.playlists) {
				playlist.TimeSync();
			}
		}
	}

	/**
	 * @param {!Schedule} schedule
	 */
	ScheduleLoaded(schedule) {
		this.parseNetworkCommands(
			schedule.networkCommands,
			schedule.networkCommandSchedules
		);
	}

	/**
	 * @param {!Object[]} alerts
	 * @returns {!Promise}
	 */
	async updateCapsAlerts(alerts) {
		alerts = alerts.map(alert => CAPS.CAPSLoader.fromJson(alert));
		const wasAlerting = this.alerts.size > 0;

		for (const [ id, alertViewer ] of this.alerts) {
			if (alerts.every(alert => alert.id !== id)) {
				// The alert has expired
				alertViewer.remove();
				this.alerts.delete(id);
			}
		}

		const displayables = alerts.filter(alert => alert.info.length > 0);
		for (const alert of displayables) {
			if (!this.alerts.has(alert.id)) {
				let found = false;
				for (const filter of this.filters) {
					if (filter.test(alert)) {
						found = true;
						// Windows timezones are unparsable and local time should be forced instead
						const timezone = this.container.type === '' || this.container.type === 'Node' ? null : this.settings.data.timezone;
						const av = new AlertViewer(alert, filter, this.settings.data.usableSize, this.alertsDiv, timezone);
						if (av.isFullscreen()) {
							this.schedule.canReload = false;
							this.schedule.stop();
						}
						this.alerts.set(alert.id, av);
						break;
					}
				}
				if (!found) {
					console.warn(`Skipping alert (id => ${alert.id}) because no filters match!`);
				}
			}
		}

		// Resume playback if no active alert is suspending it.
		if (this.alerts.size === 0) {
			this.schedule.canReload = true;
			if (wasAlerting) {
				if (this.schedule.shouldUpdate) {
					await this.updateSchedule(true);
				} else if (this.schedule.missedReloads > 0) {
					this.schedule.reloadBlock();
				} else {
					this.schedule.play();
				}
			}
		}
	}

	/**
	 * @param {!string} [ccmProvider]
	 */
	setBackgroundMusicProviders(ccmProvider = '') {
		if (ccmProvider.length > 0) {
			if (this.backgroundMusicService.add(ccmProvider)) {
				this.backgroundMusicService.play();
			}
		} else {
			this.backgroundMusicService.clear();
		}
	}

	/**
	 * @param {!NetworkCommand} command
	 * @param {!moment} [timestamp]
	 * @param {!number} [expires]
	 */
	networkCommand(command, timestamp, expires = 5) {
		const eligible = timestamp === undefined;
		const expired = expires < timestamp.diff(moment.unix(command.timestamp), 'minutes');
		if (!eligible || !expired) {
			Container.networkCommand(command);
		}
	}

	/**
	 * @param {!boolean} [force]
	 * @param {!boolean} [cache]
	 * @returns {!Promise}
	 */
	async updateSchedule(force = false, cache = false) {
		this.schedule.shouldUpdate = true;
		if (force || this.schedule.canReload) {
			await this.schedule.load(this.container.fonts, cache);
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async updateLineInAudio() {
		try {
			let stream = await navigator.mediaDevices.getUserMedia({
				audio: {
					echoCancellation: false,
					noiseSuppression: false,
					autoGainControl: false
				}
			});
			// Reuse the AudioContext and GainNode for every stream
			if (this.audioContext == null) {
				this.audioContext = new AudioContext();

				this.gainNode = this.audioContext.createGain();
				this.gainNode.connect(this.audioContext.destination);
			}
			if (!this.hasOwnProperty('audioSourceNode')) {
				this.audioSourceNode = null;
			}
			if (this.audioSourceNode) {
				// Disconnect so that the old source node can garbage collect
				this.audioSourceNode.disconnect(this.gainNode);
			}
			// Create a new source node for each new stream
			this.audioSourceNode = this.audioContext.createMediaStreamSource(stream);
			this.gainNode.gain.value = 1;
			this.audioSourceNode.connect(this.gainNode);
		} catch (err) {
			console.log(err.stack || err);
		}
	}

	/*
	 * Just handles audio in ducking. If anything else is playing audio then duck the audio in.
	 */
	updateDucking(duck) {
		if (this.gainNode) {
			if (duck) {
				this.gainNode.gain.value = Math.max(this.gainNode.gain.value - 0.01, (this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('audioinminvolume')) ? this.settings.data.widgetData.audioinminvolume : 0);
			} else {
				this.gainNode.gain.value = Math.min(this.gainNode.gain.value + 0.01, 1);
			}
		}
	}

	destroy() {
		this.timers.clear();
		this.timeSync.destroy();
		if (this.schedule !== null) {
			this.schedule.destroy();
			this.schedule = null;
		}
	}

	/**
	 * @param {!string} [previewPlaylistName]
	 * @returns {!Promise}
	 */
	async start(previewPlaylistName = '') {
		if (this.isPreview) {
			if (previewPlaylistName == null || '')
			{
				console.log('previewPlaylistName is null. This should not be possible.');
			}
			else
			{
				try {
					await this.schedule.loadPreview(previewPlaylistName);
				} catch (err) {
					console.log(`Skipping loading playlist '${previewPlaylistName}' due to error: ${err.stack || err}`);
				}
			}
		} else {
			await this.schedule.load(this.container.fonts, true);
			this.timers.on('Checkin', this.checkinDelay, this._checkin);
		}
	}

	/**
	 * @param {?Object} networkCommands
	 * @param {?Object} networkCommandSchedule
	 */
	parseNetworkCommands(networkCommands, networkCommandSchedules) {
		if (this.networkCommandSchedules !== null) {
			for (const networkCommandSchedule of this.networkCommandSchedules) {
				clearTimeout(networkCommandSchedule.timer)
			}
		}
		if (networkCommands === null || networkCommandSchedules === null) {
			this.networkCommands = null;
			this.networkCommandSchedules = null;
		} else {
			this.networkCommands = networkCommands;
			this.networkCommandSchedules = [];
			for (const networkCommandSchedule of Object.values(networkCommandSchedules)) {
				for (const scheduledCommand of networkCommandSchedule.data.scheduledcommands) {
					// Network commands can be deleted and aren't checked if they're used in a schedule. They can also be 'none' which will cause them not to be sent to specific players
					if (this.networkCommands.hasOwnProperty(scheduledCommand.command)) {
						const playerScheduledCommand = {
							schedule: new BreezeSchedule(this, scheduledCommand),
							command: this.networkCommands[scheduledCommand.command],
							timer: null
						};
						this.networkCommandSchedules.push(playerScheduledCommand);
						const scheduleNetworkCommand = () => {
							const delay = playerScheduledCommand.schedule.delayFromNow();
							if (delay !== Infinity) {
								playerScheduledCommand.timer = setTimeout(() => {
									Container.networkCommand(playerScheduledCommand.command);
									scheduleNetworkCommand();
								}, delay);
							}
						};
						scheduleNetworkCommand();
					}
				}
			}
		}
	}

	/**
	 * @returns <!Promise>
	 */
	async dailyUpdate() {
		try {
			await this.updateSchedule(false, false);
		} catch (err) {
			console.log(err.stack || err);
		}
	}

	/**
	 * @param {?number] delay
	 */
	scheduleDailyUpdate(delay) {
		this.timers.off('DailyUpdate');
		if (delay !== null) {
			this.timers.on(
				'DailyUpdate',
				delay,
				this.dailyUpdate.bind(this)
			);
		}
	}

	setDailyUpdate() {
		const dailyUpdate = this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('dailyUpdate') ? this.settings.data.widgetData.dailyUpdate : '';
		const delay = this.calculateDelayToDailyTime(dailyUpdate === undefined ? '' : dailyUpdate);
		
		this.scheduleDailyUpdate(delay);
	}

	/**
	 * @param {?number] delay
	 */
	scheduleDailyReboot(delay) {
		this.timers.off('DailyReboot');
		if (delay !== null) {
			this.timers.on(
				'DailyReboot',
				delay,
				ContainerAPI.RequestHardReboot
			);
		}
	}

	setDailyReboot() {
		const dailyReboot = this.settings.data.widgetData != null && this.settings.data.widgetData.hasOwnProperty('dailyReboot') ? this.settings.data.widgetData.dailyReboot : '';
		const delay = this.calculateDelayToDailyTime(dailyReboot === undefined ? '' : dailyReboot);
		this.scheduleDailyReboot(delay);
	}

	/**
	 * Calculates the delay to a daily reoccurring time.
	 * @param {!string} time Empty indicates no reboot, otherwise in "hh:mm:ss" format.
	 * @returns {?number}
	 */
	calculateDelayToDailyTime(time) {
		if (time === '') {
			return null;
		}
		const currentTime = this.currentTime;
		let local = this.timezone ? moment.tz(time, 'HH:mm:ss', this.timezone) : moment(time, 'HH:mm:ss').local();
		if (local.diff(currentTime) < 0) {
			local.add(1, 'day');
		}
		return local.diff(currentTime);
	}

	/**
	 * @returns {!HTMLElement}
	 */
	static createPlaceholder() {
		return $.div({ css: {
			position: 'absolute',
			width: '100%',
			height: '100%',
			pointerEvents: 'none'
		} });
	}

	resetIdle() {
		if (this.schedule) {
			for (const playlist of this.schedule.playlists) {
				playlist.resetIdle();
			}
		}
	}
}
