/* exported PlayerSettings */
/* global ContainerAPI, Dexie, Minibus, Size, Network */

const PlayerSettings = (() => {
	class PlayerSettingsData {
		/**
		 * @param {!string} key
		 * @param {!string} [account]
		 * @param {?string} [timezone]
		 * @param {?Object} [widgetData]
		 * @param {!Size} [size]
		 * @param {!Size} [usableSize]
		 * @param {!Size} [thumbnailSize]
		 * @param {!boolean} [stretchToFit]
		 * @param {!AlertFeed[]} [alertFeeds]
		 */
		constructor(
			key,
			account = '',
			timezone = null,
			widgetData = null,
			size,
			usableSize,
			thumbnailSize,
			stretchToFit,
			alertFeeds = []
		) {
			this.key = key;
			this.account = account;
			this.timezone = timezone;
			this.widgetData = widgetData;
			this.size = size;
			this.usableSize = usableSize;
			this.thumbnailSize = thumbnailSize;
			this.alertFeeds = alertFeeds;

			/**
			 * When true, the player will scale playlists to fit the
			 * player resolution, and when false the playlist will render 
			 * at a 1:1 pixel scale. This value is deduced from the
			 * passed in resolution, if either value is null it is
			 * assumed autoScale will be true (and the player resolution 
			 * is auto-detected) otherwise autoScale will be false.
			 * Always set to false if the player is in preview (the
			 * playlist will always render at the playlist resolution in
			 * preview.)
			 * @type {boolean}
			 */
			this.stretchToFit = stretchToFit;
		}
	}

	return class PlayerSettings {
		/**
		 * @param {!string} key
		 * @param {!string} [account]
		 */
		constructor(key, account) {
			this.data = new PlayerSettingsData(key, account);
			this.database = PlayerSettings.getDatabase();
			this.bus = Minibus.create();
		}

		/**
		 * @returns {!Promise}
		 */
		save() {
			return this.database.playerSettingsData.put(this.data);
		}

		/**
		 * @returns {!Promise}
		 */
		async load() {
			const data = await PlayerSettings.load(this.database, this.data.key);
			if (data !== undefined) {
				this.data = data;
			}
		}

		/**
		 * Validates that all of the basic required settings are properly 
		 * set.
		 * @returns {!boolean}
		 */
		validate() {
			return this.data.key != null &&
				this.data.account != null &&
				this.data.timezone != null &&
				this.data.widgetData != null &&
				this.data.size != null;
		}

		/**
		 * @param {!string} hostUrl
		 * @returns {!Promise}
		 */
		async initialize(hostUrl) {
			let shouldRequest = false;
			try {
				await this.load();
			} catch (err) {
				shouldRequest = true;
			}
			if (shouldRequest || !this.validate()) {
				await this.requestSettings(hostUrl);
			}
		}

		/**
		 * Requests only the players settings, should not clear queued
		 * player commands on the server as they will not be processed.
		 * @param {!string} hostUrl
		 * @returns {!Promise<!Response<!CheckinSettings>>}
		 */
		fetchSettings(hostUrl) {
			const url = `${hostUrl}get/playersettings.php?${Network.getParam('account', this.data.account)}&${Network.getParam('key', this.data.key)}`;
			return Network.fetch(url, 'json');
		}

		/**
		 * @param {!string} hostUrl
		 * @returns {!Promise}
		 */
		async requestSettings(hostUrl) {
			const { account } = await ContainerAPI.QueryRegistration();
			this.data.account = account;
			const response = await this.fetchSettings(hostUrl);
			if (!response.success) {
				throw new Error(response.error);
			}
			await this.updateWidgetData(response.widgetdata);
			await this.modifyTimezone(response.timezone);
			await this.updateDisplay(response.screenwidth, response.screenheight);
		}

		/**
		 * @param {!string} localIp
		 * @param {!string} containerVersion
		 * @param {!object} serial
		 * @returns {!Promise}
		 * @emits PlayerSettings#LocalsUpdated
		 */
		async updateLocals(localIp, containerVersion, serial) {
			this.data.localIp = localIp;
			this.data.containerVersion = containerVersion;
			this.data.serialnumber = serial ? serial.serialnumber : null;
			this.data.serialnumbermodifiedtime = serial ? serial.serialnumbermodifiedtime : null;
			await this.save();
			this.bus.emit('LocalsUpdated', this);
		}

		/**
		 * @param {?string} [timezone]
		 * @returns {!Promise}
		 */
		async modifyTimezone(timezone) {
			if (timezone != null && this.data.timezone != timezone) {
				this.data.timezone = timezone;
				await this.save();
				await ContainerAPI.UpdateTimezone(this.data.timezone);
				return true;
			}
			return false;
		}

		/**
		 * @param {!Object} widgetData
		 * @returns {!Promise}
		 */
		async updateWidgetData(widgetData) {
			const changed = this.data.widgetData == null || JSON.stringify(this.data.widgetData) != JSON.stringify(widgetData);
			// Handle serial number changes
			if (this.data.hasOwnProperty('serialnumber') && widgetData.hasOwnProperty('serialnumber') && this.data.hasOwnProperty('serialnumbermodifiedtime') && widgetData.hasOwnProperty('serialnumbermodifiedtime') &&
			    widgetData.serialnumbermodifiedtime > this.data.serialnumbermodifiedtime) {
				await ContainerAPI.UpdateSerialNumber(widgetData.serialnumber, widgetData.serialnumbermodifiedtime);
			}
			this.data.widgetData = widgetData;
			if (this.data.widgetData != null && this.data.widgetData.hasOwnProperty('keyboardoverlay') && this.data.widgetData.keyboardoverlay) {
				ShowKeyboard();
			} else {
				HideKeyboard();
			}
			await this.save();
			return changed;
		}

		/**
		 * @param {?number} width
		 * @param {?number} height
		 * @returns {!Promise}
		 */
		async updateDisplay(width, height) {
			// The preview player must NEVER use stretchToFit!
			const stretchToFit = width === null || height === null;
			this.data.size = new Size(width, height);
			this.data.stretchToFit = stretchToFit;
			
			const usableWidth = stretchToFit ? window.innerWidth : width;
			const usableHeight = stretchToFit ? window.innerHeight : height;
			
			let reloadSchedule = false;
			if (this.data.usableSize == undefined || this.data.usableSize.width != usableWidth || this.data.usableSize.height != usableHeight)
			{
				this.data.usableSize = new Size(usableWidth, usableHeight);
				reloadSchedule = true;
			}
			this.data.thumbnailSize = PlayerSettings.calculateThumbnailSize(usableWidth, usableHeight);
			await this.save();
			return reloadSchedule;
		}

		/**
		 * @returns {!Promise}
		 */
		delete() {
			return PlayerSettings.delete(this.database, this.data.key);
		}

		/**
		 * @returns {!Dexie}
		 */
		static getDatabase() {
			const db = new Dexie('PlayerSettingsData');
			db.version(1).stores({
				playerSettingsData: '&key'
			});
			db.version(2).stores({
				playerSettingsData: '&key'
			});
			db.playerSettingsData.mapToClass(PlayerSettingsData);
			return db;
		}

		/**
		 * Returns the specified PlayerSettingsData object if it exists,
		 * otherwise returns undefined.
		 * @param {!Dexie} db
		 * @param {!string} key
		 * @returns {!Promise<PlayerSettingsData>}
		 */
		static load(db, key) {
			return db.playerSettingsData.where('key').equals(key).first();
		}

		/**
		 * @param {!Dexie} db
		 * @param {!string} key
		 * @returns {!Promise}
		 */
		static delete(db, key) {
			return db.playerSettingsData.delete(key);
		}

		/**
		 * @param {!number} width
		 * @param {!number} height
		 * @returns {!Size}
		 */
		static calculateThumbnailSize(width, height) {
			const maxWidth = 320;
			const maxHeight = 320;
			const constrainedWidth = Math.min(width, maxWidth);
			const constrainedHeight = Math.min(height, maxHeight);
			
			return constrainedWidth / width > constrainedHeight / height ? new Size(
				Math.min(maxWidth, Math.round(width / height * constrainedHeight)),
				constrainedHeight
			) : new Size(
				constrainedWidth,
				Math.min(maxHeight, Math.round(height / width * constrainedWidth))
			);
		}
	};
})();
