/* exported Playlist */
/* global Messenger, Utility, Size, Zone, WidgetsSchema, Timer, FileInfo, Type, moment */

class Playlist {
	/**
	 * @param {!Object} data
	 * @param {!Player} player
	 * @param {?HTMLElement} $parent
	 */
	constructor(data, player, $parent) {
		this.player = player;
		this.uniqueId = this.player.playlistUniqueIds++; // This is a unique id used in places between playlist loads where an old and new playlist can be active with the same id
		this.id = data.id;
		this.name = data.name;
		this.size = new Size(data.pagewidth, data.pageheight);
		
		this.isLoaded = false;
		
		this.contenttree = data.contenttree;
		this.metadata = data.metadata;

		this.player = player;

		/*
		 * Prevent rebinding the update function every frame.
		 */
		this._update = this.update.bind(this);

		/**
		 * The amount of time in milliseconds that the cursor should remain visible
		 * after the mouse has been moved.
		 * @type {!number}
		 */
		this.CURSOR_HIDE_DELAY = 750;

		/**
		 * The delay in milliseconds between {@link Playlist#updateTicker} updates.
		 * @type {!number}
		 */
		this.TICKER_DELAY = 1000;

		/**
		 * The delay in milliseconds between audio ducking updates, see {@link 
		 * Playlist#updateDucking}.
		 * @type {!number}
		 */
		this.DUCKING_DELAY = 1000 / 25;

		this.SYNC_DELAY = 5000;

		this.LED_OFF_COLOR = 'rgb(0,0,0)';

		/**
		 * The current color that the LEDs (if available) should be displayed as. 
		 * See {@link Container#setLedColor} for additional information.
		 * @type {!string}
		 */
		this.ledColor = this.LED_OFF_COLOR;

		/**
		 * The list of Media currently in a transition.
		 * @type {!TransitionEvent[]}
		 */
		this.transitionEvents = [];

		/** @type {!Media[]} */
		this.audioTracks = [];

		/** @type {!InteractiveButton[]} */
		this.buttons = [];

		/**
		 * The timestamp of the last time a user interacted with the Playlist.
		 * @type {!number}
		 */
		this.lastUserInteraction = 0;

		/**
		 * The timestamp of the last action (either a user action or an automated 
		 * action, e.g. user tapped the screen or a button idle action fired.)
		 * @type {!number}
		 */
		this.lastAction = 0;

		/**
		 * The last button pressed. Used to stop the global and idle button from running in a loop
		 * @type {!object}
		 */
		this.lastButton = null;

		/**
		 * The timestamp of the last time a user moved the mouse. Used for showing 
		 * or hiding the cursor.
		 * @type {!number}
		 */
		this.lastPointerMove = 0;

		/** @type {!boolean} */
		this.isCursorShown = false;

		/**
		 * The amount of time in milliseconds that the Playlist must be idle (i.e. 
		 * no user interaction) before the {@link Playlist#idleAction} is executed.
		 * @type {!number}
		 */
		this.idleDelay = Infinity;

		/**
		 * The {@link InteractiveButton#name} of the InteractiveButton to run when 
		 * the Playlist has been idle for {@link Playlist#idleDelay}.
		 * @type {?string}
		 */
		this.idleAction = null;
		
		this.globalIdleDelay = Infinity;
		
		this.lastUserInteractionClickHandler = null;
		
		/**
		 * The last timestamp passed to the requestAnmiationFrame callback 
		 * {@link Playlist#update}.
		 * @type {!number}
		 */
		this.stamp = 0;

		/**
		 * Set to a timestamp in milliseconds when the playlist should call the 
		 * {@link Playlist#onPlayFinish} callback, if set. This is set when the 
		 * last Media in the playlist starts playing (see 
		 * {@link Playlist#flagLastMedia}).
		 * @type {?number}
		 */
		this.playlistFinishedStamp = null;

		/**
		 * The timestamp in milliseconds that {@link Playlist#updateZones} last 
		 * ran.
		 * @type {!number}
		 */
		this.lastUpdate = 0;

		/**
		 * The delay in milliseconds until the next MediaEvent occurs.
		 * @type {!number}
		 */
		this.delay = 0;

		/**
		 * The timestamp in milliseconds that {@link Playlist#updateTicker} last 
		 * ran.
		 * @type {!number}
		 */
		this.lastTickerUpdate = 0;

		/**
		 * The timestamp in milliseconds that {@link Playlist#updateDucking} last ran.
		 * @type {!number}
		 */
		this.lastDuckingUpdate = 0;

		this.lastSyncUpdate = 0;

		/**
		 * The requestAnimationFrame callback ID.
		 * @type {!number}
		 */
		this.rafId = -1;

		/**
		 * Represents whether the Playlist is currently playing or not.
		 * @type {!boolean}
		 */
		this.playing = false;

		/** @type {!boolean} */
		this.addVisibleClass = false;

		/** @type {!boolean} */
		this.destroyed = false;

		/**
		 * Whether zones equal to the playlist runlength should loop
		 * @type {!boolean}
		 */
		this.disableLoopOnZones = false;

		const playerSize = this.player.settings.data.stretchToFit ? this.player.settings.data.usableSize : this.player.settings.data.size;

		if (this.player.settings.data.stretchToFit) {
			this.scale = Playlist.calculateScale(this.size, playerSize);
		} else {
			this.scale = new Size(1, 1);
		}

		/**
		 * An optional callback to run when the Playlist finishes playing one 
		 * iteration of the Playlist.
		 * @type {?function}
		 */
		this.onPlayFinish = null;

		this.routes = new RouteMap(Messenger.bus);
		this.routes.on(
			'FrontLedColorChange',
			this.onFrontLedColorChange.bind(this)
		);

		this.elId = `playlist-${this.id}-${Utility.getUID()}`;
		this.el = Playlist.generateDOM(
			this.elId,
			this.size.width,
			this.size.height
		);

		this.setTransform();
		this.hide();

		if ($parent) {
			$parent.appendChild(this.el);
		}
		this.el.addEventListener('pointermove', this.onPointerMove.bind(this));

		this.zones = data.zones.map(data => new Zone(this, data));
		this.signwaveIds = this.getAllMedia().filter(data => data.isSignwave()).map(media => media.id);
	}
	setTransform() {
		if (this.player.settings.data.stretchToFit) {
			this.el.style.transform = `scale(${this.scale.width}, ${this.scale.height})`;
		} else {
			this.el.style.removeProperty('transform');
		}
	}
	static async get(player, name, loadFromCache) {
		const url = `${player.container.hostUrl}${player.isPreview ? 'preview' : 'playlist'}/${player.settings.data.account}/${name}.json?d=${Date.now()}`;
		const file = new FileInfo(`playlists/${name}.json`, false);
		return Utility.load(player.container.features.caching, loadFromCache, file, url, 'json');
	}

	/**
	 * Gets the list of font filenames excluding extensions used in the Playlist.
	 * @param {!Object} widgetsSchema
	 * @returns {!string[]}
	 */
	getUsedFonts(widgetsSchema) {
		const fontSchemas = WidgetsSchema.getSchemasByDesign(widgetsSchema, 'fontfamily');
		return this.zones
			.map(zone => zone.media)
			.reduce((flat, val) => flat.concat(val), [])
			.map(media => media.getFonts(fontSchemas))
			.filter(x => x.length > 0)
			.reduce((flat, val) => flat.concat(val), []);
	}

	/**
	 * @returns {!Media[]}
	 */
	getAllMedia() {
		return this.zones.reduce((arr, zone) => arr.concat(zone.media), []);
	}

	/**
	 * @returns {!Promise}
	 */
	static async cleanCache() {
		const dir = new FileInfo('playlists', true);
		await dir.delete();
	}

	/**
	 * @returns {!Promise<!Promise[]>}
	 */
	async load() {
		const media = this.getAllMedia();
		for (const zone of this.zones) {
			zone.attach(this.el);
		}
		const waitForReady = Utility.flat(this.zones.map(zone => zone.initialize()));
		const waitForCacheRequests = media.map(
			media => Promise.race([media.cacheRequestedPromise, media.cachePromise, media.destroyPromise])
		);
		console.log(`start: await Promise.allSettled(waitForCacheRequests); Length: ${waitForCacheRequests.length}`);
		const allSettled = promises =>
			Promise.all(promises.map(promise => promise
				.then(value => ({ state: 'fulfilled', value }))
				.catch(reason => ({ state: 'rejected', reason }))
		));
		const results = await allSettled(waitForCacheRequests);
		// throw if any of these isn't just a media cache failure
		for (const result of results) {
			if (result.state == 'rejected' && result.reason.message != 'Failed twice to download and/or verify the hash.') {
				throw result.reason;
			}
		}
		
		console.log('end: await Promise.allSettled(waitForCacheRequests);');
		
		// Do the updateView once to ensure media events are populated
		for (const zone of this.zones) {
			zone.updateActiveMedia();
		}
		const maximumZoneRunLength = this.zones.reduce((max, zone) => Math.max(max, zone.duration), 0);
		for (const zone of this.zones) {
			zone.updateView(zone.startStamp, false, zone.stopOnLastMedia, maximumZoneRunLength);
		}
		
		// Schedule all zones to call updateView after they finish initializing.
		// Do not await on this as load() should return as soon as possible after awaiting 'waitForCacheRequests'
		// Note: Brandon thinks this only exists here so that the then inserts onto the task and executes after the await in playerschedule.js
		// So this code could be moved after the playlist load code there I think.
		const waitForAllReady = Promise.all(waitForReady);
		waitForAllReady.then(() => {
			console.log('start: all zones updateView');
			for (const zone of this.zones) {
				zone.updateActiveMedia();
			}
			const maximumZoneRunLength = this.zones.reduce((max, zone) => Math.max(max, zone.duration), 0);
			for (const zone of this.zones) {
				zone.updateView(zone.startStamp, false, zone.stopOnLastMedia, maximumZoneRunLength);
			}
			console.log('end: all zones updateView');
		});
		return waitForAllReady;
	}

	TimeSync() {
		for (const zone of this.zones) {
			zone.TimeSync();
		}
	}

	/**
	 * @returns {!Promise<number>}
	 */
	async getLoadingProgress() {
		const media = this.getAllMedia();
		const mediaProgressPromises = media.map(media => new Promise(async resolve => resolve({ media, progress: await media.getLoadingProgress() })));
		const mediaProgress = await Promise.all(mediaProgressPromises);
		console.log(`${this.name} Loaded Media: ${JSON.stringify(mediaProgress.filter(m => m.progress == 100).map(m => m.media.name))}`);
		console.log(`${this.name} Loading Media: ${JSON.stringify(mediaProgress.filter(m => m.progress != 100).map(m => m.media.name))}`);
		const totalProgress = mediaProgress.reduce((sum, media) => sum + media.progress, 0);
		return totalProgress / (media.length * 100);
	}

	/**
	 * Marks the end of the playlist by setting a flag on the first Media object 
	 * that ends when the Playlist does. Ignores cloned media.
	 */
	flagLastMedia() {
		for (const zone of this.zones) {
			for (const media of zone.activeMedia) {
				media.lastMediaInPlaylist = false;
			}
		}

		const longestZoneData = this.zones.map((zone, index) => {
			let duration = zone.duration;
			return {
				index: index,
				duration: duration
			};
		}).reduce((prev, curr) => {
			if (prev === null) {
				return curr;
			}
			if (prev.duration >= curr.duration) {
				return prev;
			}
			return curr;
		}, null);

		if (longestZoneData) {
			const zone = this.zones[longestZoneData.index];
			if (zone.activeMedia.length != 0) {
				zone.activeMedia[zone.activeMedia.length - 1].lastMediaInPlaylist = true;
			}
		}
	}

	reset() {
		if (this.el) {
			this.el.classList.remove('playlist-visible');
		}
		if (this.zones) {
			for (const zone of this.zones) {
				zone.reset();
			}
		}
	}

	/**
	 * Calls {@link Zone#destroy} on each contained Zone. Releases all Container 
	 * resources linked to this playlistElementId. Removes all children DOM 
	 * Elements from the DOM and removes this Playlists DOM Element from the DOM. 
	 * Removes all Messenger routes. Resets playback state.
	 */
	destroy() {
		if (!this.destroyed) {
			this.destroyed = true;
			this.stop();
			this.routes.clear();
			if (this.lastUserInteractionClickHandler != null) {
				document.removeEventListener('click', this.lastUserInteractionClickHandler);
				this.lastUserInteractionClickHandler = null;
			}
			if (this.zones) {
				for (const zone of this.zones) {
					zone.destroy();
				}
			}
			if (this.el) {
				console.log('PlaylistDestroyed');
				ContainerAPI.PlaylistDestroyed(this.uniqueId);
				this.el.innerHTML = '';
				this.el.remove();
				this.el = null;
			}
		}
	}

	/**
	 * Records the timestamp of the last mouse movement and shows
	 * the cursor if it's hidden.
	 */
	onPointerMove() {
		this.lastPointerMove = performance.now();
		if (!this.isCursorShown) {
			this.showCursor();
		}
	}

	/**
	 * Shows the cursor on the page via CSS.
	 */
	showCursor() {
		this.isCursorShown = true;
		if (this.el) {
			this.el.style.cursor = 'default';
		}
	}

	/**
	 * Hides the cursor on the page via CSS.
	 */
	hideCursor() {
		this.isCursorShown = false;
		if (this.el) {
			this.el.style.cursor = 'none';
		}
	}

	/**
	 * Shows the Playlist in the DOM.
	 */
	show() {
		this.el.style.visibility = 'visible';
		this.setTransform();

		if (this.lastUserInteractionClickHandler == null) {
			document.addEventListener('click', this.lastUserInteractionClickHandler = () => {
				this.resetIdle();
			});
		}
	}

	/**
	 * Hides the Playlist in the DOM.
	 */
	hide() {
		this.el.style.visibility = 'hidden';
		// Fix for a compositing issue over videos on Chrome 85
		this.el.style.transform = 'translate(-10000px,-10000px)';
		
		if (this.lastUserInteractionClickHandler != null) {
			document.removeEventListener('click', this.lastUserInteractionClickHandler);
			this.lastUserInteractionClickHandler = null;
		}
	}

	/**
	 * @param {!string} rgb
	 * @param {!string|!number} mediaId
	 */
	onFrontLedColorChange(rgb, mediaId) {
		if (this.signwaveIds.indexOf(mediaId) != -1) {
			this.ledColor = rgb;
			if (this.playing && this.player.container.features.led) {
				this.player.container.setFrontLedColor(rgb);
			}
		}
	}

	/**
	 * @param {} button
	 * @param {HTMLElement} $button
	 */
	AddButton(button, $button) {
		this.buttons.push(button);
		if (button.hasOwnProperty('globalidle') && button.globalidle && button.hasOwnProperty('globalidledelay')) {
			this.globalIdleDelay = button.globalidledelay * 1000;
		}
		if (button.clickable) {
			$button.addEventListener('click', () => {
				this.resetIdle();
				this.UpdateView(button);
			});
		}
	}

	resetIdle() {
		this.lastUserInteraction = performance.now();
		this.lastAction = this.lastUserInteraction;
	}

	/**
	 * @param {} button
	 */
	UpdateView(button = null) {
		if (button) {
			this.lastButton = button;
			for (const otherButton of this.buttons) {
				if (otherButton.buttongroup === button.buttongroup) {
					ApplyButtonStyle(this.player.container, otherButton.$button, otherButton, false);
				}
			}
			ApplyButtonStyle(this.player.container, button.$button, button, true);
			this.resetIdleAction();
			if (button.idleactionbutton !== 'none') {
				this.idleAction = button.idleactionbutton;
				this.idleDelay = button.idledelay * 1000;
			}
			this.stamp = performance.now();
			button.zones.forEach((zoneView, zoneIndex) => {
				const zone = this.zones[zoneIndex];
				if (zoneView.statechange) {
					const trimLeft = zoneView.trimleft;
					let trimRight = zoneView.trimright === -1 ? zone.media.length : zoneView.trimright;
					if (!zoneView.show) {
						trimRight = -1;
					}
					zone.media.forEach((media, mediaIndex) => {
						if (mediaIndex >= trimLeft && mediaIndex <= trimRight) {
							if (button.hasOwnProperty('metadata')) {
								for (const metadata of button.metadata) {
									const mediaMetadata = media.metadata.find(m => m.name == metadata.name && m.type == metadata.type);
									if (mediaMetadata) {
										mediaMetadata.value = metadata.value;
									} else {
										media.metadata.push(Object.assign({}, metadata));
									}
								}
							}
							media.enable(false);
						} else {
							media.disable(false);
						}
					});
				}
			});
			if (button.hasOwnProperty('networkcommands')) {
				for (const networkCommandId of button.networkcommands) {
					if (this.player.networkCommands.hasOwnProperty(networkCommandId)) {
						Container.networkCommand(this.player.networkCommands[networkCommandId]);
					}
				}
			}
			if (button.hasOwnProperty('apicommands')) {
				for (const apiCommandId of button.apicommands) {
					const url = `${this.player.container.hostUrl}post/apicommand.php?${Network.getParam('key', this.player.settings.data.key)}&${Network.getParam('apicommand', apiCommandId)}`;
					Network.fetch(url, 'json');
				}
			}
		}
		for (const zone of this.zones) {
			zone.updateActiveMedia();
		}
		const maximumZoneRunLength = this.zones.reduce((max, zone) => Math.max(max, zone.duration), 0);
		this.zones.forEach((zone, zoneIndex) => {
			if (button) {
				const buttonZone = button.zones[zoneIndex];
				if (buttonZone.statechange) {
					zone.updateView(this.stamp, buttonZone.restart, buttonZone.stop, maximumZoneRunLength);
				}
			} else {
				zone.updateView(this.stamp, false, false, maximumZoneRunLength);
			}
		});
		this.flagLastMedia();
		this.updateZones();
		if (this.player.isPreview) {
			this.postUpdateZoneViews();
		}
	}

	postUpdateZoneViews() {
		window.parent.postMessage({
			event: 'updatezoneviews',
			zones: this.zones.map(zone =>
				zone.media.map(media => media.enabled)
			)
		}, '*');
	}

	/**
	 * @returns {!number} Returns the largest {@link Zone#firstEventOffset} value 
	 * across all active Zones.
	 */
	getInitialStartTime() {
		return this.zones.reduce((offset, zone) => Math.max(offset, zone.firstEventOffset), 0);
	}

	resetIdleAction() {
		this.idleAction = null;
		this.idleDelay = Infinity;
		this.lastAction = performance.now();
	}

	runIdleAction() {
		const button = this.buttons.find(button => this.idleAction === button.name);
		if (button && this.lastButton !== button) {
			this.UpdateView(button);
		}
	}

	runGlobalIdleAction() {
		const button = this.buttons.find(button => button.hasOwnProperty('globalidle') && button.globalidle);
		if (button && this.lastButton !== button) {
			this.UpdateView(button);
		}
	}

	/**
	 * @param {!number} startTime
	 */
	start(startTime) {
		console.log('Playlist start');
		this.flagLastMedia();
		this.stamp = performance.now();
		this.lastUpdate = this.stamp;
		this.lastTickerUpdate = this.stamp - this.TICKER_DELAY;
		this.lastDuckingUpdate = this.stamp - this.DUCKING_DELAY;
		this.lastSyncUpdate = this.stamp - this.SYNC_DELAY;
		this.delay = 0;
		const initialStartTime = this.getInitialStartTime();
		console.log('initialStartTime: ', initialStartTime);
		if (startTime > 0) {
			console.log('startTime > 0');
			const goalStamp = this.stamp;
			this.stamp = goalStamp - startTime - initialStartTime;
			for (const zone of this.zones) {
				zone.start(this.stamp, initialStartTime - zone.firstEventOffset);
			}
			while (this.stamp <= goalStamp) {
				this.delay = null;
				for (const zone of this.zones) {
					this.updateZone(zone);
				}
				this.stamp += this.delay;
			}
			this.stamp -= this.delay;
			const extraTime = goalStamp - this.stamp;
			this.delay = this.delay - extraTime;
			for (const zone of this.zones) {
				zone.nextEventStamp -= extraTime;
			}
			for (const transitionEvent of this.transitionEvents) {
				transitionEvent.media.endTransition(transitionEvent.type);
			}
			this.transitionEvents = [];
			for (const zone of this.zones) {
				const zoneStartTime = startTime % zone.duration;
				for (const media of zone.activeMedia) {
					if (media.isShowing()) {
						const deltaTime = zoneStartTime - media.entry.event.defaultTime;
						media.beginAnimation(this.stamp - deltaTime);
						if (deltaTime < media.entry.transition.duration) {
							const transitionDelta = zoneStartTime - media.entry.event.defaultTime;
							const transitionStamp = this.stamp - transitionDelta;
							media.beginTransition(Type.PlaybackEvent.ENTRY, transitionStamp);
							media.show();
							this.addTransitionEvent(media, Type.PlaybackEvent.ENTRY, transitionStamp);
						}
						if (zoneStartTime >= media.exit.event.defaultTime) {
							const transitionDelta = zoneStartTime - media.exit.event.defaultTime;
							const transitionStamp = this.stamp - transitionDelta;
							media.beginTransition(Type.PlaybackEvent.EXIT, transitionStamp);
							this.addTransitionEvent(media, Type.PlaybackEvent.EXIT, transitionStamp);
						}
						if (media.isPlaying()) {
							const seekTime = zoneStartTime - media.startTime;
							media.seek(seekTime / 1000);
						}
						media.render();
					}
				}
			}
		} else {
			console.log('before updateZones');
			for (const zone of this.zones) {
				zone.start(this.stamp, initialStartTime - zone.firstEventOffset);
			}
			this.updateZones();
		}

		if (this.player.isPreview) {
			if (startTime === 0) {
				this.lastTickerUpdate += initialStartTime;
			}
		}
		for (const button of this.buttons) {
			if (button.default) {
				this.UpdateView(button);
				if (this.audioTracks.length > 1) {
					for (const media of this.audioTracks) {
						media.duckVolume();
					}
				}
			}
		}

		// If there are multiple audio tracks they should all start ducked.
		if (this.audioTracks.length > 1) {
			for (const media of this.audioTracks) {
				media.duckVolume();
			}
		}

		this.rafId = requestAnimationFrame(this._update);
	}

	/**
	 * Begin playback of the Playlist.
	 * @param {!number} [startTime]
	 */
	play(startTime = 0) {
		console.log('playlist play');
		// Super naive hack to fix the while loop below in this.start( from running a really long time
		const maximumZoneRunLength = this.zones.reduce((max, zone) => Math.max(max, zone.duration), 0);
		if (maximumZoneRunLength != 0) {
			startTime %= maximumZoneRunLength;
		} else {
			this.playlistFinishedStamp = 0;
		}
		this.player.backgroundMusicService.ActivePlaylistChanged(this.id);
		this.kill = false;
		this.finished = false;
		this.playing = true;
		console.log('start: ', startTime);
		this.start(startTime);
		if (this.player.container.features.led) {
			this.player.container.setFrontLedColor(this.ledColor);
		}
		if (this.player.isPreview) {
			this.postUpdateZoneViews();
		}
		(async () => {
			try {
				const initialStartTime = this.getInitialStartTime();
				await Timer.sleep(initialStartTime);
				this.addVisibleClass = true;
			} catch (err) {
				console.log(err);
			}
		})().catch((err) => {
			console.log(err);
		});
	}

	/**
	 * Stops all playback and resets playback state.
	 */
	stop() {
		this.playing = false;
		this.kill = true;
		if (this.rafId > -1) {
			cancelAnimationFrame(this.rafId);
		}

		this.transitionEvents = [];
		this.audioTracks = [];
		this.ledColor = Utility.ledOffColor;
		this.rafId = -1;
		this.playlistFinishedStamp = null;
		this.reset();
	}

	/**
	 * @param {!Media} media
	 */
	addAudioTrack(media) {
		this.audioTracks.push(media);
		if (this.audioTracks.length === 1) {
			media.resetVolume();
		} else {
			media.duckVolume();
		}
	}

	/**
	 * @param {!Media} media
	 */
	removeAudioTrack(media) {
		for (let itr = this.audioTracks.length - 1; itr >= 0; --itr) {
			if (this.audioTracks[itr].id === media.id) {
				this.audioTracks.splice(itr, 1);
				break;
			}
		}
		media.stop();
		media.seek(0);
	}

	/**
	 * @param {!Media} media
	 */
	ResetMedia(media) {
		for (let itr = this.transitionEvents.length - 1; itr >= 0; --itr) {
			const transitionEvent = this.transitionEvents[itr];
			if (transitionEvent.media.id == media.id) {
				this.transitionEvents.splice(itr, 1);
				break;
			}
		}
	}

	/**
	 * @param {!Media} media
	 * @param {!Type.PlaybackEvent} type
	 * @param {!number} stamp
	 */
	addTransitionEvent(media, type, stamp) {
		let end = stamp;
		let play = null;
		if (type === Type.PlaybackEvent.ENTRY) {
			end += media.entry.transition.duration;
			if (media.type === Type.Resource.AUDIO ||
				media.type === Type.Resource.VIDEO) {
				play = stamp + media.entry.transition.duration / 2;
			}
		} else if (type === Type.PlaybackEvent.EXIT) {
			end += media.exit.transition.duration;
		}
		this.transitionEvents.push({
			media: media,
			type: type,
			start: stamp,
			end: end,
			play: play
		});
	}

	/**
	 * Removes the given Media object from playback. Runs after the Medias EXIT 
	 * transition is finished (if an EXIT transition exists) or after the Media has
	 * played for its full duration. Does not call {@link Media#render}.
	 * @param {!Media} media
	 */
	removeMedia(media) {
		if (media.type === Type.Resource.AUDIO ||
			media.type === Type.Resource.VIDEO) {
			this.removeAudioTrack(media);
		}
		media.reset();
	}

	/**
	 * @param {} evt
	 */
	playMediaEvent(evt) {
		if (evt === undefined) {
			return;
		}
		if (evt.type === Type.PlaybackEvent.ENTRY) {
			if (!evt.media.isShowing()) {
				if (this.onPlayFinish !== null && evt.media.lastMediaInPlaylist) {
					this.playlistFinishedStamp = this.stamp + evt.media.duration + evt.media.entry.transition.duration / 2;
				}
				if (evt.media.type === Type.Resource.AUDIO ||
					evt.media.type === Type.Resource.VIDEO ||
					evt.media.type === Type.Resource.WIDGET && (evt.media.widgettype == 'audiostreaming' || evt.media.widgettype == 'videostreaming' && evt.media.attributes.format == 'hls')) {
					if (evt.media.maxVolume > 0) {
						this.addAudioTrack(evt.media);
					}
					if (evt.media.entry.transition.type === Type.Transition.NONE) {
						evt.media.play();
					}
				}
				if (evt.media.entry.transition.type !== Type.Transition.NONE) {
					evt.media.beginTransition(Type.PlaybackEvent.ENTRY, this.stamp);
					this.addTransitionEvent(
						evt.media,
						Type.PlaybackEvent.ENTRY,
						this.stamp
					);
				}
				evt.media.beginAnimation(this.stamp);
				evt.media.show();
				evt.media.render();
			} else {
				// The Media is already added, so simply reset the Animation and playback states.
				evt.media.beginAnimation(this.stamp);
				if (evt.media.isPlaying()) {
					evt.media.seek(0);
					evt.media.play();
					evt.media.render();
				}
			}
		} else if (evt.type === Type.PlaybackEvent.EXIT) {
			if (evt.media.exit.transition.type !== Type.Transition.NONE) {
				evt.media.beginTransition(Type.PlaybackEvent.EXIT, this.stamp);
				evt.media.render();
				this.addTransitionEvent(
					evt.media,
					Type.PlaybackEvent.EXIT,
					this.stamp
				);
			} else {
				this.removeMedia(evt.media);
			}
		}
	}

	updateZones() {
		this.delay = null;
		for (const zone of this.zones) {
			this.updateZone(zone);
		}
		this.lastUpdate = this.stamp;
	}

	/**
	 * @param {!Zone} zone
	 */
	updateZone(zone) {
		zone.update(this.stamp);
		if (zone.currentEvents !== null) {
			for (const event of zone.currentEvents) {
				this.playMediaEvent(event);
			}
		}
		if (this.delay === null || this.delay > zone.delay) {
			this.delay = zone.delay;
		}
	}

	syncMedia() {
		for (const zone of this.zones) {
			for (const media of zone.activeMedia) {
				media.sync();
			}
		}
	}

	updateTransitions() {
		for (let itr = this.transitionEvents.length - 1; itr >= 0; --itr) {
			const transitionEvent = this.transitionEvents[itr];
			if (transitionEvent.play !== null && this.stamp >= transitionEvent.play && !transitionEvent.media.isPlaying()) {
				transitionEvent.media.play();
				transitionEvent.media.render();
			}
			if (this.stamp >= transitionEvent.end) {
				if (transitionEvent.type === Type.PlaybackEvent.ENTRY) {
					transitionEvent.media.endTransition(transitionEvent.type);
					transitionEvent.media.render();
				} else if (transitionEvent.type === Type.PlaybackEvent.EXIT) {
					this.removeMedia(transitionEvent.media);
				}
				this.transitionEvents.splice(itr, 1);
			}
		}
	}

	updateTicker() {
		window.parent.postMessage({
			event: 'time',
			times: this.zones.map(zone => {
				let zoneDuration = zone.duration;
				if (zone.stopped) {
					return zoneDuration / 1000;
				} else {
					const zonePlayedTime = this.stamp - zone.startStamp;
					return Math.round((zonePlayedTime % zoneDuration) / 1000);
				}
			})
		}, '*');
	}

	/**
	 * @param {!Media} media
	 */
	updateDucking(media) {
		if (this.audioTracks.length > 1) {
			media.duckDown();
		} else {
			media.duckUp();
		}
	}

	/**
	 * @param {!number} stamp The timestamp passed to the
	 * requestAnimationFrame callback.
	 */
	update(stamp) {
		if (this.kill) {
			return;
		}
		if (this.playlistFinishedStamp !== null && this.stamp >= this.playlistFinishedStamp) {
			if (this.onPlayFinish !== null) {
				this.playlistFinishedStamp = null;
				this.onPlayFinish();
				return;
			}
		}
		this.rafId = requestAnimationFrame(this._update);
		this.stamp = stamp;
		if (this.stamp - this.lastAction > this.idleDelay) {
			this.runIdleAction();
		}
		if (this.stamp - this.lastAction > this.globalIdleDelay) {
			this.runGlobalIdleAction();
		}
		if (this.isCursorShown && this.stamp - this.lastPointerMove > this.CURSOR_HIDE_DELAY) {
			this.hideCursor();
		}
		if (stamp >= this.lastSyncUpdate + this.SYNC_DELAY) {
			this.syncMedia();
			this.lastSyncUpdate = stamp;
		}
		this.updateTransitions();
		if (stamp >= this.lastDuckingUpdate + this.DUCKING_DELAY) {
			for (const media of this.audioTracks) {
				this.updateDucking(media);
			}
			this.player.updateDucking(this.audioTracks.length > 0);
			if (this.audioTracks.length > 0) {
				this.player.backgroundMusicService.DuckAudioDown(this.id);
			} else {
				this.player.backgroundMusicService.DuckAudioUp(this.id);
			}
			this.lastDuckingUpdate = stamp;
		}
		if (stamp >= this.lastUpdate + this.delay) {
			this.updateZones();
		}
		if (this.addVisibleClass) {
			this.el.classList.add('playlist-visible');
			this.addVisibleClass = false;
		}
		if (this.player.isPreview && stamp >= this.lastTickerUpdate + this.TICKER_DELAY) {
			this.updateTicker();
			this.lastTickerUpdate = stamp;
		}
	}

	/**
	 * @param {!string} id
	 * @param {!number} width
	 * @param {!number} height
	 * @returns {!HTMLElement}
	 */
	static generateDOM(id, width, height) {
		return $.div({ id, class: 'playlistContainer', css: {
			position: 'absolute',
			left: '0',
			top: '0',
			width: `${width}px`,
			height: `${height}px`,
			transformOrigin: 'top left'
		} });
	}

	/**
	 * @param {!Size} playlistSize
	 * @param {!Size} playerSize
	 * @returns {!Size}
	 */
	static calculateScale(playlistSize, playerSize) {
		return new Size(playerSize.width / playlistSize.width, playerSize.height / playlistSize.height);
	}
}

/**
 * @typedef {Object} InteractiveButton
 * @property {boolean} default Sets the default InteractiveButton view
 * @property {boolean} globalidle
 * @property {number} globalidledelay
 * one InteractiveButton per Playlist will be set to the default.
 * @property {number} idledelay The amount of time in seconds without a 
 * user interaction before switching to the InteractiveButton view set by 
 * the {@link InteractiveButton#idleactionbutton} property.
 * @property {string} idleactionbutton The {@link InteractiveButton#name} 
 * of the InteractiveButton to run if no user interaction occurs before 
 * {@link InteractiveButton#idledelay} seconds pass.
 * @property {string} name A Playlist-wide unique InteractiveButton name.
 * @property {ZoneView[]} zones An array detailing the ZoneView of each 
 * Zone in the playlist that should be used when this InteractiveButton is
 * run.
 */

/**
 * @typedef {Object} ZoneView
 * @property {boolean} restart Determines if the Zone should start on its 
 * first Media item or continue as it was. Zones implicitly restart if they
 * had no items in their {@link Zone#activeMedia} list.
 * @property {boolean} show Determines if the display status of the Zone. 
 * If this is set to false, the Zones {@link Zone#activeMedia} list will be
 * empty.
 * @property {boolean} stop If true, the Zone will play until the it plays 
 * the last Media item in the Zones {@link Zone#activeMedia} list, where it
 * will then stop (it will not play the end transition of the last piece of
 * Media).
 * @property {number} trimleft If {@link ZoneView#show} is set to true, 
 * then trimleft will either be -1 (to indicate no trim) or a 1-based index
 * into {@link Zone#media} determining where Media should begin to be 
 * enabled. E.g., if trimleft is 1 then the first Media in the 
 * {@link Zone#media} list at index 0 will be disabled, and Media starting 
 * from trimleft (1) will be enabled.
 * @property {number} trimright If {@link ZoneView#show} is set to true,
 * then trimright will either be -1 (to indicate no trim) or a 1-based 
 * index into {@link Zone#media} determining where Media should end being 
 * enabled. E.g., if trimright is 5 then all Media after index 5 will be 
 * disabled.
 */

/**
 * Metadata information about Media transitions, used by the player to track 
 * which Media are preforming what transition and when they should start or 
 * stop.
 * @typedef {Object} TransitionEvent
 * @property {!Media} media
 * @property {!Type#PlaybackEvent} type
 * @property {!number} start
 * @property {!number} end
 * @property {?number} play The timestamp in milliseconds when the Media should 
 * begin audio/video playback. Set to null if the Media does not support audio 
 * or video playback, or if type is Type.PlaybackEvent.EXIT.
 */
