/* exported BreezeSchedule */
/* global moment */

class BreezeSchedule {
	/**
	 * Parses Breeze scheduling data and provides useful methods for testing times 
	 * against the schedules and determining schedule status and delay to schedule 
	 * changes.
	 * @param {!Object} scheduleData The scheduling data generated by the server. May
	 * be a JSON Media object or an object from the 'scheduledcommands' array 
	 * inside of a schedule.json file.
	 * @constructor
	 */
	constructor(player, scheduleData) {
		if (!scheduleData.hasOwnProperty('utc')) {
			scheduleData.utc = true;
		}
		this.player = player;
		this.utc = scheduleData.utc;
		this.parseDateRange(scheduleData);
		this.parseDateTime(scheduleData);
		this.parseDayParting(scheduleData);
		this.enabled = this.daterange.enabled || this.datetime.enabled || this.dayparting.enabled;
	}

	/**
	 * A DateRange schedule provides a DateTime range where the item is never 
	 * scheduled prior to the DateRange.start DateTime, and the item is never 
	 * scheduled after the DateRange.end DateTime.
	 * DateRange can be combined with Dayparting to restrict the dates and times 
	 * between the DateRange.start DateTime adnd the DateRange.end DateTime.
	 * @param {!Object} data The scheduling data generated by the server.
	 */
	parseDateRange(data) {
		// MEDIA SCHEDULING:
		// 		1. daterange disabled (startdate and enddate are null)
		// 		2. daterange enabled (startdate and enddate are not null)
		//
		// NETWORK CMD SCHEDULING:
		// 		1. daterange disabled (daystartdatetime and dayenddatetime are null)
		// 		2. daterange enabled (daystartdatetime and dayenddatetime are not null)
		//
		this.daterange = {
			enabled: false,
			start: null,
			end: null
		};
		if (data.startdate != null && data.enddate != null) {
			this.daterange.enabled = true;
			this.daterange.start = data.utc ? this.player.applyTimezone(moment.utc(data.startdate)) : (this.player.timezone ? moment.tz(data.startdate, this.player.timezone) : moment(data.startdate).local());
			this.daterange.end = data.utc ? this.player.applyTimezone(moment.utc(data.enddate)) : (this.player.timezone ? moment.tz(data.enddate, this.player.timezone) : moment(data.enddate).local());
		} else if (data.daystartdatetime != null && data.dayenddatetime != null) {
			this.daterange.enabled = true;
			this.daterange.start = data.utc ? this.player.applyTimezone(moment.utc(data.daystartdatetime)) : (this.player.timezone ? moment.tz(data.daystartdatetime, this.player.timezone) : moment(data.daystartdatetime).local());
			this.daterange.end = data.utc ? this.player.applyTimezone(moment.utc(data.dayenddatetime)) : (this.player.timezone ? moment.tz(data.dayenddatetime, this.player.timezone) : moment(data.dayenddatetime).local());
		}
	}
	
	/**
	 * A DateTime scheduling feature is an instantaneous point in time at which the
	 * item is 'triggered' to activate. If the schedule enables the DateTime 
	 * scheduling feature, then it ignores the DateRange and DayParting scheduling 
	 * features.
	 * @param {!Object} data The scheduling data generated by the server.
	 */
	parseDateTime(data) {
		// NETWORK CMD SCHEDULING:
		// 		5. datetime disabled (datetime is null)
		// 		6. datetime enabled (datetime is not null)
		//
		this.datetime = {
			enabled: false,
			start: null
		};
		if (data.datetime != null) {
			this.datetime.enabled = true;
			this.datetime.start = data.utc ? this.player.applyTimezone(moment.utc(data.datetime)) : (this.player.timezone ? moment.tz(data.datetime, this.player.timezone) : moment(data.datetime).local());
		}
	}

	/**
	 * Dayparting scheduling allows the user to enable or disable the scheduling 
	 * by the day of the week (monday, tuesday, wednesday, etc) and by a time range
	 * (5:00am-6:00pm) on those days.
	 * @param {!Object} data The scheduling data generated by the server.
	 */
	parseDayParting(data) {
		// MEDIA SCHEDULING:
		// 		1. day parting disabled (all days set to false)
		// 		2. day parting enabled (at least one day is true, and daypartstarttime and daypartendtime become valid)
		//
		// NETWORK CMD SCHEDULING:
		// 		1. day parting disabled (all days set to false)
		// 		2. day parting enabled (at least one day is true, daytime becomes valid)
		//
		this.dayparting = {
			days: [data.sunday, data.monday, data.tuesday, data.wednesday, data.thursday, data.friday, data.saturday],
			enabled: false,
			start: null,
			end: null
		};
		this.dayparting.enabled = !this.dayparting.days.every(day => !day);
		if (this.dayparting.enabled) {
			this.dayparting.start = data.daypartstarttime ? data.daypartstarttime : data.daytime;
			this.dayparting.end = data.daypartstarttime ? data.daypartendtime : data.daytime;
		}
	}
	
	/**
	 * @param {!moment} time The given time.
	 * @returns {!boolean} Returns true if this item is scheduled to play at the 
	 * given time, otherwise returns false. If the schedule has the DateTime 
	 * scheduling feature enabled, this function will ALWAYS return false.
	 */
	isScheduledAt(time) {
		if (!this.enabled) {
			return true;
		}
		let scheduled = true;
		if (this.daterange.enabled) {
			scheduled = time.isBetween(this.daterange.start, this.daterange.end);
		}
		if (scheduled && this.dayparting.enabled) {
			scheduled = this.dayparting.days[time.day()];
			if (scheduled) {
				const startTime = this.formatDayparting(time, this.dayparting.start);
				const endTime = this.formatDayparting(time, this.dayparting.end);
				if (startTime.isBefore(endTime)) {
					scheduled = time.isBetween(startTime, endTime);
				} else {
					scheduled = time.isBetween(endTime.clone().startOf('day'), endTime) || time.isBetween(startTime, startTime.clone().endOf('day'));
				}
			}
		}
		return scheduled;
	}

	/**
	 * Utility function for calling {@link BreezeSchedule#isScheduledAt} using the 
	 * current timestamp.
	 * @returns {!boolean}
	 */
	isScheduledNow() {
		return this.isScheduledAt(this.player.currentTime);
	}

	/**
	 * @param {!moment} time The given time.
	 * @returns {!number} Returns the amount of milliseconds from the given time to 
	 * the next schedule change event. A schedule change event is whenever the 
	 * schedule item changes from being scheduled to not being scheduled, or from 
	 * not being scheduled to being scheduled. This function will return Infinity 
	 * for cases where there are no schedule change events after the given time, 
	 * e.g. if time comes after a schedules {@link BreezeSchedule#datetime#start},
	 * or if the schedule does not have any scheduling features enabled, etc.
	 */
	delayFrom(time) {
		if (!this.enabled) {
			return Infinity;
		}
		if (this.daterange.enabled) {
			if (time.isBefore(this.daterange.start)) {
				return this.daterange.start.diff(time);
			} else if (time.isBefore(this.daterange.end)) {
				if (this.dayparting.enabled) {
					const dayPartingDelay = this.getDelayToNextDaypartingChangeFrom(time);
					const dateRangeDelay = this.daterange.end.diff(time);
					return Math.min(dayPartingDelay, dateRangeDelay);
				} else {
					return this.daterange.end.diff(time);
				}
			} else {
				return Infinity;
			}
		} else if (this.dayparting.enabled) {
			return this.getDelayToNextDaypartingChangeFrom(time);
		} else if (this.datetime.enabled) {
			if (time.isBefore(this.datetime.start)) {
				return this.datetime.start.diff(time);
			} else {
				return Infinity;
			}
		} else {
			console.log('BreezeSchedule.delayFrom() called on schedule that appears to be enabled but has no scheduling features enabled.');
			return Infinity;
		}
	}

	/**
	 * Utility function for calling {@link BreezeSchedule#delayFrom} using the 
	 * current timestamp.
	 * @returns {!number}
	 */
	delayFromNow() {
		return this.delayFrom(this.player.currentTime);
	}

	/**
	 * @param {!moment} time The given time.
	 * @returns {!number} Returns the delay in milliseconds from the given time to  the next dayparting schedule change.
	 */
	getDelayToNextDaypartingChangeFrom(time) {
		if (!this.dayparting.days.some(value => value)) {
			return Infinity;
		}
		let day = time.clone();
		while (true) {
			if (this.dayparting.days[day.day()]) {
				const startTime = this.formatDayparting(day, this.dayparting.start);
				const endTime = this.formatDayparting(day, this.dayparting.end);
				const min = moment.min(startTime, endTime);
				const max = moment.max(startTime, endTime);
				if (time.isBefore(min)) {
					return min.diff(time);
				} else if (time.isBefore(max)) {
					return max.diff(time);
				}
			}
			day.add(1, 'day');
		}
	}

	/**
	 * @param {!moment} day The day to format the dayparting onto. This function calls day.startOf('day') and then adds the dayparting hour and minutes.
	 * @param {string} dayparting The dayparting information used to create the dayparting timestamp.
	 * @returns {!moment} Returns a moment timestamp that represents the dayparting time at the given day.
	 */
	formatDayparting(day, dayparting) {
		return this.utc ?
			moment.utc(`${day.clone().startOf('day').format('YYYY-MM-DD')} ${dayparting}`, 'YYYY-MM-DD HH:mm:ss') :
			(this.player.timezone ? moment.tz(`${day.clone().startOf('day').format('YYYY-MM-DD')} ${dayparting}`, 'YYYY-MM-DD HH:mm:ss', this.player.timezone) : moment(`${day.clone().startOf('day').format('YYYY-MM-DD')} ${dayparting}`, 'YYYY-MM-DD HH:mm:ss').local());
	}
}
