/* exported PlayerSchedule */
/* global BreezeLib, ScheduleLoader, SchedulePreloader, LoadingPlaylist, TimerMap, Utility, Container, LoadingScreen, moment */

class PlayerSchedule {
	/**
	 * @param {!Player} player
	 * @param {!number} startTime
	 * @param {!HTMLElement} $playlists
	 */
	constructor(player, startTime, $playlists) {
		this.player = player;
		this.startTime = startTime;
		this.$playlists = $playlists;
		this.timers = new TimerMap();
		this.schedule = null;
		this.canReload = true;
		this.missedReloads = 0;
		this.loadingScreen = null;
		this.loadingPlaylists = [];
		this.playlists = [];
		this.currentPlaylistIndex = 0;
		this.lastReload = this.player.currentTime;
		this.currentBlock = null;
		this.shouldUpdate = false;
		this.blockStart = null;
		this.preloadScheduleContentReject = null;
		this.schedulePreloader = null;

		this.lastProgress = null;
	}

	/**
	 */
	initialize() {
	}

	destroy() {
		this.timers.clear();
		if (this.playlists) {
			for (const playlist of this.playlists) {
				playlist.destroy();
			}
			this.playlists = null;
		}
		this.resetLoadingState();
	}

	/**
	 * @param {!string} name
	 * @returns {!Promise}
	 */
	async loadPreview(name) {
		const loadingPlaylist = new LoadingPlaylist(name, false, this.player, this.$playlists, null);
		this.loadingPlaylists.push(loadingPlaylist);
		await loadingPlaylist.preloadMetadata();
		await this.createLoadingScreen(loadingPlaylist.size.width);
		await loadingPlaylist.load(false);
		this.finishLoading();
	}

	/**
	 * @param {!BreezeFonts} fonts
	 * @param {!boolean} [loadFromCache]
	 * @returns {!Promise}
	 */
	async load(fonts, loadFromCache = false) {
		let loader = new ScheduleLoader(this.player);
		if (!loadFromCache) {
			if (this.preloadScheduleContentReject) {
				this.preloadScheduleContentReject(new Error('Aborted due to schedule update.'));
			}
			if (this.schedulePreloader) {
				await this.schedulePreloader.abort();
				this.schedulePreloader = null;
			}
			
			// Clear metadata cache to force a fresh download in case the player is closed before it is able to cache the new schedule.
			await this.cleanCache();
		}
		try {
			await loader.load(this.player.container.features.caching, loadFromCache);
			this.schedule = new BreezeLib.Schedule(loader.data, this.player.timezone);
			this.player.container.contenttree = this.schedule.contenttree;
			this.player.container.metadata = this.schedule.metadata;
			this.ScheduleLoading();
			await fonts.downloadCustomFonts(this.schedule.customFonts);
			this.ScheduleLoaded(this.schedule);
			this.player.ScheduleLoaded(this.schedule);
		} catch (err) {
			this.shouldUpdate = true;
			console.warn(`Download and update schedule failed, will retry later.  ${err}`);
		}
	}

	/*
	 * @returns {!string}
	 */
	getActiveScheduleName() {
		if (this.schedule && this.schedule.name) {
			return this.schedule.name;
		}
		return 'None';
	}

	/**
	 * @returns {!Promise}
	 */
	async reloadBlock() {
		try {
			if (!this.canReload) {
				this.missedReloads++;
				return;
			}
			this.missedReloads = 0;

			this.lastReload = this.player.currentTime;
			const scheduleBlock = this.schedule.getBlockAt(this.lastReload);
			if (scheduleBlock === null) {
				// Blank schedule detected - remove all playlists
				for (const playlist of this.playlists) {
					playlist.destroy();
				}
				this.playlists = [];
				this.resetLoadingState();
				if (this.player.container.features.led) {
					this.player.container.disableFrontLed();
				}
				this.player.backgroundMusicService.ActivePlaylistChanged(null);
				return;
			}

			if (this.currentBlock === null) {
				this.currentBlock = scheduleBlock;
				// Calculate how long this schedule block will play for in seconds
				const scheduledRunLength = this.calcBlockScheduledRunLength(this.lastReload);
				await this.loadAllScheduleBlockContent(scheduleBlock, scheduledRunLength);
			} else {
				const idChanged = this.currentBlock.id !== scheduleBlock.id;
				const isDefault = this.currentBlock.type === 'Fail Safe';
				const isContinuous = isDefault || this.currentBlock.schedule.schedule.continuous;
				this.currentBlock = scheduleBlock;
				if (this.shouldUpdate || idChanged || (!idChanged && !isContinuous)
				) {
					const scheduledRunLength = this.calcBlockScheduledRunLength(this.lastReload);
					await this.loadAllScheduleBlockContent(scheduleBlock, scheduledRunLength);
				}
			}
			this.shouldUpdate = false;
			console.log('completed: loadAllScheduleBlockContent');

			const delay = this.calcTimeToBlockEnd(this.lastReload);
			if (delay) {
				this.timers.on('ReloadBlock', delay, this.reloadBlock.bind(this));
			}
		} catch (err) {
			console.log(`Unexpected error thrown while reloading block: ${err}`);
		}
	}

	/**
	 * @param {!moment} from
	 * @returns {?number}
	 */
	calcBlockScheduledRunLength(from) {
		const blockStart = this.schedule.getBlockStartAt(from);
		const blockEnd = this.schedule.getBlockEndAt(from);
		if (blockStart === null || blockEnd === null) {
			return null;
		}
		return blockEnd.diff(blockStart);
	}

	/**
	 * @param {!moment} from
	 * @returns {?number}
	 */
	calcTimeToBlockEnd(from) {
		const blockEnd = this.schedule.getBlockEndAt(from);
		if (blockEnd === null) {
			return null;
		}
		return blockEnd.diff(from);
	}

	resetLoadingState() {
		if (this.loadingScreen) {
			this.loadingScreen.destroy();
			this.loadingScreen = null;
		}
		this.timers.off('LoadingTick');
		for (const lp of this.loadingPlaylists) {
			lp.destroy();
		}
		this.loadingPlaylists = [];
	}

	/**
	 * @param {!number} width
	 */
	async createLoadingScreen(width) {
		this.loadingScreen = new LoadingScreen(width, 110);
		this.loadingScreen.show();
		await this.loadingTick();
	}

	async loadingTick() {
		this.timers.on('LoadingTick', 100, this.loadingTick.bind(this));
		// Calculate progress of all loading playlists
		const loadingProgressPromises = this.loadingPlaylists.filter(lp => lp.isLoaded).map(lp => lp.playlist).map(p => p.getLoadingProgress());
		const totalProgress = (await Promise.all(loadingProgressPromises)).reduce((sum, x) => sum + x, 0);
		const totalPlaylists = this.loadingPlaylists.length;
		if(totalPlaylists === 0){
			console.info("There aren't any playlists to load.")
		}
		const progress = totalPlaylists > 0 ? totalProgress / totalPlaylists : 100;
		// Debug information
		if (this.lastProgress != progress) {
			this.lastProgress = progress;
			for (const lp of this.loadingPlaylists) {
				console.log(`Playlist: '${lp.name}', Loaded: ${JSON.stringify(lp.isLoaded)} Media Load Progress: ${lp.isLoaded ? (await lp.playlist.getLoadingProgress()).toFixed(1) : '0'}%`);
			}
			console.log(`Progress: ${progress}%`);
		}
		
		this.loadingScreen.progress(progress);
	}

	finishLoading() {
		if (this.loadingScreen) {
			this.loadingScreen.finish(() => {
				this.playLoadedPlaylists();
			});
		}
		this.timers.off('LoadingTick');
	}

	/**
	 */
	ScheduleLoading() {
		this.timers.off('ReloadBlock');
	}

	/**
	 * @param {!Schedule} schedule
	 */
	ScheduleLoaded(schedule) {
		this.reloadBlock();
	}

	TimeSync() {
		this.syncBlockStart(this.player.currentTime);
	}

	/**
	 * @param {!moment} now
	 */
	syncBlockStart(now) {
		if (this.schedule) {
			console.log(`syncBlockStart: ${now.unix()}, ${now.format()}`);
			this.blockStart = this.schedule.getBlockStartAt(now);
			console.log(`this.blockStart: ${this.blockStart.unix()}, ${this.blockStart.format()}`);
		} else {
			console.log('syncBlockStart skipped. Schedule not loaded');
		}
	}

	playNextPlaylist() {
		console.log('playNextPlaylist');
		const currentPlaylist = this.playlists[this.currentPlaylistIndex];
		this.currentPlaylistIndex = ++this.currentPlaylistIndex % this.currentBlock.playlists.length;
		const nextPlaylist = this.playlists[this.currentPlaylistIndex];

		const changed = currentPlaylist.id !== nextPlaylist.id;
		if (!changed) {
			currentPlaylist.stop();
		}
		nextPlaylist.play(0);
		if (changed) {
			nextPlaylist.show();
			currentPlaylist.hide();
			currentPlaylist.stop();
		}
	}

	async loadAllScheduleBlockContent(scheduleBlock, scheduledRunLength) {
		console.log('resetLoadingState');
		this.resetLoadingState();
		console.log('start: createLoadingScreen');
		await this.createLoadingScreen(this.player.settings.data.usableSize.width);
		console.log('end: createLoadingScreen');
		// Gets an array of all playlists without duplicates in a schedule block
		const playlists = Array.from(new Set(scheduleBlock.playlists));
		await Promise.all(playlists.map(async p => {
			const loadFromCache = !this.shouldUpdate;
			try {
				const loadingPlaylist = new LoadingPlaylist(p.name, loadFromCache, this.player, this.$playlists);
				this.loadingPlaylists.push(loadingPlaylist);
				console.log('start: await loadingPlaylist.load();');
				await loadingPlaylist.load();
				if (scheduledRunLength == scheduleBlock.totalDwell) {
					// This only applies to zones that have a runlength equal to the playlist runlength which is checked later
					loadingPlaylist.playlist.disableLoopOnZones = true;
				}
				console.log('end: await loadingPlaylist.load();');
			} catch (err) {
				console.log(`Skipping loading playlist '${p.name}' due to error: ${err.stack || err}`);
			}
		}));
		console.log('finished all loadingPlaylists');
		const loadedPlaylists = this.loadingPlaylists.filter(lp => lp.isLoaded);
		if (loadedPlaylists.length > 1) {
			for (const p of loadedPlaylists.map(lp => lp.playlist)) {
				p.onPlayFinish = this.playNextPlaylist.bind(this);
			}
		}
		console.log('start: finishLoading');
		this.finishLoading();
		console.log('end: finishLoading');
	}

	/**
	 * @param {!number} id
	 * @returns {?Playlist}
	 */
	getLoadedPlaylistById(id) {
		const loadingPlaylist = this.loadingPlaylists.filter(lp => lp.isLoaded).map(lp => lp.playlist).find(p => p.id === id);
		return loadingPlaylist == undefined ? null : loadingPlaylist;
	}

	playLoadedPlaylists() {
		console.log('playLoadedPlaylists');
		const loadedPlaylists = this.loadingPlaylists.filter(lp => lp.isLoaded).map(lp => lp.playlist);
		loadedPlaylists.forEach(p => p.hide());
		for (const playlist of this.playlists) {
			playlist.destroy();
		}
		if (this.currentBlock) {
			this.playlists = this.currentBlock.playlists.map(playlist => this.getLoadedPlaylistById(playlist.id)).filter(playlist => playlist != null);
		} else {
			this.playlists = loadedPlaylists.splice(0);
		}
		this.loadingPlaylists = [];

		this.resetLoadingState();
		this.currentPlaylistIndex = 0;
		console.log('Playlists Load Complete');
		for (const playlist of this.playlists) {
			playlist.TimeSync();
		}
		if (!this.player.isPreview) {
			const now = this.player.currentTime;
			this.syncBlockStart(now);
			this.startTime = now.diff(this.blockStart);
		} else {
			this.startTime *= 1000;
		}
		console.log('playlist length: ', this.playlists.length);
		if (this.playlists.length > 0) {
			this.playlists[0].show();
			this.playlists[0].play(this.startTime);
		}
		if (this.player.isPreview) {
			this.startTime = 0;
		} else {
			// Begin preloading the remainder of the schedule block content, if any exists
			this.preloadScheduleContent();
		}
	}

	play() {
		if (this.playlists[this.currentPlaylistIndex]) {
			this.playlists[this.currentPlaylistIndex].show();
			this.playlists[this.currentPlaylistIndex].play();
		}
	}

	stop() {
		const playlist = this.playlists[this.currentPlaylistIndex];
		if (playlist && playlist !== undefined) {
			playlist.hide();
			playlist.stop();
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async preloadScheduleContent() {
		try {
			await new Promise(async (resolve, reject) => {
				this.preloadScheduleContentReject = reject;
				let toPreload = [];
				if (this.currentBlock) {
					this.schedule.blocks.filter(block => block.id !== this.currentBlock.id).forEach(block => toPreload.push(...block.playlists));
				}
				const mediaIds = new Set();
				this.playlists.map(p => p.getAllMedia()).map(media => media.id).forEach(id => mediaIds.add(id));
				this.schedulePreloader = new SchedulePreloader(this.schedule, this.player, toPreload, mediaIds);
				await this.schedulePreloader.preload();
				this.schedulePreloader = null;
				resolve();
			});
		} catch (err) {
			console.log(`Unexpected error thrown while preloading schedule content: ${err}`);
		}
	}


	/**
	 * @returns {!Promise}
	 */
	async cleanCache() {
		if (this.player.container.features.caching) {
			await ScheduleLoader.cleanCache();
			await Playlist.cleanCache();
			await Media.cleanCache();
		}

		// TODO: Somehow media data (static webpage images and widget data) should be cleared right before new playlists begin loading while stopping the older widgets from saving new files
		// to the cache. This is needed because if a playlist was modified to have a static webpage point to a new url, and then an update player was sent the old static webpage may save the old image
		// into the cache and the newly loading widget will see it already exists and then load it instead of downloading a fresh image.
	}
}
