/* exported Schedule */
/* global FileInfo, Utility, moment */

class Schedule {
	/**
	 * @param {!string} hostUrl
	 * @param {!string} account
	 * @param {!string} key
	 */
	constructor(hostUrl, account, key) {
		this.url = `${hostUrl}schedule/${account}/${key}.json`;
		this.file = new FileInfo('schedules/latest.schedule', false);

		this.customFontsExtra = [];
		this.networkCommands = null;
		this.networkCommandSchedules = null;

		this.scheduleBlocks = {};
		this.failSafeBlock = null;
		this.isBlank = false;
		this.name = null;

		/**
		 * List of all playlists in the schedule, there may be
		 * duplicated entries.
		 * @type {Object[]}
		 */
		this.playlists = null;
	}

	/**
	 * @param {!boolean} cacheSupport
	 * @param {!boolean} loadFromCache
	 * @returns {!Promise}
	 */
	async load(cacheSupport, loadFromCache) {
		const url = `${this.url}?d=${Date.now()}`;
		const scheduleData = await Utility.load(
			cacheSupport,
			loadFromCache,
			this.file,
			url,
			'json'
		);
		if (Schedule.validate(scheduleData)) {
			this.initialize(scheduleData);
		}
	}

	/**
	 * @param {!Object} data
	 */
	initialize(data) {
		const extraFonts = data.customfontsextra;
		if (extraFonts === null || extraFonts === undefined) {
			this.customFontsExtra = [];
		} else {
			this.customFontsExtra = extraFonts;
		}
		if (data.networkcommands !== undefined) {
			this.networkCommands = data.networkcommands;
		}
		if (data.networkcommandschedules !== undefined) {
			this.networkCommandSchedules = data.networkcommandschedules;
		}

		if (data.id == -1) {
			this.isBlank = true;
		} else {
			this.name = data.name;
			this.playlists =
				Schedule.getAllPlaylists(data.scheduleBlocks);
			this.initializeBlocks(data);
		}
	}

	/**
	 * @param {!Object} data
	 */
	initializeBlocks(data) {
		for (const scheduleBlock of data.scheduleBlocks) {
			if (scheduleBlock.type === 'Fail Safe') {
				this.failSafeBlock = scheduleBlock;
				continue;
			}
			this.parseDateTimes(scheduleBlock);
			this.splitToDayBlocks(scheduleBlock);
		}
	}

	/**
	 * Any schedule block that spans more than ONE day (as in, it
	 * crosses into the next or previous day), should be split up into
	 * day-sized blocks with the same block id. This greatly simplifies
	 * working with the schedule blocks.
	 * @param {!Object} scheduleBlock
	 */
	splitToDayBlocks(scheduleBlock) {
		const days = [
			'sunday',
			'monday',
			'tuesday',
			'wednesday',
			'thursday',
			'friday',
			'saturday'
		];
		scheduleBlock.continuous = true;
		for (let itr = 0; itr < days.length; ++itr) {
			if (scheduleBlock.scheduleDate[days[itr]]) {
				scheduleBlock.continuous = false;
				break;
			}
		}

		const timespan = 1 + scheduleBlock.endDateTime.diff(
			scheduleBlock.startDateTime, 'days'
		);
		for (let itr = 0; itr < timespan; ++itr) {
			const startDateTime = moment(scheduleBlock.startDateTime)
				.add(itr, 'days');
			const endDateTime = moment(scheduleBlock.startDateTime)
				.add(itr, 'days').startOf('day')
				.hours(scheduleBlock.endDateTime.hours())
				.minutes(scheduleBlock.endDateTime.minutes());

			if (
				!scheduleBlock.continuous &&
				scheduleBlock.scheduleDate[days[startDateTime.day()]]
			) {
				this.pushDayBlock(startDateTime, {
					id: scheduleBlock.id,
					priority: scheduleBlock.priority,
					playlists: scheduleBlock.playlists,
					continuous: scheduleBlock.continuous,
					startDateTime: startDateTime,
					endDateTime: endDateTime
				});
			} else if (scheduleBlock.continuous) {
				const priorDay = moment(startDateTime)
					.subtract(1, 'days');
				const scheduledToPlayPreviousDay =
					scheduleBlock.startDateTime
						.isSameOrBefore(priorDay, 'day');
				if (scheduledToPlayPreviousDay) {
					startDateTime.hour(0).minute(0).second(0);
				}
				const nextDay = moment(endDateTime).add(1, 'days');
				const scheduledToPlayNextDay = scheduleBlock
					.endDateTime
					.isSameOrAfter(nextDay, 'day');
				if (scheduledToPlayNextDay) {
					endDateTime
						.hour(23)
						.minute(59)
						.second(59)
						.millisecond(999);
				}
				this.pushDayBlock(startDateTime, {
					id: scheduleBlock.id,
					priority: scheduleBlock.priority,
					playlists: scheduleBlock.playlists,
					continuous: scheduleBlock.continuous,
					startDateTime: startDateTime,
					endDateTime: endDateTime
				});
			}
		}
	}

	/**
	 * @param {!moment} dateTime
	 * @returns {?Object}
	 */
	getDayBlocks(dateTime) {
		const year = dateTime.year().toString();
		const month = dateTime.month().toString();
		const date = dateTime.date().toString();
		this.scheduleBlocks[year] = this.scheduleBlocks[year] || {};
		this.scheduleBlocks[year][month] = 
			this.scheduleBlocks[year][month] || {};
		this.scheduleBlocks[year][month][date] =
			this.scheduleBlocks[year][month][date] || [];
		return this.scheduleBlocks[year][month][date];
	}

	/**
	 * @param {!moment} dateTime
	 * @param {?Object} dayBlock
	 */
	pushDayBlock(dateTime, dayBlock) {
		const dayBlocks = this.getDayBlocks(dateTime);
		let insertIndex = 0;
		for (let itr = 0; itr < dayBlocks.length; ++itr) {
			if (dayBlock.priority < dayBlocks[itr].priority) {
				break;
			}
			++insertIndex;
		}
		dayBlocks.splice(insertIndex, 0, dayBlock);
	}

	/**
	 * @param {!Object} scheduleBlock
	 */
	parseDateTimes(scheduleBlock) {
		scheduleBlock.startDateTime =
			Schedule.parseDateTime(
				scheduleBlock.scheduleDate.startDate,
				scheduleBlock.scheduleDate.startTime
			);
		scheduleBlock.endDateTime =
			Schedule.parseDateTime(
				scheduleBlock.scheduleDate.endDate,
				scheduleBlock.scheduleDate.endTime
			);
	}

	/**
	 * @param {!moment} time
	 * @returns {?Object}
	 */
	getScheduleBlockAt(time) {
		let scheduleBlock = null;
		if (!this.isBlank) {
			const dayBlocks = this.getDayBlocks(time);
			for (let itr = 0; itr < dayBlocks.length; ++itr) {
				const dayBlock = dayBlocks[itr];
				if (
					time.isSameOrAfter(dayBlock.startDateTime) &&
					time.isBefore(dayBlock.endDateTime)
				) {
					scheduleBlock = dayBlock;
					break;
				}
			}
			if (scheduleBlock === null) {
				scheduleBlock = this.failSafeBlock;
			}
		}
		return scheduleBlock;
	}

	/**
	 * @param {!Object} currentScheduleBlock
	 * @param {!moment} time
	 * @returns {?Object}
	 */
	getNextScheduleBlockFrom(currentScheduleBlock, time) {
		// If the current schedule block is null there is no failsafe
		// block, which means there is no schedule, and thus no next
		// block.
		if (currentScheduleBlock === null) {
			return null;
		}

		let nextBlock = null;
		const currentDayBlocks = this.getDayBlocks(time);
		const playingFailSafe =
			currentScheduleBlock.type === 'Fail Safe';
		for (let itr = 0; itr < currentDayBlocks.length; ++itr) {
			const scheduleBlock = currentDayBlocks[itr];
			let elgible = false;
			if (playingFailSafe) {
				if (scheduleBlock.startDateTime.isAfter(time)) {
					elgible = true;
				}
			} else {
				if (
					scheduleBlock.priority < currentScheduleBlock.priority &&
					scheduleBlock.startDateTime.isAfter(time) &&
					scheduleBlock.startDateTime.isBefore(currentScheduleBlock.endDateTime)
				) {
					elgible = true;
				}
			}
			if (elgible) {
				if (nextBlock !== null) {
					if (scheduleBlock.startDateTime.isBefore(nextBlock.startDateTime)) {
						nextBlock = scheduleBlock;
					} else if (
						scheduleBlock.startDateTime.isSame(nextBlock.startDateTime) &&
						scheduleBlock.priority < nextBlock.priority
					) {
						nextBlock = scheduleBlock;
					}
				} else {
					nextBlock = scheduleBlock;
				}
			}
		}
		if (!playingFailSafe && nextBlock === null) {
			nextBlock = this.getScheduleBlockAt(
				currentScheduleBlock.endDateTime
			);
		}
		return nextBlock;
	}

	/**
	 * @param {!Object} currentScheduleBlock
	 * @param {!moment} time
	 * @returns {!number}
	 */
	getDelayToNextScheduleBlockFrom(currentScheduleBlock, time) {
		if (currentScheduleBlock === null) {
			return Number.NaN;
		}
		const nextBlock = this.getNextScheduleBlockFrom(
			currentScheduleBlock,
			time
		);
		let delay;
		if (nextBlock === null) {
			delay = moment().endOf('day').valueOf() - time.valueOf();
		} else if (
			nextBlock.type !== 'Fail Safe' &&
			nextBlock.startDateTime.isAfter(time)
		) {
			delay = nextBlock.startDateTime.valueOf() - time.valueOf();
		} else {
			delay = currentScheduleBlock
				.endDateTime.valueOf() - time.valueOf();
		}
		return delay;
	}

	/**
	 * Gets the set of all playlist names from a given array of
	 * playlists.
	 * @param {!Object[]} playlists
	 * @returns {!Set<!string>}
	 */
	static getPlaylistNames(playlists) {
		const set = new Set();
		playlists
			.map(playlist => playlist.name)
			.forEach(name => set.add(name));
		return set;
	}

	/**
	 * @param {!moment} currentTime
	 * @returns {?ScheduleBlock}
	 */
	getCurrentScheduleBlock(currentTime) {
		if (this.isBlank) {
			return null;
		}
		return this.getScheduleBlockAt(currentTime);
	}

	/**
	 * @returns {!moment}
	 */
	static getCurrentTime() {
		return moment();
	}
	
	/**
	 * @returns {!Promise}
	 */
	static async cleanCache() {
		const dir = new FileInfo('schedules', true);
		await dir.delete();
	}

	/**
	 * @param {!string} date mm/dd/yyyy
	 * @param {!string} time hh:mmaa
	 * @returns {!moment}
	 */
	static parseDateTime(date, time) {
		return moment(`${date}T${time}`);
	}

	/**
	 * Get the list of all playlists in a schedule.
	 * @param {!Object[]} scheduleBlocks
	 * @param {!Object[]} scheduleBlocks.playlists
	 * @returns {!string[]} Returns an array of all playlists in
	 * scheduleBlocks.
	 * There may be duplicated entries.
	 */
	static getAllPlaylists(scheduleBlocks) {
		return scheduleBlocks.reduce((playlists, scheduleBlock) =>
			playlists.concat(scheduleBlock.playlists),
		[]);
	}

	/**
	 * @param {?Object} data
	 * @returns {!boolean}
	 */
	static validate(data) {
		return data !== undefined &&
			data !== null &&
			data.id !== undefined;
	}
}
