/* exported SchedulePreloader */
/* global Utility, Timer */

class SchedulePreloader {
	/**
	 * @param {?Schedule} schedule
	 * @param {!Player} player
	 * @param {!Object[]} playlistsToPreload
	 * @param {!Set<!number>} loadedMediaIds
	 */
	constructor(
		schedule,
		player,
		playlistsToPreload,
		loadedMediaIds
	) {
		this.player = player;
		this.name = schedule == null ? null : schedule.name;
		this.aborted = false;
		this.failures = new Map();
		this.downloads = new Map();

		this.playlists = Array.from(new Set(playlistsToPreload));
		this.ids = new Set(loadedMediaIds);
	}

	/**
	 * @returns {!Promise}
	 */
	async abort() {
		this.aborted = true;
		for (const [id, md] of this.downloads) {
			this.player.downloader.abort(md.file);
			this.downloads.delete(id);
		}
	}

	/**
	 * @returns {!Promise}
	 */
	async preload() {
		try {
			const datas = await this.loadPlaylists();
			for (const data of datas) {
				await this.loadFonts(data);
			}
			const preloading = [];
			for (const data of datas) {
				const media = data.getAllMedia();
				for (const md of media) {
					if (!this.ids.has(md.id)) {
						this.ids.add(md.id);
						preloading.push(md);
					}
				}
			}
			while (!this.aborted && preloading.length > 0) {
				const md = preloading.pop();
				try {
					this.downloads.set(md.id, md);
					console.log('start preload cache: ', md.name);
					await md.cache();
					console.log('end preload cache: ', md.name);
				} catch (err) {
					if (this.aborted) {
						return;
					}
					preloading.push(md);
					let count = 0;
					if (this.failures.has(md.id)) {
						count = this.failures.get(md.id);
						await Timer.sleep(30 * 1000);
					}
					this.failures.set(md.id, ++count);
					console.log(`Unexpected error while preloading media (id => ${md.id}, has failed ${count} times): ${err}.  Will retry.`);
				}
				this.downloads.delete(md.id);
			}
		} catch (err) {
			console.log(`Unexpected error thrown while preloading schedule '${this.name}': ${err}`);
		}
	}

	/**
	 * @returns {!Promise<!playlist[]>}
	 */
	async loadPlaylists() {
		const datas = [];
		for (const playlist of this.playlists) {
			try {
				const data = await Playlist.get(this.player, playlist.name, true);
				datas.push(new Playlist(data, this.player, this.$playlists));
			} catch (err) {
				console.log(`Skipping preload of playlist data ${playlist.name} due to error: ${err}`);
			}
		}
		return datas;
	}

	/**
	 * @param {!Playlist} playlist
	 * @returns {!Promise}
	 */
	async loadFonts(playlist) {
		try {
			const fonts = playlist.getUsedFonts(this.player.container.widgetsSchema);
			await this.player.container.fonts.using(fonts);
		} catch (err) {
			console.log(`Failed to preload fonts associated with playlist: '${playlist.name}' (id => ${playlist.id}): ${err}`);
		}
	}
}
