class Timer {
	/**
	 * @param {!number} delay In milliseconds.
	 * @returns {!Promise}
	 */
	static sleep(delay) {
		return new Promise(resolve => setTimeout(resolve, delay));
	}

	/**
	 * Waits until predicate(action()) returns true, or until a maximum number
	 * of calls to action().  Returns the result from the last call to action.
	 * @param {!Function} action
	 * @param {!Function} predicate
	 * @param {!number} [delay=5000] Delay to wait before each try.
	 * @param {!number} [attempts=3] Number of tries before giving up.
	 * @returns {!Promise<?Object>}
	 */
	static async wait(action, predicate, delay = 5000, attempts = 3) {
		let tries = 0;
		let result;
		do {
			await Timer.sleep(delay);
			result = await action();
			if (++tries >= attempts) {
				break;
			}
		} while (!predicate(result));
		return result;
	}

	/**
	 * Waits until predicate(action()) returns true, or until a maximum number
	 * of calls to action().  Returns the result from the last call to action.
	 * Pads an escalating delay between calls to action().
	 * @param {!Function} action
	 * @param {!Function} predicate
	 * @param {!number} [initialDelay]
	 * @param {!number} [maxDelay]
	 * @param {!number} [maxAttempts]
	 * @returns {!Promise<?Object>}
	 */
	static async escalatingWait(action, predicate, initialDelay = 10 * 1000, maxDelay = 10 * 60 * 1000, maxAttempts = 30) {
		let delay = initialDelay;
		let delayPadding = initialDelay;
		let attempts = 0;
		let result = await action();
		while (!predicate(result)) {
			await Timer.sleep(delay);
			delay = Math.min(maxDelay, delay += delayPadding);
			delayPadding = Math.min(maxDelay, delayPadding * 2);
			result = await action();
			if (++attempts >= maxAttempts) {
				break;
			}
		}
		return result;
	}
}
