/* exported TimerMap */
/* global Minibus */

const TimerMap = (() => {
	class Timer {
		/**
		 * @param {!Minibus} bus
		 * @param {!string} label
		 * @param {!number} ms
		 * @param {!function} callback
		 */
		constructor(bus, label, ms, callback) {
			this.label = label;
			this.bus = bus;
			this.callback = callback;
			// Naively cap the delay
			let totalMS = ms;
			const callWithLargeDelay = () => {
				if (totalMS > 0x7FFFFFFF) {
					totalMS -= 0x7FFFFFFF;
					this.id = window.setTimeout(() => {
						callWithLargeDelay();
					}, 0x7FFFFFFF);
				} else {
					this.id = window.setTimeout(() => {
						this.expire();
						callback();
					}, totalMS);
				}
			};
			callWithLargeDelay();
		}

		cancel() {
			window.clearTimeout(this.id);
			this.expire();
		}

		expire() {
			this.bus.emit('TimerExpired', this.label);
		}
	}

	return class TimerMap {
		constructor() {
			this.timers = new Map();
			this.bus = Minibus.create();
			this.bus.on('TimerExpired', this.onExpired.bind(this));
		}

		/**
		 * @param {!string} label
		 */
		onExpired(label) {
			this.timers.delete(label);
		}

		/**
		 * @param {!string} label
		 * @param {!number} ms
		 * @param {!function} callback
		 */
		on(label, ms, callback) {
			// Disallow rebinding of a timer without explicit removal
			// (timers are automatically removed when they expire.)
			if (this.timers.has(label)) {
				throw new Error(`Callback is already set for timer '${label}'!`);
			}
			const timer = new Timer(this.bus, label, ms, callback);
			this.timers.set(label, timer);
		}

		/**
		 * @param {!string} label
		 */
		off(label) {
			const timer = this.timers.get(label);
			if (timer !== undefined) {
				timer.cancel();
			}
		}

		clear() {
			const timers = this.timers.values();
			for (const timer of timers) {
				timer.cancel();
			}
		}
	};
})();
