/* exported TimeSync */
/* global Minibus */

class TimeSync {
	constructor(callback = () => {}) {
		/**
		 * 
		 * @type {!number}
		 */
		this.diff = 0;
		/**
		 * Time of last clock sync request
		 * @type {!number}
		 */
		this.lastSyncAt = 0;
		
		/**
		 * @type {!number}
		 */
		this.timer = null;
		
		
		this.callback = callback;
	}

	destroy() {
		clearTimeout(this.timer);
	}

	/**
	 * Attempts to sync the local and server clocks.
	 * @param {!string} hostUrl
	 */
	async sync(hostUrl) {
		this.lastSyncAt = performance.now();
		const samples = 20;
		const diffs = await TimeSync.sync(hostUrl, samples);
		if (diffs.length > 0) {
			// Remove outliers by sorting and then removing the start and end diffs. (Could use statistics here instead to remove outside of a standard deviation).
			diffs.sort();
			while (diffs.length > samples - 4) {
				diffs.shift();
				diffs.pop();
			}
			let totalDiff = diffs.reduce((sum, sample) => sum + sample, 0);
			// Average the new sample(s) into the diff
			//totalDiff = totalDiff + (this.diff * this.samples);
			//this.diff = totalDiff / (this.samples + samples);
			// Changed to just take the average
			this.diff = totalDiff / diffs.length;
			this.callback(this.diff);
		}
		clearTimeout(this.timer);
		this.timer = setTimeout(this.sync.bind(this, hostUrl), TimeSync.TIME_SYNC_DELAY);
	}

	/**
	 * Attempts to calculate an approximation of the difference between the server and client clocks.
	 * Returns null if unable to connect to server.
	 * @param {!string} hostUrl
	 * @param {!number} [samples]
	 * @returns {!number[]}
	 */
	static async sync(hostUrl, samples) {
		const MAX_FAILURES = 3;
		let failures = 0;
		const diffs = [];
		while (diffs.length < samples) {
			if (failures > MAX_FAILURES) {
				break;
			}
			const sample = await TimeSync.sample(`${hostUrl}get/time.php`);
			if (sample === null) {
				failures++;
				continue;
			}
			diffs.push(sample);
		}
		return diffs;
	}

	/**
	 * Samples the approximate difference between local and server time, used for players that do not support modern performance timing.
	 * @param {!string} timeApi
	 * @returns {?number}
	 */
	static async sample(timeApi) {
		try {
			const start = performance.now();
			const serverResponse = await Promise.race([fetch(timeApi), new Promise((resolve, reject) => setTimeout(reject, 2000))]);
			const serverResponseTime = parseInt(await serverResponse.text(), 10);
			const roundTripTime = performance.now() - start;
			const clientTime = Date.now();
			const serverTime = serverResponseTime - (roundTripTime / 2);
			const diff = serverTime - clientTime;
			return diff;
		} catch (err) {
			console.log(err);
		}
		return null;
	}

	/**
	 * Samples the difference between local and server time.
	 * @param {!string} timeApi
	 * @returns {?number}
	 */
	/*
	static async samplePrecise(timeApi) {
		try {
			const response = await fetch(timeApi);
			const clientTime = Date.now();
			const body = await response.text();
			const perf = performance.getEntriesByName(timeApi).pop();
			const rtt = perf.responseStart - perf.fetchStart;
			const serverTime = Number(body) - (rtt / 2);
			const diff = serverTime - clientTime;
			return diff;
		} catch (err) {
			console.log(err);
		}
		return null;
	}
	*/
	static get TIME_SYNC_DELAY() { return 5 * 60 * 1000; }
}
