class Transition {
	/**
	 * @param {!Element} el
	 * @param {!number} duration
	 * @param {!Size} size
	 * @param {!Point} position
	 * @param {!Size} parentSize
	 * @param {!Point} parentPosition
	 * @param {!Type.Transition} type
	 * @param {!Type.PlaybackEvent} playbackEventType
	 */
	constructor(el, duration, size, position, parentSize, parentPosition, type, playbackEventType) {
		this.duration = duration;
		this.type = type;
		this.playbackEventType = playbackEventType;
		this.animation = null;
		this.willChange = '';

		if (el !== null && type !== Type.Transition.NONE) {
			let keyframes;
			const options = {
				duration: duration
			};
			options.direction = this.playbackEventType === Type.PlaybackEvent.ENTRY ? 'normal' : 'reverse';
			if (this.type === Type.Transition.FADE) {
				keyframes = Transition.fade();
				this.willChange = 'opacity';
			} else if (this.type === Type.Transition.FADE_TO_BLACK || this.type === Type.Transition.FADE_TO_WHITE) {
				keyframes = Transition.fade();
				options.easing = this.playbackEventType === Type.PlaybackEvent.ENTRY ? 'ease-out' : 'ease-in';
				this.willChange = 'opacity';
			} else if (this.type === Type.Transition.SLIDE_L) {
				keyframes = Transition.slide(
					options.direction === 'normal' ? 'left' : 'right',
					size,
					position,
					parentSize
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.SLIDE_R) {
				keyframes = Transition.slide(
					options.direction === 'normal' ? 'right' : 'left',
					size,
					position,
					parentSize
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.SLIDE_T) {
				keyframes = Transition.slide(
					options.direction === 'normal' ? 'up' : 'down',
					size,
					position,
					parentSize
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.SLIDE_B) {
				keyframes = Transition.slide(
					options.direction === 'normal' ? 'down' : 'up',
					size,
					position,
					parentSize
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.EXPAND_L) {
				keyframes = Transition.expand(
					options.direction === 'normal' ? 'left' : 'right'
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.EXPAND_R) {
				keyframes = Transition.expand(
					options.direction === 'normal' ? 'right' : 'left'
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.EXPAND_T) {
				keyframes = Transition.expand(
					options.direction === 'normal' ? 'up' : 'down'
				);
				this.willChange = 'transform';
			} else if (this.type === Type.Transition.EXPAND_B) {
				keyframes = Transition.expand(
					options.direction === 'normal' ? 'down' : 'up'
				);
				this.willChange = 'transform';
			}
			options.fill = options.direction !== 'normal' ? 'backwards' : 'forwards';
			this.animation = el.animate(keyframes, options);
			this.animation.finish();
		}
	}

	/**
	 * @param {!number} stamp
	 */
	start(stamp) {
		if (this.animation !== null) {
			this.animation.startTime = stamp;
			this.animation.play();
		}
	}

	finish() {
		if (this.animation !== null) {
			this.animation.finish();
		}
	}
	destroy() {
		if (this.animation !== null) {
			this.animation.cancel();
			this.animation = null;
		}
	}

	/**
	 * @param {!function} handler
	 */
	onFinish(handler) {
		if (this.animation !== null) {
			this.animation.onfinish = handler;
		}
	}

	/**
	 * @param {!function} handler
	 */
	onCancel(handler) {
		if (this.animation !== null) {
			this.animation.oncancel = handler;
		}
	}

	/**
	 * @returns {!Keyframe[]}
	 */
	static fade() {
		return [
			{
				opacity: 0
			},
			{
				opacity: 1
			}
		];
	}

	/**
	 * @param {!string} direction e.g. 'left', 'right', 'up', 'down'
	 * @param {!Size} size
	 * @param {!Point} position
	 * @param {!Size} parentSize
	 * @returns {!Keyframe[]}
	 */
	static slide(direction, size, position, parentSize) {
		const initial = new Point(0, 0);
		if (direction === 'left') {
			initial.x = parentSize.width;
		} else if (direction === 'right') {
			initial.x = -(position.x + size.width);
		} else if (direction === 'up') {
			initial.y = parentSize.height;
		} else if (direction === 'down') {
			initial.y = -(position.y + size.height);
		}
		return [
			{
				transform: `translate(${initial.x}px,${initial.y}px)`
			},
			{
				transform: 'translate(0,0)'
			}
		];
	}

	/**
	 * @param {!string} direction e.g. 'left', 'right', 'up', 'down'
	 */
	static expand(direction) {
		const initial = new Point(1, 1);
		const target = new Point(1, 1);
		let transformOrigin = '';
		if (direction === 'left') {
			initial.x = 0;
			transformOrigin = 'right';
		} else if (direction === 'right') {
			initial.x = 0;
			transformOrigin = 'left';
		} else if (direction === 'up') {
			initial.y = 0;
			transformOrigin = 'bottom';
		} else if (direction === 'down') {
			initial.y = 0;
			transformOrigin = 'top';
		}
		return [
			{
				transform: `scale(${initial.x}, ${initial.y})`,
				transformOrigin: transformOrigin
			},
			{
				transform: `scale(${target.x}, ${target.y})`,
				transformOrigin: transformOrigin
			}
		];
	}

	/**
	 * Merges multiple willChange strings (may be single or multiple willChange values,
	 * e.g. 'opacity' or 'opacity, transform'.)
	 * @param {!string} ...willChanges
	 * @returns {!string}
	 */
	static mergeWillChange(...willChanges) {
		const props = willChanges.map(wc => wc.split(','));
		const flattened = Utility.flat(props);
		const set = new Set(flattened);
		let willChange = '';
		for (const wc of set) {
			if (wc.length > 0) {
				willChange += `${willChange.length > 0 ? ', ' : ''}${wc}`;
			}
		}
		return willChange;
	}
}

class Animation {
	/**
	 * @param {!Element} el
	 * @param {!number} duration
	 * @param {!Type.Animation} type
	 */
	constructor(el, duration, type) {
		this.duration = duration;
		this.type = type;
		this.animation = null;
		this.willChange = '';

		if (el !== null && type !== Type.Animation.NONE) {
			let keyframes;
			const options = {
				duration: duration,
				fill: 'both'
			};
			if (this.type === Type.Animation.ZOOM_PAN_STANDARD) {
				keyframes = Animation.zoomPan(this.duration);
				this.willChange = 'transform';
			}
			this.animation = el.animate(keyframes, options);
			this.animation.finish();
		}
	}

	/**
	 * @param {!number} stamp
	 */
	start(stamp) {
		if (this.animation !== null) {
			this.animation.startTime = stamp;
			this.animation.play();
		}
	}

	finish() {
		if (this.animation !== null) {
			this.animation.finish();
		}
	}

	destroy() {
		if (this.animation !== null) {
			this.animation.cancel();
			this.animation = null;
		}
	}

	/**
	 * @param {!number} duration
	 */
	static zoomPan(duration) {
		const initial = 1;
		const rate = 0.25 / 900;
		const target = 1 + rate * duration / 1000 * 60;
		return [
			{
				transform: `translate(50%, 50%) scale(${initial}) translate(-50%, -50%)`
			},
			{
				transform: `translate(50%, 50%) scale(${target}) translate(-50%, -50%)`
			}
		];
	}
}
