/* exported Utility */
window.Utility = (() => {
	const textEncoder = new TextEncoder();
	const textDecoder = new TextDecoder();
	let uid = 0;
	return class Utility {
		/** @returns {!string} */
		static get ledOffColor() {
			return 'rgb(0,0,0)';
		}

		/**
		 * Returns a number that is unique across all of this instances getUID calls.
		 * @returns {!number}
		 */
		static getUID() {
			return uid++;
		}

		/**
		 * Deletes all entries from a map-like Object.
		 * @param {!Object} dict
		 */
		static clear(dict) {
			for (const key of Object.keys(dict)) {
				delete dict[key];
			}
		}

		/**
		 * Returns val converted to an integer Number. Throws an error if val is not a
		 * natural number, or if val lies outside the optional min and max constraints.
		 * @param {!string} val
		 * @param {?number} [min]
		 * @param {?number} [max]
		 * @returns {!number}
		 */
		static toInteger(val, min = null, max = null) {
			if (val === null || val === '') {
				throw new Error(`Expected integer value, received non-integer value '${val}'!`);
			}
			const x = Number(val);
			if (!Number.isInteger(x)) {
				throw new Error(`Expected integer value, received non-integer value '${val}'!`);
			}
			if (min !== null && x < min) {
				throw new Error(`Value is smaller than min constraint: ${x} < ${min}!`);
			}
			if (max !== null && x > max) {
				throw new Error(`Value is larger than max constraint: ${x} > ${max}!`);
			}
			return x;
		}

		/**
		 * @param {!string} hex e.g. '#aaa' or '#ffffff'
		 * @returns {!{r: !number, b: !number, g: !number}}
		 */
		static hexToRgb(hex) {
			hex = hex.replace('#', '').trim();
			const digits = hex.length === 6 ? 2 : 1;
			return {
				r: parseInt(hex.substr(0, digits), 16),
				g: parseInt(hex.substr(digits, digits), 16),
				b: parseInt(hex.substr(2 * digits, digits), 16)
			};
		}

		/**
		 * @param {!string} cdnUrl
		 * @returns {!string}
		 */
		static getFilename(cdnUrl) {
			let index = cdnUrl.lastIndexOf('/');
			if (index > 0) {
				cdnUrl = cdnUrl.substring(index + 1);
			}
			cdnUrl = cdnUrl.replace(/\W+/g, '_');
			index = cdnUrl.lastIndexOf('_');
			if (index > 0) {
				cdnUrl = cdnUrl.substring(0, index) + '.' + cdnUrl.substring(index + 1);
			}
			return cdnUrl;
		}

		/**
		 * @param {!Error} e
		 */
		static unhandledError(e) {
			throw e;
		}

		/**
		 * @param {!number} ms
		 * @returns {!string}
		 */
		static formatMilliseconds(ms) {
			const totalSeconds = ms / 1000;
			const seconds = Math.round(totalSeconds) % 60;
			const totalMinutes = (totalSeconds - seconds) / 60;
			const minutes = Math.round(totalMinutes) % 60;
			const totalHours = (totalMinutes - minutes) / 60;
			const hours = Math.round(totalHours) % 60;

			return `${hours}:${minutes}:${seconds}s`;
		}

		/**
		 * @param {!string} url
		 * @returns {!string}
		 */
		static getUrlPathname(url) {
			const fileUrl = new URL(url);
			return fileUrl.pathname.replace('/', '');
		}

		/**
		 * @param {?Object} o
		 * @returns {!string}
		 */
		static toValuesString(o) {
			if (o === null) {
				return 'null';
			}
			if (typeof o !== 'object') {
				return o;
			}
			if (Array.isArray(o) || o instanceof Object) {
				const keys = Object.keys(o);
				if (keys.length === 0) {
					return o.toString();
				} else {
					return keys
						.map(key => Utility.toValuesString(o[key]))
						.join('');
				}
			}
		}

		/**
		 * @param {!string} str
		 * @returns {!string}
		 */
		static asciiToHex(str) {
			let hex = '';
			for (let itr = 0; itr < str.length; ++itr) {
				const ch = str.charCodeAt(itr).toString(16);
				hex += ch.length === 1 ? '0' + ch : ch;
			}
			return hex;
		}

		/**
		 * @param {!string} str
		 * @returns {!string}
		 */
		static hexToAscii(str) {
			str = str.replace(/ /g, '');
			let ascii = '';
			for (var itr = 0; itr < str.length; itr += 2) {
				ascii += String.fromCodePoint(Number(`0x${str.substring(itr, itr + 2)}`));
			}
			return ascii;
		}

		/**
		 * @param {!string} str
		 * @param {!string} [algo]
		 * @returns {!Promise<!string>}
		 */
		static async hash(str, algo = 'SHA-1') {
			const buffer = await crypto.subtle.digest(
				algo,
				textEncoder.encode(str)
			);
			const ascii = textDecoder.decode(buffer);
			return Utility.asciiToHex(ascii);
		}

		/**
		 * Convenience wrapper around fetch that throws an error on non-2xx responses and
		 * optionally returns the response in the desired format.
		 * @param {!string} url
		 * @param {!Object} [options]
		 * @param {!Object} [options.init] Passed directly to fetch as the optional init argument.
		 * @param {!String} [options.dataType] If set to 'json', 'text', 'arrayBuffer', 'blob',
		 * or 'formData' then resolves the response as the desired type.
		 * @returns {!Promise<(!Response|?Object)>}
		 */
		static async fetch(url, options = {}) {
			const response = await fetch(url, options.init);
			if (!response.ok) {
				throw new Error(`Response ${response.status} - ${response.statusText}`);
			}
			if (options.dataType === 'json') {
				return response.json();
			} else if (options.dataType === 'text') {
				return response.text();
			} else if (options.dataType === 'arrayBuffer') {
				return response.arrayBuffer();
			} else if (options.dataType === 'blob') {
				return response.blob();
			} else if (options.dataType === 'formData') {
				return response.formData();
			} else {
				return response;
			}
		}

		/**
		 * Returns a unix timestamp.
		 * @param {(!string|!number)} mtime Either a unix timestamp or an ISO 8601
		 * datetime string.
		 * @returns {!number}
		 */
		static parseMtime(mtime) {
			const unix = Number(mtime);
			if (Number.isNaN(unix)) {
				// ISO 8601 datetime strings must be converted
				return moment(mtime).unix();
			}
			return unix;
		}

		/**
		 * Inserts a new style sheet with the given CSS rules and optional id.
		 * @param {!string} rules contents of the style tag
		 * @param {!string} [id]
		 * @returns {!Element}
		 */
		static insertStyleSheet(rules, id) {
			return document.head.appendChild($.style({ id, html: rules }));
		}

		/**
		 * Disables and deletes a given style sheet.
		 * @param {?Element} style the styles DOM element
		 */
		static deleteStyleSheet(style) {
			if (style !== null) {
				style.disabled = true;
				style.parentNode.removeChild(style);
			}
		}

		/**
		 * MDN flattenDeep function
		 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/flat
		 * @param {!Object[]} input
		 * @returns {!Object[]}
		 */
		static flat(input) {
			const stack = [...input];
			const res = [];
			while (stack.length) {
				const next = stack.pop();
				if (Array.isArray(next)) {
					stack.push(...next);
				} else {
					res.push(next);
				}
			}
			return res.reverse();
		}

		/**
		 * @param {!boolean} predicate
		 * @param {!string} msg
		 */
		static assert(predicate, msg) {
			if (!predicate) {
				throw new Error(`Assertion failure: ${msg}`);
			}
		}

		/**
		 * @param {!boolean} cacheSupport
		 * @param {!boolean} loadFromCache
		 * @param {?FileInfo} file
		 * @param {?string} [url]
		 * @param {?string} [dataType]
		 * @returns {!Promise<?Object>}
		 */
		static async load(cacheSupport, loadFromCache, file, url = null, dataType = 'text') {
			if (loadFromCache && cacheSupport) {
				try {
					const data = await file.load();
					if (data !== null) {
						if (dataType === 'json') {
							return JSON.parse(data);
						}
						return data;
					}
				} catch (err) {
					console.log(`Couldn't read local file '${file}', falling back to downloading from server: ${err}`);
				}
			}
			const data = await Network.download(url, dataType);
			if (cacheSupport) {
				try {
					await file.save(
						dataType === 'json' ?
							JSON.stringify(data) :
							data
					);
				} catch (err) {
					console.log(`Unable to write data to '${file}': ${err}.`);
				}
			}
			return data;
		}
	};
})();

class Size {
	/**
	 * @param {?number} width
	 * @param {?number} height
	 */
	constructor(width, height) {
		this.width = width;
		this.height = height;
	}
}

class Point {
	/**
	 * @param {?number} x
	 * @param {?number} y
	 */
	constructor(x, y) {
		this.x = x;
		this.y = y;
	}
}
