/* exported Zone */
/* global Size, Point */

class Zone {
	/**
	 * @param {!number} volume
	 * @returns {!number}
	 */
	static parseVolume(volume) {
		return volume === undefined ? 0 : volume / 100;
	}

	/**
	 * @param {!Playlist} playlist
	 * @param {!Object} data
	 */
	constructor(playlist, data) {
		this.player = playlist.player;
		this.playlist = playlist;
		/** @type {!number} */
		this.id = data.id;
		
		this.maxVolume = Zone.parseVolume(data.volume);

		this.position = new Point(data.x, data.y);
		this.size = new Size(data.width, data.height);
		
		this.scale = playlist.scale;
		
		this.elId = `zone-${this.playlist.id}-${this.id}-${Utility.getUID()}`;
		this.compositeElementId = `composites-${this.playlist.id}-${this.id}-${Utility.getUID()}`;
		this.$zoneLayer = $.div({ id: this.elId, class: 'zoneContainer', css: {
			pointerEvents: 'none',
			position: 'absolute',
			width: `${this.size.width}px`,
			height: `${this.size.height}px`,
			left: `${this.position.x}px`,
			top: `${this.position.y}px`,
			overflow: 'hidden'
		} });
		
		this.$compositesLayer = $.div({ id: this.compositeElementId, class: 'compositesContainer', css: {
			pointerEvents: 'none',
			position: 'absolute',
			width: `${this.size.width}px`,
			height: `${this.size.height}px`,
			left: `${this.position.x}px`,
			top: `${this.position.y}px`,
			overflow: 'hidden'
		} });
		
		/**
		 * The amount of time in milliseconds before the next Media Event is ready to play (see {@link Zone#events}), updated each time {@link Zone#update} is called.
		 * @type {!number}
		 */
		this.delay = 0;

		/**
		 * The timestamp in milliseconds when the next Media Event should play (see {@link Zone#events}).
		 * @type {!number}
		 */
		this.nextEventStamp = 0;

		/**
		 * The timestamp in milliseconds when the first Media Event was played.
		 * @type {!number}
		 */
		this.startStamp = 0;

		/**
		 * The list of current Media Events that should be played, updated each time {@link Zone#update} is called.
		 * @type {?MediaEvents[]}
		 */
		this.currentEvents = null;

		/**
		 * The sum total time of all {@link Zone#activeMedia} discounting the Media transition durations.
		 * @type {!number}
		 */
		this.duration = 0;

		/**
		 * The amount of extra time this Zone requires during the first iteration of playback (due to entry tranisitons at the start of the Zone.)
		 * Alternatively, this can be thought of as the "time to zero" of the Zone. It will ususally be equal to half of the first Medias entry transition duration, but e.g. the first Media could have no entry transition and a
		 * duration of 1, and the second Media could have an entry transition duration of 3000, meaning the first Media Event will be the second Media entry, and then this offset would be equal to some portion of that.
		 * @type {!number}
		 */
		this.firstEventOffset = 0;

		/**
		 * If set to true, then the Zone will set a delay of Infinity before reaching the final event in {@link Zone#events}.
		 * @type {!boolean}
		 */
		this.stopOnLastMedia = false;

		/**
		 * When {@link Zone#stopOnLastMedia} is enabled and the Zone has stopped playing due to reaching the end of the Zone, this will be set to true.
		 * @type {!boolean}
		 */
		this.stopped = false;

		/**
		 * The activeMedia is a subset of the {@link Zone#media} list containing only the Media which are currently enabled (see {@link Media#enabled}.)
		 * This list must be recalculated each time a {@link Media#enabled} property changes, with the exception that interactive frames cache the activeMedia view assossciated with each button.
		 * @type {!Media[]}
		 */
		this.activeMedia = [];

		/**
		 * The ordered list of Media Events (see {@link Media#entryTime} and {@link Media#exitTime}). Represents a single iteration of the Zone. Only generates Media Events for enabled Media (see {@link Media#enabled} and {@link Zone#activeMedia}).
		 * @type {!{media: !Media, type: !Type.PlaybackEvent, defaultTime: ?number, nthTime: ?number}[]}
		 */
		this.events = [];

		/**
		 * Points to an event in the {@link Zone#events} list.
		 * @type {!number}
		 */
		this.eventCursor = 0;

		/**
		 * The full list of children Media that belong to this Zone. This list is only used for calculating the {@link Zone#activeMedia} view.
		 * @type {!Media[]}
		 */
		this.media = data.media.map(data => new Media(this, data));
	}

	/**
	 * @param {!HTMLElement} $parent
	 */
	attach($parent) {
		$parent.appendChild(this.$compositesLayer);
		$parent.appendChild(this.$zoneLayer);
	}

	/**
	 * Performs all the necessary tasks for the Zone and associated Media objects to be 
	 * ready to show.
	 * @returns {!Promise[]}
	 */
	initialize() {
		return this.media.map(media => {
			media.attach(this.$zoneLayer);
			media.attachTransitionalMedia(this.$compositesLayer);
			return media.initialize();
		});
	}

	TimeSync() {
		for (const media of this.media) {
			media.TimeSync();
		}
	}

	// This updates the activeMedia array and was pulled out of updateView below so that all the zone durations could be calculated and the maximum run length would be known when running updateView
	updateActiveMedia() {
		const oldActiveMedia = this.activeMedia.slice();
		
		// Reset media that was previously enabled but is now disabled
		for (const media of this.activeMedia) {
			if (!media.enabled) {
				media.reset();
				this.playlist.ResetMedia(media);
			}
		}
		this.activeMedia = this.media.filter(media => media.enabled);
		this.activeMediaChanged = oldActiveMedia.length !== this.activeMedia.length || oldActiveMedia.some((media, index) => media.id != this.activeMedia[index].id);
		this.duration = this.activeMedia.reduce((duration, media) => duration + media.duration, 0);
	}

	/**
	 * Syncs the Zone properties and Media Events list to the current state of active Media. This should be called whenever a child Media object changes its {@link Media#enabled} flag, in order for that change to be reflected in Zone playback.
	 * @param {!number} stamp The current timestamp in milliseconds.
	 * @param {!boolean} restart If true, restarts the Zone playback to the first event in the newly generated {@link Zone#events} list.
	 * @param {!boolean} stopOnLastMedia If true, stops the Zone from looping once the end of the Zones playback is reached.
	 */
	updateView(stamp, restart, stopOnLastMedia, maximumZoneRunLength) {
		console.log('LOGS:', this.playlist.name, this.playlist.disableLoopOnZones, this.duration, maximumZoneRunLength);
		stopOnLastMedia = stopOnLastMedia || this.playlist.disableLoopOnZones && this.duration == maximumZoneRunLength;
		// Force a restart if the zone has stopped but should no longer stop.
		if (this.stopOnLastMedia && !stopOnLastMedia && this.stopped) {
			restart = true;
		}
		this.stopOnLastMedia = stopOnLastMedia;
		
		// Restart if the active media has changed
		restart = restart || this.activeMediaChanged;
		
		// NOTE: At the moment cloned media is disabled, instead in any situation where a cloned media would be generated and used, instead the zone will enable the stopOnLastMedia flag. This solution isn't great because it prevents zones that have a single media item from having repeat transitions or animations, but it seems like this is a better default.
		// Audio zones are exempt from this rule and must loop even if only one media item is active (as audio files cannot have transitions currently anyway.)
		// Brandon removed this code after talking to John M.
		if (this.activeMedia.length == 1) {
			if (this.activeMedia[0].type != Type.Resource.AUDIO && this.activeMedia[0].type != Type.Resource.VIDEO) {
				this.stopOnLastMedia = true;
			}
		}
		
		this.generateMediaEvents();
		
		if (restart) {
			this.restartZone(stamp);
		}
	}

	/**
	 * Restarts playback to the beginning of the zone.
	 * @param {!number} stamp
	 */
	restartZone(stamp) {
		for (const media of this.activeMedia) {
			media.reset();
			this.playlist.ResetMedia(media);
		}
		this.eventCursor = 0;
		this.nextEventStamp = stamp;
		this.startStamp = stamp + this.firstEventOffset;
		this.stopped = false;
	}

	/**
	 * Generates and sorts the list of Media Events that describes the order that Media Events occur in this Zone. See {@link Zone#events} for additional information.
	 */
	generateMediaEvents() {
		console.log('generateMediaEvents');

		// Generate the Media Events for the first playback iteration of the Zone.
		this.events = [];
		let startTime = 0;
		console.log(`this.activeMedia: ${this.activeMedia.length}`);
		for (const media of this.activeMedia) {
			media.startTime = startTime;
			media.updateMediaEvents();
			this.events.push(media.entry.event, media.exit.event);
			startTime += media.duration;
		}

		this.events.sort((a, b) => a.defaultTime - b.defaultTime);

		this.firstEventOffset = this.events.length > 0 ? Math.abs(this.events[0].defaultTime) : 0;

		if (this.events.length === 0) {
			// empty zone
			return;
		}

		// Bind all Media Events to the interval [0, this.duration).
		const offset = Math.abs(this.events[0].defaultTime);
		for (const evt of this.events) {
			// Time shift all Media Events over to account for the first Media Event possibly being negative (it is guaranteed to be <= 0).
			evt.nthTime = evt.defaultTime + offset;
			evt.nthTime %= this.duration;
		}

		// Sort the Media Events again in ascending order
		this.events.sort((a, b) => a.nthTime - b.nthTime);

		// There is a special case where the Zone has only a single active Media without an ENTRY or EXIT transition that causes two playback issues: the first being that the events list is out of order and the second is that it's pointless to remove the Media if it will be immediately added back (which causes issues with things like video Input).
		// A simple fix to both problems comes from removing the EXIT event and leaving only the ENTRY event. The player detects if the Media is already added when it plays ENTRY events, and will simply reset the playback state making this work.
		// NOTE: This is the ONLY case where there are ONLY two Media Events.
		if (this.events.length === 2) {
			this.events = [this.activeMedia[0].entry.event];
		}
	}

	selectNextEvent() {
		this.eventCursor = ++this.eventCursor % this.events.length;
	}

	getCurrentEvent() {
		return this.events[this.eventCursor];
	}

	/**
	 * @param {!number} stamp The current timestamp in milliseconds.
	 */
	update(stamp) {
		if (!this.hasOwnProperty('logLimit')) {
			this.logLimit = 0;
		}
		const limitedLog = message => {
			if (this.logLimit < 10) {
				console.log(message);
				this.logLimit++;
			}
		};
		this.currentEvents = null;
		if (stamp >= this.nextEventStamp) {
			limitedLog(`stamp: ${stamp}`);
			const timingError = stamp - this.nextEventStamp;
			this.currentEvents = [];
			do {
				if (this.stopOnLastMedia) {
					let shouldStop = stamp > this.startStamp + this.duration + this.getCurrentEvent().defaultTime;
					if (shouldStop) {
						this.stopped = true;
						this.currentEvents = null;
						break;
					}
				}
				this.currentEvents.push(this.getCurrentEvent());
				this.delay = this.getDelayToNextEvent();
				limitedLog(`this.delay first: ${this.delay}`);
				this.selectNextEvent();
			} while (this.delay === 0 && this.currentEvents.length < this.events.length);
			limitedLog(`timingError: ${timingError}`);
			this.nextEventStamp = stamp + this.delay - timingError;
			limitedLog(`this.nextEventStamp: ${this.nextEventStamp}`);
		} else {
			this.delay = this.nextEventStamp - stamp;
			limitedLog(`this.delay second: ${this.delay}`);
		}
	}

	/**
	 * @param {!number} stamp The starting timestamp in milliseconds.
	 * @param {!number} initialDelay The amount of time in milliseconds that the 
	 * Zone should wait before it's first event (to synchronize multiple Zones with different starting times).
	 */
	start(stamp, initialDelay) {
		console.log(`zone start, initial delay: ${initialDelay}`);
		this.stopped = false;
		this.delay = 0;
		this.eventCursor = 0;
		console.log(`this.events.length: ${this.events.length}`);
		if (this.events.length === 0) {
			this.nextEventStamp = Infinity;
			this.startStamp = Infinity;
		} else {
			console.log(`this.nextEventStamp = stamp + initialDelay; stamp: ${stamp} initialDelay: ${initialDelay}`);
			this.nextEventStamp = stamp + initialDelay;
			this.startStamp = this.nextEventStamp + this.firstEventOffset;
		}
	}

	/**
	 * @returns {!number} Returns the delay from the the current event pointed to by
	 * {@link Zone#eventCursor} to the next event in the {@link Zone#events} list.
	 */
	getDelayToNextEvent() {
		if (this.events.length === 0 || this.stopped) {
			return Infinity;
		}
		let eventCursor = this.eventCursor;
		const currentEvent = this.events[eventCursor];
		let eventCursorWrapped = false;
		if (++eventCursor >= this.events.length) {
			eventCursor %= this.events.length;
			eventCursorWrapped = true;
		}
		const nextEvent = this.events[eventCursor];
		if (eventCursorWrapped) {
			const extraTime = this.duration - currentEvent.nthTime;
			return nextEvent.nthTime + extraTime;
		} else {
			return nextEvent.nthTime - currentEvent.nthTime;
		}
	}

	/**
	 * A clone is only required when the Zone loops playback, has only one active media
	 * item, and that active media item has an entry or an exit transition. Additionally,
	 * SignWave widgets shall not be cloned.
	 * @returns {!boolean}
	 */
	shouldClone() {
		const playbackLoops = !this.stopOnLastMedia;
		const isUnit = this.activeMedia.length == 1;
		if (playbackLoops && isUnit) {
			const media = this.activeMedia[0];
			const hasTransition = media.entry.transition.type !== Type.Transition.NONE || 
				media.exit.transition.type !== Type.Transition.NONE;
			return hasTransition && !media.isSignwave();
		}
		return false;
	}

	/**
	 * Calls {@link Media#reset} on each contained Media.
	 */
	reset() {
		for (const media of this.media) {
			media.reset();
			this.playlist.ResetMedia(media);
		}
	}

	/**
	 * Calls {@link Media#destroy} on each contained Media.
	 */
	destroy() {
		for (const media of this.media) {
			media.destroy();
		}
	}
}
