const serialClientBus = new EventEmitter();

class SerialClient {
	static serialListDevices() {
		return new Promise((resolve, reject) => {
			chrome.serial.getDevices(devices => {
				if (chrome.runtime.lastError !== undefined) {
					reject({ reason: chrome.runtime.lastError.message });
				} else {
					resolve(devices);
				}
			});
		});
	}
	static serialConnect(port, bitrate, dataBits, parityBits, stopBits, ctsFlowControl) {
		return new Promise((resolve, reject) => {
			chrome.serial.connect(port, {
				bitrate: bitrate,
				dataBits: dataBits === 7 ? 'seven' : 'eight',
				parityBit: parityBits === 'none' ? 'no' : parityBits,
				stopBits: stopBits === 2 ? 'two' : 'one',
				ctsFlowControl: ctsFlowControl
			}, connectionInfo => {
					if (chrome.runtime.lastError !== undefined) {
						reject({ reason: chrome.runtime.lastError.message });
					} else {
						resolve(connectionInfo.connectionId);
					}
				}
			);
		});
	}
	static serialDisconnect(connectionId) {
		return new Promise(resolve => {
			chrome.serial.disconnect(connectionId, result => {
				resolve(result);
			});
		});
	}
	static serialWrite(connectionId, data) {
		const utf8decoder = new TextDecoder('utf-8');
		console.log('writing data:');
		console.log(utf8decoder.decode(data));
		return new Promise(resolve => {
			chrome.serial.send(connectionId, data, sendInfo => {
				resolve(sendInfo);
			});
		});
	}
	static serialFlush(connectionId) {
		return new Promise(resolve => {
			chrome.serial.flush(connectionId, result => {
				resolve(result);
			});
		});
	}
	static serialRead(info) {
		const utf8decoder = new TextDecoder('utf-8');
		serialClientBus.emit('serialRead' + info.connectionId, utf8decoder.decode(info.data));
	}
	static serialReceiveError(info) {
		serialClientBus.emit('serialReceiveError' + info.connectionId, info.error);
	}
	/**
	 * Opens and writes to a serial port, flushes the connection and then closes.
	 * This will resolve to any output or errors read from the connection while 
	 * it is open, but, it is unreliable to assume you will receive ALL of the 
	 * output from the connection as no blocking or waiting is done before closing
	 * the connection.
	 * @returns {!Promise}
	 */
	static async request(portName, command, format, baudRate = 9600, dataBits = 8, parity = 'none', stopBits = 1, flowControl = false) {
		try {
			const devices = await SerialClient.serialListDevices();
			console.log('--- List of Ready Serial Devices ---');
			for (const device of devices) {
				console.log(`path: ${device.path}, vendorId: ${device.vendorId}, productId: ${device.productId}, displayName: ${device.displayName}`);
			}
		} catch (exception) {
			console.log(exception.reason);
		}
		
		try {
			let output = '';
			const onRead = data => {
				output = output + data;
			};
			const onReceiveError = error => {
				output = output + error;
			};
			console.log(`Attempting to open serial connection to ${portName}...`);
			let ttyConnId = await SerialClient.serialConnect(portName, baudRate, dataBits, parity, stopBits, flowControl);
			serialClientBus.on('serialRead' + ttyConnId, onRead);
			serialClientBus.on('serialReceiveError' + ttyConnId, onReceiveError);
			console.log(`Opened serial connection to ${portName}, connection id: ${ttyConnId}...`);
			console.log(`Writing to serial connection ${portName}: ${command}`);
			let buffer;
			if (format === 'hex') {
				buffer = Buffer.from(command, 'hex');
			} else {
				buffer = Buffer.from(command, 'ascii');
			}
			let sendInfo = await SerialClient.serialWrite(ttyConnId, buffer.buffer.slice(buffer.byteOffset, buffer.byteOffset + buffer.byteLength));
			if (sendInfo.error) {
				console.log(`Encountered error ${sendInfo.error} while sending data on serial connection ${portName}...`);
			} else {
				console.log(`Wrote ${sendInfo.bytesSent} bytes to serial connection on ${portName}...`);
			}
			let flushStatus = await SerialClient.serialFlush(ttyConnId);
			if (!flushStatus) {
				console.log(`Could not flush serial connection on ${portName}...`);
			} else {
				console.log(`Successfully flushed serial connection on ${portName}...`);
			}
			// chrome.serial is deprecated and flush no longer works. Wait 1 second.
			// TODO: Change this all to use Web Serial API
			await new Promise(resolve => setTimeout(resolve, 1000));
			
			let disconnectStatus = await SerialClient.serialDisconnect(ttyConnId);
			if (!disconnectStatus) {
				console.log(`Could not disconnect serial connection from ${portName}...`);
			} else {
				console.log(`Succesfully disconnected serial connection on ${portName}...`);
			}
			serialClientBus.removeListener('serialRead' + ttyConnId, onRead);
			serialClientBus.removeListener('serialReceiveError' + ttyConnId, onReceiveError);
			return Promise.resolve(output);
		} catch (err) {
			if (err.reason !== undefined) {
				console.log(`Serial connection interrupted: ${err.reason}`);
			} else {
				throw err;
			}
		}
	}
}

chrome.serial.onReceive.addListener(SerialClient.serialRead);
chrome.serial.onReceiveError.addListener(SerialClient.serialReceiveError);