/**
 * @param {!string} versionString
 * @returns {!Promise}
 */
async function createUpdateLock(versionString) {
	const filepath = path.join(paths.DIRECTORY, 'update.lock');
	try {
		await fs.writeFile(filepath, JSON.stringify(versionString));
	} catch (err) {
		console.log(err);
		throw err;
	}
}

/**
 * @param {!string} bin
 * @param {!string[]} args
 * @param {!string} cwd
 * @returns {!Promise}
 */
async function runDetachedUpdater(bin, args, cwd) {
	const filepath = path.join(paths.CONFIG, 'update.log');
	const log = await fs.open(filepath, 'w');
	let child = spawn(bin, args, {
		detached: true,
		stdio: ['ignore', log, log],
		cwd: cwd
	});
	child.unref();
}

/**
 * @param {!string} cmd
 * @param {!Option} [opts]
 * @returns {!Promise<!number, !string>}
 */
function sudo(cmd, opts) {
	return execute('sudo', ['/bin/sh', '-c', cmd], opts);
}

/**
 * @param {!string} bin
 * @param {!string[]} args
 * @param {!Option} [opts]
 * @returns {!Promise<!number, !string>}
 */
function execute(bin, args, opts, timeout = null) {
	return new Promise((resolve, reject) => {
		if (!opts) {
			opts = {};
		}
		const controller = new AbortController();
		opts.signal = controller.signal;
		const proc = spawn(bin, args, opts);
		console.log(`Spawning process: ${bin} ${args}`);
		proc.stdout.on('data', (data) => {
			console.log(`stdout: ${data}`);
		});
		proc.stderr.on('data', (data) => {
			console.log(`stderr: ${data}`);
		});
		proc.on('exit', (code, signal) => {
			console.log(`${bin} ${args} closed with code ${code}, and signal ${signal}`);
			resolve(code, signal);
		});
		proc.on('error', (err) => {
			console.log(`${bin} ${args} encountered an error`);
			reject(err);
		});
		proc.stdin.end();
		if (timeout) {
			setTimeout(() => {
				controller.abort();
				reject(new Error('Timed out'));
			}, timeout);
		}
	});
}

const LinuxShell = {
	reboot: function() {
		sudo('reboot').catch(e => { throw e; });
	},
	/**
	 * @param {!string} timezone
	 * @returns {!Promise}
	 */
	updateTimezone: async timezone => {
		await sudo('mv /etc/localtime /etc/localtime.old');
		await sudo(`echo "${timezone}" | sudo tee /etc/timezone`);
		await sudo('dpkg-reconfigure --frontend noninteractive tzdata');
		// TODO: Will nodejs use the new timezone without restarting the process or does the tz env variable need to be set as well?
		await sudo('dpkg-reconfigure -f noninteractive tzdata');
	},
	/**
	 * @param {!string} updateFilepath
	 * @param {!string} dstFollder
	 * @param {!string} versionString
	 * @returns {!Promise}
	 */
	runUpdate: async (updateFilepath, dstFolder, versionString) => {
		const unzipFilepath = path.join(paths.UPDATES, dstFolder);
		await execute('unzip', [updateFilepath, '-d', unzipFilepath]);
		const dirs = await FileIO.enumerateDirs(unzipFilepath);
		const updatePath = path.join(unzipFilepath, dirs[0]);
		const scriptPath = path.join(updatePath, 'breeze2nwjscontainer/scripts/updater.sh');
		await createUpdateLock(versionString);
		await sudo(`chmod +x ${scriptPath}`);
		await runDetachedUpdater('/bin/sh', ['-c', scriptPath], updatePath);
	},
	/**
	 * @param {!string} outputFilepath
	 * @returns {!Promise}
	 */
	captureScreenshot: function(outputFilepath) {
		return execute('/usr/local/bin/maim', [outputFilepath]);
	},
	/**
	 * @param {!number} r
	 * @param {!number} g
	 * @param {!number} b
	 * @returns {!Promise}
	 */
	setFrontLedColor: async function(r, g, b) {
		// Not supported on Linux
	}
};

// TODO: Update to use the new execute() format
const WindowsShell = {
	reboot: function() {
		execute('shutdown', ['/r']).catch(e => { throw e; });
	},
	/**
	 * @param {!string} timezone
	 */
	updateTimezone: function(timezone) {
		execute('tzutil', ['/s', timezone]).catch(e => { throw e; });
	},
	/**
	 * @param {!string} updateFilepath
	 * @param {!string} dstFolder
	 * @param {!string} versionString
	 * @returns {!Promise}
	 */
	runUpdate: async (updateFilepath, dstFolder, versionString) => {
		const unzipFilepath = updateFilepath.substring(0, updateFilepath.length - ('.zip').length);
		await new Promise((resolve, reject) => {
			extract(updateFilepath, {dir: unzipFilepath}, (err) => {
				if (err !== undefined) {
					console.log('Unexpected error with runUpdate:');
					console.log(err);
					reject(err);
				} else {
					resolve();
				}
			});
		}).catch(e => { throw e; });
		const dirs = await FileIO.enumerateDirs(unzipFilepath);
		const updatePath = path.join(unzipFilepath, dirs[0]);
		const scriptPath = path.join(updatePath, 'breeze2nwjscontainer/scripts/updater.ps1');
		await createUpdateLock(versionString);
		// powershell.exe cannot be detached. It seemingly instantly destroys itself if detached.
		// https://github.com/nodejs/node-v0.x-archive/issues/8795
		await runDetachedUpdater('cmd', ['/c', 'powershell', '-ExecutionPolicy', 'ByPass', '-File', scriptPath], updatePath);
	},
	/**
	 * @param {!string} outputFilepath
	 * @returns {!Promise}
	 */
	captureScreenshot: async function(outputFilepath) {
		// Do not use the boxcutter -f (fullscreen) flag because it does not correctly account
		// for resolutions that do not use standard (100%) Desktop DPI.
		let w = 0, h = 0;
		nw.Screen.screens.forEach(v => {
			w += v.bounds.width * v.scaleFactor;
			h += v.bounds.height * v.scaleFactor;
			console.log('Screen [' + v.id + '] resolution: ' + v.bounds.width + ', ' + v.bounds.height);
		});
		try {
			return await execute(
				path.join(paths.ASSETS, '/boxcutter/boxcutter.exe'),
				['-c', `0,0,${w},${h}`, outputFilepath], {}, 2000
			);
		} catch (exception) {
			// Do nothing, probably timed out
		}
	},
	/**
	 * @param {!number} r
	 * @param {!number} g
	 * @param {!number} b
	 * @returns {!Promise}
	 */
	setFrontLedColor: async function(r, g, b) {
		try {
			await MeetingMachines.setFrontLedColor(r, g, b);
		} catch (err) {
			console.log(`Error setting front led color: ${err.stack}`);
		}
	}
};

let shell;
let platform;
switch (process.platform) {
	case 'win32':
		Shell = WindowsShell;
		platform = 'Windows';
		break;
	case 'linux':
	default:
		Shell = LinuxShell;
		platform = 'Linux';
		break;
}
