class Version {
	/**
	 * @param {!string} [version] e.g. "major.minor.patch" or "major.minor.patch.buildnumber".
	 * The major, minor, and patch fields must be exactly two digits each. Buildnumber
	 * may be any valid integer.
	 */
	constructor(version = '00.00.00.0') {
		this.major = 0;
		this.minor = 0;
		this.patch = 0;
		this.buildnumber = 0;

		const split = version.split('.');
		if (split.length >= 3) {
			this.major = parseInt(split[0], 10);
			this.minor = parseInt(split[1], 10);
			this.patch = parseInt(split[2], 10);
		}
		if (split.length >= 4) {
			this.buildnumber = parseInt(split[3], 10);
		}
	}

	/**
	 * Adds a leading zero to single digit version numbers.
	 * @param {!number} component major, minor or patch number
	 * @return {!string}
	 */
	pad(component) {
		let str = component.toString();
		return `${str.length === 1 ? '0' : ''}${str}`;
	}

	/**
	 * @param {!string} [delimiter]
	 * @return {!string}
	 */
	format(delimiter = '.') {
		return this.pad(this.major) + delimiter +
			this.pad(this.minor) + delimiter +
			this.pad(this.patch) + delimiter + 
			this.buildnumber;
	}

	/**
	 * Determines if this is a non-zero version number.
	 * @return {!boolean}
	 */
	isValid() {
		return !(this.major === 0 &&
			this.minor === 0 &&
			this.patch === 0 &&
			this.buildnumber === 0);
	}

	/**
	 * Compares two instances of {@code Version} against one another.
	 * @param {!Version} a
	 * @param {!Version} b
	 * @return {!number} Returns:
	 * <ul>
	 * <li> -1 when `a` represents a lower version number than `b`
	 * <li> 0 when `a` represents the same version number as `b`
	 * <li> 1 when `a` represents a higher version number than `b`
	 * </ul>
	 */
	static compare(a, b) {
		// make sure versions are valid
		if(a.major === 0 && a.minor === 0 && a.patch === 0 && a.buildnumber === 0){ return 0; }
		if(b.major === 0 && b.minor === 0 && b.patch === 0 && b.buildnumber === 0){ return 0; }

		if (a.major > b.major) {
			return 1;
		} else if (a.major < b.major) {
			return -1;
		}

		if (a.minor > b.minor) {
			return 1;
		} else if (a.minor < b.minor) {
			return -1;
		}

		if (a.patch > b.patch) {
			return 1;
		} else if (a.patch < b.patch) {
			return -1;
		}

		if (a.buildnumber > b.buildnumber) {
			return 1;
		} else if (a.buildnumber < b.buildnumber) {
			return -1;
		}

		return 0;
	}
}
